package displayPack;

import java.awt.event.ActionEvent;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import learnerPack.Learner;
import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.LearningStage;
import modelPack.NodeType;
import modelPack.Ops;
import modelPack.Slot;
import modelPack.WordToLearn;
import utilPack.Util;

public class Actions {
	
	DrawArea drawArea;

	VFrame vf;
	
	LanguageModel model;
	
	LanguageModel model() 
	{
		if (model != null) return model;
		return vf.getModel();
	}
	
	// needed to service learning replay requests for auto-learnt words
	Learner learner = null;
	public void setLearner(Learner learner) {this.learner = learner;}
	
	// the sentence selected from the sentences menu
	FSNode selectedSentence = null;
	
	// steps in unification of current sentence, or in learning
	Vector<FSNode> steps = new Vector<FSNode>();
	
	// Vector of words, in the order needed to unify
	Vector<String> unifyWords = new Vector<String>();
	
	// current step of unification
	int unifyStep = 0;
	
	// number of generalisations done in learning
	int generalisations = 0;
	
	// partially unified learning examples
	Vector<FSNode> partials = new Vector<FSNode>();
	
	/**
	 * to alternate steps between 'retrieving' a word or learning example from memory 
	 * (i.e showing it beside the partially unified sentence)
	 * and unifying it
	 */
	boolean retrieveStep = true;
		
	// result of the last valid generalisation - which gave enough information in its result scene
	private FSNode lastValidGenResult = null;
	
	// the type of action to do
	int actionType = 0;
		
	// allowed action types
	public static int UNDERSTAND = 1;
	public static int GENERATE = 2;
	public static int LEARN = 3;
	
	// selective tracing
	private boolean tracing = false;	
	private void trace(String s) {if (tracing) Util.message(s);}
	
	  //------------------------------------------------------------------------------------------------
	  //                              constructors and startup
	  //------------------------------------------------------------------------------------------------
	
	
	/**
	 * constructor for the workbench
	 * @param vf
	 * @param drawArea
	 */
	public Actions(VFrame vf, DrawArea drawArea)
	{
		this.vf = vf;
		this.drawArea = drawArea;
	}
	
	/**
	 * constructor for the server
	 * @param model
	 */
	public Actions(LanguageModel model)
	{
		this.model = model;
	}
	
	
	//-------------------------------------------------------------------------------
	//                        Starting demonstrations
	//-------------------------------------------------------------------------------
	
	/**
	 * actions for each button
	 * 
	 * Learning demos are started direct from a menu.
	 * 
	 * Understanding and Generation demos are started by choosing a sentence from a menu, 
	 * then using the Understand or Generate button
	 * 
	 * All demos then use the Step button
	 * 
	 * @param buttonId
	 */
	void doButtonAction(int buttonId)
	{
		try
		{
			if (buttonId == 0) startUnderstandOrGenerate(UNDERSTAND);
			if (buttonId == 1) startUnderstandOrGenerate(GENERATE);
			if (buttonId == 2) doStep();
			if (buttonId == 3) doShowHide();
			if (buttonId == 4) vf.doAcceptEdit();
			if (buttonId == 5) vf.doCancelEdit();
			if (buttonId == 6) vf.doAutoLearn();
			if (buttonId == 7) vf.doSaveClass();
		}
		catch (Exception ex)
		{
			explain(ex.getMessage());
		}
	}
	
	/**
	 * do the next step for each kind of demonstration
	 */
	void doStep()
	{
		vf.showHideButton.setEnabled(true);
		drawArea.setIgnoreScrollBar(true);
		
		if (actionType == UNDERSTAND) doUnderstandOrGenerateStep(UNDERSTAND);
		if (actionType == GENERATE) doUnderstandOrGenerateStep(GENERATE);
		if (actionType == LEARN) doLearnStep();
	}
	
	
	//-------------------------------------------------------------------------------
	//               Demonstration of Understanding or Generation
	//-------------------------------------------------------------------------------
	
	
	/**
	 * start a demonstration of understanding or generation
	 * @param type = UNDERSTAND or GENERATE
	 */
	void startUnderstandOrGenerate(int type) throws Exception
	{
		actionType = type;
		
		// do not allow a sentence FS to be edited any more
		vf.editLanguage.setShownFSType(EditLanguage.NONE);
		
		// set the first step of the unification - sentence sounds only, or meaning only
		steps = new Vector<FSNode>();
		FSNode currentStage = null;
		if (type == UNDERSTAND)
		{
			currentStage = selectedSentence.soundOnly();
			currentStage.setDescription("Sounds only - as heard");
			try {Util.message("Made unify order string: " + makeUnifyOrderString(currentStage.getName()));}
			catch (Exception ex) {ex.printStackTrace();}
		}
		else if (type == GENERATE)
		{
			currentStage = selectedSentence.meaningOnly();
			currentStage.setDescription("Meaning only - before generation");
		}
		steps.add(currentStage);
		
		// set up the first word to unify
		unifyWords = selectedSentence.getUnifyWords();
		// reverse unification order for generation
		if (type == GENERATE) unifyWords = Util.reverse(unifyWords);
		unifyStep = 0;
		if (unifyWords.size() == 0) throw new Exception("No words defined in unify order.");
		String firstWordName = unifyWords.get(0);
		
		FSNode firstWord = model().getWords().get(firstWordName);
		
		// draw the inputs for the first step
		drawArea.setIgnoreScrollBar(true);
		drawArea.addNewRow();
		drawArea.drawFS(false, firstWord, currentStage, null);

		// the first word has already been retrieved
		retrieveStep = false; 
		
		// enable buttons
		vf.stepButton.setEnabled(true);
		vf.understandButton.setEnabled(false);
		vf.generateButton.setEnabled(false);		
	}
	
	/**
	 * 
	 * @param type = UNDERSTAND or GENERATE
	 */
	void doUnderstandOrGenerateStep(int type)
	{
		int step = unifyStep;
		Vector<String> words = unifyWords;
		if (step < words.size())
		{
			String wordName = unifyWords.get(unifyStep);
			FSNode word = new FSNode(model().getWords().get(wordName));
			
			if (word != null) try
			{

				FSNode previousStage = steps.get(unifyStep);
				
				// retrieve step - show the next word to be unified , to the left of the sentence unified so far - on a new row
				if (retrieveStep)
				{
					drawArea.addNewRow();
					drawArea.drawFS(false, word, previousStage, null);
					retrieveStep = false;
				}
				
				// non-retrieve step; do the unification, and show the result
				else if (!retrieveStep)
				{
					int opType = 0;
					if (type == UNDERSTAND) opType = Ops.UNDERSTAND;
					if (type == GENERATE) opType = Ops.GENERATE;
					Util.message("step " + step);
					FSNode currentStage = Ops.unify(word,previousStage, opType);
					
					// now I try to display the result even if it is not viable
					{
						steps.add(currentStage);
						unifyStep++;
						retrieveStep = true;

						currentStage.setName("Result of step " + unifyStep);
						if (!currentStage.isViable()) currentStage.setName("Unify failed at step " + unifyStep);
						drawArea.extendLastRow(currentStage);
						
					}
					if (!currentStage.isViable()) vf.stepButton.setEnabled(false);
					if (step == words.size() - 1) vf.stepButton.setEnabled(false);
				}
				
			}
			catch (Exception ex) {Util.message(ex.getMessage());ex.printStackTrace();}			
		}
	}
	
	/**
	 * 
	 * @param toUnderstand a sentence composed by the user; words separated by '.'
	 * @return
	 * @throws Exception
	 */
	public Vector<Vector<FSNode>> understandComposed(String toUnderstand) throws Exception
	{
		// make a sounds-only FS to understand
		FSNode sentence = makeSoundSentence(toUnderstand);
		// try to understand it
		return generateOrUnderstandAllSteps(sentence,false);
	}

	
	
	/**
	 * used to return the complete results of generation or understanding to the web client
	 * @param sentenceIndex index of the sentence to be gnerate or understood
	 * @param doGenerate
	 * @return
	 */
	public Vector<Vector<FSNode>> generateOrUnderstandForClient(int sentenceIndex, boolean doGenerate) throws Exception
	{
		FSNode sentence = model().getSentence(sentenceIndex);		
		return generateOrUnderstandAllSteps(sentence,doGenerate);
	}
	
	/**
	 * used either to return all results to the web client, 
	 * or to compute a sentence meaning in the language editor
	 * @param sentence
	 * @param doGenerate
	 * @return
	 * @throws Exception
	 */
	public Vector<Vector<FSNode>> generateOrUnderstandAllSteps(FSNode sentence, boolean doGenerate) throws Exception
	{
		Vector<Vector<FSNode>> result = new Vector<Vector<FSNode>>();

		// set up the words to unify
		unifyWords = sentence.getUnifyWords();
		// reverse unification order for generation
		if (doGenerate) unifyWords = Util.reverse(unifyWords);
		int steps = unifyWords.size();

		FSNode firstStage = null;
		if (doGenerate)
		{
			firstStage = sentence.meaningOnly();
			firstStage.setDescription("Meaning only - before generation of sounds");
		}
		
		else
		{
			firstStage = sentence.soundOnly();
			firstStage.setDescription("Sounds only - as heard");
		}
		
		Vector<FSNode> stages = new Vector<FSNode>();
		stages.add(firstStage);
		
		boolean stepSucceeded = true;
		for (int step = 0; step < steps; step++) if (stepSucceeded)
		{
			// set up the two feature structures to unify in this step
			String wordName = unifyWords.get(step);			
			FSNode word = model().getWords().get(wordName);
			FSNode currentStage = stages.get(step);
			
			// unify, and record the result to use in the next step (if there is one)
			FSNode res = Ops.unify(word,currentStage, opType(doGenerate));
			res.setName("result of step " + (step + 1));
			stages.add(res);
			
			// if the step failed, do not try any more steps
			stepSucceeded = res.isViable();
			
			// record the two inputs and the result
			result.add(resultRow(word,currentStage, res, step));
		}
				
		return result;
		
	}
	
	/**
	 * make a result row to send to the web client
	 * @param word
	 * @param currentStage
	 * @param res
	 * @return
	 */
	private Vector<FSNode> resultRow(FSNode word, FSNode currentStage, FSNode res, int rowNo)
	{
		Vector<FSNode> row = new Vector<FSNode>();
		
		FSNode wc = new FSNode(word);
		wc.setRow(rowNo);
		wc.setColumn(0);
		wc.setAllTruePositions();
		row.add(wc);
		
		FSNode cc = new FSNode(currentStage);
		cc.setRow(rowNo);
		cc.setColumn(1);
		cc.setAllTruePositions();
		row.add(cc);
		
		FSNode rc = new FSNode(res);
		rc.setRow(rowNo);
		rc.setColumn(2);
		rc.setAllTruePositions();
		row.add(rc);
		
		return row;
	}

	
	
	/**
	 * 
	 * @param doGenerate
	 * @return
	 */
	private int opType(boolean doGenerate)
	{
		int opType = 0;
		if (doGenerate) opType = Ops.GENERATE;
		else opType = Ops.UNDERSTAND;
		return opType;
	}
	
	/**
	 * 
	 * @param doc
	 * @param result
	 * @return
	 * @throws Exception
	 */
	public void makeResponse(Document doc, Vector<Vector<FSNode>> result, String type) throws Exception
	{
		boolean learn = (type.equals("learn"));

		Element el = doc.createElement("result");
		el.setAttribute("type",type);
		doc.appendChild(el);

		int step = 0;
		for (Vector<FSNode> row: result)
		{
			String wordName = "";
			if (!learn) wordName = unifyWords.get(step);
			
			Element stepEl = doc.createElement("step");
			stepEl.setAttribute("stepNo", Integer.valueOf(step).toString());
			for (int i = 0; i < 3 ; i++)
			{
				FSNode r = row.get(i);
				boolean forClient = true;
				Element rEl = r.xmlForm(doc,forClient);
				// add the name attribute to the script node of the FS, so the client can write it.
				rEl.setAttribute("name", r.getName());
				Element wrapper = doc.createElement("wrapper");
				wrapper.appendChild(rEl);
				if (i == 0) wrapper.setAttribute("wordName",wordName);
				stepEl.appendChild(wrapper);	
			}

			el.appendChild(stepEl);
			step++;
		}
		
	}

	/**
	 * 
	 * @param toUnderstand
	 * @return a sounds-only sentence from words composed by the user
	 */
	private FSNode makeSoundSentence(String toUnderstand) throws Exception
	{
		FSNode script = new FSNode(model(),"script");
		
		String name = "";
		
		// set all the sounds of the script
		StringTokenizer st = new StringTokenizer(toUnderstand,".");
		while (st.hasMoreTokens())
		{
			String word = st.nextToken();
			name = name + word + " ";
			FSNode fsWord = model().getWords().get(word);
			// if you cannot find the word, use the sounds of the first sense of the word you find
			if (fsWord == null)
			{
				Vector<FSNode> senses = model().getSenses(word);
				if (senses != null) fsWord = senses.get(0);
			}
			if (fsWord == null) throw new Exception("Cannot find word '" + word  + "' to build sentence sounds");
			
			// add clones of all sounds from the word to the script
			for (FSNode wordChild: fsWord.getChildren())
				if (wordChild.getType().equals("sound"))
					script.addChild(new FSNode(wordChild));
		}
		
		// set the name and the unify order
		script.setName(name);
		script.setUnifyOrder(makeUnifyOrderString(name));
		
		return script;
	}
			


	//-------------------------------------------------------------------------------------------------------
	//                                   Demonstrations of Learning
	//-------------------------------------------------------------------------------------------------------
	
	/**
	 * On selecting a word to learn, start learning it
	 * @param ev
	 */
	void startLearn(ActionEvent ev, String stageName)
	{
		drawArea.clearAllRows();
		drawArea.setVerticalScrollValue(0);

		String wordName = ev.getActionCommand();
		LearningStage stage = model().getLearningStageTable().get(stageName);
		WordToLearn wLearn = stage.getWordLearn(wordName);
		
		// there must be at least two learning examples
		if (wLearn.getExamples().size() < 2) Util.message("Not enough learning examples for "+ wordName);
		
		actionType = LEARN;
		generalisations = 0;
		
		// explain the learning demo
		explainLearning(wLearn);
		
		vf.stepButton.setEnabled(true);
		
		// display the target word which is to be learned, on the first row
		if (wLearn.getWordFS() != null)
		{
			FSNode target = new FSNode(wLearn.getWordFS());
			target.setName(target.getName() + " - to be learned");
			drawArea.drawFS(false, target, null, null);
		}
				
		// set up all partially understood learning examples
		try {partials = setUpPartials(wLearn);}
		catch (Exception ex) {Util.message(ex.getMessage());}
		
		// display the first two learning examples on the next row below the target
		steps = new Vector<FSNode>();
		// first step is the first learning example
		FSNode first = partials.get(0);
		steps.add(first);
		// the second element of steps is the second learning example
		FSNode second = partials.get(1);
		steps.add(second);
		// later elements of steps will be the results of generalisation
		
		drawArea.addNewRow();
		drawArea.drawFS(false, first, second, null);
		retrieveStep = false;				

	}
	
	/**
	 * 
	 */
	void doLearnStep()
	{
		// special case with generalisations = 0; the first generalise
		int nextPartial = 0; // index of the next partially unified example, in partials
		
		if (generalisations > 0) nextPartial = generalisations + 1;
		else if (generalisations == 0) lastValidGenResult = steps.get(1);

		FSNode example = partials.get(nextPartial);
				
		if (retrieveStep)
		{
			drawArea.addNewRow();
			drawArea.drawFS(false, example, lastValidGenResult, null);
			retrieveStep = false;
		}
		
		else if (!retrieveStep) try
		{
			FSNode genResult = Ops.generalise(example, lastValidGenResult);
			steps.add(genResult);
			
			int[] bits = genResult.getInfoContent();
			
			boolean valid = validResult(lastValidGenResult,genResult);
			String invalidString = "";
			if (!valid) invalidString = " (learning example rejected)";
			
			generalisations++;
			genResult.setName("Result of step " + generalisations + invalidString);
			drawArea.drawFS(false, example, lastValidGenResult, genResult);
			
			if (valid) lastValidGenResult = genResult;
			retrieveStep = true;				
		}
		catch (Exception ex) {ex.printStackTrace();}

	}
	
	/**
	 * for all generalisations except the first one
	 * (all cases where there is a previous valid generalisation, which has the information content of its result defined) 
	 * @param previousResult
	 * @param result
	 * @return
	 */
	private boolean validResult(FSNode previousResult,FSNode result)
	{
		boolean valid = true; // remains valid if the previous result has no defined information content (i.e was a learning example)
		
		// this generalisation is valid if it does not halve the meaning result information of the previous generalisation
		if ((previousResult.getInfoContent() != null) && (result.getInfoContent()  != null))
		{
			int current = result.getInfoContent()[1];
			int past = previousResult.getInfoContent()[1];
			valid = (2*current > past);
		}
			
		return valid;
	}

	
	/**
	 * 
	 * @param wLearn
	 */
	private Vector<FSNode> setUpPartials(WordToLearn wLearn)
	{
		Vector<FSNode> partials = new Vector<FSNode>();
		
		for (FSNode example: wLearn.getExamples())
		{
			// make a clone for progressive unification
			FSNode current = new FSNode(example);
			String trueName = example.getName();

			String wName = "";
			String unifyOrder  = example.getUnifyOrder();
			if ((unifyOrder != null) && (unifyOrder.length() > 0))
			{
				StringTokenizer st = new StringTokenizer(unifyOrder, ", ");
				while (st.hasMoreTokens()) try
				{
					wName = st.nextToken();
					// for cycleNo = 0, the word is from the static language model; for cycleNo > 0, it is a learnt word
					FSNode word = wLearn.getWord(wName);
					
					// only try to unify if you can find the word (it may have been un-learned since it was learned)
					FSNode next = null;					
					if (word == null) Util.message("For partial unifying a learning example, cannot find word " + wName);
					else next = Ops.unify(word,current, Ops.UNDERSTAND);
					
					if ((next != null) && (!next.isViable())) Util.message("Failed Unification of word " + wName + " with example " + trueName);
					else if (next != null) current = next;
				}
				catch (Exception ex) {Util.message("Exception in Unification of word " + wName + " with example " + trueName);ex.printStackTrace();}
				
				current.setName(trueName);
				current.setDescription("Partly understood, using learned words: " + unifyOrder );
			}
			
			partials.add(current);
			
		}
		
		return partials;
	}
	
	/**
	 * 
	 * @param stageIndex
	 * @param wordIndex
	 * @return
	 */
	public Vector<Vector<FSNode>> learnForClient(int stageIndex,int wordIndex, int cycleNo) throws Exception
	{
		Vector<Vector<FSNode>> result = new Vector<Vector<FSNode>>();
		
		LearningStage stage = null;
		
		// for cycle 0, use the static language model
		if (cycleNo == 0) stage = model().getLearningStages().get(stageIndex);
		// for all other cycles, use the words automatically learnt up to this cycle
		else if (cycleNo > 0) {stage = learner.getAllStages().get(cycleNo - 1).get(stageIndex);}

		WordToLearn wl = stage.getLearnWords().get(wordIndex);
		// partial unification of learning examples, using known words
		Vector<FSNode> partials = setUpPartials(wl);
		
		int steps = partials.size() - 1;

		// intermediate results which have been accepted
		Vector<FSNode> intermediates = new Vector<FSNode>();
		
		for (int step = 0; step < steps; step++)
		{
			Vector<FSNode> row = new Vector<FSNode>();

			// first step - get the first two examples
			if (step == 0)
			{
				row.add(partials.get(0));
				row.add(partials.get(1));
			}
			
			// other steps - get the next example, and the last accepted generalise result
			else if (step > 0)
			{
				row.add(partials.get(step + 1));
				row.add(intermediates.get(intermediates.size() - 1));				
			}
			
			// compute the generalisation
			FSNode res = Ops.generalise(row.get(0),row.get(1));

			// test if the result is to be accepted or rejected (compare its result scene information content to the previous generalisation result)
			boolean valid = validResult(row.get(1),res);
			String invalidString = "";
			if (!valid) invalidString = " (learning example rejected)";
			
			res.setName("result of step " + (step + 1) + invalidString);
			
			// store the result if it has not been rejected
			if (valid) intermediates.add(res);
			
			row.add(res); // now there are 3 FSNode in row
			// mark the 3 FS with their row number (step) and columns
			result.add(resultRow(row.get(0),row.get(1),row.get(2),step));			
		}
				
		return result;
		
	}


	
	/**
	 * 
	 * @param wLearn
	 */
	void explainLearning(WordToLearn wLearn)
	{
		String learnEx = "";
		
		learnEx = ("Learning stage: " + wLearn.getStageName());
		learnEx = learnEx + ("\nWord to learn: " + wLearn.getWordName() + "\n");
		
		int exNo = 1;
		boolean someWordsKnown = false;
		for (FSNode example: wLearn.getExamples())
		{
			String eLabel = "\nLearning example " + exNo + ": " + example.getName();
			String knownWords = example.getUnifyOrder();
			if (knownWords == null) eLabel = eLabel + " (no words known yet)";
			else 
			{
				someWordsKnown = true;
				eLabel  = eLabel + " (words already known: " + knownWords + ")";
			}
			learnEx = learnEx + (eLabel);
			exNo++;
		}
		
		learnEx = learnEx + "\n\nFor each learning example, the child hears the sounds (upper nodes) and observes a meaning (bottom node).";
		if (someWordsKnown) learnEx = learnEx + "\nAs some words are already known, the child uses them (by unifying) to partially understand the sounds.";
		else learnEx = learnEx + "\nAs no other words in the learning examples are known, syntax cannot be learned - only the link between sounds and meaning.";
		learnEx = learnEx + "\n\nStep1: Generalise example 1 and example 2";		
		explain(learnEx);
		
	}
	
	
	//-------------------------------------------------------------------------------------------
	//                       showing or hiding used nodes        
	//-------------------------------------------------------------------------------------------
	
	
	/**
	 * 
	 */
	void doShowHide()
	{
		// toggle the state of showing or hiding used nodes
		drawArea.setShowUsedNodes(!drawArea.showUsedNodes());
		drawArea.repaint();
		
		// toggle the button label
		if (vf.showHideButton.getLabel().equals("Show")) vf.showHideButton.setLabel("Hide");
		else if (vf.showHideButton.getLabel().equals("Hide")) vf.showHideButton.setLabel("Show");
	}
	

	
	//-------------------------------------------------------------------------------
	//                    Actions -support for editing, or showing 
	//-------------------------------------------------------------------------------
	
	/**
	 * show a node type description, possibly for editing
	 * @param nType
	 */
	void showNode(NodeType nType, boolean doEdit, EditModel em)
	{
		vf.showHideButton.setEnabled(true);
		explain(nType.getExplanation());
		if (doEdit)
		{
			drawArea.clearAllRows();
			em.setSelectedNodeType(nType);
			em.setEditState(EditModel.MODEL);
		}
	}
	

	
	/**
	 * show a slot
	 * @param s
	 */
	void showSlot (Slot s)
	{
		explain(s.getExplanation());
	}


	
	/**
	 * show a topic and leave it ready for editing
	 * @param topicId
	 */
	protected void showTopic(String topicId, boolean withTitle)
	{
		Topic top = model().getLeafTopic(topicId);		
		if (top != null) 
		{
			explain(top.getCompleteContent(withTitle));
			
			vf.getEditModel().setSelectedTopic(top);
			vf.getEditModel().refreshEditModelMenus(false,null);
		}
	}
	
	/**
	 * clear the explanation area and show something new
	 * disable editing the previous topic 
	 * @param explanation
	 */
	public void explain(String explanation)
	{
		vf.explanationArea.setText("");
		vf.explanationArea.append(explanation);
		vf.explanationArea.setCaretPosition(0); // moves scrollbar to the top
		vf.getEditModel().setSelectedTopic(null);
		vf.getEditModel().refreshEditModelMenus(false,null);
		vf.validate();
	}
	

	
	/**
	 * draw a Feature Structure - a word or sentence
	 * @param value
	 * @param showType 1 for words, 2 for sentences
	 */
	void showFS(String value, int showType, int editActionType)
	{
		// do I always want to clear the drawing first? Not when editing (draw an FS to copy)
		if (!vf.getEditLanguage().isEditing()) drawArea.clear();
		
		FSNode toDraw = null;
		if (showType == EditLanguage.WORD) toDraw = model().getWords().get(value);
		else if (showType == EditLanguage.SENTENCE) toDraw = model().getSentences().get(value);
		
		if (editActionType == EditLanguage.SHOW) showTheFS(toDraw,showType, false);
		else if (editActionType == EditLanguage.EDIT) showTheFS(toDraw,showType, true);
		
		else if (editActionType == EditLanguage.REMOVE)
		{
			boolean doIt = Util.getYesNo("Remove feature structure " + toDraw.getName() + "?" , vf);
			if (doIt)
			{
				if (showType == EditLanguage.WORD) model().getWords().remove(toDraw.getName());
				else if (showType == EditLanguage.SENTENCE) model().getSentences().remove(toDraw.getName());
				
				vf.editLanguage.refreshEditLanguageMenus(null, null,false);
			}
		}
		
	}
	
	/**
	 * show a feature structure of any type, and set up for editing etc.
	 * @param toDraw
	 * @param showType
	 * @param forEdit
	 */
	public void showTheFS(FSNode toDraw, int showType, boolean forEdit)
	{
		
		// draw this FS to the right of an existing drawn FS (for copy & paste) if 
		// (a) that FS is being edited (b) this FS is not being shown for editing
		boolean keepExistingFS = vf.editLanguage.canEdit() && !forEdit;

		if (!keepExistingFS) drawArea.clearAllRows();
		drawArea.setVerticalScrollValue(0);
		vf.stepButton.setEnabled(false);
		vf.editLanguage.setShownFSType(showType);

		if (forEdit)
		{
			toDraw.setEdited(true);
			vf.editLanguage.setEditState(showType);
			// null, null : no node has been selected or copied yet
			vf.editLanguage.refreshEditLanguageMenus(null, null,true);			
		}

		if (toDraw != null) drawArea.drawFS(keepExistingFS, toDraw, null, null);
		else explain("Cannot find Feature structure.");
		
		if (showType == EditLanguage.SENTENCE)
		{
			if (!forEdit)
			{
				vf.understandButton.setEnabled(true);
				vf.generateButton.setEnabled(true);
			}
			selectedSentence = toDraw;
		}
	}
	
	/**
	 * show a Feature Structure  - a learning example
	 * @param stageName
	 * @param wordName
	 * @param exampleName
	 * @param forEdit
	 */
	void showLearningExample(String stageName, String wordName, String exampleName, int editAction)
	{
		vf.showHideButton.setEnabled(true);
		LearningStage stage = model().getLearningStage(stageName);
		WordToLearn wl = stage.getWordLearn(wordName);
		FSNode example = wl.getExample(exampleName);
		
		if (editAction == EditLanguage.SHOW) showTheFS(example,EditLanguage.LEARNINGEXAMPLE,false);

		else if (editAction == EditLanguage.EDIT) showTheFS(example,EditLanguage.LEARNINGEXAMPLE,true);
		
		else if (editAction == EditLanguage.REMOVE)
		{
			boolean doIt = Util.getYesNo("Remove learning example " + example.getName() + "?" , vf);
			if (doIt)
			{
				wl.getExamples().remove(example);				
				vf.editLanguage.refreshEditLanguageMenus(null, null,false);
			}
		}
	}
	

	
	/**
	 * remove a word to learn from a stage, with all its learning examples
	 * @param stageName
	 * @param wlName
	 */
	public void removeWordToLearn(String stageName,String wlName)
	{
		LearningStage stage = model().getLearningStage(stageName);
		WordToLearn toRemove  = stage.getWordLearn(wlName);

		boolean doIt = Util.getYesNo("Remove word to learn '" + wlName + "' with its learning examples?" , vf);
		if (doIt)
		{
			stage.getLearnWords().remove(toRemove);				
			vf.editLanguage.refreshEditLanguageMenus(null, null,false);
		}		
	}

	  //------------------------------------------------------------------------------------------------
	  //                            User Guide Actions
	  //------------------------------------------------------------------------------------------------

	
	/**
	 *  show a user guide topic on startup
	 * @throws Exception
	 */
	public void showInitialTopic() throws Exception
	{
		if (model() == null) throw new Exception("No model read");
		if (model().getStartTopicId() != null) showTopic(model().getStartTopicId(),true);
		else Util.message("No initial user guide topic");
	}


	  //------------------------------------------------------------------------------------------------
	  //                      Compute the unify order for a sentence
	  //------------------------------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @param sentence
	 * @return single string unify order for understanding
	 * @throws Exception
	 */
	public String makeUnifyOrderString(String sentence) throws Exception
	{
		String result = "";
		Vector<String> words = makeUnifyOrder(sentence);
		for (String word: words) result = result + word + " ";		
		return result;
	}

	/**
	 * 
	 * @param sentence
	 * @return
	 * @throws Exception
	 */
	public Vector<String> makeUnifyOrder(String sentence) throws Exception
	{
		Vector<Vector<FSNode>> allSenses = new Vector<Vector<FSNode>>();

		// find all the different senses of each word
		StringTokenizer st = new StringTokenizer(sentence," '");
		while (st.hasMoreTokens())
		{
			Vector<FSNode> senses = new Vector<FSNode>();
			String wordStem = st.nextToken();
			
			// the word is already confined to one sense in brackets
			if (wordStem.contains("(")) senses.add(model().getWords().get(wordStem));
			// there may be several different senses of this word stem
			else senses = model().getSenses(wordStem);
			
			allSenses.add(senses);
		}
		
		// find all combinations of senses of the words
		Vector<Vector<FSNode>> combinations = Util.cartesianProduct(allSenses);
		
		// try out each combination of word senses to find the most complete unify order (lowest count of child nodes in the end)
		int bestCount = 100;
		int bestGaps = 100;
		Vector<String> bestOrder = null;
		
		// for each combination of word senses
		for (Vector<FSNode> comb: combinations)
		{
			// build up the full sentence of words with senses in brackets, e.g her(poss)
			String fullSentence = "";
			for (FSNode sense: comb)
			{
				String fullName = sense.getName();
				fullSentence = fullSentence + fullName + " ";
			}
			
			// find a unify order for this combination of word senses - with the smallest possible number of gaps
			Vector<Vector<String>> result = makeUnifyOrderForSenses(fullSentence);
			Vector<String> order = result.get(0);
			Vector<String> gaps = result.get(1);
			int gapCount = countGaps(gaps); // number of words with any gap
			Vector<String> afterUnify = result.get(2); // result scenes after unify
			int count = afterUnify.size(); // should be 1 if the unify order is successful
			
			// choose the smallest count of scenes after unification
			if (count < bestCount)
			{
				bestCount = count;
				bestGaps = gapCount;
				bestOrder = order;
			}
			
			// if the number of scenes is tied, choose the smallest number of words with gaps
			if ((count == bestCount) && (gapCount < bestGaps))
			{
				bestCount = count;
				bestGaps = gapCount;
				bestOrder = order;
			}
		}
		
		return bestOrder;		
	}
	
	/**
	 * 
	 * @param sentence
	 * @return a Vector of three Vectors of strings:
	 *  - a Vector of full word senses, in unify order for understanding
	 *  - a Vector of number of gaps per word, 'zero' or 'one', in the unify order
	 *  - a Vector of the resulting nodes [a] or [t]  - should have only one element for a successful parse
	 * @throws Exception
	 */
	public Vector<Vector<String>> makeUnifyOrderForSenses(String sentence) throws Exception
	{
		Vector<String> order = new Vector<String>(); // will the first part of the result
		Vector<String> gaps = new Vector<String>(); // will be the second part of the result
		Vector<String> allWords = new Vector<String>(); // vector of word or <scene>, to refine; will be the third part of the result
		Vector<String> current = new Vector<String>();
		
		Vector<FSNode> nouns = new Vector<FSNode>(); // nouns in the sentence (can be unified standalone first)
		Vector<FSNode>  nonNouns = new Vector<FSNode>(); // other words in the sentence

		// this call is just to fill the two words lists, so as to then prepare the gaps tables
		prepareParse(sentence, allWords, nouns, nonNouns);
		
		Vector<Vector<FSNode>> perms = Util.permutations(nonNouns);
		
		// prepare all gap tables
		Vector<Hashtable<String,String>> gapTables = makeGapTables(nonNouns);
		
		// iterate over gap tables, in order of increasing number of gaps (the table with no gaps is first), until you find a successful parse
		boolean stopLooking = false;
		for (Hashtable<String,String> gapTable: gapTables) if (!stopLooking)
		{
			// tracing for this set of word senses, and this set of gaps
			traceGaps(sentence, gapTable);
			
			// re-initialise the result unify order for this set of gaps
			current = (Vector<String>)allWords.clone();
			order = new Vector<String>();
			
			// unify all nouns first - the order does not matter, and success is guaranteed
			for (FSNode noun: nouns)
			{
				String[] pattern = pattern(noun); //trivial pattern - no meaning scenes to match
				String name = noun.getName();
				/// change current and extend order
				current = applyPattern(name,pattern,current);
				order.add(name);
			}
			
			// try out all orders of unifying non-nouns with possible gaps, until you get an order that completes unification for all words
			boolean success = false;
			for (Vector<FSNode> perm: perms) if (!success)
			{
				tracePerm(perm);
				
				Vector<String> tryCurrent = (Vector<String>)current.clone();
				Vector<String> tryOrder = (Vector<String>)order.clone();
				
				// try all non-nouns in one order, cutting failures short
				boolean stillGoing = true;
				for (FSNode w:perm) if (stillGoing)
				{
					String word  = w.getName();
					String gap = gapTable.get(word);
					boolean gappable = (gap.equals("one"));
					Vector<String[]> patterns = patterns(w,gappable);
					
					// try all gap or non-gap patterns for this word to see if any work
					boolean pWorks = false;
					for (String[] pattern: patterns) if (!pWorks)
					{
						Vector<String> gCurrent = (Vector<String>)tryCurrent.clone();
						gCurrent = applyPattern(word,pattern,gCurrent);
						
						// this pattern worked - matched tryCurrent
						if (gCurrent.size() > 0)
						{
							pWorks = true;
							tryCurrent = gCurrent;
							tryOrder.add(word);
						}						
					}
					
					// still going if any pattern of this word worked - try the next word
					stillGoing = pWorks;
					
				} // end of iteration over non-nouns in a permutation
				
				// managed to find a unify position for all non-nouns
				if (stillGoing)
				{
					current = tryCurrent;
					order = tryOrder;
					gaps = makeGapVector(order, gapTable);
					
					success = (current.size() == 1); // so you do not try any more permutations
					stopLooking = (current.size() == 1); // so you do not try any more gap configurations
				}
				
			} // end of iteration over permutations of non-nouns			
			
		} //end of iteration over gap tables
							
		// return a Vector of three Vectors of strings
		Vector<Vector<String>> result = new Vector<Vector<String>>();
		result.add(order); // word senses in unify order
		result.add(gaps); // words with gaps in this unify order
		result.add(current); // the shorter the better
		
		return result;
	}
	
	/**
	 * apply the unify pattern of  a word to the current unify state, and return the next unify pattern.
	 * return an empty vector if failed.
	 * 
	 * @param word
	 * @param pattern
	 * @param current vector of <t>, <a>, or word
	 * 
	 * @return the next such vector after this word has been unified, or an empty vector if it cannot unify
	 */
	private Vector<String> applyPattern(String word, String[] pattern, Vector<String> current)
	{
		Vector<String> res = new Vector<String>();
		boolean failed = true;
		
		// find the first position of the word in the current unify state
		int pos = -1;
		for (int i = 0; i < current.size(); i++) if (pos == -1)
		{
			String w = current.get(i);
			if (w.equals(word)) pos = i;
		}
		
		// if the word can be found
		if (pos > -1)
		{
			int startPos = pos - pattern[0].length(); // first element of current which must match the word pattern
			int endPos = pos + pattern[1].length(); // last element of current which must match the word pattern
			
			// if the word pattern does not overshoot either end of the current pattern
			if ((startPos > -1) && (endPos < current.size()))
			{
				failed = false; 
				
				// add parts of current before the start of the word pattern
				for (int c = 0; c < startPos; c++) res.add(current.get(c));
				
				// check parts of the word pattern before the word
				for (int c = startPos; c < pos; c++)
				{
					int pb = c - startPos; // 0 up to pattern[0].length()
					String toMatch = "<" + pattern[0].substring(pb,pb+1) + ">";
					if (!toMatch.equals(current.get(c))) failed = true;
				}
				
				// add the word result pattern
				res.add("<" + pattern[2] + ">");
				
				// check parts of the word pattern after the word
				for (int c = pos + 1; c < endPos + 1; c++)
				{
					int pb = c - pos - 1; // 0 up to pattern[1].length()
					String toMatch = "<" + pattern[1].substring(pb,pb+1) + ">";
					if (!toMatch.equals(current.get(c))) failed = true;
				}
				
				// add parts of current after the end of the word
				for (int c = endPos + 1; c < current.size(); c++) res.add(current.get(c));
			}
			
			if (failed) res = new Vector<String>();
		
			else traceApply(word,pattern,res);
	
		}
		
		return res;
	}

	
	/**
	 * 
	 * @param sentence
	 * @param allWords
	 * @param nouns
	 * @param nonNouns
	 * 
	 * effects:
	 *  - build allWords as a vector of word names
	 *  - build the vectors of words of nouns and nonNouns (even if they have been built already)
	 *  
	 * @throws Exception
	 */
	private void prepareParse(String sentence, Vector<String> allWords, Vector<FSNode> nouns, Vector<FSNode> nonNouns) throws Exception
	{
		// check that all the words exist, and make the start state allWords of the iteration
		StringTokenizer st = new StringTokenizer(sentence," '");
		while (st.hasMoreTokens())
		{
			String wordName = st.nextToken();
			FSNode word = model().getWords().get(wordName);
			if (word == null) throw new Exception("Cannot find word " + wordName);
			
			if (isNoun(word)) nouns.add(word);
			else nonNouns.add(word);
			
			allWords.add(wordName);
		}
	}
	
	/**
	 * 
	 * @param word
	 * @return true if the word is a noun (has no input meaning scenes)
	 */
	private boolean isNoun(FSNode word)
	{
		int scenes = 0;
		for (FSNode child:word.getChildren())
			if (child.getType().equals("scene")) scenes++;
		return (scenes == 1);
	}
	
	
	/**
	 * A gap table is a <Hashtable<String,String> with key = word name, value = 'zero' or 'one'.
	 * 'one' means the word is allowed one gap in its parse.
	 * @param words
	 * @return a vector of gap tables, in ascending order of number of gaps.
	 * 
	 * The first single element has no gaps.
	 * Then there are elements with one gap on some word.
	 * Then there are elements with a gap on two words.
	 * To date, the list goes no further. Three-gap parses are ignored.
	 */
	Vector<Hashtable<String,String>> makeGapTables(Vector<FSNode> words)
	{
		Vector<Hashtable<String,String>> allTables = new Vector<Hashtable<String,String>>();
		
		// start the first entry, for no gaps
		Hashtable<String,String> noGapTable = new Hashtable<String,String>();
		
		// check all words can have gaps(i.e are not nouns), and make the first entry for no gaps
		Vector<String> gappableWords = new Vector<String>();
		for (FSNode wn: words)
		{
			String word = wn.getName();
			noGapTable.put(word, "zero");
			int meanings = 0;
			for (FSNode child: wn.getChildren())
				if (child.getType().equals("scene")) meanings++;
			// if there are one or more input meaning scenes which can be gapped
			if (meanings > 1) gappableWords.add(word);
		}
		allTables.add(noGapTable);
		
		// make the entries with one gap
		for (String word: gappableWords)
		{
			Hashtable<String,String> oneGapTable = (Hashtable<String,String>)noGapTable.clone();
			oneGapTable.put(word,"one");
			allTables.add(oneGapTable);
		}
				
		// make the entries with two gaps (duplicated??  it does not matter)
		for (int i = 1; i < gappableWords.size() + 1; i++)
		{
			Hashtable<String,String> oneGapTable = allTables.get(i);
			for (String word: gappableWords) if (oneGapTable.get(word).equals("zero"))
			{
				Hashtable<String,String> twoGapTable = (Hashtable<String,String>)oneGapTable.clone();				
				twoGapTable.put(word,"one");
				allTables.add(twoGapTable);
			}			
		}
				
		return allTables;
	}

	/**
	 * trace a permutation of non-nouns
	 * @param perm
	 */
	private void tracePerm(Vector<FSNode> perm)
	{
		if (tracing)
		{
			String s = "Trying non-noun word order: ";
			for (FSNode word: perm) s = s + word.getName() + " ";
			Util.message(s);
		}
	}


	/**
	 * trace the application of a word in a unify order
	 * @param w
	 * @param pat
	 * @param current
	 */
	private void traceApply(String w,  String[] pat, Vector<String> current)
	{
		if (tracing)
		{
			String s = "Add word " + w + " [" + pat[0] + "," + pat[1] + "," + pat[2] + "]; unify state: ";
			for (String part: current) s = s + part + " ";
			Util.message(s);
		}
	}

	
	/**
	 * trace of this attempt to find a unify order,
	 * showing the gaps allowed
	 * @param sentence
	 * @param gapTable
	 */
	private void traceGaps(String sentence, Hashtable<String,String> gapTable)
	{
		if (tracing)
		{
			String ss = "Unify order for '" + sentence + "'; gaps [";
			for (Enumeration<String> en = gapTable.keys(); en.hasMoreElements();)
			{
				String word = en.nextElement();
				String gapped = gapTable.get(word);
				if (gapped.equals("one")) ss = ss + word + " ";
			}
			Util.message("\n" + ss + "]");
		}
	}
	

	
	/**
	 * make a vector of gaps, from a gap table
	 * @param sentence
	 * @param gapTable
	 * @return
	 */
	private Vector<String> makeGapVector(Vector<String> order, Hashtable<String,String> gapTable)
	{
		Vector<String> gaps = new Vector<String>();		
		for (String word:order) 
		{
			String gapValue = gapTable.get(word);
			if (gapValue != null) gaps.add(gapValue);
		}
		return gaps;
	}

	/**
	 * count the number of gaps in a gap vector
	 * @param gaps
	 * @return
	 */
	private int countGaps(Vector<String> gaps)
	{
		int gc = 0 ;
		for (String gap: gaps) if (!(gap.equals("zero"))) gc++;
		return gc;
	}
	

	/**
	 * 
	 * @param word
	 * @return an array of :
	 * - the types (t or a) of meaning scenes needed to parse a word before it
	 * - the types (t or a) of meaning scenes needed after it
	 * - the type of its result scene (t or a)
	 * 
	 * an example is [t,tt,a] for 'give'
	 */
	private String[] pattern(FSNode word)
	{
		String[] req = new String[3];
		for (int i = 0; i < 3; i++) req[i] = "";
		boolean soundStarted = false;
		
		int cn = word.getChildren().size();
		for (int i = 0; i < cn; i++)
		{
			FSNode child = word.getChildren().get(i);
			String type = child.getType();
			if (type.equals("sound")) soundStarted = true;
			else if (type.equals("scene"))
			{
				String act = child.getSlotValue("act");
				boolean isAct = ((act != null) && (act.equals("true")));
				String sceneType = "t";
				if (isAct) sceneType = "a";
				if (i < cn - 1) // not the last result scene
				{
					if (soundStarted) req[1] = req[1] + sceneType;
					else req[0] = req[0] + sceneType;
				}
				// the last result scene
				else req[2] = sceneType;
			}
			
		}						
		return req;
	}
	
	/**
	 * 
	 * @param word
	 * @param gapped
	 * if (!gapped) return just the ungapped pattern for the word
	 * if (gapped) return all the gapped patterns for the word 
	 * - each pattern missing out one <t> scene before or after the word
	 */
	private Vector<String[]> patterns(FSNode word, boolean gapped)
	{
		String wName = word.getName();
		Vector<String[]> pats = new Vector<String[]>();
		String[] ungapped = pattern(word);
		
		if (!gapped) pats.add(ungapped);
		else
		{
			// iterate over 'before' and 'after' parts of the ungapped word pattern
			for (int i = 0; i < 2; i++)
			{
				// if there are any 't' to be gapped in the before or after pattern
				if (ungapped[i].length() > 0)
				{
					// the number of gapped patterns is the number of 't' in basic[i]
					for (int p = 0; p < ungapped[i].length();p++)
					{
						String letter  = ungapped[i].substring(p,p+1);
						// gaps can only have one 't' dropped - not 'a'
						if (letter.equals("t"))
						{
							// make a copy of the ungapped pattern and empty its element 0 or 1 
							String[] gPat = new String[3];
							for (int s = 0; s < 3; s++) gPat[s] = ungapped[s];
							gPat[i] = "";

							// add all parts of the ungapped pattern except one 't' at position p
							for (int q = 0; q < ungapped[i].length();q++) if (q != p) 
							{
								String letterq  = ungapped[i].substring(q,q+1);
								gPat[i] = gPat[i] + letterq;
							}

							// add the gapped pattern to the list
							pats.add(gPat);
						}
					}
				}
			}
		}
		
		return pats;
	}



}
