package generatorPack;

import java.io.InputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import displayPack.Actions;
import displayPack.EditLanguage;
import learnerPack.Learner;
import learnerPack.LearningExample;
import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.LearningStage;
import modelPack.NodeType;
import modelPack.Slot;
import modelPack.SlotValue;
import modelPack.WordToLearn;
import utilPack.GenerateLearntDataClass;
import utilPack.LearntDataClass;
import utilPack.Util;

/**
 * generates random sentences according to a production rule model
 * 
 * each category of the model has:
 * (a) a result type 
 * (eg U = utterance, S = sentence, NP noun phrase, TV = transitive verb, IV, PP, N, A = adjective, PN = proper name, D = determiner)
 * (b) a set of production rules from other categories (eg NP ->PN, NP -> D N, NP -> D A N, NP -> NP PP) with probabilities that add up to 1
 * (c) terminal rules have a vocabulary (e.g P -> in,to, from) of N words, each with probability 1/N; and PN ->fred,anna,...
 * (d) U has U -> S and U ->  NP
 * (e) the top production is S has two rules: S -> NP IV and S -> NP TV NP
 * 
 * Start from U and apply categories with their probability, until you get to terminal categories and vocabulary,
 * assembling the utterance by serial concatenation.
 * 
 * Some sentences will not parse because their words have restrictions, e.g. 'runs' requires an animate subject, 'eats' requires a food object.
 * These will just get rejected because they do not unify.
 * 
 */

public class ExGenerator {
	
	private LanguageModel model;
	public LanguageModel getModel() {return model;}
	
	public Learner learner;
		
	private Hashtable<String,Category> categories = new Hashtable<String,Category>();
	public Category getCategory(String catName) {return categories.get(catName);}
	
	// the number of learning examples generated only for pre-learning of possible word sounds - then forgotten
	public int preLearn;
	
	// the number of learning examples to be generated in one cycle
	public int exPerCycle;
	
	// number of cycles per press of the 'Auto-Learn' button
	public int cyclesPerPress; 
	
	// number of cycles passed
	protected int cycleNo = 0;
	public int getCycleNo() {return cycleNo;}
	
	// the number of learning examples to be retained in memory for generalisation
	protected int memory;
	public int getMemory() {return memory;}
	
	// constant for exponential decay of weighted contribution of a learning example to support; set in data
	public int expLife = 0;
	
	// threshold for sum of weighted successful learning examples, for positive support; set in data
	public int threshold = 0;
	
	// all parsable learning examples in order; sentence-meaning pairs. Inner Vector = partly understood versions, with the same meaning
	private Vector<LearningExample> orderedExamples = new Vector<LearningExample>();
	public LearningExample getExample(int exNo) {return orderedExamples.get(exNo);}
	public int numberOfExamples() {return orderedExamples.size();}
	
	// learning examples (including unparsable ones) keyed by sentence to avoid duplicates
	private Hashtable<String,String> keyedExamples = new Hashtable<String,String>();
	
	// spaced real words used in generating the examples - shown as g-ur-l etc.
	public Vector<String> spacedRealWords = new Vector<String>();
	
	// mapping from spaced names to word names
	protected Hashtable<String,String> shortToSpaced = new Hashtable<String,String>();
	
	// table of feature structures for real words, keyed by spaced names b-o-y etc.
	public Hashtable<String,FSNode> realWordFS = new Hashtable<String,FSNode>();
	
	// number of bits of information for each slot:value pair, calculated from its rarity in generated meanings
	protected Hashtable<String,Double> infoCount = new Hashtable<String,Double>();
	public Hashtable<String,Double> getInfoCount() {return infoCount;}
	
	// cumulative numbers of uses of each word in generated examples
	protected Hashtable<String,Integer> wordUses = new Hashtable<String,Integer>();
	
	// number of meanings to use to compute slot rarities
	int meanings = 2000;
	
	// the maximum number of failed tries in a cycle to make new sentences
	int maxRepeats = 10000;
	
	// number of correct + distractor meanings (1 correct)  per learning example
	public int nDistractors;
	
	// a pool of random meanings to use as distractors in learning examples
	private Vector<FSNode> distractorMeanings = new Vector<FSNode>();
	private FSNode randomMeaning()
	{
		int r = (int)((distractorMeanings.size() - 1)*Math.random());
		return distractorMeanings.get(r);
	}
	
	// bits of meaning information required to start a candidate word - with and without guessed learning examples
	private int[] startCriteria = new int[2];
	public int[] getStartCriteria() {return startCriteria;}
	
	// mean number of extra slots added by the learner, to the meaning of a learning example
	private int extraSlots;
	public int getExtraSlots() {return extraSlots;}
	
	// log of all learning examples, to be written to a text file
	public Vector<String> exampleLog = new Vector<String>();

	// number of examples to be permuted for permutation learning (from data file)
	public int permutable;
	
	// number of reorderings of the examples to be tried (from data file)
	public int  permutations;
	
	
	// data created by each learning cycle, to be stored in a generated data class
	private Document learntDoc;
	private Document explainDoc;
	private Document menuDoc;

	private Element learntXML;
	private Element explainXML;
	private Element menuXML;


	
//--------------------------------------------------------------------------
//                        Constructor
//--------------------------------------------------------------------------

	
	/**
	 * 
	 * @param model
	 * @param forWeb if true, there is no generator file, so cannot start for automatic learning
	 * (the class can only be used to return the results of a previous learning session to the web client)
	 * @throws Exception
	 */
	public ExGenerator(LanguageModel model, boolean forWeb) throws Exception
	{
		this.model = model;
		
		if (!forWeb) startForLearning();
		
		// make the learner (required to hold words and learning stages from a previous learning run)
		else learner = new Learner(model, this);

	}
	
	/**
	 * read the generator file and start the generator and learner for a 
	 * run of automatic learning
	 * @throws Exception
	 */
	private void startForLearning() throws Exception
	{
		Element genRoot = model.readGenerator();
		
		preLearn = Integer.parseInt(genRoot.getAttribute("preLearn"));
		exPerCycle = Integer.parseInt(genRoot.getAttribute("exPerCycle"));
		cyclesPerPress = Integer.parseInt(genRoot.getAttribute("cyclesPerPress"));
		memory = Integer.parseInt(genRoot.getAttribute("memory"));
		expLife = Integer.parseInt(genRoot.getAttribute("expLife"));
		threshold = Integer.parseInt(genRoot.getAttribute("threshold"));
		nDistractors = Integer.parseInt(genRoot.getAttribute("distractors"));
		extraSlots = Integer.parseInt(genRoot.getAttribute("extraSlots"));
		permutable = Integer.parseInt(genRoot.getAttribute("permutable"));
		permutations = Integer.parseInt(genRoot.getAttribute("permutations"));

		// define the criteria for starting a new candidate word
		String startString = genRoot.getAttribute("startCriteria");
		StringTokenizer st = new StringTokenizer(startString,"[],");
		int pos = 0;
		while (st.hasMoreTokens())
		{
			String crit = st.nextToken();
			startCriteria[pos] = Integer.parseInt(crit);
			pos++;
		}
		
		for (Element catRoot: Util.namedChildElements(genRoot, "cat"))
		{
			Category cat = new Category(catRoot,this);
			categories.put(cat.getName(), cat);
		}
		
		Util.message(categories.size() + " learning example generator categories read.");
		
		checkCategories();
		
		// start use counts
		for (String word: spacedRealWords)
			wordUses.put(word, Integer.valueOf(0));
		
		//calculate rarities of slot values in generated meanings
		Hashtable<String,Integer> slotCount = makeSlotCount(meanings);
		double mCount = (double)meanings;
		for (Enumeration<String> en = slotCount.keys(); en.hasMoreElements();)
		{
			String svp = en.nextElement();
			double count = slotCount.get(svp).doubleValue();
			double bits = Math.log(mCount/count)/Math.log(2.0);
			// it is just possible for a slot value to occur more times than the number of meanings (number = @one)
			if (bits < 0.0) bits = 0.0;
			infoCount.put(svp, Double.valueOf(bits));
		}
		
		// make the learner
		learner = new Learner(model, this);
		
		// initialise the XML record of learning
		startLearntXML();
		
	}
	
	
	//--------------------------------------------------------------------------
	//	                    Checking Categories and recording real words
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 */
	private void checkCategories() throws Exception
	{
		for (Enumeration<Category> en = categories.elements();en.hasMoreElements();)
		{
			Category cat = en.nextElement();
			Vector<String> catWords = cat.check();
			for (String cw:catWords) spacedRealWords.add(cw);
		}
		Util.message("Checked all categories in the example generator; uses " + spacedRealWords.size() + " words");
	}
	
	/**
	 * 
	 * @param meanings number of valid meanings to be generated
	 * @return a table with key = slot:value, element = number of occurrences of the slot value
	 */
	private Hashtable<String,Integer> makeSlotCount(int meanings) throws Exception
	{
		Hashtable<String,Integer> slotCount = new Hashtable<String,Integer>();
		int nm = 0;
		Category topCat = getCategory("U");		
		
		while (nm < meanings)
		{
			ExamplePart utter = new ExamplePart(this, topCat);
			String sentence = utter.getWords();
			FSNode result = parse(sentence,true);
			
			// if the sentence can be parsed to a single meaning scene
			if (result != null)
			{
				nm++;
				addMeaningSlots(slotCount,result.getLastChild());
				
				//collect distractor meanings to put in learning examples
				distractorMeanings.add(result);
			}
			
		}
		
		return slotCount;
	}
	
	/**
	 * accumulate the count of occurrences for each slot:value pair, in FSNode f and its descendants
	 * @param slotCount
	 * @param f
	 */
	private void addMeaningSlots(Hashtable<String,Integer> slotCount, FSNode f)
	{
		for (Enumeration<String> en = f.getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			Slot s = model.getSlot(slotName);
			String slotValue = f.getSlotValue(slotName);
			SlotValue sv = s.getSlotValueWith(slotValue);
			
			// ascend the hierarchy of slot values, adding to the count for each value encountered
			while (sv != null)
			{
				String svp = slotName + ":" + sv.getValue();

				Integer count = slotCount.get(svp);
				if (count == null) count = Integer.decode("1");
				else count++;
				slotCount.put(svp, count);
				
				sv = sv.getParent();				
			}
			
		}
		
		// descend the FSNode tree
		for (FSNode child:f.getChildren()) addMeaningSlots(slotCount,child);		
	}
	
	
	//--------------------------------------------------------------------------
	//	                          Framework
	//--------------------------------------------------------------------------
	
	/**
	 * do one cycle of automatic learning.
	 * The first cycle just writes out the words used to generate learning examples.
	 * @throws Exception
	 */
	public void oneCycle() throws Exception
	{		
		if (cycleNo == 0)
		{
			// start recording XML data for the start cycle
			Element learntCycle = newCycle(0);
			Element explainCycle = newCycle(1);
			Element menuCycle = newCycle(2);
			
			// pre-learn the allowed word sounds
			boolean forPreLearn = true;
			makeExamples(forPreLearn);
			// argument is standard deviations for pass 1 of pre-learn
			learner.findAllWordSounds(4.0);			
			clearExamples();
			
			// write explanation of words used to generate learning data
			writeStartCycle(explainCycle);
			
			// add the cycle data to the data for all cycles
			addCycleData(learntCycle, explainCycle, menuCycle);
			
			cycleNo++;

		}
		
		else if (cycleNo > 0) 
		{
			int[] choice = doLearningCycle();
			Util.message("\nExamples with some known words: " + choice[0] + "; bad choices: " + choice[1] + "; all words known " + choice[2]);
			
			// write the log of examples
			Util.writeTextFile(model.exampleLogPath, exampleLog);
		}
		

	}
	
	/**
	 * save the data class generated by automatic learning
	 * @throws Exception
	 */
	public void saveLearntClass() throws Exception
	{
		if (cycleNo > 2)
		{
			Util.message("Started saving class of auto-learning data for cycle " + (cycleNo - 1));
			
			writeLearntDataClass();

			Util.message("Finished saving class of auto-learning data");		
		}
		
		else Util.message("No auto-learning data to save yet.");
	}

	
	/**
	 * from the XML created by all automatic learning stages so far,
	 * generate and save the learnt data class
	 */
	private void writeLearntDataClass() throws Exception
	{
		// convert the XML to strings that can be data in a Java class
		String learntXMLString = LanguageModel.dataFromXMLDocument(learntXML.getOwnerDocument());
		String explainXMLString = LanguageModel.dataFromXMLDocument(explainXML.getOwnerDocument());
		String menuXMLString = LanguageModel.dataFromXMLDocument(menuXML.getOwnerDocument());
		
		// make the generator class, and get the lines of the generated class from it
		GenerateLearntDataClass dataGenerator = new GenerateLearntDataClass(learntXMLString,explainXMLString,menuXMLString);
		Vector<String> lines = dataGenerator.generateDataLines();
		
		// save LearntDataClass.java
		Util.writeTextFile(model.learntDataClassFilePath, lines);
		
	}
	
	
	/**
	 * make one cycle's worth of learning examples and learn from them
	 * return int[3]:
	 *  number of examples with one or more known words, and so a margin to choose the best meaning, 
	 *  number of bad choices of the example meaning (chose a distractor meaning, when there are words)
	 *  number of examples with all words known
	 */
	private int[] doLearningCycle() throws Exception
	{
		// start recording XML data for the cycle
		Element learntCycle = newCycle(0);
		Element explainCycle = newCycle(1);
		Element menuCycle = newCycle(2);
		
		// start recording example meaning choices and mistakes for the cycle, and learning examples in which all words are known
		int[] choice = new int[3];
		for (int i = 0; i < 3; i++) choice[1] = 0;
		
		// start of new learning examples
		int min = orderedExamples.size();
		
		// make a cycle worth of examples
		boolean forPreLearn = false;
		makeExamples(forPreLearn);
		
		// write out cumulative numbers of word uses in examples so far
		// writeWordUses();
		
		// clear out old learning examples
		int clear = min - memory;
		if (clear < 0) clear = 0;
		for (int c = clear; c < min; c++)
			replaceExample(orderedExamples.get(c));
						
		// pass all examples in this cycle to the learner
		for (int i = min; i < orderedExamples.size();i++)
		{
			LearningExample lex = orderedExamples.get(i);
			// use this example to learn more words, and record it in XML
			boolean[] res = learner.useOneExample(lex,explainCycle);
			
			// if a choice of meaning was made based on words in the example
			if (res[1]) choice[0]++; // there was a margin between the best choice and the next
			if (res[0] && res[1]) choice[1]++; // a bad choice of meaning was made, even with a margin
			if (res[2]) choice[2]++; // all words in the learning example are known
		}
		
		// create a new Learning menu for the application, with the words learnt
		Vector<LearningStage> stages = learner.refreshLearningMenu(cycleNo,explainCycle);
		
		// create the XML for the new learning stages and their menus, to send to the web client
		createLearntWordXML(stages,learntCycle, menuCycle);
		
		// revise words that are not well supported
		if (cycleNo >= 3) learner.permutationImprove(cycleNo);
		
		// write an update on the cycle
		learner.writeCycleUpdate();
		
		// add the cycle data to the data for all cycles
		addCycleData(learntCycle, explainCycle, menuCycle);
		
		// make the transition to using only learning examples validated by words, if appropriate
		if ((!learner.useOnlyGoodLearningExamples) && (choice[0] > learner.useGoodExampleThreshold))
		{
			learner.preferGoodExamples(true); // set the learner to only use validated learning examples
			Util.message("\nNow using only validated learning examples, from example " + numberOfExamples());
			// learner.restartCandidateWords(); // restart all learnt words except supported words
			// Util.message("Restarting learning of all candidate words");
		}
		
		// ready for the next cycle
		cycleNo++;
		
		return choice;
		
	}
	
	/**
	 * replace a learning example by an empty learning example
	 * @param lex
	 */
	private void replaceExample(LearningExample lex)
	{
		lex = new LearningExample(this);
	}
	
	/**
	 * clear all examples after the pre-learning phase
	 */
	private void clearExamples()
	{
		keyedExamples = new Hashtable<String,String>();
		orderedExamples = new Vector<LearningExample>();
	}
	
	/**
	 * make new learning examples
	 * @param preLearn; if true, make preLearn  = 1000 examples for pre-learning of word sounds
	 * if false, make perCycle examples for one cycle of learning
	 */
	private void makeExamples(boolean forPreLearn) throws Exception
	{
		if (forPreLearn) Util.message("Making examples for pre-learn ");
		else Util.message("\nMaking examples for cycle " + cycleNo + "\n");
		Category topCat = getCategory("U");		
		int made = 0;

		int repeatSentences = 0;
		int exampleCount = exPerCycle;
		if (forPreLearn) exampleCount = preLearn;
		while (made < exampleCount)
		{
			ExamplePart utter = new ExamplePart(this, topCat);
			String sentence = utter.getWords();
			//first check that this sentence has not been made before (even if it is invalid)
			if (keyedExamples.get(sentence) != null) 
			{
				repeatSentences++;
				if (repeatSentences > maxRepeats)
					throw new Exception("More than " + maxRepeats 
							+ " repeats of a previous sentence; ran out of new sentences at " 
							+ keyedExamples.size() + ", " + orderedExamples.size() + " valid");
			}
			else if (keyedExamples.get(sentence) == null)
			{
				keyedExamples.put(sentence, "1");
				
				// the sentence must not have any words repeated, or be only one word
				if ((!repeatingWord(sentence)) && (!tooShort(sentence)))
				{
					FSNode result = parse(sentence,true);
					
					// if the sentence can be parsed to a single meaning scene
					if (result != null)
					{
						String name = "Ex" + orderedExamples.size() + ": " + sentence;
						// Util.message(name);
						result.setName(name);
						// record the meaning with no noise added
						FSNode noNoise = new FSNode(result);
						// add noise to the meaning
						addRandomMeaning(result);
						
						// make a set of distractor meanings for the learning example; the first is the true meaning and the rest are random meanings
						Vector<FSNode> distractors = new Vector<FSNode>();
						distractors.add(result);
						for (int d = 1; d < nDistractors; d++) distractors.add(makeDistractor(result));
						
						LearningExample lex = new LearningExample(this, distractors,orderedExamples.size(),noNoise);
						// for pre-learn only, each learning example must have a first unify result (with no unifications done)
						if (forPreLearn) lex.addUnifyResult(noNoise);
						orderedExamples.add(lex);
						
						// record the cumulative number of uses of each word
						recordWordUse(lex);
						
						made++;
					}
				}
			}			
		}
		
		if (forPreLearn) writeWordUses();
	}
	
	/**
	 * add random extra meaning slots to a feature structure, 
	 * which do not conflict with its existing slots.
	 * 
	 * These are meanings observed by the learner, 
	 * which are not part of the meaning the speaker expressed
	 * 
	 * @param result
	 * 
	 */
	private void addRandomMeaning(FSNode result)
	{
		// slots are be a Poisson-distributed random integer, with mean extraSlots defined in generator.xml		
		int slots  = Util.poisson((double)extraSlots);
		
		for (int s = 0; s < slots; s++)
		{
			// choose a random node in the meaning
			FSNode toAdd = randomNode(result.getLastChild());
			// slots where a value might be added to then node
			Vector<Slot> addable = addableSlots(toAdd);
			if (addable.size() > 0) // there might just be none of them..
			{
				int chosen = (int)(addable.size()*Math.random()); // rounds down to 0...addable.size() -1
				Slot a = addable.get(chosen); // choose a random slot from the list
				toAdd.setSlotValue(a.getName(), a.getRandomValue());
			}
		}
	}
	
	/**
	 * pick a random node from a meaning subtree
	 * @param meaning
	 * @return
	 */
	private FSNode randomNode(FSNode meaning)
	{
		// make a flat list of the nodes
		Vector<FSNode> flat = meaning.flatten();
		int chosen = (int)(flat.size()*Math.random()); // rounds down to 0...flat.size() -1
		return flat.get(chosen);
	}
	
	
	/**
	 * 
	 * @param toAdd
	 * @return a list of the slots valid for this node type,
	 * which do not already have a value
	 */
	private Vector<Slot> addableSlots(FSNode toAdd)
	{
		Vector<Slot> addable = new Vector<Slot>();
		for (Enumeration<Slot> en = model.getSlots().elements();en.hasMoreElements();)
		{
			Slot s = en.nextElement();
			for (NodeType t: s.getAllowedTypes())
				if ((t.getName().equals(toAdd.getType())) && (toAdd.getSlotValue(s.getName()) == null)) 
					addable.add(s);
		}
		return addable;
	}


	
	/**
	 * 
	 * @param result
	 * @return a distractor for the learning example -
	 * which has the same sounds but a random meaning
	 */
	private FSNode makeDistractor(FSNode result)
	{
		// pick a random meaning from those made earlier
		FSNode randFS = randomMeaning();
		
		// clone the correct result
		FSNode dis = new FSNode(result);
		
		// replace its meaning by the random meaning
		Vector<FSNode> newChildren = new Vector<FSNode>();
		for (FSNode child:result.getChildren())
		{
			if (child.getType().equals("sound")) newChildren.add(child);
			else newChildren.add(randFS.getLastChild());
		}
		dis.setChildren(newChildren);
		
		return dis;
	}
	
	
	/**
	 * record the words used in a learning example
	 * @param lex
	 */
	private void recordWordUse(LearningExample lex) throws Exception
	{
		// the words in the example, in short form
		String ex = lex.getName();
		
		StringTokenizer st = new StringTokenizer(ex," ");
		while (st.hasMoreTokens())
		{
			// convert short form to spaced form
			String shortName = st.nextToken();
			String spacedName = shortToSpaced.get(shortName);
			if (spacedName != null) // s
			{
				Integer uses = wordUses.get(spacedName);
				if (uses != null)
				{
					Integer newUses = Integer.valueOf(uses.intValue() + 1);
					wordUses.put(spacedName, newUses);
				}
			}
		}
	}
	
	/**
	 * write out the cumulative number of word uses in all examples made so far,
	 * grouped by category and leaf production
	 */
	public void writeWordUses()
	{
		Util.message("\nNumber of uses of words in " + orderedExamples.size() + " examples:");
		for (Enumeration<String> en = categories.keys(); en.hasMoreElements();)
		{
			String catName = en.nextElement();
			Category cat = categories.get(catName);
			String lines = catName;
			int wn = 0;
			for (Production prod: cat.productions) if (prod.isLeaf())
			{
				lines = lines + "\n   ";
				for (String word:prod.getWords())
				{
					String spacedName = shortToSpaced.get(word);
					if (spacedName != null)
					{
						Integer ii = wordUses.get(spacedName);
						if (ii != null)
						{
							wn++;
							lines = lines + word + ":" + ii.intValue() + " ";
						}
					}
				}
			}
			if (wn > 0) Util.message(wn + " words in category " + lines);
		}
		Util.message("");
	}
	
	/**
	 * on the zeroth learning cycle, write out the words used to generate learning examples
	 * grouped by productivity
	 * @param explainCycle
	 */
	private void writeStartCycle(Element explainCycle)
	{
		String header = "Words used to generate random learning examples: ";
		learner.explainLine(header, explainCycle);
		
		// different possible productivities
		String[] prodString = {"Productivity [0,0]: ","Productivity [0,1]: ","Productivity [1,0]: ","Productivity [1,1]: "};
		int[] wordNo = new int[4]; // numbers of words of each productivity
		for (int p = 0; p< 4; p++) wordNo[p] = 0;
				
		for (Enumeration<String> en = categories.keys(); en.hasMoreElements();)
		{
			String catName = en.nextElement();
			int p = getProd(catName); // -1 if the category has no words; otherwise 0,1,2,3 as above
			Category cat = categories.get(catName);
			
			if (p >= 0) for (Production prod: cat.productions) if (prod.isLeaf())
			{
				for (String word:prod.getWords())
				{
					prodString[p] = prodString[p] + word + ", ";
					wordNo[p]++;
				}
			}
		}
		
		// write the 4 lines of explanation, to the explain area and to XML for the web client
		for (int p = 0; p < prodString.length; p++)
		{
			String line = prodString[p] + "(" + wordNo[p] + " words)";
			learner.explainLine(line, explainCycle);
		}

	}
	
	/**
	 * 
	 * @param catName
	 * @return index 0..3 for the productivity of a category
	 * index to position in {"[0,0]","[0,1]","[1,0]","[1,1]"};
	 * or -1 if no terminal words in the category
	 */
	private int getProd(String catName)
	{
		// categories which contain words of different productivities:
		//                          [0,0]            [0,1]      [1,0]   [1,1]
		String [][] prodCats = {{"PN","IN","PR"},{"POSS","ADJ"},{"IV"},{"TV"}};

		int prod = -1;
		
		for (int p = 0; p < 4; p++)
		{
			String[] pCat = prodCats[p];
			for (int c = 0; c < pCat.length; c++)
				if (pCat[c].equals(catName)) prod = p;
		}
		
		return prod;
	}


	
	//--------------------------------------------------------------------------
	//	                        Making Learning Examples
	//--------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @param sentence
	 * @return
	 */
	private static boolean repeatingWord(String sentence)
	{
		boolean repeat = false;
		Hashtable<String,String> words = new Hashtable<String,String>();
		
		StringTokenizer st = new StringTokenizer(sentence," ");
		while (st.hasMoreTokens())
		{
			String word = st.nextToken();
			if (words.get(word) != null) repeat = true;
			words.put(word, word);
		}
		
		return repeat;
	}
	
	/**
	 * sentences or fragments of one word are not allowed
	 * @param sentence
	 * @return
	 */
	private static boolean tooShort(String sentence)
	{
		StringTokenizer st = new StringTokenizer(sentence," ");
		return  (st.countTokens() == 1);
	}
	
	/**
	 * 
	 * @param sentence
	 * @return
	 */
	private FSNode parse(String sentence,boolean repeat)
	{
		FSNode result = null;
		Actions act = model.getFrame().getAct();
		try
		{
			FSNode newFS = new FSNode(model,"script");
			// add the sounds of the words to the sentence
			EditLanguage.addWordSounds(newFS,sentence,model);

			// set the unify order
			String unifyOrder = act.makeUnifyOrderString(sentence);			
			newFS.setUnifyOrder(unifyOrder);

			// try to unify in the unify order
			Vector<Vector<FSNode>> understandResult = act.generateOrUnderstandAllSteps(newFS, false);
			
			// if all steps of unification succeeded
			Vector<FSNode> lastRow  = understandResult.get(understandResult.size() -1);
			if (lastRow.size() == 3)
			{
				FSNode lastResult = lastRow.get(2);
				if ((lastResult.isViable()) && (understandResult.size() == newFS.getUnifyWords().size()))
				{
					// set up the sound scenes and the unify order in a result
					result = new FSNode(newFS);

					// copy across the meaning scene of the result of unification
					FSNode meaningScene = new FSNode(lastResult.getLastChild());
					result.addChild(meaningScene);
				}
			}

		}
		catch (Exception ex) 
		{
			Util.message("Error in parsing " + sentence + ": " + ex.getMessage());
			if (repeat) parse(sentence,false);
		}
		
		return result;
	}
	
	/**
	 * 
	 * @param sounds
	 * @return all learning examples containing that sound
	 */
	public Vector<LearningExample> exVector(Vector<String> sounds)
	{
		Vector<LearningExample> exes = new Vector<LearningExample>();		
		
		for (int g = 0; g < numberOfExamples(); g++)
		{
			LearningExample lex = getExample(g);
			Vector<String> exs = lex.getExampleSounds();
			if (Util.includes(exs, sounds)) exes.add(lex);
		}
				
		return exes;
	}


	
	//---------------------------------------------------------------------------------------
	//	         Storing results of each learning cycle in a generated data class
	//---------------------------------------------------------------------------------------
	
	/**
	 * 
	 * @throws Exception
	 */
	private void startLearntXML() throws Exception
	{
		learntDoc = Util.makeOutDoc();
		explainDoc = Util.makeOutDoc();
		menuDoc = Util.makeOutDoc();
		
		learntXML = learntDoc.createElement("learntWords");
		explainXML = explainDoc.createElement("explanation");
		menuXML = menuDoc.createElement("menus");
		
		// I'm not sure if I need these
		learntDoc.appendChild(learntXML);
		explainDoc.appendChild(explainXML);
		menuDoc.appendChild(menuXML);
	}
	
	/**
	 * 
	 * @param type 0, 1, 2 for learntXML, explainXML and menuXML
	 * @return an XML element for a new cycle
	 */
	private Element newCycle(int type)
	{
		Document doc = null;
		if (type == 0) doc = learntDoc;
		if (type == 1) doc = explainDoc;
		if (type == 2) doc = menuDoc;
		
		Element el = doc.createElement("cycle");
		Integer c = Integer.valueOf(cycleNo);
		el.setAttribute("cycleNo", c.toString());
		
		if (type == 0) learntXML.appendChild(el);
		if (type == 1) explainXML.appendChild(el);
		if (type == 2) menuXML.appendChild(el);
		
		return el;
	}
	
	/**
	 * 
	 * @param stages learning stages [0,0], [0,1] etc. with newly learnt words
	 * @param learntCycle XML definition of learning stage, to be sent to the web client
	 * @param menuCycle XML definition of the new learning menu, to send to the web client
	 */
	private void createLearntWordXML(Vector<LearningStage> stages, Element learntCycle, Element menuCycle) throws Exception
	{
		Element cycleTop = learntDoc.createElement("learning");
		learntCycle.appendChild(cycleTop);

		Element menuTop = menuDoc.createElement("menuItem_1");
		menuTop.setAttribute("name", "Learning");
		menuCycle.appendChild(menuTop);
		
		for (LearningStage stage: stages)
		{
			Element stageEl = stage.xmlForm(learntDoc, true);
			cycleTop.appendChild(stageEl);
			
			Element menuStage = menuDoc.createElement("menuItem_2");
			menuStage.setAttribute("name", stage.getName());
			menuTop.appendChild(menuStage);
			
			for (WordToLearn wl: stage.getLearnWords())
			{
				Element menuWord = menuDoc.createElement("menuItem_3");
				menuWord.setAttribute("name", wl.getWordName());
				menuStage.appendChild(menuWord);
			}
		}
		
	}
	
	/**
	 * add the cycle data to the data for all cycles
	 * @param learntCycle
	 * @param explainCycle
	 * @param menuCycle
	 */
	private void addCycleData(Element learntCycle,Element explainCycle,Element menuCycle)
	{
		learntXML.appendChild(learntCycle);
		explainXML.appendChild(explainCycle);
		menuXML.appendChild(menuCycle);
	}


	
	/**
	 * 
	 * @param doc
	 * @param cycleNo
	 * @return the data for one cycle, as XML, e.g for cycle 3:
	 * 
	 * <cycleData>
	 * 	   <cycle cycleNo = '3' type='0'>
	 *      .... (4 Learning stages)
	 *     </cycle>
	 * 	   <cycle cycleNo = '3' type='1'>
	 *      .... (explanation lines)
	 *     </cycle>
	 * 	   <cycle cycleNo = '3' type='2'>
	 *      .... (new Learning Menu structure)
	 *     </cycle>
	 * </cycleData>
	 */
	public Element makeCycleData(Document doc, int cycleNo)
	{
		// make a root to attach the three types of data for the cycle
		Element root = doc.createElement("cycleData");
		// do not append the root to the document - that is done in the calling method.
		
		// three data types are: LearningStages (all words to be learnt); explanation text lines; menu structure
		for (int type = 0; type < 3; type++) try
		{
			// get data of the type for all cycles from the generated data class
			InputStream stream = LearntDataClass.getDataStream(type);
			Element mRoot = Util.getRootElement(stream);			
			Vector<Element> cycles = Util.namedChildElements(mRoot, "cycle");

			// pick out the requested cycle for the type
			if (cycleNo < cycles.size())
			{
				// check the cycle number is what you expect
				Element cycleEl = cycles.get(cycleNo);
				String cString = cycleEl.getAttribute("cycleNo");
				int cNo = Integer.parseInt(cString);
				if (cNo != cycleNo) throw new Exception("Cycle number " + cNo + " does not match " + cycleNo);

				// mark the type of data for this cycle, and mark the last cycle
				String typeString = Integer.valueOf(type).toString();
				cycleEl.setAttribute("type", typeString);
				if (cycleNo == cycles.size() -1)  cycleEl.setAttribute("last", "true");
				
				// deep import into the document to be sent to the web client
				Element cRoot = (Element)doc.importNode(cycleEl, true);
				// add all three types  of data for the cycle to the root
				root.appendChild(cRoot);
			}
			else throw new Exception("Cycle number " + cycleNo + " is out of range for type " + type);

		}
		catch(Exception ex) {Util.message(ex.getMessage());}
		
		Util.message("Sending learning data for cycle " + cycleNo);
		
		return root;
	}
	
	/**
	 * set up the data needed for the server to respond 
	 * to learning replay requests for all automatic learning cycles.
	 * 
	 * These data are:
	 * (a) for each cycle of automatic learning, a Vector of the LearningStages 
	 *     for that cycle (with their contained WordToLearn objects)
	 * (b) for each cycle of learning, a table of FSNodes for all words learnt in the cycle
	 *     (used for partial parses of learning examples, in learning replays)
	 */
	public void setupLearningCycles() throws Exception
	{

		// get data of type 0 (LearningStages)for all cycles from the generated data class
		int type = 0; 
		InputStream stream = LearntDataClass.getDataStream(type);
		Element mRoot = Util.getRootElement(stream);			
		Vector<Element> cycles = Util.namedChildElements(mRoot, "cycle");
		
		// cycle 0 only defines the words used to generate learning examples
		for (int cycleNo = 1; cycleNo < cycles.size(); cycleNo++)
		{
			Element cycle = cycles.get(cycleNo);
			Element learningEl = Util.firstNamedChild(cycle, "learning");
			Vector<Element> stageEls = Util.namedChildElements(learningEl, "stage");
			Vector<LearningStage> stagesInCycle = new Vector<LearningStage>();
			Hashtable<String,FSNode> wordsInCycle = new Hashtable<String,FSNode>();
			for (Element stageEl: stageEls)
			{
				LearningStage stage = new LearningStage(model,learner,stageEl);
				stagesInCycle.add(stage);
				for (WordToLearn wl:stage.getLearnWords())
				{
					FSNode word = wl.getWordFS();
					String spacedName = wl.getWordName();
					wordsInCycle.put(spacedName, word);
				}
			}
			
			learner.getKnownWords().add(wordsInCycle);
			learner.getAllStages().add(stagesInCycle);
		}

		
	}

}

