package learnerPack;

import java.util.Vector;

import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.Ops;

/**
 * represents a node in a binary subsumption tree
 *
 */

public class GraphNode {
	
	// Example Generator, for computing information content of FSNodes
	private ExGenerator generator;
	
	// true if this is a leaf node
	public boolean isLeaf() {return isLeaf;}
	private boolean isLeaf;
	
	// true if this is the root node
	public boolean isRoot() {return isRoot;}
	private boolean isRoot;
	
	// the FSNode, which for non-leaf nodes is the generalisation of the two child FSNodes 
	public FSNode getNodeFS() {return nodeFS;}
	private FSNode nodeFS;
	
	// Learning example, for leaf nodes
	private LearningExample example;
	
	// parent GraphNode
	public GraphNode getParent() {return parent;}
	private GraphNode parent;
	
	// 0, 1 or 2 child nodes, which this node sumsumes
	public Vector<GraphNode> getChildNodes() {return childNodes;}
	private Vector<GraphNode> childNodes = new Vector<GraphNode>();
	
	//-------------------------------------------------------------------------
	//                          Constructors
	//-------------------------------------------------------------------------
	
	/**
	 * constructor for the root node and the first learning example
	 * @param model the language model
	 */
	public GraphNode(Learner learner, LearningExample first) throws Exception
	{
		// the root FSNode is an empty script (and stays empty)
		nodeFS = new FSNode(learner.getModel(),"script");
		generator = first.generator;
		isLeaf = false;
		isRoot = true;
		
		// first child node is a leaf
		GraphNode firstChild = new GraphNode(first);
		childNodes.add(firstChild);
	}

	/**
	 * constructor for a leaf node
	 * @param fs
	 */
	public GraphNode(LearningExample lex)
	{
		example = lex;
		nodeFS = lex.getUnifyResult(0);
		generator = lex.generator;
		isLeaf = true;
		isRoot = false;
		
		// the list of child GraphNodes is empty
	}

	//-------------------------------------------------------------------------
	//                          methods
	//-------------------------------------------------------------------------
	
	/**
	 * add a learning example somewhere in the tree below this GraphNode
	 * @param fs
	 */
	public void addExample(LearningExample lex) throws Exception
	{
		// if this is a leaf node, make it non-leaf with two child leaf GraphNodes (its previous example, and the new one)
		if (isLeaf)
		{
			GraphNode g1 = new GraphNode(example);
			GraphNode g2 = new GraphNode(lex);
			nodeFS = Ops.generalise(g1.getNodeFS(),g2.getNodeFS());
			childNodes.add(g1);
			childNodes.add(g2);
			isLeaf = false;
		}
		
		// if there are 0 or 1 child nodes, add a leaf GraphNode as a child
		else if (childNodes.size() < 2) 
		{
			GraphNode oldNode = null;
			if (childNodes.size() > 0) oldNode = childNodes.get(0);

			GraphNode newNode = new GraphNode(lex);
			childNodes.add(new GraphNode(lex));

			if (!isRoot) 
			{
				if (oldNode == null) nodeFS = newNode.getNodeFS();
				else nodeFS = Ops.generalise(oldNode.getNodeFS(), newNode.getNodeFS());
			}
		}
				
		// if there are 2 child GraphNodes, find which one has the better overlap with the new example, and add the example to it
		else if (childNodes.size() == 2)
		{
			FSNode g1 = Ops.generalise(childNodes.get(0).getNodeFS(), lex.getUnifyResult(0));
			FSNode g2 = Ops.generalise(childNodes.get(1).getNodeFS(), lex.getUnifyResult(0));
			int i1 = meaningInfo(g1);
			int i2 = meaningInfo(g2);

			// add the example to the node which will have greater generalisation information content
			int chosen = 0;
			if (i2 > i1) chosen = 1;
			// if generalisation information content is equal, add the example to the node with fewer leaf descendants, for balance
			else if ((i2 == i1) && (childNodes.get(1).descLeaves() < childNodes.get(0).descLeaves())) chosen = 1;
			else {} // leave chosen at 0

			GraphNode choice = childNodes.get(chosen);
			GraphNode other = childNodes.get(1 - chosen);
			// push the example further down the tree, on the chosen side
			choice.addExample(lex);
			
			// the nodeFS may be altered by this (information content may be reduced)
			if (!isRoot) try{nodeFS = Ops.generalise(choice.getNodeFS(), other.getNodeFS());}
			catch (Exception ex) {}
		}
	}

	
	//-------------------------------------------------------------------------
	//
	//-------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @param fs
	 * @return 10* meaning scene information content
	 */
	public int meaningInfo(FSNode fs)
	{
		int info = 0;
		if (fs.getLastChild() != null)
			info = (int)(10*fs.getLastChild().getInformation(generator.getInfoCount()));	
		
		return info;
	}
	
	/**
	 * 
	 * @return number of descendant leaf nodes
	 */
	public int descLeaves()
	{
		int desc = 0;
		if (isLeaf) desc = 1;
		
		for (GraphNode child:childNodes) desc = desc + child.descLeaves();
		return desc;
	}
	
	/**
	 * 
	 * @param depth
	 * @return indented tree for subsumption graph
	 */
	public String stringForm(int depth)
	{
		String line = "";
		// indent
		for (int i = 0; i < depth;i++)  line = line + "  ";
		// size of subtree and meaning information content
		line = line + "(" + descLeaves() + "," + meaningInfo(nodeFS);
		// for leaf nodes, give the example number 
		if (isLeaf) line = line + ",Ex" + example.getExNo();
		// and a marker for a distractor meaning
		if ((isLeaf) && (example.getbestMeaning() > 0)) line = line + ",d";
		line = line + ")";
		
		// indented subtree
		for (GraphNode child: childNodes) line = line + "\n" + child.stringForm(depth + 1);
		
		return line;
	}


}
