package learnerPack;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.JMenu;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import displayPack.VFrame;
import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.LearningStage;
import modelPack.Ops;
import modelPack.WordToLearn;
import utilPack.Util;


public class Learner {
	
	private LanguageModel model;
	public LanguageModel getModel() {return model;}
	
	// for display of results to the user
	private VFrame vFrame() {return model.getFrame();}
	
	private ExGenerator generator;
	public ExGenerator generator() {return generator;}
		
	// all words that have been learnt so far, keyed by the word name
	private Hashtable<String,Vector<LearntWord>> learntWords = new Hashtable<String,Vector<LearntWord>>();
	public Vector<LearntWord> getNamedWords(String spacedName) 
	{
		Vector<LearntWord> words =  learntWords.get(spacedName);
		if (words == null) words = new Vector<LearntWord>();
		return words;
	}
	
	// for recording words as they are learnt - before they get good support
	private void addLearntWord(String spacedName, LearntWord lw)
	{
		Vector<LearntWord> words =  learntWords.get(spacedName);
		if (words == null) words = new Vector<LearntWord>();
		words.add(lw);
		learntWords.put(spacedName,words);
	}
	
	// all words that are known (learnt with good support) at the end of each cycle of automatic learning
	private Vector<Hashtable<String,FSNode>> knownWords = new Vector<Hashtable<String,FSNode>>();
	public Vector<Hashtable<String,FSNode>> getKnownWords() {return knownWords;}
	public FSNode getKnownWord(int cycleNo, String spacedName) {return knownWords.get(cycleNo -1).get(spacedName);}
	
	// all LearningStage objects, for each cycle of learning
	private Vector<Vector<LearningStage>> allStages = new Vector<Vector<LearningStage>>();
	public Vector<Vector<LearningStage>> getAllStages() {return allStages;}
	
	// real words learnt so far, to track what is new each cycle
	private Hashtable<String, String> realWordsLearnt = new Hashtable<String, String>();
	
	// words with 1 or 2 sounds isolated in learning examples
	private Hashtable<String,Vector<LearningExample>> shortWords = new Hashtable<String,Vector<LearningExample>>(); 
	
	// number of learning examples needed to start learning a short word
	private int shortWordStart = 2;
	
	// tabular summary of progress at the end of each cycle
	private Vector<String> progress = new Vector<String>();
	
	/* as soon as more than this percentage of learning examples have some word known, 
	 * (a) re-start learning all words which are not yet supported
	 * (b) use only leaning examples in which some words are known (so rejection of distractors is not just random)*/
	public int useGoodExampleThreshold = 30;
	
	// when this becomes true, use only learning examples in which one or more words are known (so rejection of distractors is not just random)
	public boolean useOnlyGoodLearningExamples = false;
	
	public void preferGoodExamples(boolean useOnlyGoodLearningExamples) {this.useOnlyGoodLearningExamples = useOnlyGoodLearningExamples;}
		
	// table of word sounds pre-learnt
	public Hashtable<Vector<String>, String> preLearntWords = new Hashtable<Vector<String>, String>();
	private boolean learnable(Vector<String> word) {return (preLearntWords.get(word) != null);}
	
	// true if only pre-leaned words are to be started
	private boolean constrained() {return true;}
	
	// level of Bayesian support for each word after each cycle
	Hashtable<String,int[]> supportLevel = new Hashtable<String,int[]>();
	
	// binary subsumption trees for each word to be learnt
	Hashtable<String,GraphNode> wordTrees = new Hashtable<String,GraphNode>();
	
	
//--------------------------------------------------------------------------
//                        Constructor
//--------------------------------------------------------------------------

	
	public Learner(LanguageModel model, ExGenerator generator) throws Exception
	{
		this.model = model;
		this.generator = generator;
	}
	
	

	//!---------------------------------------------------------------------------------
	//                     Distributional Learning of Word Sounds
	//----------------------------------------------------------------------------------
	
	/**
	 * called once for the pre-learn phase, after preLearn (e.g. 1000) examples have been made
	 * to learn the sounds of real words, in a 2-pass approach
	 * @param thresh1 statistical threshold (standard deviations) for first pass
	 * @return Vector of learned word sounds
	 */
	public Vector<Vector<String>> findAllWordSounds(double thresh1)
	{
		Vector<Vector<String>> allWords = new Vector<Vector<String>>();
		Vector<Vector<String>> nonWords = new Vector<Vector<String>>();
		Hashtable<String,Vector<String>> cands = new Hashtable<String,Vector<String>>();

		double gm = (double)generator.preLearn;
		int trials = 6000; // random pairs of learning examples
		
		
		{			
			// first pass: choose random pairs of examples in the pre-learn examples
			for (int t = 0; t < trials; t++)
			{
				int g1 = (int)(gm*Math.random());
				int g2 = (int)(gm*Math.random());
				FSNode f1 = generator.getExample(g1).getUnifyResult(0);
				FSNode f2 = generator.getExample(g2).getUnifyResult(0);
				
				Vector<String> shared = sharedSound(f1,f2);
				String sound = Util.serial(shared);
				
				// if this sound sequence has not been made before, and has three or more sounds
				if ((cands.get(sound) == null) && (shared.size() >= 3))
				{
					cands.put(sound, shared);
					boolean isWord = minStandardDev(shared,false) > thresh1;
					
					// collect candidate words
					if (isWord)
					{
						allWords.add(shared);
						// write a trace for any non-words picked up in the first pass
						if (!realWord(shared)) minStandardDev(shared, true);
					}
					
					// words missed on the first pass
					else if (!isWord) 
					{
						nonWords.add(shared);
						// write a trace for any real words missed in the first pass
						// if (realWord(shared)) minStandardDev(shared, true);
					}
				}
			} // end of loop over random pairs of learning example

			int firstPassSize = allWords.size(); // note the number of real words found in pass 1			
						
			// second pass - make partitions with words from the first pass
			// Util.message("Second pass");
			secondPass(allWords);
			
			// writing stuff out - no new results made for the rest of the method
			Util.message("\nPre-learning of word sounds; candidate sound sequences: " + cands.size());
			String mess = "\nPass 1 " + firstPassSize + ": ";	
			for (int w = 0; w < allWords.size(); w++)
			{
				Vector<String> wv = allWords.get(w);
				mess = mess + Util.serial(wv) + ", ";
				if (w == firstPassSize - 1) mess = mess + "\n\nPass 2 " + (allWords.size() - firstPassSize) + ": ";
			}
			Util.message(mess);	
			
			// count real and unreal words
			String unreals = "";
			int realCount = 0;
			int unrealCount = 0;
			int allReals = generator.spacedRealWords.size();
			Hashtable<String,String> realsFound = new Hashtable<String,String>();
			for (Vector<String> wv: allWords)
			{
				String sound = Util.serial(wv);
				if (realWord(wv)) 
				{
					realCount++;
					realsFound.put(sound.substring(0,sound.length() - 1), "1");
				}
				else 
				{
					unrealCount++;
					unreals = unreals + sound + ", ";
				}
			}
			
			String realsMissed = "";
			for (String s: generator.spacedRealWords)
			{
				if (realsFound.get(s) == null)  
					realsMissed = realsMissed + s + ", ";				
			}
			
			Util.message("\n" + (allReals - realCount) + " real words missed: " + realsMissed);
			Util.message(unrealCount + " unreal words found: " + unreals);

		}
		
		// side-effect: set up the table of pre-learnt words
		for (Vector<String> word: allWords) preLearntWords.put(word,"1");
				
		return allWords;
	}
	
	/**
	 * 
	 * @param wv
	 * @return true if the sounds are a real word
	 */
	private boolean realWord(Vector<String> wv)
	{
		boolean realWord = false;
		String sound = Util.serial(wv);
		if (sound.length() > 0) 
		{
			sound = sound.substring(0,sound.length()-1); // remove final '-'
			realWord = Util.inVector(sound, generator.spacedRealWords);	
		}
		return realWord;
	}
	
	/**
	 * 
	 * @param f1
	 * @param f2
	 * @return
	 */
	private Vector<String> sharedSound(FSNode f1, FSNode f2)
	{
		Vector<String> sounds = new Vector<String>();
		int[] shared = Ops.matchSoundSequence(f1, f2);
		for (int i = shared[0]; i < shared[0] + shared[2]; i++)
		{
			String s = f1.getChild(i).getSlotValue("sound");
			sounds.add(s);
		}
				
		return sounds;
	}
	
	
	/**
	 * 
	 * @param sounds
	 * @param write
	 * 
	 * @return the minimum standard deviation, for any any partition of the sounds into two sub-sequences
	 * with which the sequence of sounds occurs more often than independent sub-sequences
	 * if allWords is not null, the sub-sequences must be the end and start of words in AllWords
	 */
	public double minStandardDev(Vector<String> sounds, boolean write)
	{
		String spaced = Util.serial(sounds);
		double count = occCount(sounds);
		String line = spaced + "( " + count + "): ";
		double sample = (double)generator.numberOfExamples();
		double minDev = 100.0;
				
		// iterate over substrings partitioning the string,the first substring of increasing length len from 1 to size - 1
		for (int len = 1; len < sounds.size(); len++)
		{
			Vector<String> fromStart = new Vector<String>();
			for (int l = 0; l < len; l++)
				fromStart.add(sounds.get(l));

			int com = sounds.size() - len; // length of complement string in the partition
			Vector<String> fromEnd = new Vector<String>();
			for (int l = 0; l < com; l++)
			{
				int lp = l + sounds.size() - com;
				fromEnd.add(sounds.get(lp));
			}
			
			String sp1 = Util.serial(fromStart);
			String sp2 = Util.serial(fromEnd);
			
			// first pass - raw occurrence counts of strings
			double sc = occCount(fromStart);
			double ec = occCount(fromEnd);
			
			// avoid NaN if either substring count is zero. 6 standard deviations will pass any practical threshold
			double standard = 0.0;
			if ((sc < 0.5)||(ec < 0.5)) standard = 6.0;			
			else
			{
				double mean = sc*ec/sample;
				standard = (count - mean)/Math.sqrt(mean); // can be negative				
			}

			if (standard < minDev) minDev = standard;
			line = line + "[" + sp1 + " " + sc + "," + sp2 + " " + ec + "] " + standard + "  ";				
			
		}
		
		if (write) Util.message(line);
		
		return minDev;
	}
	
	/**
	 * second pass - make partitions with words from the first pass. 
	 * Use them to find more words, to add to allWords
	 * @param allWords a list of words found so far - to be added to
	 */
	private void secondPass(Vector<Vector<String>>allWords)
	{
		Hashtable<Vector<String>,Integer> leftOvers = new Hashtable<Vector<String>,Integer>();

		// match all examples against the first list of words
		for (int i = 0; i < generator.preLearn; i++)
		{
			FSNode f = generator.getExample(i).getUnifyResult(0);
			Vector<String> exSounds = f.getSounds().get(0);
			// find all sounds in the example which do not match the current list of words; false = do not keep words in the partition
			Vector<Vector<String>> leftOver = Util.bestPartition(exSounds,allWords,false);
			// count all left over fragments, not matched with a word
			for (Vector<String> frag: leftOver)
			{
				Integer c = leftOvers.get(frag);
				if (c == null) c = Integer.valueOf(0);
				c = Integer.valueOf(c.intValue() + 1);
				leftOvers.put(frag,c);
			}
		}
		
		// rearrange the table of leftovers, into groups of fragments with the same use count
		Hashtable<Integer, Vector<Vector<String>>> rightOvers = new Hashtable<Integer, Vector<Vector<String>>>();
		for (Enumeration<Vector<String>> en = leftOvers.keys(); en.hasMoreElements();)
		{
			Vector<String> frag  = en.nextElement();
			Integer c = leftOvers.get(frag);
			Vector<Vector<String>> frags = rightOvers.get(c);
			if (frags == null) frags = new Vector<Vector<String>>();
			frags.add(frag);
			rightOvers.put(c,frags);
		}
		
		// iterate over fragments (candidate words), in order of decreasing use count.
		int total = 0;
		int minOccs = 10; // minimum number of occurrences to count as a word
		int minSounds = 2; // minimum number of sounds in a word
		for (int cc = generator.preLearn; cc >= minOccs; cc--)
		{
			Integer c = Integer.valueOf(cc);
			Vector<Vector<String>> frags = rightOvers.get(c);
			if (frags != null)
			{
				String line = cc + " occurrences: ";
				for (Vector<String> frag: frags) if (frag.size() >= minSounds)
				{
					line = line + Util.serial(frag) + ", ";
					total = total + cc;
					// this counts as another word
					 allWords.add(frag);
				}
				// Util.message(line);
			}
		}
		
	}
	
	
	
		/**
		 * 
		 * @param sounds
		 * @return the number of times a sequence of sounds occurs, 
		 * in the latest generator.memory learning examples
		 */
		private double occCount(Vector<String> sounds)
		{
			return (double)generator.exVector(sounds).size();
		}
		


	//!---------------------------------------------------------------------------------
	//                        Handling one learning example
	//----------------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @param example
	 * @param explainCycle
	 * 
	 * @return an boolean[4]: 
	 * true if a distractor meaning chosen, 
	 * true if margin > 0 (to choose the real meaning over a distractor), 
	 * true if all words known in learning example, 
	 * true if learning example used
	 */
	public boolean[] useOneExample(LearningExample lex, Element explainCycle) throws Exception
	{
		boolean[] result = new boolean[4];
		for (int i = 0; i < 3; i++) result[i] = false;

		// find the best partition of the example, using supported words
		lex.findBestPartition(this);
		
		// try to choose the best meaning from the distractors, using any known words in the partition
		int best = lex.findBestDistractor();

		String explain = "Example " + lex.getName() + ": "  + lex.partitionString;
		generator.exampleLog.add("");
		explainLine(explain,explainCycle);
		
		// [true, true] is a mistake - choosing a distractor when words are known
		if (best > 0) result[0] = true;
		if (lex.getMargin() > 0) result[1] = true;
		result[2] = lex.allWordsKnown();
				
		/*  Really use this learning example if (a) you are still using all learning examples, or 
		 *  (a) this learning example can be validated by knowing one or more words in it */
		if ((!useOnlyGoodLearningExamples)||(lex.getMargin() > 0))
		{
			reallyUseExample(lex);
			result[3] = true;			
		}
				
		return result;

	}
	
	/**
	 * 
	 * @param lex
	 * @throws Exception
	 */
	private void reallyUseExample(LearningExample lex) throws Exception
	{
		// find the best partial parse of the learning example, using currently known words with some support
		ParseResult pr = lex.partialParse();
		
		// refine the words unified in the parse
		refineUnifiedWords(lex,pr);
		
		// collect learning examples for short words, and start to learn one when it has enough learning examples
		addShortWords(lex);

		// try to create more productive versions of words unified in the parse
		addMoreProductiveWords(lex);
		
		// refine other candidate words (not used in the best parse, and left as word sounds)
		Vector<LearntWord> refined = refineCandidateWords(lex);
		
		// create new un-productive candidate words from the sounds
		makeNewCandidates(lex, refined);
		
		// build subsumption trees for each word, when there are not too many distractors
		if (lex.getExNo() > 100) buildWordTrees(lex);
	}
	

	//----------------------------------------------------------------------------------------------
	//             Handling one learning example - refining unified words
	//----------------------------------------------------------------------------------------------
	
	 /**
	  *  refine the words unified in the parse
	  * @param lex
	  * @param pr
	  * @throws Exception
	  */
	 private void refineUnifiedWords(LearningExample lex,ParseResult pr) throws Exception
	 {
			// refine the words unified in the best parse
			String rString = "Refined unified words: ";
			for (int i = 0; i < pr.getWordsApplied().size(); i++)
			{
				LearntWord lw = pr.getWordsApplied().get(i);
				// choose the unified version of the example just before the word was unified
				lw.addExample(lex, i);
				rString = rString + lw.getSpacedProdName();
			}
			if (pr.someUnify()) generator.exampleLog.add(rString);
	 }
	


	//----------------------------------------------------------------------------------------------
	//             Handling one learning example - promoting a word to higher productivity
	//----------------------------------------------------------------------------------------------
	
	
	
	/**
	 * try to promote words to higher productivity
	 * @param lExample Learning Example 
	 * @throws Exception
	 */
	private void addMoreProductiveWords(LearningExample lExample) throws Exception
	{
		
		// find words in the partition that might be made with higher productivity
		for (int i = 0; i < lExample.getPartition().size(); i++)
		{
			PartOfExample we = lExample.getPartition().get(i);
			if (!we.isGap)
			{
				LearntWord lWord  = we.getLearntWord();
				String wName = lWord.getSpacedWordName(); // for debugging
				// current productivity of the word
				int[] cProd = lWord.getProductivity();
				int ctProd = cProd[0] + cProd[1];
				
				// the two sides on which productivity might be increased, if it is now zero
				for (int side = 0; side < 2; side++) if (cProd[side] == 0)
				{
					// if this learning example can increase the productivity on this side, increasing the number of meaning scenes
					int fsNo = lExample.promoter(lWord, side);
					if (fsNo > ctProd)
					{
						LearningExample previousExample = previousProductiveExample(lWord,side,lExample.getExNo());
						// if a previous example exists that can be used..
						if (previousExample != null)
						{
							// if the more productive word already exists...
							boolean already = moreProductiveWordExists(lWord,side);
							// if the two learning examples both unify some same other word, before the word to be promoted...
							boolean sharedWord = someSharedWord(previousExample,lExample,lWord,side);
							
							// if neither of these, try to promote the word
							if ((!already) && (!sharedWord))
							{
								LearntWord higher = new LearntWord(generator, previousExample,lExample,lWord,side);
								if (higher.isViable())
								{
									int versions = replaceWord(lWord, higher);
									Util.message("--- Making productive word " + higher.getSpacedProdName() + " from the best of " + versions + " versions");
								}
								else Util.message("--- Productive word was not viable: " + higher.getSpacedProdName());
							}
						}
					} // end of if (fsNo > 0)
					
				} // end of loop over sides for promotion				
				
			} // end of case !we.isGap
			
		} // end of loop over the partition
	}
	
	
	/**
	 * 
	 * @param l1
	 * @param l2
	 * @param w
	 * 
	 * @return true if examples l1 and l2 use some common word in parsing, before parsing word w,
	 * and so should not be used together in the LearntWord constructor to promote w to higher productivity 
	 */
	private boolean someSharedWord(LearningExample l1,LearningExample l2,LearntWord w, int side)
	{
		boolean shared = false;
		
		// all words used to unify in each learning example
		Vector<LearntWord> u1 = l1.getUnifiedWords();
		Vector<LearntWord> u2 = l2.getUnifiedWords();
		
		// find the required number of unifications from each learning example
		int ex1 = l1.promoter(w, side); // known to be 1 or more
		int ex2 = l2.promoter(w, side); // known to be 1 or more
		
		// check all the necessary unifications, to see if any of them use the same words
		for (int i1 = 0; i1 < ex1; i1++)
		{
			LearntWord w1 = u1.get(i1);
			for (int i2 = 0; i2 < ex2; i2++)
			{
				LearntWord w2 = u2.get(i2);
				if (w1.equalWord(w2)) shared = true;
			}
		}

		return shared;
	}

	
	
	/**
	 * replace a learnt word by a more productive version
	 * @param higher
	 */
	private int replaceWord(LearntWord less, LearntWord higher) throws Exception
	{
		String name = higher.getSpacedWordName();
		Vector<LearntWord> words = learntWords.get(name);		
		if (words == null) throw new Exception("Word list missing");
		int versions = words.size();

		words.remove(less);
		words.add(higher);
		if (versions != words.size()) Util.message("--- failed to remove less productive version.");

		learntWords.put(name, words);
		return versions;
	}


	
	/**
	 * 
	 * @param lw word whose productivity is to be increased
	 * @param side side 0 or 1 on which productivity is to be increased
	 * @param exNo example number to look  back from
	 * @return a previous LearningExample that can increase the productivity of lw on side side
	 */
	private LearningExample previousProductiveExample(LearntWord lw, int side, int exNo)
	{
		LearningExample plex = null;
		
		// look for previous learning examples that can increase the productivity
		int min  = exNo - generator.getMemory();
		if (min < 0) min = 0;
		for (int ix = min; ix < exNo; ix ++)
		{
			LearningExample ll = generator.getExample(ix);
			if (ll.promoter(lw, side) > 0) plex = ll;
		}
		
		return plex;
	}
	
	/**
	 * 
	 * @param lw
	 * @param side
	 * @return true if a more productive version of word lw exists,
	 * so a new one should not be created
	 */
	private boolean moreProductiveWordExists(LearntWord lw, int side)
	{
		boolean exists = false;
		Vector<LearntWord> lvv = learntWords.get(lw.getSpacedWordName());
		for (LearntWord lx: lvv)
			if (lw.moreProductiveWord(lx, side)) exists = true;
		
		return exists;
		
	}



	//----------------------------------------------------------------------------------------------
	//                   Handling one learning example - refining candidate words
	//----------------------------------------------------------------------------------------------
	

	
	
	/**
	 * use only the last FS in an example to refine candidate words
	 * 
	 * @param lex the learning example
	 * 
	 * @return list of learnt words, extended by those candidate words that have been refined with the last learning example of the group.
	 */
	private Vector<LearntWord> refineCandidateWords(LearningExample lex) throws Exception
	{
		FSNode lastExample = lex.getLastUnifyResult();
		String mess = "Refined candidate words: ";
		Vector<LearntWord> words = new Vector<LearntWord>();
		
		/* try out all learnt words which the example can apply to; 
		 * The canApply test checks that the number of meaning scenes in the example is sufficient for the word.*/
		int additions = 0;
		Vector<LearntWord> lwv = allWords();
		for (LearntWord lw: lwv) if (lw.canApply(lastExample))
		{						
			// try to add the example to the learnt word
			boolean[] res = lw.addExample(lex,lex.getFSCount()-1);
			
			// if the example has not been already used for this word, write it out
			if (!res[0])
			{
				words.add(lw);
				mess = mess  + lw.getSpacedWordName() + "(" + res[1] + ") ";
				additions++;
			}
			else mess = mess + " (failed addition) ";
		}
		if (additions > 0) generator.exampleLog.add(mess);
		return words;
	}


	//----------------------------------------------------------------------------------------------
	//                   Handling one learning example - making new candidate words
	//----------------------------------------------------------------------------------------------
	
	

	/**
	 * try to make new candidate words, 
	 * by generalising this learning example with other examples whose text partly matches it, 
	 * but not creating a new word if this example has been used to refine an existing word
	 * @param lex
	 * @param refined
	 */
	private void makeNewCandidates(LearningExample lex, Vector<LearntWord> refined) throws Exception
	{
		// use the last example in the group 
		FSNode partial = lex.getLastUnifyResult();				
		
		// build a list of candidate new learnt words
		Vector<LearntWord> candidates = new Vector<LearntWord>();
		Vector<LearntWord> shortCandidates = new Vector<LearntWord>();
				
		// go over all earlier learning examples, which are still remembered
		int max = generator.getMemory();
		if (lex.getExNo() < max) max = lex.getExNo();
		for (int g = 1; g < max; g++)
		{
			// go in reverse order, from just before the current example
			LearningExample previousEx = generator.getExample(lex.getExNo() - g);

			// only try the last FS in the previous example
			FSNode prevFS = null;
			try {prevFS = previousEx.getLastUnifyResult();}
			catch (Exception ex)
			{
				Util.message("\n^^^ Missing last unify result, looking back from example " + lex.getExNo() + " to example " + previousEx.getExNo());
			}
			
			// find the largest overlap in sounds between the two examples
			Vector<String> best = LearningExample.bestOverlap(partial,prevFS);
			
			// criterion for starting a new word
			boolean learnThisWord = (best.size() >= 3);
			if (constrained()) learnThisWord = ((best.size() >= 2) && (learnable(best)));
			
			// further criterion - the two learning examples must have only one word in common, in their partitions
			learnThisWord = learnThisWord && (lex.sharedsounds(previousEx).size() == 1);
			
			if (learnThisWord)
			{				
				// the 2 examples share enough sounds to match; try to generalise them together */
				Vector<LearningExample> exList = new Vector<LearningExample>();
				exList.add(previousEx);
				exList.add(lex);
				LearntWord newWord = new LearntWord(generator, exList,best);
				
				// test of a successful generalisation
				if (newWord.isViable())
				{
					// do not make a new candidate word if this example has been used to refine the same word
					boolean refinedWord = false;
					for (LearntWord lw: refined) if (newWord.equalWord(lw)) refinedWord = true;
					
					if (!refinedWord) candidates.add(newWord);

				} // end of successful creation of a new candidate word	
				
			} // end of if best.size() >= 3
			
			// learning short words (2 sounds only)
			boolean tryShortWords = false;
			if ((best.size() == 2) && tryShortWords)
			{
				// try to  find a third learning example with the same two sounds in sequence
				for (int gg = g + 1; gg < max; gg++)
				{
					// go in reverse order, from just before the matching previous example
					LearningExample thirdEx = generator.getExample(lex.getExNo() - gg);
					FSNode thirdFS = thirdEx.getLastUnifyResult();
					
					if (Util.containsSequence(thirdFS, best))
					{
						// use the three learning examples in the constructor
						Vector<LearningExample> exList = new Vector<LearningExample>();
						exList.add(previousEx);
						exList.add(lex);
						exList.add(thirdEx);
						LearntWord newShortWord = new LearntWord(generator, exList,best);
						
						if (newShortWord.isViable())
						{
							// do not make a new candidate word if this example has been used to refine the same word
							boolean refinedWord = false;
							for (LearntWord lw: refined) if (newShortWord.equalWord(lw)) refinedWord = true;
							
							if (!refinedWord) shortCandidates.add(newShortWord);
						}
						
					}
				}
				
			}
			
		} // end of loop over earlier learning examples
				

		// loop over both lists of candidates, choosing the best ones - why only one?
		addGoodWords(candidates);
		addGoodWords(shortCandidates);
					
	}
	
	
	/**
	 * loop over candidates, choosing the best one - why only one?
	 * @param candidates
	 */
	private void addGoodWords(Vector<LearntWord> candidates)
	{
		double bestInfo = 0.0;
		LearntWord bestCand = null;
		String candNames = "";
		for (LearntWord cand: candidates)
		{
			double info = cand.getInfo();
			if (info > bestInfo)
			{
				bestInfo = info;
				bestCand = cand;
			}
			candNames = candNames + cand.getShortProdName() + ", ";
		}
		
		if (bestCand != null) 
		{
			String name = bestCand.getSpacedWordName();
			Vector<LearntWord> lWords = learntWords.get(name);
			if (lWords == null) lWords = new Vector<LearntWord>();
			lWords.add(bestCand);
			learntWords.put(name, lWords);
			generator.exampleLog.add("New candidate word " + bestCand.getShortProdName() + " (number " + lWords.size() + ") from " + candidates.size() + " candidates "  + candNames);
		}
	}
	
	
	
	//-------------------------------------------------------------------------------------------------------
	// 							clean up the words that have been learnt
	//-------------------------------------------------------------------------------------------------------

	
	
	/**
	 * 
	 * @param spacedName
	 * @return the most productive and best supported word with given spaced name
	 * or null if no word has been learnt
	 */
	public LearntWord getBestWord(String spacedName)
	{
		Vector<LearntWord> words = learntWords.get(spacedName);
		LearntWord best = null;
		int[] bestProd = {0,0};
		double bestSupport = -50.0;
		
		if (words != null) for (LearntWord w: words)
		{
			if (    (w.getProductivity()[0] >= bestProd[0]) && 
					(w.getProductivity()[1] >= bestProd[1]) && 
					(w.bayesSupport() > bestSupport))
			{
				best = w;
				bestProd = w.getProductivity();
				bestSupport = w.bayesSupport();
			}
		}
		return best;
	}
	
	/**
	 * restart the learning of all words, except supported words
	 */
	public void restartCandidateWords()
	{
		for (Enumeration<String> en = learntWords.keys();en.hasMoreElements();)
		{
			String wName = en.nextElement();
			Vector<LearntWord> words = learntWords.get(wName);
			Vector<LearntWord> newWords = new Vector<LearntWord>();
			for (LearntWord word: words)
				if (word.bayesSupport() > 0) newWords.add(word);
			
			if (newWords.size() > 0) learntWords.put(wName, newWords);
			else 
			{
				learntWords.remove(wName);
				Util.message("Restarting learning word " + wName);
			}
		}
	
	}

	
	//--------------------------------------------------------------------------
	//	                          Framework
	//--------------------------------------------------------------------------
	
	
	/**
	 * all learnt words, as a Vector
	 * @return
	 */
	private Vector<LearntWord> allWords()
	{
		Vector<LearntWord> words = new Vector<LearntWord>();
		for (Enumeration<String> en = learntWords.keys();en.hasMoreElements();)
		{
			String name = en.nextElement();
			Vector<LearntWord> lWords = learntWords.get(name);
			for (LearntWord lw: lWords) words.add(lw);
		}
		
		return words;
	}
	
	/**
	 * supported words, keyed by name 
	 * @return
	 */
	public Hashtable<String, Vector<LearntWord>> getSupportedWords()
	{
		Hashtable<String, Vector<LearntWord>> words = new Hashtable<String, Vector<LearntWord>>();
		for (Enumeration<String> en = learntWords.keys();en.hasMoreElements();)
		{
			String name = en.nextElement();
			
			//all words with this name
			Vector<LearntWord> lWords = learntWords.get(name);
			
			// all supported words with this name
			Vector<LearntWord> sWords = new Vector<LearntWord>();
			for (LearntWord lw: lWords) if (lw.bayesSupport() > 0.0) sWords.add(lw);
			
			// if there are any supported words of this name, store them
			if (sWords.size() > 0) words.put(name, sWords);
		}
		
		
		return words;
	}
	

	//----------------------------------------------------------------------------------------------------
	//                          End of cycle - set up words to replay learning interactively
	//----------------------------------------------------------------------------------------------------
	
	
	
	/**
	 * 
	 * @throws Exception
	 * @return a summary of words learnt to display in the explanation area
	 */
	public Vector<LearningStage> refreshLearningMenu(int cycle, Element explainCycle) throws Exception 
	{
		// words that are known with good support at the end of this learning cycle
		Hashtable<String,FSNode> cycleKnownWords = new Hashtable<String,FSNode>();

		// learning stages for this learning cycle, named after productivities
		Vector<LearningStage> stages = new Vector<LearningStage>();
		
		Vector<String> summary = new Vector<String>();
		summary.add("\n\n");
		summary.add("Words learnt by cycle " + cycle + ": ");
		explainLine(summary.get(0),explainCycle);
		explainLine(summary.get(1),explainCycle);
		Util.message("\n\n" + summary.get(1));
		
		// remove learnt words that do not have the maximum productivity
		removeLessProductiveWords();
		
		// remove previous learning stages - hand constructed or from the previous learning cycle
		model.clearLearningStages();
		
		// classify learning stages and words to learn by productivity - the number of meaning scenes before and after the word 
		for (int before = 0; before < 2; before++)
		{
			for (int after = 0; after < 2; after++)
			{
				String stageName = "Productivity [" + before + "," + after + "]";
				String learntLine = "\n" + stageName + ": ";
				int learntProd = 0;
				String wordString = "";
				LearningStage stage = new LearningStage(model,stageName,this,cycle);
				
				for (LearntWord lw: allWords())
				{
					int[] scenes = lw.getProductivity();
					if ((lw.bayesSupport() > 0) && (scenes[0] == before) && (scenes[1] == after))
					{
						// add the word to be learnt to the correct learning stage for the cycle
						WordToLearn wl = new WordToLearn(stage,lw);
						stage.addWord(wl);
						
						// record the words learnt in this cycle
						learntProd++;
						wordString = wordString + lw.getSpacedWordName() + ", ";
						
						// record words for use in partial parsing of learning examples
						FSNode w = lw.getWord();
						cycleKnownWords.put(lw.getSpacedWordName(), w);
					}
				}
				
				learntLine = learntLine + " (" + learntProd + " words): " + wordString;
				explainLine(learntLine,explainCycle);
				Util.message(learntLine);
				
				// add the LearningStage to the model in the application
				model.addLearningStage(stage,stageName);
				
				// to be returned
				stages.add(stage);
			}			
		}
		
		explainLine("\n\n",explainCycle);
		
		// remake the Learning menu in the application
		VFrame frame = model.getFrame();
		JMenu learnMenu = frame.getLearnMenu();
		learnMenu.removeAll();
		frame.addLearnItems(model.getLearningStages(), learnMenu, null);
		
		// disable all edits, to prevent corrupting the hand-made model WordToLearn objects with automatically learned words
		frame.disableAllEdits();
		
		// record the words learnt up to this  cycle, for use in partial parses of learning examples
		knownWords.add(cycleKnownWords);
		
		return stages;
	}

	/**
	 * 
	 */
	private void removeLessProductiveWords()
	{
		for (Enumeration<String> en = learntWords.keys();en.hasMoreElements();)
		{
			String wName = en.nextElement();
			Vector<LearntWord> lwv = learntWords.get(wName);
			int max = maxSupport(lwv);
			if ((max > 0) && (lwv.size() > 1))
			{
				int maxProd = 0;
				LearntWord prodWord = null;
				for (LearntWord lw:  lwv)
				{
					int[] prod = lw.getProductivity();
					int prodSum = prod[0] + prod[1];
					if (prodSum  > maxProd)
					{
						prodWord = lw;
						maxProd = prodSum;
					}
				}
				
				if (prodWord != null)
				{
					Vector<LearntWord> newv = new Vector<LearntWord>();
					newv.add(prodWord);
					learntWords.put(wName, newv);
				}
			}
			
		}
	}


	//----------------------------------------------------------------------------------------------------
	//                          Writing out to the console and explanation area
	//----------------------------------------------------------------------------------------------------
	
	/**
	 * clear the explanation area
	 */
	public void clearExplain()
	{
		vFrame().explanationArea.setText("");
		vFrame().explanationArea.setCaretPosition(0); // moves scrollbar to the top
		
	}
	
	/**
	 * add a line of explanation to the explain area of the application,
	 * and store it as XML to be sent to the web client
	 * @param explanation
	 */
	public void explainLine(String explanation, Element explainData)
	{
		// write to the explanation area of the application
		vFrame().explanationArea.append("\n" + explanation);
		
		// store in XML to be sent to the web client
		Document doc = explainData.getOwnerDocument();
		Element lineEl = doc.createElement("line");
		lineEl.setAttribute("content", explanation);
		explainData.appendChild(lineEl);
		
		// store as lines to be written to a text file
		generator.exampleLog.add(explanation);
	}
	
	
	/**
	 * write an update at the end of every cycle
	 */
	public void writeCycleUpdate() throws Exception
	{
		// allWordSummary();
		
		// write out words created - details to a text file
		int realW = 0;
		int supported = 0;
		int realSupported = 0;
		int matchesReal = 0;
		int prod1Supported = 0;
		int prod2Supported = 0;
		Hashtable<String, String> prod0Words = new Hashtable<String, String>();
		Hashtable<String, String> prod1Words = new Hashtable<String, String>();
		Hashtable<String, String> prod2Words = new Hashtable<String, String>();
		Hashtable<String,Vector<LearntWord>> byName = new Hashtable<String,Vector<LearntWord>>();
		Vector<LearntWord> unreals = new Vector<LearntWord>();
		Vector<int[]> cycleRealWordDiffs = makeNotLearntDiffs();
		Vector<String> learntWordLines  = new Vector<String>();
		
		for (LearntWord lw: allWords())
		{
			// write evidence for all words
			lw.writeEvidence(learntWordLines);
			String name = lw.getSpacedWordName();
			boolean hasSupport = (lw.bayesSupport() > 0);
			boolean isReal = (Util.inVector(name, generator.spacedRealWords));
			
			// there was a call lw.isRejected() which has now been scrapped
			{ 
				noteSupport(lw, generator.getCycleNo());
				
				if (hasSupport) 
				{															
					if (isReal) 
					{
						FSNode realWord = generator.realWordFS.get(name);
						// compare learnt meaning information with the result information of the real word
						double learntInfo = lw.getWord().getLastChild().getInformation(generator.getInfoCount());
						double realInfo = realWord.getLastChild().getInformation(generator.getInfoCount());
						double diff = learntInfo - realInfo;
						learntWordLines.add("Information in " + name + " : learnt: " + learntInfo + " ; real word: " + realInfo + " ; difference: " + diff);
						// compare result scenes only
						int[] dc = new int[3]; // extra slots, missing slots, Bayes support
						if (Math.abs(diff) > 0.06) dc = FSNode.writeSlotDifference(lw.getWord().getLastChild(),realWord.getLastChild(),learntWordLines);
						else
						{
							learntWordLines.add("Equal slot information");
							matchesReal++;
							for (int d = 0; d < 2; d++) dc[d] = 0;
						}
						dc[2] = (int)(lw.bayesSupport());
						recordDiff(lw.getSpacedWordName(),dc,cycleRealWordDiffs);
					
					}
					else unreals.add(lw);
					
					Vector<LearntWord> named  = byName.get(name);
					if (named == null) named = new Vector<LearntWord>();
					named.add(lw);
					byName.put(name, named);
				}
				
				if (isReal) 
				{
					realW++;
					if (hasSupport)
					{
						if (lw.getProductivitySum() == 1) {prod1Words.put(lw.getShortProdName(),"1");prod1Supported++;}
						else if (lw.getProductivitySum() == 2) {prod2Words.put(lw.getShortProdName(),"1");prod2Supported++;}
						else prod0Words.put(lw.getShortWordName(),"1");
						realSupported++;
					}
				}
				if (hasSupport) supported++;
			}
			
		} // end of loop  over all learnt words

		
		//  collect the word slot discrepancy scores for this cycle
		allRealWordDiffs.add(cycleRealWordDiffs);
		
		
		// save the text file of learnt word feature structures etc.
		Util.writeTextFile(model.learntWordMeaningPath, learntWordLines);

		
		// analysis of short words (1 and 2 sounds, isolated in partitions)
		countShortWords();
				
		Util.message("\nCycle " + generator.getCycleNo() + "; words: " + allWords().size()  + "; " + supported + " supported.");
		Util.message(realW + " real words found; " + realSupported + " real and supported (" + prod1Supported + " prod=1;" + prod2Supported + " prod=2);" 
		+ " match real word meaning: " + matchesReal
		+ "\nBase: " + singleString(prod0Words) + "\nProductive: " + singleString(prod1Words));
		Util.message(differenceString(prod0Words));
		Util.message((allWords().size() - realW) + " unreal words found;" + (supported - realSupported) + " unreal and supported.");
		
		// Util.macroMemoryMessage("cycle",generator.getCycleNo());
		
		// sum over all learning examples in the cycle
		int meaningPercent = meaningPercent(generator.getCycleNo());
		
		int prod0Supported = realSupported - prod1Supported - prod2Supported;

		
		String progressLine =  generator.getCycleNo() + "," + supported + "," + realSupported + "," + matchesReal  + "," + prod0Supported 
				+ "," + prod1Supported  + "," + prod2Supported+ "," + (supported - realSupported) + "," + meaningPercent;
		progress.add(progressLine);
		
		writeProgress();
		
		writeSupportHistory();
		
		writeSubsumptionTrees();
		
				
	}
	
	/**
	 * 
	 * @param cycleNo
	 * @return percentage success measure for understanding meanings in the cycle
	 */
	private int meaningPercent(int cycleNo) throws Exception
	{
		int percent = 0;
		int num = 0;
		int denom = 0;		
		int min = generator.exPerCycle*(cycleNo -1);
		int max = generator.exPerCycle*cycleNo;
		
		for (int m = min; m < max; m++)
		{
			int[] fract = generator.getExample(m).successFraction();
			num = num + fract[0];
			denom = denom + fract[1];
		}
		
		if (denom > 0) percent = (100*num)/denom;
		return percent;
	}
	
	// for each cycle, a vector of [extra slots, missing slots, Bayesian support] for each real word that has been learnt
	private Vector<Vector<int[]>> allRealWordDiffs = new Vector<Vector<int[]>>();
	
	/* initialise the vector of [extra slots, missing slots, Bayesian support] for each real word, for one cycle 
	 * set both elements -1 to indicate that the word is not learnt at all yet.*/
	private Vector<int[]> makeNotLearntDiffs()
	{
		Vector<int[]> res = new Vector<int[]>();
		for (String rWord: generator.spacedRealWords)
		{
			int[] dc = new int[3];
			for (int d = 0; d < 3; d++) dc[d] = -1;
			res.add(dc);
		}
		return res;
	}
	
	// record the gap [extra slots, missing slots, Bayesian support] for one real word which has been learnt in one cycle
	private void recordDiff(String word, int[]dc, Vector<int[]> diffs)
	{
		for (int w = 0; w < generator.spacedRealWords.size(); w++) 
		{
			String rWord = generator.spacedRealWords.get(w);
			// inserts the new difference at the right place
			if (word.equals(rWord))
			{
				diffs.remove(w);
				diffs.insertElementAt(dc, w);
			}
		}
	}
	
	
	/**
	 * note the latest level of Bayesian support for a word
	 * @param lw
	 * @param cycleNo
	 */
	private void noteSupport(LearntWord lw, int cycleNo)
	{
		String name = lw.getSpacedWordName();
		// any previous level of support recorded for this word
		int[] supp = supportLevel.get(name);
		
		int[] nSupp = new int[cycleNo];
		for (int i = 0; i < cycleNo - 1; i++) 
		{
			if (supp == null) nSupp[i] = -500; // denotes that the word is not yet started
			else if (i < supp.length) nSupp[i] = supp[i]; // copy previous value
			else nSupp[i] = -500;	// word has been started and finished??		
		}
		
		// integer level of support at end of latest cycle
		int sLev = (int)(10.0*lw.bayesSupport());
		nSupp[cycleNo - 1] = sLev;
		supportLevel.put(name, nSupp);
	}
	
	/**
	 * write the history of support for all words over all cycles
	 */
	private void writeSupportHistory() throws Exception
	{
		Vector<String> fileLines = new Vector<String>();

		// for words which agree < 80% with the true word; [word,% agreement,support]
		Vector<String[]> badWords = new Vector<String[]>();
		
		for (Enumeration<String> en = supportLevel.keys(); en.hasMoreElements();)
		{
			String name = en.nextElement();
			int[] supp = supportLevel.get(name);
			String line = name + ",";
			
			for (int l = 0; l < supp.length; l++)
			{
				if (supp[l] == -500) 
				{
					line = line + ","; //empty cell
				}
				else line = line + supp[l] + ",";
			}
			
			
			// at the end of each row, compare information content with the real word
			LearntWord lw = getBestWord(name);
			FSNode realWord = generator.realWordFS.get(name);
			if ((lw != null) && (realWord != null))
			{
				realWord = realWord.getLastChild(); // meaning of the real word
				FSNode w  = lw.getWord().getLastChild(); // meaning of the learnt word
				FSNode g = Ops.generalise(realWord, w); // meaning match
				
				double ireal = realWord.getInformation(generator.getInfoCount())*10.0;
				double ilearn = w.getInformation(generator.getInfoCount())*10.0;
				double imax = Math.max(ireal, ilearn);
				double igen = g.getInformation(generator.getInfoCount())*10.0;
				int percent = (int)(100.0* igen/imax);
				line = line + percent + ",," + (int)ireal + "," + (int)ilearn+ "," + (int)igen;	
				
				// record bad words
				if (percent < 80)  try
				{
					String[] bad = new String[3];
					bad[0] = name;
					bad[1] = Integer.valueOf(percent).toString();
					bad[2] = Integer.valueOf(supp[supp.length - 1]).toString(); // support level at last cycle
					badWords.add(bad);
				}
				catch (Exception ex) {}
			}
			
			fileLines.add(line);
		}
		
		// write bad words (agreement < 80%) in three lines
		fileLines.add("");
		fileLines.add("bad");
		for (int i = 0; i < 3;i++)
		{
			String line = "";
			for (String[] bad:badWords) line = line + bad[i] + ",";
			fileLines.add(line);
		}
		
		Util.writeTextFile(model.supportHistoryPath, fileLines);
		
	}

	/**
	 * write out subsumption trees of  learning examples for all words
	 */
	private void writeSubsumptionTrees() throws Exception
	{
		Vector<String> fileLines = new Vector<String>();
		
		for (Enumeration<String> en = wordTrees.keys(); en.hasMoreElements();)
		{
			String wName = en.nextElement();
			String name = wName.substring(0,wName.length() - 1);
			
			int realInfo = 0;
			FSNode realWord = generator.realWordFS.get(name);
			if (realWord != null)
			{
				realWord = realWord.getLastChild(); // meaning of the real word
				realInfo = (int) (10.0*realWord.getInformation(generator.getInfoCount()));				
			}
			
			GraphNode gn = wordTrees.get(wName);
			String line = "\n" + name + " " + realInfo + "\n" + gn.stringForm(0);
			fileLines.add(line);
		}
		
		Util.writeTextFile(model.subsumptionTreePath, fileLines);
		
	}
	
	
	/**
	 * write a tabular summary of progress up to this cycle,
	 * to a csv file.
	 * 
	 * Each row has columns:
	 *  - cycle no.
	 *  - number of known (supported) words
	 *  - number of real supported words
	 *  - match 
	 *  - number with productivity [0,0]
	 *  - number with productivity [0,1] or [1,0]
	 *  - number with productivity [1,1]
	 *  - number of unreal supported words
	 *  - percent success for the cycle matching meanings of examples with meanings of known words
	 */
	private void writeProgress() throws Exception
	{
		Vector<String> fileLines = new Vector<String>();

		// start the progress spreadsheet - parameter line, followed by column headers and a first row of zeroes
		String paramLine = "threshold,expLife,start,distractors,extra," 
				+ generator.threshold + " " + generator.expLife 
				+" [" + generator.getStartCriteria()[0]+ "; " + generator.getStartCriteria()[1]
				+ "] " + generator.nDistractors + " " + generator.getExtraSlots();
		fileLines.add(paramLine);
		fileLines.add("");
 
		String header = "Cycle,Supp,Real,Match,Prod 0,Prod 1,Prod 2,Unreal,Progress %";
		String firstRow = "0,0,0,0,0,0,0,0,0";
		fileLines.add(header);
		fileLines.add(firstRow);
		for (String line: progress) fileLines.add(line);
		
		Util.writeTextFile(model.progressSummaryPath, fileLines);

	}

	/**
	 * tab '\t' in the explanation area do not work as  expected;
	 * they just introduce a fixed number of spaces.
	 * So for the console, I am using ',' instead or '\t'
	 * Useful for making csv files)
	 * 
	 * Put the correct number of spaces after each string to do proper tabbing,
	 *
	 * @param s
	 * @param tab
	 * @return
	 */
	public String makeTabs(String s, int tab)
	{
		String tabbed = "";
		
		StringTokenizer st = new StringTokenizer(s,",");
		
		while (st.hasMoreTokens())
		{
			String el = st.nextToken();
			tabbed = tabbed + el;
			for (int i = el.length(); i < tab; i++) tabbed = tabbed + " ";
		}
		
		return tabbed;
	}
	
	/**
	 * 
	 * @param latestWordsLearnt
	 * @return
	 */
	private String singleString(Hashtable<String,String> latestWordsLearnt)
	{
		String s = "";
		for (Enumeration<String> en = latestWordsLearnt.keys();en.hasMoreElements();)
		{
			String w = en.nextElement();
			s = s + w + ", ";
		}
		return s;
	}

	/**
	 * 
	 * @param latestWordsLearnt
	 * @return
	 */
	private String differenceString(Hashtable<String,String> latestWordsLearnt)
	{
		String s = "Added: ";
		for (Enumeration<String> en = latestWordsLearnt.keys();en.hasMoreElements();)
		{
			String w = en.nextElement();
			if (realWordsLearnt.get(w) == null) s = s + w + ", ";
		}
		
		s = s + "; Unlearned: ";
		for (Enumeration<String> en = realWordsLearnt.keys();en.hasMoreElements();)
		{
			String w = en.nextElement();
			if (latestWordsLearnt.get(w) == null) s = s + w + ", ";
		}
		
		// update ready for next cycle
		realWordsLearnt = latestWordsLearnt;

		return s;
	}
	
	
	
	/**
	 * 
	 * @param lwv
	 * @return 10* the best Bayesian support in a list of word senses
	 */
	private int maxSupport(Vector<LearntWord> lwv)
	{
		int max = -1000;
		for (LearntWord lw: lwv)
		{
			int supp = (int)(10.0*lw.bayesSupport());
			if (supp > max) max = supp;
		}
		return max;
	}
	

	//----------------------------------------------------------------------------------------------------
	//                       Revising words at the end of a cycle
	//----------------------------------------------------------------------------------------------------
	
	// if support is greater than this, the word can be improved
	public static double bottom = -4.0;
	
	// if support is greater than this, the word needs no improvement
	public static double top = 15.0;
	
	/**
	 * Try to improve all words in a range of success scores,
	 * by re-learning them using random permutations of their learning examples
	 */
	public void permutationImprove(int cycleNo) throws Exception
	{
		Util.message("\nImprovement of unproductive words, using learning examples permuted in random order\n");
		String improvedWords = "Improved words: ";
		for (LearntWord lw:allWords()) if (lw.getProductivitySum() == 0)
		{
			// try to improve any  word with support in a (wide) range
			if ((lw.bayesSupport() > bottom) && (lw.bayesSupport() < top))
			{
				LearntWord improved = lw.permutationLearn(cycleNo);
				
				// improved is non-null only if support has been increased, and the meaning changed
				if (improved != null)
				{
					Vector<LearntWord> lwv = new Vector<LearntWord>();
					lwv.add(improved);
					String wordName = improved.getSpacedWordName();
					learntWords.put(wordName,lwv);
					improvedWords = improvedWords + wordName + ", ";
				}				
			}
		}
		Util.message(improvedWords);
	}


	//----------------------------------------------------------------------------------------------------
	//                      Building subsumption trees for each word
	//----------------------------------------------------------------------------------------------------
	

	/**
	 *  build binary subsumption trees for each word
	 * @param lex
	 */
	private void buildWordTrees(LearningExample lex) throws Exception
	{
		for (PartOfExample part: lex.getPartition())
		{
			String sounds = Util.serial(part.getWordSounds());
			GraphNode node = wordTrees.get(sounds);
			if (node == null)
			{
				node = new GraphNode(this,lex);
				wordTrees.put(sounds, node);
			}
			else node.addExample(lex);
		}
	}

	//----------------------------------------------------------------------------------------------------
	//                       Learning short words (1 or 2 sounds)
	//----------------------------------------------------------------------------------------------------
	

	
	/**
	 * add the the collection of short words and their learning examples.
	 * @param lex
	 */
	private void addShortWords(LearningExample lex) throws Exception
	{
		Vector<PartOfExample> wv = lex.getPartition();
		for (PartOfExample we: wv) if (we.isGap)
		{
			Vector<String> gapSounds = we.getWordSounds();
			if (gapSounds.size() < 3)
			{
				// save the learning example for the short word
				String shortWord = Util.serial(gapSounds);
				Vector<LearningExample> lv = shortWords.get(shortWord);
				if (lv == null) lv = new Vector<LearningExample>();
				lv.add(lex);
				shortWords.put(shortWord, lv);
				
				// if there are the required number of learning examples, start to learn the word
				if (lv.size() == shortWordStart) beginShortWord(gapSounds,lv);
			}
		}
		
	}
	
	/**
	 * start to learn a short word from a short list of learning examples, of length shortWordStart
	 * @param sounds
	 * @param lv
	 */
	private void beginShortWord(Vector<String> sounds,Vector<LearningExample> lv) throws Exception
	{
		String wordName = Util.serial(sounds);
		// Util.message("$$$  Starting new short word " + wordName);

		// use the first two examples to learn the word (most unified FS will be used)
		LearntWord newWord = new LearntWord(generator, lv,sounds);
		
		// if generalisation succeeds, save the word and refine it with the remaining learning examples
		if (newWord.isViable())
		{
			// save the learnt word
			addLearntWord(newWord.getSpacedWordName(),newWord);
			
			// refine the word with the remaining learning examples
			for (int w = 2; w < shortWordStart; w++)
			{
				LearningExample lex = lv.get(w);
				int exNo = lex.getFSCount() - 1;
				newWord.addExample(lex, exNo);
			}
			// Util.message("$$$  Saved and refined short word " + newWord.getSpacedWordName());
			
		} 
		// else Util.message("$$$ short word is not viable");
		
	}

	/**
	 * 
	 */
	private void countShortWords()
	{
		Hashtable<String,Integer> shortWordCounts = new Hashtable<String,Integer>();
		for (int ex = 0; ex < generator.numberOfExamples();ex++)
		{
			LearningExample lex = generator.getExample(ex);
			Vector<PartOfExample> wv = lex.getPartition();
			for (PartOfExample we: wv) if (we.isGap)
			{
				Vector<String> gapSounds = we.getWordSounds();
				if (gapSounds.size() < 3)
				{
					String shortWord = Util.serial(gapSounds);
					Integer count = shortWordCounts.get(shortWord);
					if (count == null) count = Integer.valueOf(0);
					count = Integer.valueOf(count.intValue() + 1);
					shortWordCounts.put(shortWord, count);
				}
			}
		}
		
		String shortLine = "\nShort words: ";
		for (Enumeration<String> en = shortWordCounts.keys();en.hasMoreElements();)
		{
			String shortWord = en.nextElement();
			Integer count = shortWordCounts.get(shortWord);
			shortLine = shortLine + shortWord + "(" + count.toString() + ") ";
		}
		Util.message(shortLine);
	}

	

}
