package learnerPack;

import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.Ops;
import utilPack.Util;

public class LearningExample {
	
	protected ExGenerator generator;
	protected Learner learner;	

	LanguageModel model() {return generator.getModel();}
	
	// Vector of N FSNodes, all with the same sounds. Typically N = 2
	// The first of them has the true meaning of the learning example; The other (N-1) have random distractor meanings
	private Vector<FSNode> distractors;
	
	// the meaning for the learning example as originally made, with no noise added
	private FSNode noNoise;
	
	// index of the best candidate meaning 0.. N-1
	private int bestMeaning = 0;
	public int getbestMeaning() {return bestMeaning;}
	
	// margin in meaning information match score for known words, between the best candidate meaning and the next distractor
	private int margin = 0;
	public int getMargin() {return margin;}
	
	// the correct meaning of the learning example is guessed if there is not enough margin
	public boolean isGuessed() {return margin < 4;}
		
	/**
	 * results of unification of supported words with sounds; 
	 * the first element has no unifications, sounds only. 
	 * It has the meaning chosen by the learner out of the distractors
	 */
	private Vector<FSNode> unifyResults = new Vector<FSNode>();
	public void addUnifyResult(FSNode example) {unifyResults.add(example);}
	
	public FSNode getLastUnifyResult() {return unifyResults.get(unifyResults.size() -1);}
	public FSNode getUnifyResult(int p) {return unifyResults.get(p);}
	public int getFSCount() {return unifyResults.size();}
	
	/**
	 * supported words that have been unified with the example sounds
	 */
	private Vector<LearntWord> unifiedWords = new Vector<LearntWord>();
	private void addUnifiedWord(LearntWord lw) {unifiedWords.add(lw);}
	public Vector<LearntWord> getUnifiedWords() {return unifiedWords;}
	
	// name of the learning example (all distractors have the same name)
	public String getName() {return distractors.get(0).getName();}
	
	// example number
	private int exNo;
	public int getExNo() {return exNo;}
	
	// sounds of the example before any unification
	private Vector<String> sounds = new Vector<String>();
	public Vector<String> getExampleSounds() {return sounds;}
	
	// sounds of the example split into words
	private Vector<Vector<String>> splitSounds = new Vector<Vector<String>>();
	public Vector<Vector<String>> getSplitSounds() {return splitSounds;};
	
	/**
	 *  the best partition of this example into supported words and gaps. 'best' minimises gaps.
	 */
	private Vector<PartOfExample> partition = new Vector<PartOfExample>();
	public Vector<PartOfExample> getPartition() {return partition;}
	
	protected String partitionString = "";
	
	private ParseResult bestParseResult = null;
	
	// denotes an empty learning example, which has been forgotten
	private boolean empty = false;
	public boolean isEmpty() {return empty;}
	
	//--------------------------------------------------------------------------
	//                             Constructor
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param generator
	 * @param distractors Vector of N FSNode, all with the same sounds.
	 * One of them has the true meaning; (N-1) have random distractor meanings
	 * @param exNo
	 * @param noNoise
	 */
	public LearningExample(ExGenerator generator, Vector<FSNode> distractors, int exNo, FSNode noNoise)
	{
		this.generator = generator;
		this.learner = generator.learner;
		this.exNo = exNo;
		this.distractors = distractors;
		this.noNoise = noNoise;
		
		// all distractors have the same set of sounds
		sounds = distractors.get(0).getSounds().get(0);
	}
	
	/**
	 * constructor for an empty learning example, to
	 * fill out the list of examples
	 * @param generator
	 */
	public LearningExample(ExGenerator generator)
	{
		this.generator = generator;
		this.learner = generator.learner;
		
		exNo = -1;
		empty = true;		
	}
	
	
	//--------------------------------------------------------------------------
	//                    Partitions of the sounds
	//--------------------------------------------------------------------------
	
	/**
	 * find the best partition of the sounds of this example, 
	 * using only the sounds of words that have been pre-learnt
	 * @param preLearntWords
	 */
	public Vector<Vector<String>> findSoundPartition(Hashtable<Vector<String>, String> preLearntWords)
	{
		// make a list of all words whose sounds are included in the sounds of the example
		Vector<Vector<String>> wordSounds = new Vector<Vector<String>>();
		for (Enumeration<Vector<String>> en = preLearntWords.keys();en.hasMoreElements();)
		{
			Vector<String> wSound = en.nextElement();
			if (Util.includes(sounds, wSound)) wordSounds.add(wSound);
		}
		
		// remove words that are a concatenation of two other words
		wordSounds = Util.removeConcatenations(wordSounds);
		
		// make the best possible partition using these word sounds (with the fewest sounds not matched)
		Vector<Vector<String>> best = Util.bestPartition(sounds, wordSounds, true);		
		
		// set the instance variable
		splitSounds = best;
		
		return best;
	}
	
	/**
	 * 
	 * @param learner
	 */
	public void findBestPartition(Learner learner)
	{
		// find the best partition of example sounds into words
		Vector<Vector<String>> soundPart = findSoundPartition(learner.preLearntWords);
		
		// all learnt and supported words, keyed by word name
		Hashtable<String, Vector<LearntWord>> words = learner.getSupportedWords();

		for (Vector<String> part: soundPart)
		{
			String s = Util.serial(part); // spaced word name
			s = s.substring(0,s.length() - 1); // remove final '-'
			Vector<LearntWord> lwv = words.get(s);
			
			// the word has been learnt to supported level
			if (lwv != null)
			{
				// just pick the first word sense
				LearntWord lw = lwv.get(0);
				PartOfExample w = new PartOfExample(lw, learner);
				partition.add(w);
			}
			
			// the word has not been learnt
			else
			{
				PartOfExample wi = new PartOfExample(part,learner);
				partition.add(wi);				
			}
		}
		
		// set up the partition string
		for (PartOfExample w: partition) partitionString = partitionString + w.stringForm();
	}
	
	
	/**
	 * @param word
	 * @return true if the sounds of the word are included in one part of the partition
	 */
	public boolean isInOnePart(FSNode word)
	{
		if (word.getSounds().size() == 0) return false;
		Vector<String> sounds = word.getSounds().get(0);
		boolean onePart = false;
		for (PartOfExample w: partition)
			if (Util.includes(w.wordSounds, sounds)) onePart = true;
		return onePart;
	}
	
	/**
	 * 
	 * @param other
	 * @return true if this learning example has any of the same supported words as some other example -
	 * -so the two examples should not be used together the make a candidate word, 
	 * because its meaning would be contaminated by the shared word meaning.
	 */
	public boolean hasSameWord(LearningExample other)
	{
		boolean shares = false;
		Vector<PartOfExample> owe = other.getPartition();
		Vector<PartOfExample> twe = getPartition();
		
		for (PartOfExample tw:twe) if (!tw.isGap)
		{
			for (PartOfExample ow:owe) if (!ow.isGap)
				if (tw.getLearntWord().getSpacedWordName().equals(ow.getLearntWord().getSpacedWordName())) 
					shares = true;			
		}
		
		
		return shares;
	}
	
	/**
	 * 
	 * @return true if all words in this learning example have been learnt
	 */
	public boolean allWordsKnown()
	{
		boolean known = true;
		for (PartOfExample we: partition) if (we.isGap) known = false;
		return known;
	}

	
	//-------------------------------------------------------------------------------------
	//    Use the best partition to try to find the best meaning from the distractors 
	//-------------------------------------------------------------------------------------
	
	/**
	 * choose the distractor meaning with best fit to known words -
	 * or if there are no known words, choose a distractor at random.
	 * sets bestMeaning and margin.
	 * @return the index chosen (0 is the correct meaning; other are distractors)
	 * @throws Exception
	 */
	public int findBestDistractor() throws Exception
	{
		int d = distractors.size();
		int[] score = new int[d];
		bestMeaning = -1;
		double bestScore  = 0.0;
		
		// find if any of the distractors have meaning information overlap with any known word in the partition
		for (int dd = 0; dd < d; dd++) 
		{
			score[dd] = 0;
			FSNode dis = distractors.get(dd);
			
			// try out all known words in the partition
			for (PartOfExample we: partition) if (!we.isGap)
			{
				FSNode gen = Ops.generalise(dis, we.getLearntWord().getWord());
				// measure the result meaning information score of the generalisation
				double genInfo = gen.getLastChild().getInformation(generator.getInfoCount());
				score[dd] = score[dd] + (int)genInfo;
			}
			if (score[dd]> bestScore)
			{
				bestMeaning = dd;
				bestScore = score[dd];
			}
			
		}
		
		// if no best candidate has been found (e.g. because no words are known) make a random guess 0.. (d-1); and margin stays 0
		if (bestMeaning < 0) bestMeaning = (int)(d*Math.random());
		
		// make the chosen meaning the first element of unifyResults
		unifyResults.add(distractors.get(bestMeaning));
		
		// find the smallest margin between the chosen meaning and any other meaning (usually a distractor)
		margin = 50;
		for (int dd = 0; dd < d; dd++) if (dd != bestMeaning)
		{
			int marg = score[bestMeaning] - score[dd];
			if (marg < margin) margin = marg;
		}
		
		return bestMeaning;
	}
	
	/**
	 * 
	 * @return element 1 bits of matching information summed over all known words (numerator)
	 * element 2 bits of information in example meaning with no added noise (denominator)
	 * both to be summed for a cycle to compute a percentage success
	 */
	public int[] successFraction() throws Exception
	{
		int[] fract = new int[2];
		for (int i = 0; i < 2; i++) fract[i] = 0;
		
		// only add to the denominator for examples where the correct meaning 0 is chosen 
		if (bestMeaning == 0)
		{
			// the denominator is the information content of the example with no added noise
			fract[1] = (int)(noNoise.getLastChild().getInformation(generator.getInfoCount()));
			
			// calculate the numerator by summing over known words
			for (PartOfExample we: partition) if (!we.isGap)
			{
				// generalise the word with the learning example (with noise)
				FSNode gen = Ops.generalise(distractors.get(0), we.getLearntWord().getWord());
				// measure the result meaning information score of the generalisation
				double genInfo = gen.getLastChild().getInformation(generator.getInfoCount());
				fract[0] = fract[0] + (int)genInfo;
			}
		}
		
		return fract;
	}
	
	
	//--------------------------------------------------------------------------
	//                    Partial parsing using known words
	//--------------------------------------------------------------------------
	
	/**
	 * find the best partial parse of the learning example, 
	 * and add the partially unified FS to the example
	 */
	public ParseResult partialParse() throws Exception
	{
		// result if there are no words to unify; just the un-unified FS
		ParseResult bestParse = new ParseResult(getUnifyResult(0));
		double bestScore = 0.0;
				
		// for each word or sound sequence in the example, find all word senses with appreciable probability
		Vector<Vector<PartOfExample>> allWords = new Vector<Vector<PartOfExample>>();
		
		/* make a Vector of all possible words for each part of the partition.
		 * 'possible' have probability > 0.1, which is equivalent to some small negative Bayesian support */
		for(PartOfExample wex:partition)
		{
			Vector<PartOfExample> poss= wex.possibles();
			allWords.add(poss);
		}
		
		/* filter down to the word senses which match the inferred meaning well enough 
		 * (the test is: could the learning example refine the word?) This test seems to remove a lot of words - especially nouns. */
		Vector<Vector<PartOfExample>> filteredWords = filterByMeaning(allWords);
		
		// try doing without the filtering by meaning, to get more noun unifications
		filteredWords = allWords;
		
		// write out how many senses of each word remain; 'word senses' in the example log
		writeCombinations(filteredWords);
		
		// make a Cartesian product of the filtered word senses for each part
		Vector<Vector<PartOfExample>> cart = Util.cartesianProduct(filteredWords);
		
		// loop over each possible combination of word senses in the Cartesian product
		for (Vector<PartOfExample> oneComb: cart)
		{
			// parse the example as far as possible using this combination of word senses
			ParseResult result = oneParse(oneComb);
			
			if (result.score > bestScore)
			{
				bestParse = result;
				bestScore = result.score;
			}
		}
		
		// store the partially unified FS from the best parse in the example
		reflectParse(bestParse);

		return bestParse;
	}
	
	/**
	 * write out how many senses of each word remain after filtering for sense
	 * @param filteredWords
	 */
	private void writeCombinations(Vector<Vector<PartOfExample>> filteredWords)
	{
		String line = "Word senses: (";
		for (Vector<PartOfExample> wv: filteredWords)
		{
			int senses = wv.size();
			PartOfExample we = wv.get(0);
			if (we.isGap) line = line + "<" + Util.serial(we.wordSounds) + "> ";
			else line = line + "[" + we.getLearntWord().getShortWordName() + "] ";
			line = line + senses + ", ";			
		}
		generator.exampleLog.add(line + ")");
	}

	
	/**
	 * Incorporate the results of the best parse in this learning example
	 * @param pr
	 */
	private void reflectParse(ParseResult pr) throws Exception
	{
		// save the best parse result
		bestParseResult = pr;
		
		String trace = "Best Unification Results: ";
		
		// the meaning of the learning example, as inferred by the learner
		FSNode inferredMeaning = getUnifyResult(0).getLastChild();
		
		// the first FSNode is the sounds and the inferred meaning. Do not add it
		for (int i = 1; i < pr.getResults().size(); i++)
		{
			//  the word that was used to unify
			LearntWord lw = pr.getWordsApplied().get(i-1);
			trace = trace + lw.getShortProdName();
			FSNode w = lw.getWord();
			w.setAllTruePositions();
			FSNode wordMeaning = w.getLastChild();
						
			// the result of a unification
			FSNode nextResult = pr.getResults().get(i);
			
			if (nextResult.isViable())
			{
				// input to the last unification
				FSNode input = pr.getResults().get(i-1);
				input.setAllTruePositions();
				
				// repeat the unification with tracing of true positions and origins, to find the word meaning after unification
				FSNode unifiedMeaning = FSNode.findDerivedNode(w, input, wordMeaning);
				
				if (unifiedMeaning == null)
				{
					generator.exampleLog.add(w.getTextTree(0) + "\n" + input.getTextTree(0) + "\n" + wordMeaning.getTextTree(0));
					FSNode.findDerivedNode(w, input, wordMeaning);
					unifiedMeaning = wordMeaning;
				}
				
				// generalise the unified meaning with the inferred meaning of the example
				FSNode generalisedmeaning = Ops.generalise(unifiedMeaning, inferredMeaning);
				trace = trace + infoCompare(generalisedmeaning,unifiedMeaning) + ", ";
				
				// Remove all used sound and meaning nodes from the unify result
				nextResult = new FSNode(nextResult,FSNode.REMOVEUSED);
				
				// add back the inferred meaning scene (do not change its parent)
				nextResult.getChildren().add(inferredMeaning);

				// add to the example the unified FS  
				addUnifyResult(nextResult);
								
				// add to the example the LearntWord that was used in the unification
				addUnifiedWord(lw);
			}
			
			else trace = trace + "(failed) ";

		}
		
		generator.exampleLog.add(trace);
	}
	
	/**
	 * 
	 * @param f1
	 * @param f2
	 * @return
	 */
	public  String infoCompare(FSNode f1, FSNode f2)
	{
		int i1 = (int)(10.0*f1.getInformation(generator.getInfoCount()));
		int i2 = (int)(10.0*f2.getInformation(generator.getInfoCount()));
		return "(" + i1 + ":" + i2 + ")";
	}

	
	/**
	 *  filter down to the word senses which match the inferred meaning well enough 
	 *  (for the example to succeed in refining the word)
	 * @param allWords
	 * @return
	 */
	private Vector<Vector<PartOfExample>> filterByMeaning(Vector<Vector<PartOfExample>> allWords) throws Exception
	{
		Vector<Vector<PartOfExample>> filtered = new Vector<Vector<PartOfExample>>();
		
		for (Vector<PartOfExample> part: allWords)
		{
			Vector<PartOfExample> newPart = new Vector<PartOfExample>();
			for (PartOfExample wex: part)
			{
				// if there is a single sound-only part, include it
				if (wex.isGap) newPart.add(wex);

				// learnt words - test them against the example meaning
				else
				{
					LearntWord lw = wex.getLearntWord();
					FSNode before = lw.getWord();
					
					//generalise the word with the inferred meaning of the example
					FSNode after = Ops.generalise(before,getUnifyResult(0));
					// test the result; could the example successfully refine the word?
					boolean accept = lw.acceptResult(before, after,true);
					if (accept) newPart.add(wex);
				}
			}
			
			// rescue with a sounds-only part, if no learnt word senses in the part match the inferred meaning well enough
			if (newPart.size() == 0) 
			{
				Vector<String> wordSounds = part.get(0).getLearntWord().getWordSounds();
				PartOfExample rescue = new PartOfExample(wordSounds,learner);
				newPart.add(rescue);
			}
			
			filtered.add(newPart);
		}
		
		return filtered;
	}

	
	/**
	 * parse this example using one choice of word senses for each part - 
	 * unifying all unproductive words first, (unifying most noun-like words first)
	 * then trying every possible order of unifying productive words - 
	 * that is, every possible unify order for this example, given the choice of word senses 
	 * for the words known at the time
	 * 
	 * @param oneComb the chosen word senses for each part of the example
	 * 
	 * @return the parse result for the productive word unify order with best information score.
	 * 
	 * Information score is a sum of:
	 * - for each unproductive word, the information in the overlap of the unify result and the inferred meaning
	 * - for each productive word, the same information overlap, plus the information content of its input meaning scenes
	 */
	private ParseResult oneParse(Vector<PartOfExample> oneComb) throws Exception
	{
		double score = 0.0;
		
		// to record the words applied, in the order in which they were applied
		Vector<LearntWord> wordsApplied = new Vector<LearntWord>();
		
		// to keep track of the status of the unification
		Vector<int[]> status = new Vector<int[]>();
		
		// make a sounds-only version of the learning example, to start unifying words with
		FSNode starter = unifyResults.get(0).soundOnly();
		Vector<FSNode> results = new Vector<FSNode>();
		results.add(new FSNode(starter));
		
		// make a list of unproductive words, to sort with the most noun-like words first
		Vector<NounOrder> nounsFirst = new Vector<NounOrder>();
		for (int p = 0; p < oneComb.size(); p++)
		{
			PartOfExample wex = oneComb.get(p);
			
			/*  an array of elements 
			 * 0: gap or word; 1: unproductive or productive; 2 not unified or unified */
			int[] stat = new int[3];
			for (int i = 0; i < 3; i++) stat[i] = 0;

			if (wex.isGap) stat[0] = 0;
			else
			{
				stat[0] = 1;
				LearntWord lww = wex.getLearntWord();
				int[] meanings = lww.getProductivity();
				if ((meanings[0] > 0)||(meanings[1] > 0)) stat[1] = 1;
				
				// if the word is not productive
				if (stat[1] == 0)
				{
					NounOrder nord = new NounOrder(p,lww.getNounStrength());
					nounsFirst.add(nord);
				}
			}
			
			status.add(stat);
		}
		
		// sort the unproductive words to get nouns first
		Collections.sort(nounsFirst);
		
		// unify all unproductive words, in noun-first order
		boolean failed = false;
		String unified = "";
		for (NounOrder nord: nounsFirst) if (!failed)
		{
			PartOfExample wex = oneComb.get(nord.wordPos);
			LearntWord lw = wex.getLearntWord();
			
			wordsApplied.add(lw);
			FSNode word = lw.getWord();
			unified = unified + lw.getSpacedWordName() + ", ";
			try {starter = Ops.unify(word, starter, Ops.UNDERSTAND);}
			catch (Exception ex)
			{
				generator.exampleLog.add("Unification exception at "  + unified + ex.getMessage());
				generator.exampleLog.add("Word input: " + word.getTextTree(0));
				generator.exampleLog.add("Example input: " + starter.getTextTree(0));
				failed = true;
			}
			if (!starter.isViable()) failed = true;
			if (!failed)
			{
				results.add(new FSNode(starter));
				FSNode overlap = Ops.generalise(word, getUnifyResult(0));
				score = score + overlap.getInformation(generator.getInfoCount());
			}
		}
				
		// get all productive words (to be permuted in all orders)
		Vector<PartOfExample> allProds = new Vector<PartOfExample>();
		for (int i = 0; i < oneComb.size(); i++)
			if (status.get(i)[1] == 1) allProds.add(oneComb.get(i));
		
		// to record the results of the best permutation of productive words
		double bestProdScore = 0.0;
		Vector<LearntWord> bestProdWordsApplied = new Vector<LearntWord>();
		Vector<FSNode> bestProdResults = new Vector<FSNode>();

		// if there are any productive words to permute
		if ((!failed) && (allProds.size() > 0))
		{
			// make all permutations of productive words
			Vector<Vector<PartOfExample>> permedProds = Util.permutations(allProds);
			
			// try out all permutations of productive words, to find the one with the highest score from unifications			
			for (Vector<PartOfExample> perm: permedProds)
			{
				// do this permutation, and record the results
				FSNode prodStarter = starter;
				double prodScore  = 0.0;
				Vector<LearntWord> prodWordsApplied = new Vector<LearntWord>();
				Vector<FSNode> prodResults = new Vector<FSNode>();
				
				// loop through words in one permutation of productive words
				for (PartOfExample wex: perm)
				{
					LearntWord lw = wex.getLearntWord();
					unified = unified + lw.getSpacedWordName() + ", ";

					// check if the productive word can be applied to the current result of unification
					int[] meaningsNeeded = lw.getProductivity();
					int[] meanings = prodStarter.meaningSceneCount(lw.getWordSounds());
					boolean applicable = ((meanings[0] >= meaningsNeeded[0]) && (meanings[1] >= meaningsNeeded[1]));
					
					// the word can be applied. unify it an build the score for this permutation
					if (applicable) try
					{
						FSNode word = lw.getWord();
						prodStarter = Ops.unify(word, prodStarter, Ops.UNDERSTAND);
						if (prodStarter.isViable())
						{
							prodWordsApplied.add(lw);
							prodResults.add(new FSNode(prodStarter));
							FSNode overlap = Ops.generalise(word, getUnifyResult(0));
							// add the information overlap of the result, and the information content of the input meaning scenes
							prodScore = prodScore + overlap.getInformation(generator.getInfoCount()) + word.getMeaningInformation(generator.getInfoCount());
						}
					}
					catch (Exception ex)
					{
						generator.exampleLog.add(" Unification exception at: " + unified + ex.getMessage());
						generator.exampleLog.add("Productive word input: " + lw.getWord().getTextTree(0));
						generator.exampleLog.add("Example input: " + prodStarter.getTextTree(0));
					}
					
				} // end of loop over productive words in the permutation
				
				// if this is the best permutation so far, record its results
				if (prodScore > bestProdScore)
				{
					bestProdScore = prodScore;
					bestProdWordsApplied = prodWordsApplied;
					bestProdResults = prodResults;
				}
				
			} // end of loop over permutations of productive words
						
		} // end of case where there are some productive words
				
		// record the best results of unproductive and productive words together
		score = score + bestProdScore;	
		for (LearntWord lw: bestProdWordsApplied) wordsApplied.add(lw);
		for (FSNode r: bestProdResults) results.add(r);
		ParseResult result = new ParseResult(wordsApplied,results, score);
		
		return result;
	}
	
	
	/**
	 * 
	 * @param lw a learnt word, which is to be promoted (its productivity increased by 1)
	 * @param side = 0 if the new input meaning scene is to be before the word sound;
	 * side = 1 if it is to be after it
	 * 
	 * @return if this example is capable of being one of two learning examples 
	 * used in the constructor to make a new promoted (higher productivity) version of the word lw,
	 * return the index of the partly unified version of this learning example 
	 * which is to be generalised to make the new word.
	 * 
	 * This index cannot be 0, as some other word must have been unified first, to make a productive learning example.
	 * 
	 * return -1 if this example is not capable of promoting the word to higher productivity
	 */
	public int promoter(LearntWord lw, int side)
	{
		// if strict is true, require that unified words  be exact neighbours of the word to promote, in the partition
		// (strict = false not yet coded)
		boolean strict = false;
		
		// start at failure value
		int promoter = -1;
		
		// productivity of  word - old, and promoted 
		int[] oldProd = lw.getProductivity();
		int[] newProd = new int[2];
		for (int i = 0; i < 2; i++) newProd[i] = oldProd[i];
		newProd[side]++;
		
		// position in FSNode list required to have a unified word before and after the word to be promoted
		int[] pos = new int[2];
		for (int i = 0; i < 2; i++) pos[i] = 0;
		
		// list of learnt words before and after the word to be promoted (only for strict = false)
		Vector<LearntWord> lPre = new Vector<LearntWord>();
		Vector<LearntWord> lPost = new Vector<LearntWord>();
		boolean after = false;
		if (!strict) for (int p = 0; p < partition.size(); p++)
		{
			PartOfExample we  = partition.get(p);			
			if (!we.isGap)
			{
				LearntWord ll = we.getLearntWord();
				if (lw.equalWord(ll)) after = true;
				else if (!after) lPre.add(ll);
				else lPost.add(ll);
			}
		}

		// find the word to be promoted in the partition (it must have been unified in its less productive form)
		for (int i = 0; i < partition.size(); i++)
		{
			PartOfExample we  = partition.get(i);
			
			// if this is the word to be promoted
			if ((!we.isGap) && (lw.equalWord(we.getLearntWord())))
			{
				LearntWord lp = we.getLearntWord(); //  not needed
				
				// loop over sides 0 = before and 1 = after, at which unified words need to be found
				for (int s = 0; s < 2; s++) if (newProd[s] > 0)
				{
					/* find the neighbouring learnt word in the partition, at position j = i - 1  or j = i + 1,
					 * which is to provide an input meaning scene for the promoted word */
					int j = i - 1 + 2*s;
					
					// if the neighbouring position in the partition exists
					if ((j >= 0) && (j < partition.size()))
					{
						PartOfExample wn  = partition.get(j);
						if (!wn.isGap)
						{
							LearntWord ln = wn.getLearntWord();
							
							// find out the positions 0..N in the unify order when the word to be promoted and the neighbour word were unified
							int promPos = -1; // promoted word
							int nPos = -1; // neighbour word
							if (bestParseResult != null) for (int p = 0; p < bestParseResult.getWordsApplied().size(); p++)
							{
								LearntWord lx = bestParseResult.getWordsApplied().get(p);
								if (lw.equalWord(lx)) promPos = p;
								if (ln.equalWord(lx)) nPos = p;
							}
							
							// the neighbouring learnt word must have been unified before the word to be promoted, in the best parse
							if (promPos > nPos) pos[s] = nPos + 1; // nPos is >= 0, thisPos is >= 1.
							
						} // end of if wn at the neighbouring position is not a gap
						
					} // end of if the neighbouring position exists	
					
				} // end of loop over s	for before and after			
					
			} // end of if this is the word to be promoted

		} // end of loop over words in the partition
			
		promoter = Math.max(pos[0], pos[1]);
		return promoter;
	}
	
	
	//--------------------------------------------------------------------------
	//    Inner class for sorting unproductive words into nouns-first order
	//--------------------------------------------------------------------------
	
	class NounOrder implements Comparable<NounOrder>
	{

		public int compareTo(NounOrder other)
		{
			int res = -1;
			double diff = nounLike - other.nounLike;
			if (diff < 0) res = 1;
			return res;
		}

		int wordPos; // position of the LearntWord in the array
		double nounLike; // a measure of how noun-like it is
		
		NounOrder(int wordPos, double nounLike)
		{
			this.wordPos = wordPos;
			this.nounLike = nounLike;
		}
	}
	
	
	//--------------------------------------------------------------------------
	//                             Other Methods
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param other
	 * @return word sounds that this example shares with another example
	 */
	public Vector<Vector<String>> sharedsounds(LearningExample other)
	{
		Vector<Vector<String>> shared = new Vector<Vector<String>>();
		Vector<Vector<String>> oSounds = other.getSplitSounds();
		for (Vector<String> part:getSplitSounds())
		{
			boolean share = false;
			for (Vector<String> oPart:oSounds)
				if (Util.equalString(oPart, part)) share = true;
			
			if (share) shared.add(part);			
		}		
		return shared;
	}
	
	
	//--------------------------------------------------------------------------
	//                             Static Methods
	//--------------------------------------------------------------------------
	
	
	
	
	/**
	 * 
	 * @param f1
	 * @param f2
	 * @return
	 */
	public static Vector<String> bestOverlap(FSNode f1, FSNode f2)
	{
		Vector<String> best = new Vector<String>();
		if ((f1== null)||(f2==null)) return best;

		Vector<Vector<String>> p1 = f1.getSounds();
		Vector<Vector<String>> p2 = f2.getSounds();
		return bestOverlap(p1,p2);
	}
	
	/**
	 * 
	 * @param ps
	 * @param qs
	 * @return
	 */
	public static Vector<String> bestOverlap(Vector<Vector<String>> ps,Vector<Vector<String>> qs)
	{
		Vector<String> best = new Vector<String>();
		for (Vector<String> p:ps)
		{
			for (Vector<String> q:qs)
			{
				Vector<String> over = longestOverlap(p,q);
				if (over.size() > best.size())  best = over;
			}			
		}
		return best;
	}

	
	/**
	 * 
	 * @param p Vector of sounds 's', 'e', 't' and so on
	 * @param q Vector of sounds 's', 'e', 't' and so on
	 * @return the longest sequence they have in common
	 */
	public static Vector<String> longestOverlap(Vector<String> p,Vector<String> q)
	{
		Vector<String> s = new Vector<String>();
		
		for (int i = 0; i < p.size(); i++)
		{
			for (int j = 0; j < q.size(); j++)
			{
				Vector<String> t = new Vector<String>();
				int k = 0;
				boolean matched = true;
				while ((i + k < p.size()) && (j + k < q.size()) && (matched))
				{
					matched = (p.get(i + k).equals(q.get(j + k)));
					if (matched) t.add(p.get(i+k));
					k++;
				}
				if (t.size() > s.size()) s = t;
			}
		}		
		return s;
	}

	

}
