package learnerPack;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.Ops;
import utilPack.Util;

/**
 * a learnt word in the automatic learning model
 *
 */

public class LearntWord implements Comparable<LearntWord> {
	
	/**
	 * compare two learnt words by their Bayesian support
	 * @param other
	 * @return
	 */
	public int compareTo(LearntWord other)
	{
		int res = -1;
		double diff = bayesSupport() - other.bayesSupport();
		if (diff < 0) res = 1;
		return res;
	}
	
		
	// information about learning examples used to make and refine this word.
	private Vector<ExampleWord> examplePositions = new Vector<ExampleWord>();
	public Vector<ExampleWord> getExamplePositions() {return examplePositions;}
	
	// true if the constructor made a viable word
	public boolean isViable() {return viable;}
	private boolean viable = false;
	
	public ExGenerator generator;
	
	// the word learnt by generalisation of valid learning examples
	private FSNode wordFS;
	public FSNode getWord() {return wordFS;}
	
	// feature structure for the real word, if there is one
	public FSNode realWord()
	{
		FSNode realWord = null;
		if (wordSounds.size() > 0)
		{
			String wName = Util.serial(wordSounds); // t-i-m-
			wName = wName.substring(0, wName.length() -1); // t-i-m
			realWord = generator.realWordFS.get(wName);
		}
		return realWord;
	}
	
	// The name of the word - made up from its phonemes separated by '-'
	private String spacedWordName ="";
	public String getSpacedWordName() {return spacedWordName;}
	
	// short name = spaced name without '-'
	public String getShortWordName()
	{
		String sn = "";
		StringTokenizer st = new StringTokenizer(spacedWordName,"-");
		while (st.hasMoreTokens()) sn = sn + st.nextToken();
		return sn;
	}
		
	// the phoneme sounds of the word
	private Vector<String> wordSounds = new Vector<String>();
	public Vector<String> getWordSounds() {return wordSounds;}
	
	// the number of input meaning scenes required before and after the word - its productivity
	public int[] getProductivity() 
	{
		int[] scenes = new int[2];
		for (int i = 0; i < 2; i++) scenes[i] = -1;
		if (wordFS == null) return scenes;
		return wordFS.meaningSceneCount(getWordSounds());
	}
	
	public String productivityString() {return ("[" + getProductivity()[0] + "," + getProductivity()[1] + "]");}
	
	public int getProductivitySum() {return wordFS.getInputMeaningScenes().size();}

	// the word name and its productivity
	public String getShortProdName() {return getShortWordName() + productivityString();}	
	public String getSpacedProdName() {return (spacedWordName + productivityString());}
	
	
	/**
	 * @param lw
	 * @return true if lw has the same sounds and productivity
	 */
	public boolean equalWord(LearntWord lw) {return (getSpacedProdName().equals(lw.getSpacedProdName()));}

	/**
	 * @param lw
	 * @param side 0 for before , 1 for after
	 * 
	 * @return true if lw has the same sounds 
	 * and has 1 more input meaning scene on side side
	 */
	public boolean moreProductiveWord(LearntWord lw, int side)
	{
		boolean equal = Util.equalString(getWordSounds(), lw.getWordSounds());
		equal = equal && (getProductivity()[side] + 1 == lw.getProductivity()[side]); 
		equal = equal && (getProductivity()[1 - side] == lw.getProductivity()[1 - side]); 
		return equal;
	}
	
	
		
	// information content of the result
	public int getInfo()
	{
		double info = wordFS.getLastChild().getInformation(generator.getInfoCount());	
		return (int)info;
	}
	
	// text form of the word input and output meaning trees 
	public String getTextMeanings()
	{
		String meaning = "";
		for (FSNode child:wordFS.getChildren()) 
			if (child.getType().equals("scene"))
			   meaning = meaning + "$" + child.getTextTree(0) + " ";
		return meaning;
	}
	
	// the learning example which last promoted this word to higher productivity
	private int[] promotedExNo = {0,0};
	public int[] getpromotedExNo() {return promotedExNo;}
	
	// cycles at which permutation learning was successfuly used for this word
	private Vector<Integer> permutationCycles = new Vector<Integer>();
	public void setPermutationCycles(Vector<Integer> permutationCycles) {this.permutationCycles = permutationCycles;}
	
	
	//--------------------------------------------------------------------------
	//	                      First Constructor - non-productive
	//--------------------------------------------------------------------------
	
	/**
	 * start a learnt word with two or more learning examples, 
	 * only if their generalisation passes (has enough information).
	 * Otherwise, the list of example positions has no members, 
	 * and so the learnt word is not valid and will not be stored
	 * 
	 * This constructor cannot learn productive words with input meaning scenes or subsumption links
	 * 
	 * @param generator gives access to learning examples
	 * @param exList 2 or more learning examples
	 * @param sounds if not null, constrained sequence of sounds of the word
	 * @throws Exception
	 */
	public LearntWord(ExGenerator generator, Vector<LearningExample> exList, Vector<String> sounds) throws Exception
	{
		this.generator = generator;

		if (exList.size() < 2) throw new Exception("To start a learnt word, there must be at least 2 learning examples.");
		
		// always use the first (least unified) forms of the two learning examples
		LearningExample l1 = exList.get(0);
		LearningExample l2 = exList.get(1);		
		FSNode ex1 = l1.getUnifyResult(0);
		FSNode ex2 = l2.getUnifyResult(0);
				
		// generalise the first two examples 
		wordFS = Ops.generalise(ex1, ex2);
		boolean accepted = wordFS.getSounds().size() > 0;
		
		// generalise any more learning examples
		for (int i = 2; i < exList.size(); i++)
		{
			FSNode ex  = exList.get(i).getUnifyResult(0);
			wordFS = Ops.generalise(wordFS, ex);
			// in case this removes all sounds from the result
			accepted  = accepted & (wordFS.getSounds().size() > 0);
		}
		
		
		if (accepted)
		{
			// try to constrain the sounds of the word to be the sequence; this might just fail
			accepted = constrainSounds(wordFS,sounds);

			// do not keep the generalisation expression (which could get very big)
			wordFS.clearExpressions();
			
			// remove any input meaning scenes and subsumption links from the word FS
			makeWordNonProductive();
			
			// the word sound must be contained in one part of the partition, for each learning example
			for (LearningExample lex: exList) accepted = accepted && lex.isInOnePart(wordFS);
			
			// the first test of meaning acceptability has no previous word to go on
			boolean strict = true;
			if ((!l1.isGuessed()) && (!l2.isGuessed())) strict = false;
			accepted = accepted && acceptResult(null,wordFS,strict);
			
			// final check of subsumption links - there should not be any
			accepted = accepted && wordFS.checkWordLinks();
		}
		
		// now after all checks it is really possible to make a word
		if (accepted)
		{
			viable = true;

			// record all examples
			for (LearningExample lex:exList)
			{
				ExampleWord xw = new ExampleWord(lex,this, 0);
				xw.setSuccess(true);
				examplePositions.add(xw);
			}
			
			// record the word name and sounds in instance variables
			spacedWordName = wordFS.getSpacedName();
			wordSounds = wordFS.getSounds().get(0);
			wordFS.setName(spacedWordName);
		}
	}
	
	/**
	 * constrain the sounds of a non-productive word FS to be a sequence of sounds, 
	 * by removing sound scenes from the word (before or after the sounds) if necessary
	 * @param word a non-productive word FS
	 * @param sounds a sequence of sounds from a pre-learned word, to be found amongst the sounds of the word FS
	 * @return true if the word FS can be constrained to have the sounds of the pre-learned word
	 */
	private boolean constrainSounds(FSNode word,Vector<String> sounds)
	{
		boolean constrainable  = false;
		Vector<String> allSounds = new Vector<String>();
		if (word.getSounds().size()> 0) allSounds  = word.getSounds().get(0);
		
		// main criterion for being able to constrain the word sounds - they include the pre-learnt word sounds
		if (Util.includes(allSounds, sounds))
		{
			constrainable = true;
			Vector<FSNode> newChildren = new Vector<FSNode>();
			int ps = 0; // position in the sounds
			for (int pw = 0; pw < word.getChildren().size();pw++)
			{
				FSNode child = word.getChild(pw);
				// ignore sound scenes of the word after the last sound of the sequence
				if ((child.getType().equals("sound")) && (ps < sounds.size()))
				{
					// look for a match of the first sound of sounds, the a later sound
					boolean matched = (sounds.get(ps).equals(child.getSlotValue("sound")));
					if (matched)
					{
						newChildren.add(child);
						ps++; //step on in the sounds
					}
					
					// ignore sound scenes before the start of the sequence; and flag any mismatch after an initial sound match
					// (if the initial sound occurs in the word before the matching sequence, this can cause failure; it is rare)
					else if (ps > 0) constrainable = false;
				}
				
				// add the final meaning scene
				else if (child.getType().equals("scene")) newChildren.add(child);
			}
			
			// replace all child nodes of the word
			if (constrainable) word.setChildren(newChildren);
		}
		
		return constrainable;
	}
	
	/**
	 *  remove any input meaning scenes and subsumption links from the word FS
	 *  (which may have been found from input meaning scenes in the learning example FS)
	 * @param wordFS
	 */
	private void makeWordNonProductive() 
	{
		Vector<FSNode> newChildren = new Vector<FSNode>();
		
		// accumulate input sounds only - no input meanings
		for (int i = 0; i < wordFS.getChildren().size() - 1; i++) // not the meaning result
		{
			FSNode child = wordFS.getChild(i);
			if (child.getType().equals("sound")) newChildren.add(child);
		}
		
		// add the meaning result, with all links removed
		FSNode newResult = wordFS.getLastChild();
		
		if (newResult != null)
		{
			newResult.removeAllLinks();
			newChildren.add(newResult);
		}
		else Util.message("word has no child nodes when making it unproductive");
		
		wordFS.setChildren(newChildren);
	}
	
	
	//--------------------------------------------------------------------------
	//	             Second Constructor - Promote to more productive word
	//--------------------------------------------------------------------------
	

	
	/**
	 * constructor to make a more productive form of a word
	 * @param generator
	 * @param l1 learning example
	 * @param l2 learning example
	 * @param w existing (less productive) word
	 * @param side = 0,1 for new input meaning scene before, after sounds
	 * @throws Exception
	 */
	public LearntWord(ExGenerator generator, LearningExample l1, LearningExample l2, LearntWord w, int side) throws Exception
	{
		this.generator = generator;
		Util.message(" Making a more productive version of word " + w.getShortProdName());
		
		// for debugging
		int[] startProds = w.getProductivity();
		int gross = startProds[0] + startProds[1];
				
		// choose the correct input FS from each learning example
		int ex1 = l1.promoter(w, side); // known to be 1 or more
		int ex2 = l2.promoter(w, side); // known to be 1 or more
		FSNode ex1FS = l1.getUnifyResult(ex1);
		FSNode ex2FS = l2.getUnifyResult(ex2);
				
		// generalise the two example feature structures (finding subsumption links)
		FSNode newWord = Ops.generalise(ex1FS,ex2FS);
		newWord.clearExpressions();
		
		// clean up the subsumption links
		boolean accepted = newWord.cleanLinks();
		
		// trim the sounds of the new word to be only the sounds of the old word, if it can be done
		FSNode oldWord = w.getWord();
		boolean trimmable = trimSounds(oldWord,newWord);
		
		// the trimmed word sound must be contained in one part of the partition, for each learning example
		accepted = accepted && (l1.isInOnePart(newWord) && l2.isInOnePart(newWord));

		// generalisation should not lose too much information
		boolean strict = true;
		if ((!l1.isGuessed())&&(!l2.isGuessed())) strict = false;
		accepted = accepted && (trimmable && acceptResult(oldWord,newWord,strict));
		
		// promote the old word to have higher productivity
		FSNode promoted = promote(oldWord,newWord);
		accepted = accepted && (promoted != null);
				
		// final check of subsumption links
		accepted = accepted && promoted.checkWordLinks();
		
		if (accepted)
		{
			viable = true;
			
			// register the word
			wordFS = promoted;
			
			// import the last N successful learning examples from the less productive word. N = 6
			importPositions(w);
			
			// add the two latest examples to the list of example positions; 
			ExampleWord xw1 = new ExampleWord(l1,this,l1.getFSCount() - 1);
			xw1.setSuccess(true);
			ExampleWord xw2 = new ExampleWord(l2,this,l2.getFSCount() - 1);
			xw2.setSuccess(true);
			
			// record them
			examplePositions.add(xw1);
			examplePositions.add(xw2);			
			
			// record the example positions (for this word) of the examples which gave the latest promotion
			promotedExNo[0] = examplePositions.size();
			promotedExNo[1] = l1.getExNo();
			
			// record the word name and sounds in instance variables
			spacedWordName = wordFS.getSpacedName();
			if (wordFS.getSounds().size() == 0)
			{
				Util.message("--- word '" + spacedWordName + "' , which appeared to be made successfully, has no sounds");
				Util.message(wordFS.getTextTree(0));
				wordSounds = new Vector<String>();
				wordSounds.add("word with no sounds");
				viable = false;
			}
			else wordSounds = wordFS.getSounds().get(0);
			wordFS.setName(spacedWordName);
			
		}
	}
	
	/**
	 * debugging promotion
	 * @param in1
	 * @param in2
	 * @param res
	 */
	private void traceFS(FSNode in1, FSNode in2, FSNode res)
	{
		Util.message("\nPromote test: ");
		Util.message("Input 1: " + in1.getTextTree(0));
		Util.message("Input 2: " + in2.getTextTree(0));
		Util.message("Generalisation: " + res.getTextTree(0));
	}
	
	
	/**
	 * promote oldWord to higher productivity, using newWord which is the result of generalising 2 learning examples.
	 * newWord has one more subsumption link than oldWord.
	 * This link comes from one new input meaning scene, at child position pos.
	 * 
	 * oldWord has output meaning scene A, and newWord has output meaning scene C. A must subsume C. i.e. A > C
	 * 
	 * In the promoted version of oldWord:
	 * 
	 *  - add a link target to one node in A (use C to find out which node 'a' in A to add it to)
	 *  - alter the new input meaning scene in newWord, changing its subtree under the link source to 'a', so the link will be valid in oldWord
	 *  - add the new input meaning scene from newWord to oldWord, in the same position pos
	 *  
	 * @param oldWord
	 * @param newWord
	 * @return promoted version of oldWord
	 */
	private FSNode promote(FSNode oldWord, FSNode newWord) throws Exception
	{
		// make the promoted word from a clone of the old word
		FSNode promoted = new FSNode(oldWord);
		
		// old and new word output meaning scenes
		FSNode A = promoted.getLastChild();
		FSNode C = newWord.getLastChild();
		// I don't know why this is ever necessary
		if (C == null) return null;
		// set all true positions on C, so the node 'a' in A with a new link target can be identified
		C.setAllTruePositions();

		// check that the old word output meaning subsumes the new word output meaning
		boolean reduce = Ops.subsumes(A, C);
		if (!reduce) return failureMessage("--- Meaning of the old less productive word does not subsume new word meaning.",oldWord,newWord);
		
		// find all link sources and targets in oldWord and newWord
		Hashtable<String,FSNode[]> oldLinks = oldWord.linkPairs();
		Hashtable<String,FSNode[]> newLinks = newWord.linkPairs();
		int nNew = newLinks.size();
		int nOld = oldLinks.size();
		if (nNew != (nOld + 1)) return failureMessage("--- New word has " + nNew + " links, but old word has " + nOld + " links.",oldWord,newWord);
					
		// find the id of the new link from its source input meaning tree in newWord
		int newSourcePos = newChildPos(oldWord,newWord);
		if (newSourcePos == -1) return null;
		FSNode newInput = newWord.getChild(newSourcePos);
		if (newInput.getLinkRoots(true).size() == 0) return failureMessage("--- Cannot find the new link sources",oldWord,newWord);

		FSNode newLinkRoot  = newInput.getLinkRoots(true).get(0);
		String newLinkId = newLinkRoot.getSubsumingId();
		if (newLinkId == null) return failureMessage("--- Cannot find the id of the new link source",oldWord,newWord);
		
		// find the new link target 'c' in C
		FSNode c = newLinks.get(newLinkId)[1];
		if (c == null) return failureMessage("--- Could not find a link target node 'c' in the new word meaning",oldWord,newWord);
		
		// find the subtree 'a' of the old word meaning A which corresponds to the link target c in the new word meaning 
		FSNode a = FSNode.findPairedNode(C, A, c); // a is not a clone
		if (a == null) return failureMessage("--- Could not find a node 'a' in the old word meaning to put the new link target on",oldWord,newWord);

		FSNode aClone = new FSNode(a);
		a.setSubsumedId(newLinkId); // link target set in A in the promoted word
		aClone.setSubsumingId(newLinkId); // link source will be reset in the new input meaning scene newInput

		// replace newLinkRoot (the part of the input meaning scene newInput with a link source) by a(cloned) so the link will still work
		boolean replace = newInput.replaceBy(newLinkRoot, aClone);
		if (!replace) return failureMessage("--- Failed to replace source of new subsumption link",oldWord,newWord);
		
		// add the new input meaning scene to the promoted word
		Vector<FSNode> newChildren = new Vector<FSNode>();
		for (int cc = 0; cc < promoted.getChildren().size();cc++)
		{
			if (cc == newSourcePos) newChildren.add(newInput);
			newChildren.add(promoted.getChild(cc));
		}
		promoted.setChildren(newChildren);
						
		// successful completion
		Util.message("--- Promotion to " + newLinks.size() + " links succeeded");
		
		return promoted;
	}
	
	/**
	 * explain the reason for any failure in promotion
	 * @param reason
	 * @param f1
	 * @param f2
	 * @return
	 */
	private static FSNode failureMessage(String reason, FSNode f1,FSNode f2)
	{
		// short message
		Util.message(reason);
		
		// long message, with feature structures
		// Util.message(reason + "\n" + f1.getTextTree(0)+ "\n" + f2.getTextTree(0));
		
		return null;
	}
	
	/**
	 * 
	 * @param oldWord
	 * @param newWord
	 * @return the position pos of the new input meaning scene in newWord, compared with oldWord
	 * @throws Exception
	 */
	private int newChildPos(FSNode oldWord, FSNode newWord)
	{
		// the new word should have 1 more child node than the old
		int cNew = newWord.getChildren().size();
		int cOld = oldWord.getChildren().size();
		if (cNew != (cOld +1))
		{
			Util.message("--- Old word has " + cOld + " child nodes; new word has " + cNew + ", not " +(cOld + 1));
			// Util.message(oldWord.getTextTree(0) + "\n" + newWord.getTextTree(0));
			
			return -1;
		}
		
		/* there are only two cases: 
		 * (a) the new meaning scene is before the sounds at scene 0; or 
		 * (b) it is after the sounds at scene cOld - 1 */
		int pos = 0;
		if (oldWord.getChild(0).getType().equals(newWord.getChild(0).getType())) pos = cOld - 1;
		
		return pos;
	}

	
	/**
	 *  trim the sounds of the new promoted word to be only the sounds of the old less productive word
	 * @param oldWord
	 * @param newWord
	 * @return true if trimming is possible
	 */
	private boolean trimSounds(FSNode oldWord,FSNode newWord) throws Exception
	{
		if (newWord.getSounds().size() == 0) 
		{
			Util.message("--- Generalisation result has no sounds before trimming");
			return false;
		}
		
		Vector<String> sounds = oldWord.getSounds().get(0);
		Vector<String> moreSounds = newWord.getSounds().get(0);
		if (!Util.includes(moreSounds, sounds)) 
		{
			Util.message("--- Promoted word sounds " + Util.serial(moreSounds) +  " do not include base word sounds " + Util.serial(sounds));
			return false;
		}		

		// find the position index where sounds starts in moreSounds
		int start = Util.startIndex(moreSounds, sounds);
		int sPos = 0;
		Vector<FSNode> newChildren  = new Vector<FSNode>();
		
		for (FSNode child: newWord.getChildren())
		{
			if (child.getType().equals("scene")) newChildren.add(child);
			else if (child.getType().equals("sound"))
			{
				// copy over only the sound nodes in list sounds
				if ((sPos >= start) && (sPos < start + sounds.size())) newChildren.add(child);
				sPos++;
			}
		}
		
		newWord.setChildren(newChildren);
		
		if (newWord.getSounds().size() == 0) 
		{
			Util.message("--- Generalisation result has no sounds after trimming");
			return false;
		}
		
		return true;
	}

	
	/**
	 * import the latest up to N successful example positions from word w
	 * (N = 6)
	 * @param w
	 */
	private void importPositions(LearntWord w)
	{
		int maxImport = 6;
		int imp = 0;
		Vector<ExampleWord> back  = new Vector<ExampleWord>();
		Vector<ExampleWord> pns = w.getExamplePositions();
		
		// make a backward list of up to N successful example positions in w
		for (int p = 0; p < pns.size();p++)
		{
			ExampleWord xw = pns.get(pns.size() - p - 1);
			if ((xw.succeeds()) && (imp < maxImport + 1))
			{
				back.add(xw);
				imp++;
			}
		}
		
		// reverse the backward list into this word's list
		for (int b = 0; b < back.size();b++)
		{
			ExampleWord ex = back.get(back.size() - b - 1);
			examplePositions.add(ex);
		}
	}

	
	//--------------------------------------------------------------------------
	//	                   adding a learning example
	//--------------------------------------------------------------------------
	
	/**
	 * try to add a new learning example, and if it is new, try to generalise it with the current word
	 * @param example
	 * 
	 * @return a boolean array [used, accepted]
	 * If used = true, the example has been used before, and will not be added again
	 * if accepted = true, the example has not been used before, and matches the meaning of the word enough
	 * 
	 * @throws Exception
	 */
	public boolean[] addExample(LearningExample lex, int fsNo) throws Exception
	{
		/* first element of return array; true if the example is already used in this word, 
		 * and so the example will not be added */
		boolean alreadyUsed = checkExample(lex);
		
		// second element of return array; can only be set true if alreadyUsed is false
		boolean accepted = false;
		
		if (!alreadyUsed)
		{			
			// do the generalisation		
			FSNode example = lex.getUnifyResult(fsNo);
	
			FSNode newWord = Ops.generalise(wordFS, example);
			newWord.clearExpressions();
			
			// for debugging, to inspect meaning trees
			String s = wordFS.getLastChild().getTextTree(0);
			String t = example.getLastChild().getTextTree(0);
			String u = "generalise failure";
			try {u = wordFS.getLastChild().getTextTree(0);} catch(Exception ex) {}
			
			// check that the generalisation does not increase information content of the word
			checkInfo(example,wordFS,newWord);
			
			//clean up the subsumption links
			accepted = newWord.cleanLinks();
						
			// the word sound must be contained in one part of the partition of the learning example
			accepted = accepted && lex.isInOnePart(newWord);			
			
			// does the generalisation result have an acceptable reduction of meaning information
			boolean strict = lex.isGuessed();
			accepted = accepted && acceptResult(wordFS,newWord,strict);
			
			// final check of subsumption links
			accepted = accepted && newWord.checkWordLinks();
			
			ExampleWord xw;
			if (accepted)
			{
				wordFS = newWord; // change the word
				xw = new ExampleWord(lex, this, fsNo);
				xw.setSuccess(true);
			}			
			else xw = new ExampleWord(lex, this, fsNo);
			
			// add the example position, as the example has not been previously used for this word
			examplePositions.add(xw);
			// checkExamplePositions(s,t,u);

		}
				
		boolean[] result = new boolean[2];
		result[0] = alreadyUsed;
		result[1] = accepted;
		return result;
	}
	
	/**
	 * check various things that should not happen
	 * when adding an example
	 */
	private void checkExamplePositions(String sm ,String tm, String um)
	{
		boolean fault = false;
		int s = examplePositions.size();
		ExampleWord xw1 = examplePositions.get(s-1);
		ExampleWord xw2 = examplePositions.get(s-2);
		String ws = Util.serial(wordSounds);
		
		// increase in information content
		if (xw1.getInfo() > xw2.getInfo())
		{
			fault = true;
			Util.message("(1):In word " + ws + " example " + xw1.getExNo() + " information " + xw1.getInfo() + " is more than example " + xw2.getExNo() + " information " + xw2.getInfo());
		}
		
		// recovering from a loser (lost slots from real word)
		if ((!xw1.isLoser()) && (xw2.isLoser()))
		{
			fault = true;
			Util.message("(2):In word " + ws + " example " + xw1.getExNo() + " recovers full word meaning from a loser");
		}
		
		// making a new loser, not from a distractor meaning
		if ((xw1.isLoser())  && (!xw2.isLoser()) && (!xw1.isDistractor()))
		{
			fault = true;
			Util.message("(3):In word " + ws + " example " + xw1.getExNo() + " is not a distractor, but loses slots from the real word");
		}
		
		/*
		if (fault)
		{
			Util.message("\nexample " + tm);
			Util.message("\nword " + sm);
			Util.message("\ngen " + um);
		}
		*/
	}

	
	/**
	 * check that the generalisation does not increase information content
	 * @param example
	 * @param word
	 * @param newWord
	 */
	private void checkInfo(FSNode example,FSNode word,FSNode newWord)
	{
		double iword = word.getLastChild().getInformation(generator.getInfoCount())*10.0;
		double inew = 0.0;
		if (newWord.getLastChild() != null)
			inew = newWord.getLastChild().getInformation(generator.getInfoCount())*10.0;
		
		// by generalisation, this should never happen
		if (inew > iword + 0.0001)
		{
			Util.message(example.getTextTree(0));
			Util.message("\nInformation in old word " + iword);
			Util.message(word.getTextTree(0));
			Util.message("\nInformation in new word " + inew);
			Util.message(newWord.getTextTree(0));

			Util.message("Stopping point");
		}
		
	}

	
	
	/**
	 *  check if an example has been used already
	 * @param example
	 * @return
	 */
	private boolean checkExample(LearningExample lex)
	{
		//check for examples already used in any version on this word , successfully or not
		boolean alreadyUsed = false;		
		for (ExampleWord xw: examplePositions)
			if (xw.getExNo() == lex.getExNo()) alreadyUsed = true;

		return alreadyUsed;
	}
	

	
	/**
	 * 
	 * @param oldWord
	 * @param newWord
	 * @param strict: if true, use strict criterion; if false use loose criterion
	 * @return
	 */
	public boolean acceptResult(FSNode oldWord, FSNode newWord, boolean strict)
	{
		// set the information criterion as strict or loose
		int criterion = generator.getStartCriteria()[1];
		if (strict) criterion = generator.getStartCriteria()[0];
			
		// calculate the information in the new result
		double info = 0.0;
		FSNode newMeaning = newWord.getLastChild();
		if (newWord.getChildren().size() > 0) info = newMeaning.getInformation(generator.getInfoCount());
		boolean accept = info > criterion;
		
		// compare with previous generalisation result
		if ((oldWord != null) && (accept))
		{
			// if N slots are lost, and the remaining information content is > 4.0*N, it is acceptable
			FSNode oldMeaning = oldWord.getLastChild();
			Vector<String> diffSlots = FSNode.getSlotDifference(oldMeaning, newMeaning);
			Vector<String> lostSlots = new Vector<String>();
			for (String s:diffSlots) if (s.startsWith("+")) lostSlots.add(s);
			accept = (info > 4.0*(lostSlots.size()));
		}
		
		return accept;

	}
	
	
	
	//---------------------------------------------------------------------------------------------
	//                       Improve this word by permutation learning
	//---------------------------------------------------------------------------------------------
	
	/**
	 * 
	 * @return a better version of this word, 
	 * learnt by making random permutations of recent learning examples, 
	 * and choosing the most popular/successful version learnt from them;
	 * 
	 * or null if a better version cannot be found
	 */
	public LearntWord permutationLearn(int cycleNo) throws Exception
	{
		// candidates to replace this LearntWord, grouped by word meaning string
		Hashtable<String,Vector<LearntWord>> candidates = new Hashtable<String,Vector<LearntWord>>();
		
		// string form of the meaning and productivity of this word, to compare with meanings of candidates
		String serialMeaningString = getWord().getLastChild().getTextTree(0) + productivityString();
		// success score of the re-learnt word which has the same meaning and productivity as this word, if there is one
		int matchedSupport = 0;
		
		// collect the last permutable examples - avoiding likely distractors
		Vector<LearningExample> examples = new Vector<LearningExample>();
		int min = examplePositions.size() - generator.permutable;
		if (min < 0) min = 0;
		for (int e = min; e < examplePositions.size(); e++)
		{
			LearningExample lex = examplePositions.get(e).getLex();
			if (!lex.isGuessed()) examples.add(lex);
		}
		
		// sometimes there are not enough examples
		int nExamples = examples.size();
		if (nExamples < 10) return null;
		
		// make random permutations of the examples, not making more than 1000 attempts to start a word (they might all fail)
		int permNo = 0;
		int tryNo = 0;
		while ((permNo < generator.permutations) && (tryNo < 1000))
		{
			tryNo++;
			int[] perm = Util.randomPermutation(nExamples);

			LearningExample l1 = examples.get(perm[0]);
			LearningExample l2 = examples.get(perm[1]);
			Vector<LearningExample> firstTwo = new Vector<LearningExample>();
			firstTwo.add(l1);
			firstTwo.add(l2);
			
			// if the first two chosen learning examples have no words shared except the word being learned
			if (l1.sharedsounds(l2).size() == 1)
			{
				LearntWord cand = new LearntWord(generator, firstTwo, wordSounds);
				
				if (cand.isViable())
				{
					permNo++;
					// refine the candidate with the remaining examples, permuted
					for (int e = 2; e < examples.size();e++)
					{
						LearningExample lex = examples.get(perm[e]);
						//as the candidate is unproductive, the non-unified version of the example will do
						cand.addExample(lex, 0);
					}
					
					// collect candidates, grouped by the string form of their meaning and productivity
					String meaning = cand.getWord().getLastChild().getTextTree(0) + cand.productivityString();
					Vector<LearntWord> cands = candidates.get(meaning);
					if (cands == null) cands = new Vector<LearntWord>();
					cands.add(cand);
					candidates.put(meaning, cands);
				}
				
			}			
			
		}
		
		
		// start a summary line for the word
		int serialWordMeaningInfo = (int)(10*wordFS.getLastChild().getInformation(generator.getInfoCount())); // serial learnt meaning information content
		int serialWordSupport = (int)(bayesSupport(nExamples)*10.0); // support of the serially learned version of the word, from the number of examples the permuted words have

		int trueMeaningInfo = 0; // true word meaning content
		if (realWord() != null) trueMeaningInfo = (int)(10*realWord().getLastChild().getInformation(generator.getInfoCount()));
		String summary = "\n" + getSpacedProdName() + " (s" + serialWordSupport + ",m" + serialWordMeaningInfo + ",tm" + trueMeaningInfo + "): ";

		int trueMeaningGroupSize = 0; // the size of the group that has the true word meaning
		int trueMeaningSupport = 0; //average support of the group with the true meaning
		int maxAvgGroupSupport = -100; // the highest average support of any group

		//choose the best candidate, which has the highest average support of any group with the same meaning, of group size >= 4 
		LearntWord bestWord = null;
		int bestWordMeaningInfo = -100;
		for (Enumeration<String> en = candidates.keys(); en.hasMoreElements();)
		{
			String groupMeaningString = en.nextElement();
			Vector<LearntWord> cands = candidates.get(groupMeaningString);
			// only count meanings that occur in more than some minimum number of  permutations
			if (cands.size() >= 4)
			{
				double sumSupport = 0.0; // sum to compute average support in the group
				for (LearntWord cand:cands)
				{
					// build the average support for the group
					double support = cand.bayesSupport()*10.0;
					sumSupport = sumSupport + support;
					
				} // end of loop over words in the group
				
				// information content of meaning
				int groupMeaningInfo = (int)(10*cands.get(0).getWord().getLastChild().getInformation(generator.getInfoCount()));

				// average support for the group
				int avgGroupSupport = (int)(sumSupport/cands.size());

				// if the group has the best average support of any group
				if (avgGroupSupport > maxAvgGroupSupport) 
				{
					maxAvgGroupSupport = avgGroupSupport;
					bestWord = cands.get(0);
					bestWordMeaningInfo = groupMeaningInfo;
				}

				// note this group's support, if the group meaning matches the serially learned word meaning
				if(groupMeaningString.equals(serialMeaningString)) matchedSupport = avgGroupSupport;				
				
				// if this group has the same meaning as the true word (measured by information content)
				if (groupMeaningInfo == trueMeaningInfo)
				{
					trueMeaningGroupSize = cands.size(); // record the size of the group
					trueMeaningSupport = avgGroupSupport;// record the average support of this group
				}
				
				summary = summary + " [s"  + avgGroupSupport + ",n"+ cands.size() + ",m" + groupMeaningInfo + "], ";
				
			} // end of if(group is large enough)
			
		} // end of loop over groups
		
		int diff = maxAvgGroupSupport - serialWordSupport;
		Util.message(summary + "; increase " + diff + " from " + nExamples + " examples");
		if (matchedSupport != 0) Util.message("Matched " + matchedSupport + ": best permuted " + maxAvgGroupSupport);
		
		boolean trueMeaningHasBestSupport = (maxAvgGroupSupport == trueMeaningSupport);
		Util.message("Group size with correct meaning: " + trueMeaningGroupSize + "; has best support of any group: " + trueMeaningHasBestSupport);
		
		// if the best group has better support than the serially learnt word, and different meaning, record the cycle
		if ((diff > 0) && (bestWordMeaningInfo != serialWordMeaningInfo))
		{
			permutationCycles.add(Integer.valueOf(cycleNo));
			bestWord.setPermutationCycles(permutationCycles);
		}
		else bestWord = null;		
		
		return bestWord;
	}
	
	
	
	//---------------------------------------------------------------------------------------------
	//
	//---------------------------------------------------------------------------------------------
	
	
	
	/**
	 * explanatory power = information content of result * support from good examples
	 * @return
	 */
	public double explanPower()
	{
		return  getInfo()*((double)getExampleCount());				
	}
	
		
	/**
	 * the probability of the word rule being correct, 
	 * calculated from the Bayesian support;
	 *  
	 * near 1.0 when support is very positive
	 * near 0.0 when support is very negative
	 * 0.5 when support is 0.0
	 * 
	 * with a scale factor to make it not too sharp
	 * @return
	 */
	public double getProbability()
	{
		double scale = 12.0;
		double support = bayesSupport()/scale;
		double positive = Math.pow(2.0, support);
		double negative  = 1.0/positive;
		return (positive/(positive + negative));
	}
	
	/**
	 * 
	 * @return level of Bayesian support
	 */
	public double bayesSupport()
	{
		return bayesSupport(examplePositions.size());
	}

	
	/**
	 * 
	 * @return level of Bayesian support, 
	 * taking account of only some maximum number of examples
	 */
	public double bayesSupport(int maxExamples)
	{
				
		// newer decay algorithm, trying to be more fair for rare words
		// - depends on the gap in examples for this word
		// sum the number of successful examples, with exponential decay for old examples
		double exampleSum = - generator.threshold;
		// do not go back more than permutable (= 20?) examples
		int min = Math.min(maxExamples,generator.permutable);
		for (int i = 0; i < min;i++)
		{
			// work backwards from the latest example, increasing the gap i
			int j = examplePositions.size() - i - 1;
			ExampleWord xw = examplePositions.get(j);
			// if the example was successful
			if (xw.succeeds())
			{
				// age of each successful example, in examples for this word
				double gap = (double)i;

				/* exponential decay 1/e in a gap of expLife examples of this word; 
				 * expLife is 40, scaled up for more productive words */
				int meaningScenes = wordFS.getInputMeaningScenes().size() + 1;
				// extend lifetime for examples where the word is isolated
				if (xw.isolated()) meaningScenes = 3*meaningScenes;
				double quotient = (double)(generator.expLife*meaningScenes);
				double factor = Math.exp(-gap/quotient);
				
				exampleSum = exampleSum + factor;
			}
		}
		
		
		// return info*exampleSum2;
		return exampleSum;
	}
	
	/**
	 *  level of support by valid learning examples
	 * @return
	 */
	public int getExampleCount()
	{
		int ex = 0;
		for (ExampleWord pos:examplePositions) if (pos.succeeds()) ex++;
		return ex;
	}
	
	//--------------------------------------------------------------------------------------
	//                         writing out evidence of the learning process
	//--------------------------------------------------------------------------------------	

	/**
	 * write a summary of the word and the evidence for it, to a list of lines to go to a file
	 */
	public void writeEvidence(Vector<String> learntWordLines)
	{
		
		// summary and text tree of word feature structure
		learntWordLines.add("\n" + getSpacedWordName() + " (" + getProductivity()[0] + "," + getProductivity()[1] + "): " 
				+ ": successes per cycle  " + expPerCycleString() + "; tree:\n" + wordFS.getTextTree(0));
		
		// cycles in which permutation learning was used successfully
		if (permutationCycles.size() > 0)
		{
			String perms = "Permutation learning improvement at cycles: ";
			for (Integer ii:permutationCycles) perms = perms + ii.toString() + ", ";
			learntWordLines.add(perms);
		}
		
		// string summary of all examples used
		String examples = "";
		// times the example number decreases. Allow up to 2, for promotions; otherwise it is caused by permutation learning
		int reversals = 0;
		int previousExNo = -1;
		// to check that word  information content does not increase
		int lastInfo = 1000;
		boolean increase = false;
		// to insert line breaks in the example log
		int breaker = 0;
		
		// list all examples used
		for (int i = 0; i < examplePositions.size() ; i++) 
		{
			ExampleWord xw = examplePositions.get(i);
			
			// if examples are not in ascending order on more than 2 occasions, they have been permuted
			if (xw.getExNo() < previousExNo) reversals++;;
			previousExNo = xw.getExNo();
			
			// test if word meaning information ever increases when refining with an example 
			if (xw.getInfo() > lastInfo) increase = true;
			lastInfo = xw.getInfo();
			
			// write the example
			examples = examples + xw.stringForm() + ", ";
			
			// line break every 6 examples
			breaker++;
			if (breaker == 6)
			{
				examples = examples + "\n";
				breaker = 0;
			}
		}

		String permuted = ": ";
		if (reversals > 2) permuted = "(permuted): ";

		// other information, followed by the list summary of the examples used
		learntWordLines.add("Noun strength: " + getNounStrength() + "; support " + bayesSupport() 
				+ "; promoted by example (" + getpromotedExNo()[0] + ","  + getpromotedExNo()[1] + ")"
				+ "\n" + examplePositions.size()  + " examples" + permuted + examples);
		
		if (increase) learntWordLines.add("*** Meaning information increased on refinement  - this should not happen.");
		
	}
	
	
	/**
	 * 
	 * @param example
	 * @return true if the learnt word can be unified with this example,
	 * because its sound appears in one of the unused sound segments of the example,
	 * and the example has sufficient meaning scenes before and after the word
	 */
	public boolean canApply(FSNode example)
	{
		Vector<String>  sv = getWordSounds();
		boolean apply  = false;
		
		// one of the segments in the example must contain the word name, in a form like b-o-y
		for (Vector<String> sw: example.getSounds())
			if (Util.includes(sw, sv)) apply = true;
		
		/* only apply an example to a learnt word if the example has 
		 * at least as many input meaning scenes before and after the word sound as the word has */
		int[] exLink = example.meaningSceneCount(sv);
		int[] wLink = getWord().meaningSceneCount(sv);
		for (int i = 0; i < 2; i++) if (exLink[i] < wLink[i]) apply = false;

		return apply;
	}
	

	/**
	 * @return the noun strength of the learnt word
	 */
	public double getNounStrength()
	{
		FSNode meaning = wordFS.getLastChild();
		double nounInfo = meaning.getNounInformation(generator.getInfoCount());
		int things = meaning.thingNodes();
		// transitive verbs can have two things in their meaning
		if (things != 1) return 0.0;
		return nounInfo;
	}

	
	/**
	 * 
	 * @return true if the word has any verb-like qualities:
	 * - act = true on the scene node
	 * - more than one thing node
	 * Do not make it verb-like from just an m_role.
	 */
	public boolean isVerbLike()
	{
		boolean verbLike = false;
		
		FSNode meaning = wordFS.getLastChild();
		String act = meaning.getSlotValue("act");
		if ((act != null) && (act.equals("true"))) verbLike = true;
		
		int things = 0;
		for (FSNode child: meaning.getChildren())
			if (child.getType().equals("thing")) things++;
		if (things == 2) verbLike = true;
		
		return verbLike;
	}
	
	
	/**
	 * 
	 * @return a string like '[3/4, 6/6,... ]'
	 * of successful examples / all examples for each cycle
	 * in ascending order of cycles
	 */
	public String expPerCycleString()
	{
		String s = "[";
		for (int[] examples: examplesPerCycle())
			s = s +  examples[0] + "/" + examples[1] + ", ";
		s = s + "]";
		return s;
	}
	
	/**
	 * 
	 * @return for each cycle so far, an int[2] of
	 * - number of successful examples in the cycle
	 * - number of examples in the cycle
	 */
	public Vector<int[]> examplesPerCycle()
	{
		Vector<int[]> examples = new Vector<int[]>();
		
		for (int c = 0; c < generator.getCycleNo(); c++)
		{
			int min = generator.exPerCycle*c;
			int max = generator.exPerCycle*(c + 1);
			int[] exes = new int[2];
			for (int i = 0; i < 2; i++) exes[i] = 0;
			
			for (ExampleWord exUse: examplePositions)
			{
				if ((exUse.getExNo() > min - 1) && (exUse.getExNo() < max))
				{
					exes[1]++;
					if (exUse.succeeds()) exes[0]++;
				}
			}
			examples.add(exes);
		}				
		return examples;
	}
	

}
