package mainPack;

import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import displayPack.Actions;
import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.LanguageModel;
import utilPack.Util;


public class LanguageServlet   extends HttpServlet{

	// this is to satisfy the compiler
	final static long serialVersionUID = 123457;
	
	private LanguageModel model;
	
	private ExGenerator generator;
	
	private Actions act;
	
	private boolean initialised = false;



    
    //---------------------------------------------------------------------------------------
	//                       Mechanics of the connection:
    //                     Branching to provide different types of service
    //---------------------------------------------------------------------------------------
	
    /**
     *  web service got by GET to addresses like 'langDemo?function=start' ,
     */
	public void doGet(HttpServletRequest request,
            HttpServletResponse response)
    		throws ServletException, IOException 
    {
		String function = request.getParameter("function");
		Util.message("received request " + function);
		
		// this should not happen, and I do not understand why it does happen from time to time
		if (!initialised)
		{
			init();
			Util.message("Needed to re-initialise the language model on receiving start request from the client - reason unknown.");
		}
		
		try
		{
			// will be needed for any response
			Document responseDoc = Util.makeOutDoc();

			if (function.startsWith("start")) makeStartResponse(responseDoc);
			
			// generate - client request is 'generateN'
			else if (function.startsWith("generate"))
			{
				int sentIndex = getIndex(function.substring("generate".length()));
				Vector<Vector<FSNode>> result = act.generateOrUnderstandForClient(sentIndex, true);
				act.makeResponse(responseDoc, result,"generate");
				Util.message("generate response sent");
			}
			
			// understand - client request is 'understandN' or 'understand.word1.word2'
			else if (function.startsWith("understand"))
			{
				Vector<Vector<FSNode>> result  = new Vector<Vector<FSNode>>();
				String toUnderstand = function.substring("understand".length());
				
				// user-composed sentence
				if (toUnderstand.startsWith("."))
				{
					result = act.understandComposed(toUnderstand);					
				}
				
				// menu-selected sentence
				else
				{
					int sentIndex = getIndex(toUnderstand);
					result = act.generateOrUnderstandForClient(sentIndex, false);
				}
				
				act.makeResponse(responseDoc, result,"understand");
				Util.message("understand response sent");
			}
			
			// learn - client request is 'learnMwordNcycleP' M = learning stage; N = word number in stage; P = cycle number
			else if (function.startsWith("learn"))
			{
				String part = function.substring("learn".length()); // 'MwordNcycleP'
				StringTokenizer st = new StringTokenizer(part,"word"); // tokens 'M' and 'NcycleP'
				int stageIndex = getIndex(st.nextToken()); // M
				StringTokenizer su = new StringTokenizer(st.nextToken(),"cycle"); // tokens 'N' and 'P' 
				int wordIndex = getIndex(su.nextToken()); // N
				int cycleNo = getIndex(su.nextToken()); // P
				Vector<Vector<FSNode>> result = act.learnForClient(stageIndex,wordIndex,cycleNo);
				act.makeResponse(responseDoc, result,"learn");
			}
			
			// learning cycle data - client request is 'cycleN'
			else if (function.startsWith("cycle"))
			{
				String cNo = function.substring("cycle".length());
				int cycleNo = Integer.parseInt(cNo);
				makeCycleResponse(responseDoc,cycleNo);				
			}

			// send the response
			sendResponse(response,responseDoc,function);							
		}
		catch (Exception ex) {Util.message("Exception processing GET: " + ex.getMessage());ex.printStackTrace();}

    }
	
	/**
	 * 
	 * @param s
	 * @return
	 * @throws Exception
	 */
	private int getIndex(String s) throws Exception
	{
		return Integer.parseInt(s);
	}
	
	
    /**
     * normal method to handle any request on the service
     */
    public void doPost(HttpServletRequest request,
            HttpServletResponse response)
    		throws ServletException, IOException 
    {
    	doGet(request,response);
    }
 
	
	/**
	 * 
	 */
	private void sendResponse(HttpServletResponse response,Document responseDoc,String function) throws IOException,Exception
	{
		response.setHeader("Cache-Control", "no-cache");
		response.setHeader("Pragma", "no-cache");

		
		// XML response to all requests (not formatted)
		{
			response.setContentType("text/xml");			
	        ServletOutputStream so = response.getOutputStream();
	        Util.writeToStream(responseDoc, so, false);
		}
		Util.message("Server sent response with top element tag name " + Util.getLocalName(responseDoc.getDocumentElement()));
	}
	
	
	//------------------------------------------------------------------------------------------------
	//                        Initialisation
	//------------------------------------------------------------------------------------------------
	
	
	/**
	 * done before the first request from the client
	 */
	public void init()
	{
		Util.message("Server initialising model");
		try
		{
			// set up the language model, with a null VFrame
			model = new LanguageModel(null,"");
			
			// after reading the model data, the model can return data to the client. Read from the data class, not a file
			model.readModel(true);
			
			// set up the actions class, with null VFrame and drawArea
			act = new Actions(model);
			
			// create an example generator, to send recorded learning cycle data to the web client
			boolean forWeb = true;
			generator = new ExGenerator(model,forWeb);
			
			// set up the data in class Learner needed to replay word learning for the web client
			generator.setupLearningCycles();
			
			// tell the Actions instance about the Learner
			act.setLearner(generator.learner);
			
			// note that initialisation has happened
			initialised = true;
		}
		catch(Exception ex) {ex.printStackTrace();};
	}
	

	
	//------------------------------------------------------------------------------------------------
	//                         Building responses to requests
	//------------------------------------------------------------------------------------------------
	
	/**
	 * make the response document to the start request
	 * @param doc
	 */
	private void makeStartResponse(Document doc)
	{
		if (model != null)
		{
			try
			{
				Element modelRoot = model.clientModel(doc);
				doc.appendChild(modelRoot);
			}
			catch (Exception ex) {Util.message("Starting exception: ");ex.printStackTrace();}
		}
		else Util.message("no language model");
	}
	
	/**
	 * make the response document to a request for data from a cycle of auto-learning
	 * @param doc an empty document
	 * @param cycleNo
	 */
	private void makeCycleResponse(Document doc, int cycleNo)
	{
		if (generator != null)
			try
			{
				Element cycleRoot = generator.makeCycleData(doc,cycleNo);
				doc.appendChild(cycleRoot);
			}
			catch (Exception ex) {Util.message("Exception getting cycle data: ");ex.printStackTrace();}
		else Util.message("no example generator");
	}


}
