package modelPack;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import utilPack.Util;

public class FSNode {
	
	//----------------------------------------------------------------------------------
	//                        Instance variables
	//----------------------------------------------------------------------------------
	
	// properties of any node
	private LanguageModel model;
	private String type;
	private int depth;
	private Hashtable<String,String> ownSlotValues = new Hashtable<String,String>();
	
	// tree structure
	private Vector<FSNode> children = new Vector<FSNode>();
	private FSNode parent;
	private int childIndex = 0;
	private int truePosition = 0; // position from a tree walk
	
	// relevant only on top script nodes
	private String name;
	private String description = null;
	private Vector<int[]> freeOrder = new Vector<int[]>();
	private Vector<Vector<String>> segments = null;
	private int[] infoContent = null; // information content in inputs and final meaning scene
	
	// operations - unify, generalise, subsume
	private String unifyOrder = null;
	private boolean isResult = false;
	private boolean isUsed = false;
	private String expression = null;
	private boolean viable = true;
	private int[] origin = null;
	private String genId = "";
	private int fullListPosition = -1;
	
	// subsumption links
	private String subsumedId = null;
	private String subsumingId = null;
	private String fromId = null;
	private String fromSubsumesId = null;
	
	// language demos
	private WordToLearn wordLearnExampleParent = null;
	private WordToLearn wordLearnToLearnParent = null;
	
	// drawing - to send to the web client 
	private int row = 0; // drawn row for step of unify or generalise
	private int column = 0; // 0, 1, 2 for 2 inputs and result
	
	// drawing - dynamic 
	private int highlightedPosition = -1;
	private Hashtable<String,int[]> linkEnds = new Hashtable<String,int[]>();	
	private int fontWidth = 7;  // initial setting	 - will be reset
	private int fontDrop = 16; // initial setting- will be reset
	
	// drawing - static
	private static String[] MAIN_NODETYPES = {"script","scene","sound","thing","prop","place","path","manner"};	// boxes for main node types (other node types can be added)		
	private static Color[] boxColours = {Color.WHITE,Color.ORANGE,Color.YELLOW,Color.GREEN,Color.BLUE,Color.CYAN,Color.CYAN,Color.LIGHT_GRAY}; // node box colours	
	private static int[] hGaps = {50,30,0,20,20,20,20,20}; // horizontal spaces for links to child nodes	
	private static int[][] typeSizes = {{12,7,16},{16,8,18},{18,9,20}}; // for each type size an array [font size, horizontal pixels per letter, vertical pixels per line]	
	private static int boxHeight = 20; // height of boxes
	private static int gapHeight = 8; // gap between FS heights 
		
	// editing
	private boolean edited = false;
	
	// subsumption graph, for experiments with better learning
	private Vector<FSNode> subChildren = new Vector<FSNode>();
	public Vector<FSNode> getSubChildren() {return subChildren;}
		
	//--------------------------------------------------------------------------
	//                        Constructors
	//--------------------------------------------------------------------------
	
	/**
	 * basic constructor for a FS with one top node (of an allowed type) and nothing else
	 * @param topNodeType
	 */
	public FSNode(LanguageModel model, String topNodeType) throws Exception
	{
		this.model = model;
		if (model.getNodeType(topNodeType) == null) 
			throw new Exception("Node type '" + topNodeType + ", is not in the model");
		type = topNodeType;
	}
	
	/**
	 * Cloning Constructor, used for Unification and Generalisation and copy/paste
	 * @param f
	 */
	public FSNode(FSNode f)
	{
		// clone all properties of the top node
		cloneTopNode(f);
		
		// clone child FS
		for (FSNode fc: f.getChildren())
		{
			//clone the child
			FSNode fcc = new FSNode(fc);
			
			// add it to the children, giving them parent links
			addChild(fcc);		
		}
	}
	
	// types of cloning constructor
	public static int FULL = 0;
	public static int REMOVEUSED = 1;
	public static int REMOVELAST = 2;
	
	/**
	 * cloning constructor, removing used nodes
	 * @param f
	 * @param removeUsed (is ignored; used only to distinguish this constructor)
	 */
	public FSNode(FSNode f, int type)
	{
		// clone all properties of the top node
		cloneTopNode(f);
		
		// clone child FS, if it is not used
		int c = 0;
		for (FSNode fc: f.getChildren()) 
		{
			boolean keepChild = false;
			if (type == FULL) keepChild = true;
			if ((type == REMOVEUSED) && (!fc.isUsed())) keepChild = true;
			if ((type == REMOVELAST) && (c < f.getChildren().size() -1)) keepChild = true;

			//clone the child
			FSNode fcc = new FSNode(fc);
			
			// add it to the children, giving them parent links
			if (keepChild) addChild(fcc);
			
			c++;
		}
	}
	
	/**
	 * to apply the result of a subsumption link unification
	 * clone  f, except for the node which has the same subsuming or subsumed id as g 
	 * - which is to be replaced by g
	 * 
	 * UNDERSTAND - find the node in f that has a subsumed id, and replace it by the node in g that has the same subsuming id
	 * GENERATE - find the node in f that has a subsuming id, and replace it by the node in g that has the same subsumed id

	 * @param f
	 * @param g
	 */
	public FSNode(FSNode f, FSNode g, int type) throws Exception
	{
		String fid = null;
		String gid = null;
		
		if (type == Ops.UNDERSTAND)
		{
			fid  = f.getSubsumedId();
			gid = g.getSubsumingId();
		}
		
		if (type == Ops.GENERATE)
		{
			fid  = f.getSubsumingId();
			gid = g.getSubsumedId();
		}
		
		if (gid == null) throw new Exception("Cannot find subsumption link on head node to be substituted");
			

		if ((fid != null) && (fid.equals(gid)))
		{
			// clone all properties of the top node of g
			cloneTopNode(g);
			
			// clone child FS of g
			for (FSNode gc: g.getChildren())
				addChild(new FSNode(gc));		
		}
		
		else
		{
			// clone all properties of the top node
			cloneTopNode(f);
			
			// clone child FS, checking for the id of g
			for (FSNode fc: f.getChildren())
				addChild(new FSNode(fc,g,type));		
		}
	}
	
	/**
	 * 
	 * @param f
	 */
	private void cloneTopNode(FSNode f)
	{
		model = f.getModel();
		type = f.getType();
		name = f.getName();
		depth = f.getDepth();
		subsumedId = f.getSubsumedId();
		subsumingId = f.getSubsumingId();
		childIndex = f.getChildIndex();
		isResult = f.isResult();
		viable = f.isViable(); 
		isUsed = f.isUsed();
		genId = f.getGenId();
		expression = f.getExpression();
		description = f.getDescription();
		truePosition = f.getTruePosition();
		origin = f.getOrigin();
		freeOrder = f.getFreeOrder();
		fullListPosition = f.getFullListPosition();
		infoContent = f.getInfoContent();
		
		// cannot simply clone the parent ?
		
		// do not set 'from' fields which are temporary
		
		// clone all slots
		for (Enumeration<String> en = f.getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			ownSlotValues.put(slotName, f.getSlotValue(slotName));
		}
	}
	
	/**
	 * replace an old node by a new node, 
	 * the first time it occurs with identical slot values to the old node in its subtree,
	 * with child nodes assumed to be in the same order
	 * (cannot yet replace the whole FSNode this way; can only replace parts of the tree)
	 * @param oldNode
	 * @param newNode
	 * @return true if a replacement is made
	 */
	public boolean replaceBy(FSNode oldNode,FSNode newNode)
	{
		
		boolean replace = false;
		Vector<FSNode> newChildren = new Vector<FSNode>();
		
		// make only one replacement somewhere in the subtree
		for (FSNode child: children)
		{
			// if there has been a replacement already, pass across the child
			if (replace) newChildren.add(child);

			else if (!replace)
			{
				// if the child equals the old node, replace it
				if (child.slotEquals(oldNode)) 
				{
					newChildren.add(newNode);
					replace = true;
				}
				
				// otherwise, look for matches below the child
				else 
				{
					boolean cRep = child.replaceBy(oldNode, newNode);
					if (cRep) replace = true;					
					newChildren.add(child);
				}
			}
		}
		
		children = new Vector<FSNode>();
		for (FSNode newChild: newChildren) addChild(newChild);
		
		return replace;
	}
	
	/**
	 * 
	 * @param other
	 * @return true if this node has equal slot values to other,
	 * and there are equal slot values on all descendants
	 */
	public boolean slotEquals(FSNode other)
	{
		boolean equals = false;
		
		// check equality of slot values
		if (ownSlotValues.size() == other.getOwnSlotValues().size())
		{
			equals = true;
			for (Enumeration<String> en = ownSlotValues.keys(); en.hasMoreElements();)
			{
				String slotName = en.nextElement();
				String val = ownSlotValues.get(slotName);
				String otherVal = other.getOwnSlotValues().get(slotName);
				if ((otherVal == null)||(!(val.equals(otherVal)))) equals = false;
			}
		}
		
		if (children.size() == other.getChildren().size())
		{
			for (int c = 0; c < children.size();c++) if (equals)
				equals = equals && getChild(c).slotEquals(other.getChild(c));
		}
		
		else equals = false;
		
		return equals;
	}


	/**
	 * constructor from XML input
	 * All feature structures have the same name as their top-level feature structure
	 * @param model
	 * @param name
	 * @param structureEl
	 */
	public FSNode(LanguageModel model, String name, Element structureEl, int depth, FSNode parent) throws Exception
	{
		viable = true;
		this.model = model;
		this.name = name;
		type = structureEl.getTagName();
		this.depth = depth;
		this.parent = parent;

		Hashtable<String,String> allValues = Util.getAllAttributes(structureEl);

		//save values of non-slots as instance variables, and remove them from the table
		treatNonSlots(allValues);
		ownSlotValues = allValues;
		
		validateSlots(); // null means no parent FS
		
		for (Element childEl:Util.childElements(structureEl))
			children.add(new FSNode(model,name,childEl,depth + 1,this));
		
		for (int i = 0; i < children.size();i++) children.get(i).setChildIndex(i);
		
	}
	
	
	//--------------------------------------------------------------------------
	//                   Reading from XML
	//--------------------------------------------------------------------------

	
	/**
	 * XML attributes that are not treated as ordinary slots
	 * @param allValues
	 */
	private void treatNonSlots(Hashtable<String,String> allValues) throws Exception
	{
		String desc = allValues.get("description");
		if (desc != null)
		{
			description = desc;
			allValues.remove("description");
		}
		
		String id = allValues.get("id");
		if (id != null)
		{
			subsumedId = id;
			allValues.remove("id");
		}
		
		String subsumes = allValues.get("subsumes");
		if (subsumes != null)
		{
			subsumingId = subsumes;
			allValues.remove("subsumes");
		}
		
		String result = allValues.get("result");
		if (result != null)
		{
			isResult = true;
			allValues.remove("result");
		}
		
		String unifyOrder = allValues.get("unifyOrder");
		if (unifyOrder != null)
		{
			this.unifyOrder = unifyOrder;
			allValues.remove("unifyOrder");
		}
		
		String freeOrderString = allValues.get("freeOrder");
		if (freeOrderString != null)
		{
			setFreeOrder(getFreeOrderGroups(freeOrderString));
			allValues.remove("freeOrder");
		}
				
		String infoContentString = allValues.get("infoContent");
		if (infoContentString != null)
		{
			StringTokenizer st = new StringTokenizer(infoContentString,"_");
			int i1 = Integer.parseInt(st.nextToken());
			int i2 = Integer.parseInt(st.nextToken());
			setInfoContent(i1, i2);
			allValues.remove("infoContent");
		}		

		
		// non-slots that are only to send to the web client; not to read from a data file
		String[] webClientFields = {"segments","origin","row","column","truePosition"};
		for (String nonSlot: webClientFields)
		{
			String nonSlotString = allValues.get(nonSlot);
			if (nonSlotString != null) allValues.remove(nonSlot);
		}
	}
	
	public void setFreeOrder(Vector<int[]> order) {freeOrder = order;}
	
	
	
	//--------------------------------------------------------------------------
	//                   Writing to XML
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param doc
	 * @return XML representation of the feature structure
	 */
	public  Element xmlForm(Document doc,boolean forClient) throws Exception
	{
		// root element named after the type
		Element root = Util.newElement(doc, getType());
		
		// attributes for slot values
		for (Enumeration<String> en = ownSlotValues.keys();en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String slotValue = ownSlotValues.get(slotName);
			root.setAttribute(slotName, slotValue);
		}
		
		// attributes for non-slot values
		writeNonSlots(root,forClient);
		
		// child FS
		for (FSNode child: children) root.appendChild(child.xmlForm(doc,forClient));
				
		return root;
	}
	
	/**
	 * 
	 * @param root
	 */
	private void writeNonSlots(Element root, boolean forClient) throws Exception
	{
		// for general saving
		if (!empty(subsumedId)) root.setAttribute("id", subsumedId);
		if (!empty(subsumingId)) root.setAttribute("subsumes", subsumingId);
		if (isResult) root.setAttribute("result", "yes");
		if (!empty(unifyOrder)) root.setAttribute("unifyOrder", unifyOrder);	
		if (!empty(description)) root.setAttribute("description", description);	
		if (freeOrder.size() > 0) root.setAttribute("freeOrder", freeOrderString());
		if (getSegments().size() > 0) root.setAttribute("segments", segmentString());
		
		// for sending to the web client
		if (forClient)
		{
			if (!empty(expression)) root.setAttribute("expression", expression);
			if (isUsed) root.setAttribute("used", "true");
			if (!isViable()) root.setAttribute("unviable", "true");
			if (origin != null) root.setAttribute("origin", origin[0] + "_" + origin[1]);
			if (infoContent != null) root.setAttribute("infoContent", infoContent[0] + "_" + infoContent[1]);
			root.setAttribute("row", Integer.toString(row));
			root.setAttribute("column", Integer.toString(column));
			root.setAttribute("truePosition", Integer.toString(truePosition));
		}
		
	}
	
	/**
	 * 
	 * @return
	 */
	public String freeOrderString()
	{
		String freeString = "";
		for (int[] sequence: freeOrder)
			freeString = freeString + "(" + sequence[0] + "-" + sequence[sequence.length - 1] + ") ";
		return freeString;
	}
	
	/**
	 * 
	 * @param s
	 * @return
	 */
	private boolean empty(String s)
	{
		if (s == null) return true;
		if (s.length() == 0) return true;
		return false;
	}
	
	/**
	 * string version of sound segments th,u,/b,o,y,/
	 * in a format that is easy to split in the client
	 * @return
	 */
	private String segmentString() throws Exception
	{
		String segString = "";
		Vector<Vector<String>> segs = getSegments();
		for (Vector<String> seg: segs)
		{
			for (String sou: seg) segString = segString + sou + ",";
			segString = segString + "|";
		}		
		return segString;
	}

	/**
	 * segments of sound, for compact drawing, and for sending in XML to the client to draw.
	 * @return
	 */
	public Vector<Vector<String>> getSegments() throws Exception
	{
		// segments are only applicable to script nodes
		if (!(type.equals("script"))) return new Vector<Vector<String>>();
		
		// for the results of unification and generation, the segments have been set in the operations
		if (segments != null) return segments;
		
		// for words, sentences and learning examples, segments are set dynamically
		Vector<Vector<String>> segs = new Vector<Vector<String>>();
		
		// sentences (all sounds are in words to be unified) and learning examples (some sounds are in words to be unified)
		if (unifyOrder != null)
		{
			StringTokenizer st = new StringTokenizer(unifyOrder,", ");
			while (st.hasMoreTokens())
			{
				String wordName = st.nextToken();
				FSNode word = model.getWords().get(wordName);
				if (word == null) 
				{
					// Util.message("When making segments from unifyOrder, cannot find word " + wordName);
				}
				// if the word has several segments, the sentence inherits them
				else for (Vector<String> wSeg: word.getSegments()) segs.add(wSeg);
			}
		}
		
		// words, and learning examples with no words known
		if (unifyOrder == null)
		{
			Vector<String> seg = new Vector<String>();
			// form a segment for each contiguous sequence of sounds
			for (FSNode child:children)
			{
				// accumulate sounds from sound child nodes (which have sounds defined)
				if (child.getType().equals("sound")) 
				{
					String ss = child.getSlotValue("sound");
					if (ss != null) seg.add(ss);
				}
				// for any other type of child node (scene), store any non-empty segment and start a new one
				else if (seg.size() > 0)
				{
					segs.add(seg);
					seg = new Vector<String>();
				}
			}
			// no scene after the last sound (hardly ever happens)
			if (seg.size() > 0) segs.add(seg);
		}

		
		return segs;
	}
	
	//--------------------------------------------------------------------------
	//                   Validation
	//--------------------------------------------------------------------------




	/**
	 * check that all slots and values are valid
	 * @throws Exception
	 */
	public void validateSlots() throws Exception
	{
		String nodeType = this.getType();
		NodeType type = model.getNodeType(nodeType);
		if (type == null) noteError("Node type is not allowed in the model.");
		
		if (parent != null)
		{
			String parentType = parent.getType();
			NodeType pNodeType = model.getNodeType(parentType);
			if (!pNodeType.isValidChild(nodeType)) 
				noteError("Node type cannot be a child of node type " + parentType);
		}
		
		for (Enumeration<String> en = ownSlotValues.keys();en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String slotValue = ownSlotValues.get(slotName);
			
			{
				Slot sl = model.getSlot(slotName);
				if (sl == null) noteError("Invalid slot '" + slotName + "'");
				else if (sl != null)
				{
					if (!sl.hasValue(slotValue)) 
						noteError("Invalid value '" + slotValue + "' for slot '" + slotName + "'");
					if (!sl.isAllowedNodeType(getType())) 
						noteError("Slot " + slotName + " cannot appear on node of this type ");
				}
						
			}
		}
		
		// check all descendants
		for (FSNode child: children)  child.validateSlots();
	}
	
	/**
	 * note validation errors for later display
	 * @param error
	 */
	private void noteError(String error)
	{
		String fullError = "In FS '" + getName() + "' on a " + getType() + " node: " + error;
		model.addError(fullError);
	}

	
	/**
	 * Decode and check the order constraints in the attribute 'order'
	 */
	public Vector<int[]> getFreeOrderGroups(String freeOrderString) throws Exception
	{
		Vector<int[]> freeOrder = new Vector<int[]>();
		if (freeOrderString != null)
		{
			StringTokenizer brackets = new StringTokenizer(freeOrderString,"()",true);
			if (brackets.countTokens() < 3) throw new Exception("Ranges for free orders must be in brackets (..)");
			boolean inBracket = false;
			while (brackets.hasMoreTokens())
			{
				String symb = brackets.nextToken();
				if ((symb.equals("("))||(symb.equals(")"))) inBracket = !inBracket;
				else if (inBracket)
				{
					StringTokenizer nums = new StringTokenizer(symb,"- ");
					if (nums.countTokens() != 2) throw new Exception("Ranges in brackets must be of the form 'x-y', not " + symb);

					int min = 0;
					int max = 0;
					try 
					{
						Integer iVal = Integer.decode(nums.nextToken());
						min = iVal.intValue();
						Integer jVal = Integer.decode(nums.nextToken());
						max = jVal.intValue();
					}
					catch (Exception ex) {throw new Exception("Free order ranges must use integers, not '" + symb + "'");}
					if ((max < min)|(min < 1)|(max > children.size())) throw new Exception("Invalid free order range " + symb);

					int[] oneConstraint = new int[max - min + 1];
					for (int i = 0; i < max - min + 1; i++) oneConstraint[i] = i + min;
					freeOrder.add(oneConstraint);
				}
			}
		}
		return freeOrder;
	}
	
	//--------------------------------------------------------------------------
	//                      Various methods
	//--------------------------------------------------------------------------
	
	/**
	 * Vector of words to unify to understand a sentence
	 * @return
	 */
	public Vector<String> getUnifyWords()
	{
		Vector<String> res = new Vector<String>();
		if (unifyOrder != null)
		{
			StringTokenizer st = new StringTokenizer(unifyOrder,", ");
			while (st.hasMoreTokens())
			{
				String word = st.nextToken();
				res.add(word);
			}
		}
		
		 return res;
	}
	
	/**
	 * 
	 * @return an FS with only the sound child nodes
	 */
	public FSNode soundOnly()
	{
		FSNode sounds = new FSNode(this);		
		sounds.keepSounds(true);		
		return sounds;
	}
	
	 /** 
	 * @return an FS with no sound nodes, but meaning nodes retained
	 */
	public FSNode meaningOnly()
	{
		FSNode meaning = new FSNode(this);		
		meaning.keepSounds(false);		
		return meaning;
	}
	
	/**
	 * find the start position of a string of sounds in a script FSNode
	 * @param sounds
	 * @return
	 * @throws Exception
	 */
	public int getStartPosition(Vector<String> sounds) throws Exception
	{
		int start = -1;
		boolean matched = false;
		
		// try all possible values of the start position of the sequence of sounds
		for (int s = 0; s < children.size() - sounds.size(); s++) if (!matched)
		{
			matched = true;
			for (int i = 0; i < sounds.size(); i++)
			{
				FSNode child = children.get(s+i);
				if (child.getType() != "sound") matched = false;
				else if (!sounds.get(i).equals(child.getSlotValue("sound"))) matched = false;
			}
			if (matched) start = s;
		}
		
		if (start == -1) throw new Exception("could not match sound");
		return start;
	}
	
	/**
	 * remove sound scenes with child index first - last inclusive
	 * and replace them by one sound '*'
	 * @param first
	 * @param last
	 */
	public void removeSounds(int first, int last) throws Exception
	{
		if (!"script".equals(getType())) throw new Exception("Can only call 'removeSounds' for a script");
		Vector<FSNode> newChildren = new Vector<FSNode>();
		int c = 0;
		for (FSNode child:children)
		{
			if ((c < first)||(c > last)) newChildren.add(child);
			else if (c == first)
			{
				FSNode sound = new FSNode(model,"sound");
				sound.setSlotValue("sound", "*");
				newChildren.add(sound);
			}
			c++;
		}
		setChildren(newChildren);		
	}
	
	/**
	 * 
	 */
	public void keepSounds(boolean keepSound)
	{
		Vector<FSNode> newChildren = new Vector<FSNode>();
		for (FSNode child: children)
		{
			if (child.getType().equals("sound")) {if (keepSound) newChildren.add(child);}
			else if (!keepSound) newChildren.add(child);
		}
		
		children = new Vector<FSNode>();
		for (FSNode child:newChildren) addChild(child);
	}
	
	/**
	 * for scripts only, return a vector of vectors of the unused sound
	 * sequences, separated by meaning scenes
	 * @return
	 */
	public Vector<Vector<String>> getSounds()
	{
		Vector<Vector<String>> sounds = new Vector<Vector<String>>();
		Vector<String> current = new Vector<String>();
		
		for (FSNode child: children)
		{
			// capture the current sequence of unused sounds, and start a new one
			if (child.getType().equals("scene"))
			{
				if (current.size() > 0) sounds.add(current);
				current = new Vector<String>();
			}
			
			// build up the current sequence of unused sounds
			else if ((child.getType().equals("sound")) && (!child.isUsed()))
				current.add(child.getSlotValue("sound"));			
		}
		
		// probably this never happens, because the last scene is a meaning scene
		if (current.size() > 0) sounds.add(current);
				
		return sounds;
	}
	

	
	/**
	 * get the position of this node 1..N in a tree walk of the whole FS
	 * @return
	 */
	public int getPosition()
	{
		Vector<FSNode> fullWalk = getRootAncestor().treeWalk();
		int pos = 0;
		for (int i = 0; i < fullWalk.size(); i++)
			if (fullWalk.get(i).equals(this)) pos = i + 1;
		return pos;
	}
	
	/**
	 * 
	 * @return a tree walk of this FS and its descendants
	 */
	public Vector<FSNode> treeWalk()
	{
		Vector<FSNode> walk = new Vector<FSNode>();
		walk.add(this);
		
		for (FSNode child:children)
		{
			Vector<FSNode> childWalk = child.treeWalk();
			for (FSNode cw: childWalk) walk.add(cw);
		}		
		return walk;
	}
	
	/**
	 * 
	 * @param treePosition
	 * @return the node at a given tree position in a full script FS
	 */
	public FSNode getNode(int treePosition)
	{
		FSNode result = null;
		Vector<FSNode> fullWalk = getRootAncestor().treeWalk();
		
		if ((treePosition > 0) && (treePosition < fullWalk.size() + 1))
			result = fullWalk.get(treePosition - 1);
		
		return result;
	}
	
	
	//--------------------------------------------------------------------------
	//                      Information  content
	//--------------------------------------------------------------------------
	
	/**
	 * set the information content of (1) all but the final meaning scene; (2) the final meaning scene
	 * @param i1
	 * @param i2
	 */
	public void setInfoContent(int i1, int i2)
	{
		int[] content = new int[2];
		content[0] = i1;
		content[1] = i2;
		infoContent = content;		
	}
	
	/*
	 * get the information content of (1) all but the final meaning scene; (2) the final meaning scene
	 */
	public int[] getInfoContent() {return infoContent;}
	
	/**
	 * 
	 * @return information content of this FS and its descendants,
	 * counting information in all slots
	 */
	public double getInformation(Hashtable<String,Double> infoCount)
	{
		double info = getSlotInformation(false,infoCount);
		
		for (FSNode child: children)
			info = info + child.getInformation(infoCount);
		
		return info;
	}
	
	/**
	 * 
	 * @return information content of this FS and its descendants
	 * counting information only in preferentially noun-like slots (t_type and identity)
	 */
	public double getNounInformation(Hashtable<String,Double> infoCount)
	{
		double info = getSlotInformation(true,infoCount);
		
		for (FSNode child: children)
			info = info + child.getNounInformation(infoCount);
		
		return info;
	}
	
	/**
	 * for a script FS
	 * @return information in all the meaning scenes of a productive word
	 */
	public double getMeaningInformation(Hashtable<String,Double> infoCount)
	{
		double info = 0.0;
		
		// ignore the final output meaning scene
		for (int i = 0; i < getChildren().size() -1; i++)
		{
			FSNode child = getChild(i);
			if (child.getType().equals("scene"))
				info = info + child.getInformation(infoCount);
		}
		
		return info;
	}
	
	
	/**
	 * 
	 * @return number of thing nodes
	 */
	public int thingNodes()
	{
		int things = 0;
		if (getType().equals("thing")) things = 1;
		for (FSNode child: children) things = things + child.thingNodes();
		return things;
	}
	
	/**
	 * @param nounSlotsOnly if true, count information only on preferentially noun-like slots
	 * @param infoCount if non-null, measured information count per slot value
	 * @return information content of the slots of this FS
	 */
	private double getSlotInformation(boolean nounSlotsOnly,Hashtable<String,Double> infoCount)
	{
		double info = 0.0;
		
		for (Enumeration<String> en = ownSlotValues.keys();en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			if ((!nounSlotsOnly)||(isNounSlot(slotName)))
			{
				String slotValue = ownSlotValues.get(slotName);
				if (infoCount == null)
				{
					Slot slot = model.getSlot(slotName);				
					info = info + slot.getInformation(slotValue);
				}
				else if (infoCount != null)
				{
					String svp = slotName + ":" + slotValue;
					Double inc = infoCount.get(svp);
					if (inc != null) info = info + inc.doubleValue();
				}
			}
		}
		
		return info;
	}
	
	/**
	 * 
	 * @param slotName
	 * @return true if the slot is one which preferentially appears on nouns
	 */
	private boolean isNounSlot(String slotName)
	{
		boolean noun = false;
		String[] nSlots = {"t_type","identity"};
		for (String ns:nSlots) if (slotName.equals(ns)) noun = true;		
		return noun;
	}

	/**
	 * 
	 * @return list of all slit:value pairs
	 */
	public Vector<String> getAllSlots()
	{
		Vector<String> allSlots = new Vector<String>();
		
		for (Enumeration<String> en = getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String value = getOwnSlotValues().get(slotName);
			String svp = slotName + ":" + value;
			allSlots.add(svp);
		}
		
		for (FSNode child:getChildren())
		{
			Vector<String> cSlots = child.getAllSlots();
			for (String s:cSlots) allSlots.add(s);
		}
		
		return allSlots;
	}
	
	/**
	 * prepare to write out the differences in slot:value pairs of two FSNodes, to a file
	 * @param f1
	 * @param f2
	 * @return int[3] array:
	 * number of extra slots
	 * number of missing slots
	 * bayesian support of word (will be set later)
	 */
	public static int[] writeSlotDifference(FSNode f1, FSNode f2, Vector<String> learntWordLines)
	{
		Vector<String> sl1 = f1.getAllSlots();
		Vector<String> sl2 = f2.getAllSlots();
		int[] dc = new int[3];
		for (int i = 0; i <3; i++) dc[i] = 0;
		
		String diff = "differences: ";
		
		for (String s1:sl1) if (!Util.inVector(s1, sl2)) 
		{
			diff = diff + "+" + s1 + " ";
			dc[0]++;
		}
		for (String s2:sl2) if (!Util.inVector(s2, sl1)) 
		{
			diff = diff + "-" + s2 + " ";
			dc[1]++;
		}
		
		learntWordLines.add(diff);
		
		return dc;
	}

	/**
	 * 
	 * @param f1
	 * @param f2
	 * @return a list of slots that differ
	 */
	public static Vector<String> getSlotDifference(FSNode f1, FSNode f2)
	{
		Vector<String> sl1 = f1.getAllSlots();
		Vector<String> sl2 = f2.getAllSlots();
		
		Vector<String> diff = new Vector<String>();
		
		for (String s1:sl1) if (!Util.inVector(s1, sl2)) diff.add("+" + s1);
		for (String s2:sl2) if (!Util.inVector(s2, sl1)) diff.add("-" + s2);
		
		return diff;
	}
	
	//--------------------------------------------------------------------------
	//              Changing slot names, for global model changes
	//--------------------------------------------------------------------------
	
	/**
	 * change all occurrences of an old slot to a new slot with the same value
	 * @param oldName
	 * @param newName
	 */
	public void changeSlotName(String oldName, String newName)
	{
		String oldSlotValue = getSlotValue(oldName);
		if (oldSlotValue != null)
		{
			ownSlotValues.remove(oldName);
			ownSlotValues.put(newName, oldSlotValue);
		}
		
		for (FSNode child: children) 
			child.changeSlotName(oldName, newName);
	}
	
	//--------------------------------------------------------------------------
	//                       Recursive drawing
	//--------------------------------------------------------------------------


	/**
	 * 
	 * @return colour for the box, depending on its type
	 */
	private Color getBoxColour() 
	{
		Color col = Color.WHITE;
		int t = -1;
		t = Util.indexOf(MAIN_NODETYPES, getType()); // returns -1 if type is not a main type
		if (t > -1) col = boxColours[t];
		return col;
	}
	
	
	/**
	 * 
	 * @param g
	 * @param size
	 * @param bold
	 */
	private void setTypeSize(Graphics g, int size, boolean bold)
	{
		int [] typeSize = null;
		for (int[] ts: typeSizes) 
			if (ts[0] == size) typeSize = ts;
		if (typeSize == null) Util.message("Parameters not yet defined for type size " + size);
		
		else
		{			
			fontWidth = typeSize[1];
			fontDrop = typeSize[2];
			
			if (g != null)
			{
				if (bold) g.setFont(new Font(Font.DIALOG,Font.BOLD,size));
				else g.setFont(new Font(Font.DIALOG,Font.PLAIN,size));
			}
		}		
	}
	
	//---------------------------------------------------------------------------
	//                           actual drawing
	//---------------------------------------------------------------------------
	
	/**
	 * 
	 * @param left
	 * @param top
	 * @param g
	 */
	public void draw(int left, int top, Graphics g, boolean showUsedNodes) throws Exception
	{
		/* test of width calculation
		Util.message(getName() + "-"  + type + ": " + left + "," + top + " " + getDrawingWidth());
		g.drawLine(left, top, left + getDrawingWidth(), top); 
		*/


		// height of the top of the root node
		int rootBoxTop = top + getTopNodeDrop(showUsedNodes);
		// height of the base of the first row of slot value
		int slotTextBase = rootBoxTop + boxHeight;
		// height of the top of the first child FS
		int childTop = top;
		
		boolean isScript = (type.equals("script"));

		// draw a title for the top FS only
		if (isScript)
		{
			// large bold font
			setTypeSize(g,18,true);
			// red name if the FS is being edited
			if (edited) g.setColor(Color.RED);

			int textBase = top + boxHeight - 5;
			g.drawChars(name.toCharArray(),0,name.length(),left + 5,textBase);
			g.setColor(Color.BLACK);
			
			// lower everything to allow for the title text
			rootBoxTop = rootBoxTop + boxHeight + fontDrop;
			slotTextBase = slotTextBase + boxHeight + fontDrop;
			childTop = childTop + boxHeight + fontDrop;
			
			// re-zero the list of link end coordinates,since the last draw
			linkEnds = new Hashtable<String,int[]>();
		}
		
		// small plain font
		setTypeSize(g,12,false);

		// draw the root node box
		int boxWidth = drawRootBox(g, left, rootBoxTop);
		
		// for drawing lines between levels
		int  startX = left + boxWidth;
		int startY = rootBoxTop + boxHeight/2;
				
		// draw the slots below the root node box, left aligned
		int slotLeft = left;
		if (type.equals("sound")) slotLeft = left + boxWidth + 10;
		else slotTextBase = slotTextBase + fontDrop + 2;
		drawSlots(slotLeft,slotTextBase,g);
		
		// draw the  FS children and the lines to them
		int childLeft = left + boxWidth + getHorizGap();
		if (getSlotWidth() > boxWidth) childLeft = left + getSlotWidth() + 10;
		for (FSNode child: getSegmentedChildren(showUsedNodes))
			if ((showUsedNodes)|(!child.isUsed))
			{
				// draw child FS
				child.draw(childLeft, childTop, g, showUsedNodes);
				
				// draw line from root node to child FS root node
				int endX = childLeft;
				int endY = childTop + child.getTopNodeDrop(showUsedNodes) + boxHeight/2; // centre height of root node box of child
				g.setColor(child.lineColour()); // grey if child has been used
				g.drawLine(startX, startY, endX, endY);
				
				//set the top of the next child
				childTop = childTop + child.getDrawingHeight(false,showUsedNodes);			
			}
		
		if (isScript) 
		{
			drawLinkageArcs(g);

			int paraTop = top + getDrawingHeight(false,showUsedNodes) + 60;
			int paraWidth = getDrawingWidth();
			setTypeSize(g,16,true);
			
			// if there is a description, draw it below the diagram with up to the same width
			if (getDescription() != null)
				paraTop = paraTop + drawTextParagraph(left, paraTop, paraWidth, getDescription(), g,true) + 20;
			
			// if there is an expression, draw it below the diagram and the description with up to the same width
			if (getExpression() != null)
				drawTextParagraph(left, paraTop, paraWidth, "Expression: " + getExpression(), g, true);
		}
		
	}
	
	/**
	 * when showUsedNodes is false, return  sound child nodes for drawing in compact segmented form
	 * @param showUsedNodes
	 * @return
	 */
	public Vector<FSNode> getSegmentedChildren(boolean showUsedNodes) throws Exception
	{
		if ((!showUsedNodes) && (getType().equals("script")) && (getSegments().size() > 0))
		{
			Vector<FSNode> segmentedChildren = new Vector<FSNode>();
			
			//segments are usually calculated dynamically, except for operation results
			Vector<Vector<String>> segments = getSegments();
			Vector<FSNode> seg = new Vector<FSNode>();
			for (int c = 0; c < children.size();c++)
			{
				FSNode child = children.get(c);

				// for sound nodes, build up a possible segment  and try to match it against any segment
				if (child.getType().equals("sound"))
				{
					seg.add(child);
					Vector<String> removable = tryToMatch(seg,segments,segmentedChildren);
					if (removable != null) 
					{
						segments.remove(removable);
						seg = new Vector<FSNode>();
					}

				}

				// for non-sound nodes, consume any unmatched sound nodes, then add the non-sound node
				else
				{
					if (seg.size() > 0)
					{
						for (FSNode sou: seg) segmentedChildren.add(sou);
						seg = new Vector<FSNode>();
					}
					segmentedChildren.add(child);
				}
				
			}
			
			// consume any last unmatched sounds (not expected to happen, as the last child is usually non-sound)
			if (seg.size() > 0) for (FSNode sou: seg) segmentedChildren.add(sou);
					
			return segmentedChildren;
		}
		
		else return children;
	}
	
	/**
	 * try to match the current segment seg to any of the segments,
	 * If it matches, add one composite sound node to segmentedChildren, and return the segment to be removed
	 * If there is no match, do nothing.
	 * @param seg
	 * @param segments
	 * @param segmentedChildren
	 */
	private Vector<String> tryToMatch(Vector<FSNode> seg, Vector<Vector<String>>segments, Vector<FSNode> segmentedChildren)
	{
		boolean matched = false;
		Vector<String> removable = null;
		for (Vector<String> segment: segments) if (!matched)
		{
			if (matches(seg,segment))
			{
				FSNode mergedSound = makeOneSoundNode(seg);
				// can be null if the segment matches all used sounds
				if (mergedSound != null) segmentedChildren.add(mergedSound);
				removable = segment;
				matched = true;
			}
		}
		
		return removable;
		
	}
	
	/**
	 * Test if a sequence of sound FSNodes matches a sequence of sound strings
	 * @param seg
	 * @param segment
	 * @return
	 */
	private boolean matches (Vector<FSNode> seg, Vector<String> segment)
	{
		boolean matches = false;		
		if (seg.size() == segment.size())
		{
			matches = true;
			for (int s = 0; s < seg.size(); s++)
			{
				FSNode sou = seg.get(s);
				String ss = sou.getSlotValue("sound");
				if (ss == null) matches = false;
				else if (!(segment.get(s).equals(ss))) matches = false;
			}
		}
		
		return matches;
	}
	
	/**
	 * make one composite sound node, out of all the sounds in a sequence
	 * @param seg
	 * @return
	 */
	private FSNode makeOneSoundNode(Vector<FSNode> seg)
	{
		// cloning is required here! Or it modifies the underlying FSNode
		FSNode newNode = new FSNode(seg.get(0));
		String segSound = "";
		for (FSNode sou: seg) 
			if (!sou.isUsed())
				segSound = segSound + sou.getSlotValue("sound") + ",";

		if (segSound.length() > 0)
		{
			segSound = segSound.substring(0, segSound.length() - 1);
			newNode.setSlotValue("sound", segSound);
			return newNode;
		}
		
		// no unused sounds - return nothing
		else return null;
	}


	
	/**
	 * 
	 * @param g
	 * @param left
	 * @param rootBoxTop
	 * 
	 * @return the width of the box
	 */
	private int drawRootBox(Graphics g, int left, int rootBoxTop)
	{
		// set line colour black, or light grey if isUsed
		g.setColor(lineColour());
				
		// draw the box for the root node with the type name
		int nChars = type.length();
		int boxWidth = fontWidth*(nChars + 1);
		g.setColor(getBoxColour());
		g.fillRect(left, rootBoxTop, boxWidth, boxHeight);
		g.setColor(lineColour());
		g.drawRect(left, rootBoxTop, boxWidth, boxHeight);
				
		// highlight selected node, or node which is an origin of the selected node on another result FS
		FSNode clicked = model.getFrame().getDrawArea().getJustClickedNode();		
		if (((clicked != null) && (clicked.equals(this)))||(highlightedPosition == getPosition()))
			g.drawRect(left - 1, rootBoxTop -1, boxWidth + 2, boxHeight + 2);

		int boxTextBase = rootBoxTop + boxHeight - 5;
		g.drawChars(type.toCharArray(),0,type.length(),left + 5,boxTextBase);
		
		// subsumption links - store start points and end points to draw arcs
		String linkId = "";
		boolean start = true;
		if (getSubsumingId() != null) linkId = getSubsumingId();
		if (getSubsumedId() != null) {linkId = getSubsumedId();start = false;}
		if (linkId.length() > 0) storeLink(linkId,start, left + boxWidth, rootBoxTop + boxHeight/2);
		
		// record this node for mouse clicking
		Rectangle clickable = new Rectangle(left, rootBoxTop,boxWidth, boxHeight);
		model.addNodeLocation(clickable, this);
		
		return boxWidth;
		
		
	}
	
	/**
	 * on the ancestor script node, store the coordinates (x,y) for the 
	 * start or end of a subsumption link
	 * 
	 * @param key key of the subsumption link
	 * @param start if true, the hash key is key + "_start"; if false, the hash key is key + "_end"
	 * @param x right hand x of the node box
	 * @param y halfway down the node nbox
	 */
	private void storeLink(String key, boolean start, int x, int y)
	{
		String hashKey = key + "_start";
		if (!start) hashKey = key + "_end";
		int[] pos = new int[2];
		pos[0] = x;
		pos[1] = y;
		
		getRootAncestor().getLinkEnds().put(hashKey, pos);
	}
	

	
	/**
	 * 
	 * @param left
	 * @param base
	 */
	private void drawSlots(int left, int base, Graphics g)
	{
		int rowDrop = 0;

		for (Enumeration<String> en = ownSlotValues.keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String slotValue = ownSlotValues.get(slotName);
			String slotText = slotName+ ":" + slotValue;

			//simple large layout for sounds 
			if (slotName.equals("sound")) 
			{
				slotText = "'" + slotValue + "'";
				setTypeSize(g, 16, false);
			}
			
			g.drawChars(slotText.toCharArray(),0,slotText.length(),left,base + rowDrop);
			rowDrop = rowDrop + fontDrop;				
			setTypeSize(g, 12, false);
		}
	}
	
	//-----------------------------------------------------------------------------------------
	//                               Draw arcs for links
	//-----------------------------------------------------------------------------------------
	
	/**
	 * draw curved arcs for subsumption links
	 */
	private void drawLinkageArcs(Graphics g)
	{
		for (Enumeration<String> en = linkEnds.keys();en.hasMoreElements();)
		{
			String startKey = en.nextElement();
			if (startKey.endsWith("_start"))
			{
				String endKey = startKey.substring(0,startKey.length() - 6) + "_end";
				int[] startPos = linkEnds.get(startKey);
				int[] endPos  = linkEnds.get(endKey);
				
				if (endPos != null) drawOneArc(g,startPos,endPos);
			}
		}
	}
	
	/**
	 * 
	 * @param g
	 * @param startPos
	 * @param endPos
	 */
	private void drawOneArc(Graphics g,int[] startPos,int[] endPos)
	{
		// find the displacement from start to finish, and the distance (= radius of the circle, for a 60 degree arc)
		int[] disp = new int[2]; // expected [+ or -, +]
		double dSquared = 0.0;
		for (int i = 0; i < 2; i++)
		{
			disp[i] = endPos[i] - startPos[i];
			dSquared = dSquared + disp[i]*disp[i];
		}
		int len = (int)Math.sqrt(dSquared);
		int width = 2*len;
		
		// find the centre of the circle
		// perpendicular bisector of the line from start to finish, with the length from the line to the centre
		int[] perp = new int[2]; // expected [-, + or -]
		double scale = Math.sqrt(3.0)/2.0;
		perp[0] = -(int)(scale*disp[1]);
		perp[1] = (int)(scale*disp[0]);
		int[] centre = new int[2];
		int[] centreToStart = new int[2]; // expected [ +, -]
		for (int i = 0; i < 2; i++)
		{
			centre[i] = startPos[i] + disp[i]/2 + perp[i];
			centreToStart[i] = startPos[i] - centre[i]; 
		}
		int x = centre[0] - len; // x,y required for drawArc
		int y = centre[1] - len;
		
		// find the angle from the centre to the start, away from 3 o'clock
		double tan = - (double)centreToStart[1]/(double)centreToStart[0]; // expected +
		double angle = Math.atan(tan); // expected between 0 and pi/2
		int degrees = (int)(angle*180.0/Math.PI); // expected between 0 and 90
		
		g.setColor(Color.LIGHT_GRAY);
		int thickness = 4;
		for (int t = 0; t < thickness; t++)
			g.drawArc(x + t - thickness/2, y, width, width, degrees, -60);
		g.setColor(Color.BLACK);
	}

	
	//---------------------------------------------------------------------------
	//                   multi-Line text captions to whole FS
	//---------------------------------------------------------------------------
	
	/**
	 * draw text within a width (in pixels)
	 * and return the depth in pixels.
	 * Font size is set outside this method.
	 * 
	 * @param top
	 * @param left
	 * @param width
	 * @param text
	 * @param g
	 * @param doDraw if false, do not draw but return the height
	 * @return
	 */
	private int drawTextParagraph(int left, int top, int width, String text, Graphics g, boolean doDraw)
	{
		int charsPerLine = width/fontWidth;
		
		Vector<String> lines = new Vector<String>();
		String line = "";
		
		// divide the text into lines, breaking on spaces
		StringTokenizer st = new StringTokenizer(text," ");
		while (st.hasMoreTokens())
		{
			String word = st.nextToken();
			String newLine = line + word + " ";
			if (newLine.length() > charsPerLine)
			{
				lines.add(line);
				line  = word + " ";
			}
			else line = newLine;
		}
		lines.add(line);
		
		// write the lines
		if (doDraw)
		{
			g.setColor(Color.BLACK);
			int height = top + fontDrop;
			for (String l: lines)
			{
				g.drawChars(l.toCharArray(),0,l.length(),left,height);
				height = height + fontDrop;			
			}
		}
		
		int depth = lines.size()*fontDrop;
		return depth;
	}
	
	
	//---------------------------------------------------------------------------
	//                           depth computations
	//---------------------------------------------------------------------------
	
			
	/**
	 * 
	 * @return depth of slot box, in pixels
	 */
	private int getSlotDepth() {return getOwnSlotValues().size()*fontDrop;}
	
	/**
	 * 
	 * @return drawing depth in pixels of this FS and all its children
	 * @param addCapton if true, add the height of the caption text below the tree diagram
	 */
	public int getDrawingHeight(boolean addCaption, boolean showUsedNodes) throws Exception
	{
		// special case of sound slots - drawn to the right of the box
		if (type.equals("sound")) return (boxHeight + gapHeight);
		
		// depth of node box  and slot box and vertical gap
		int depth =	boxHeight + getSlotDepth() + 4*gapHeight;
		
		// sum of depths of all child FS
		int childDepth = 0;
		for (FSNode child: getSegmentedChildren(showUsedNodes)) 
			if ((showUsedNodes)|(!child.isUsed))
				childDepth = childDepth + child.getDrawingHeight(false,showUsedNodes);
		
		// take the larger of the two
		if (childDepth > depth) depth = childDepth;
		
		// extra height of caption below a whole Feateur structure
		if (addCaption)
		{
			boolean someText = false;
			int paraWidth = getDrawingWidth();
			setTypeSize(null,16,true);
			
			// if there is a description, add its height below the diagram with up to the same width
			if (getDescription() != null)
			{
				someText = true;
				depth = depth + drawTextParagraph(0, 0, paraWidth, getDescription(), null, false);
			}
			
			// if there is an expression, add its height below the diagram and the description with up to the same width
			if (getExpression() != null)
			{
				someText = true;
				depth = depth + drawTextParagraph(0, 0, paraWidth, "Expression: " + getExpression(), null, false);
			}
			
			// safety first
			if (someText) depth = depth + 80;
						
			setTypeSize(null,12,true);
		}
		
		
		return depth;
	}
	
	/**
	 * 
	 * @return drop in pixels from the top of the FS, to the top of the root node box
	 */
	public int getTopNodeDrop(boolean showUsedNodes) throws Exception
	{
		// if there are no children, there is no drop
		int drop = 0;
		
		// if there are child FS, the centre of the root node box is the mean of the centres of the root node boxes of the child FS
		if (children.size() > 0)
		{
			int denom = 0;
			int total = 0;
			int childTop = 0;
			for (FSNode child: getSegmentedChildren(showUsedNodes)) if ((showUsedNodes)|(!child.isUsed))
			{
				// add the drop to the centre of the child root node
				total = total + childTop + child.getTopNodeDrop(showUsedNodes) + boxHeight/2;
				// top of next child FS
				childTop = childTop + child.getDrawingHeight(false,showUsedNodes);
				
				denom++;
			}
			
			if (denom > 0) drop = (total/denom) - boxHeight/2;
		}
		
		return drop;
	}
	
	
	//---------------------------------------------------------------------------
	//                           width computations
	//---------------------------------------------------------------------------
	
	/**
	 * horizontal gap to child nodes depends on node type, in pixels
	 * @return
	 */
	private int getHorizGap()
	{
		int gap =20; //default
		int typeIndex = -1;
		typeIndex = Util.indexOf(MAIN_NODETYPES, getType()); // stays -1 if thype not in array
		if (typeIndex > -1) gap = hGaps[typeIndex];
		
		return gap;
	}
	
	/**
	 * 
	 * @return width of slot box, in pixels
	 */
	private int getSlotWidth()
	{
		int maxWidth = 0;
		for (Enumeration<String> en = ownSlotValues.keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String slotValue = ownSlotValues.get(slotName);
			String slotText = slotName+ ":" + slotValue;
			if (slotName.equals("sound")) slotText = "'" + slotValue + "'";	
			int slotWidth = fontWidth*slotText.length();
			
			if (slotWidth > maxWidth) maxWidth = slotWidth;			
		}
		return maxWidth;
	}

	
	/**
	 * 
	 * @return width in pixels when drawn
	 */
	public int getDrawingWidth()
	{

		// width of title for top FS only
		int titleWidth = 0;
		if (type.equals("script")) 
		{
			setTypeSize(null,18,true);
			titleWidth = name.length()*fontWidth;
		}
				
		
		// width of the box for the root node with the type name
		setTypeSize(null,12,false);
		int nChars = type.length();
		int boxWidth = fontWidth*(nChars + 1);
		int width = boxWidth;
				
		// if the maximum width of any slot row is wider, use that
		int slotWidth =  getSlotWidth();
		if (slotWidth > width) width = slotWidth;
		
		// maximum width of any child FS
		int maxChildWidth  = 0;
		for (FSNode child: children)
		{
			// track the maximum child width
			int childWidth = child.getDrawingWidth();
			if (childWidth > maxChildWidth) maxChildWidth = childWidth;
		}
		
		// if there are children, add the maximum child width and the horizontal gap
		if (maxChildWidth > 0) width = width + maxChildWidth + getHorizGap();
				
		// if the title is wider than the rest of the FS, use that
		if (titleWidth > width) width = titleWidth;

		return width;
	}
	
	
	
	//--------------------------------------------------------------------------
	//                       subsumption links
	//--------------------------------------------------------------------------
	
	
	/**
	 * starting at any node in the FeatureStructure, get a unique link id
	 * for the next link to be added
	 * @return
	 */
	public String getNextLinkId()
	{
		// search the whole feature structure for existing link ids
		Hashtable<String, String> ids = new Hashtable<String, String>();
		getRootAncestor().addLinkIds(ids);
		
		String linkId = "";
		int index = 1;
		boolean newId = false;
		
		while (!newId)
		{
			linkId = "x_" + index;
			newId = (ids.get(linkId) == null);
			index++;
		}
		
		return linkId;
	}
	
	/**
	 * 
	 * @param ids
	 */
	public void addLinkIds(Hashtable<String, String> ids)
	{
		String id = getSubsumingId();
		if (id == null) id = getSubsumedId();
		if (id != null) ids.put(id, id);
		
		for (FSNode child:children) child.addLinkIds(ids);
	}
	
	/**
	 * remove all subsumption links
	 */
	public void removeAllLinks()
	{
		setSubsumingId(null);
		setSubsumedId(null);
		
		for (FSNode child: children) child.removeAllLinks();
	}
	
	/**
	 * remove both ends of a link - should return 0 or 2
	 * @param linkId
	 */
	public int removeLink(String linkId)
	{
		int rem = 0;
		String id = getSubsumingId();
		if ((id != null) && (id.equals(linkId))) {setSubsumingId(null);rem++;}
		
		id = getSubsumedId();
		if ((id != null) && (id.equals(linkId))) {setSubsumedId(null);rem++;}
		
		for (FSNode child:children) rem = rem + child.removeLink(linkId);
		
		return rem;

	}
	
	/**
	 * true if this node or its descendants has a subsuming link
	 * @return
	 */
	public boolean hasLink()
	{
		boolean link = (getSubsumingId() != null);
		for (FSNode child: children) if (child.hasLink()) link = true;		
		return link;
	}
	
	/**
	 * count the number of subsumption links in to this FS
	 * @return
	 */
	public int getLinksIn()
	{
		int links = 0;		
		if (getSubsumedId() != null) links++;
		
		for (FSNode child:children) links = links + child.getLinksIn();		
		return links;
	}
	
	
	/**
	 * count the number of subsumption links out of this FS
	 * @return
	 */
	public int getLinksOut()
	{
		int links = 0;		
		if (getSubsumingId() != null) links++;
		
		for (FSNode child:children) links = links + child.getLinksOut();		
		return links;
	}
	
	/**
	 * 
	 * @param isSource if true, look for link sources; if false, look for link targets
	 * @return all nodes which are subsumption link sources, or which are link targets
	 */
	public Vector<FSNode> getLinkRoots(boolean isSource)
	{
		Vector<FSNode> roots = new Vector<FSNode>();
		
		// does this node have a link of the kind we are looking for?
		boolean hasLink = ((isSource && (getSubsumingId() != null))||(!isSource && (getSubsumedId() != null)));
		if (hasLink) roots.add(this);
		
		for (FSNode child: children)
		{
			Vector<FSNode> cRoots = child.getLinkRoots(isSource);
			for (FSNode root:cRoots) roots.add(root);
		}
		
		return roots;
	}
	/**
	 * 
	 * @param isSource if true, look for link sources; if false, look for link targets
	 * @return all nodes which are subsumption link sources, or which are link targets
	 */
	public Hashtable<String,FSNode> getLinkRootTable(boolean isSource)
	{
		Hashtable<String,FSNode>  roots = new Hashtable<String,FSNode> ();
		
		// does this node have a link of the kind we are looking for?
		String linkId = null;
		if (isSource) linkId = getSubsumingId();
		else linkId = getSubsumedId();		
		if (linkId != null) roots.put(linkId,  this);
		
		for (FSNode child: children)
		{
			Hashtable<String,FSNode> cRoots = child.getLinkRootTable(isSource);
			for (Enumeration<String> en = cRoots.keys(); en.hasMoreElements();)
			{
				String cid = en.nextElement();
				roots.put(cid, cRoots.get(cid));
			}
		}
		
		return roots;
	}
	
	/**
	 * 
	 * @param linkId
	 * @param isSource
	 * @return the node which is a source or a target with the link id;
	 * or null if none is found
	 */
	public FSNode getLinkRoot(String linkId, boolean isSource)
	{
		FSNode res = null;
		Vector<FSNode> roots = getLinkRoots(isSource);
		
		for (FSNode root: roots)
		{
			String id = root.getSubsumingId();
			if (!isSource) id = root.getSubsumedId();
			if ((id != null) && (id.equals(linkId))) res = root;
		}
		
		return res;
	}
	
	
	
	/**
	 * 
	 * @return a table of link id  => [source tree, target tree]
	 */
	public Hashtable<String,FSNode[]> linkPairs()
	{
		Hashtable<String,FSNode[]> trees = new Hashtable<String,FSNode[]>();
		
		Vector<FSNode> sources = getLinkRoots(true);
		for (FSNode source: sources)
		{
			String key = source.getSubsumingId();
			FSNode[] st = new FSNode[2];
			st[0] = source;
			st[1] = getLinkRoot(key, false);
			trees.put(key,st);
		}
				
		return trees;		
	}
	
	/**
	 * check that subsumption links are correct for a word FS:
	 * 
	 * - each input meaning scene has exactly one link source, with distinct link ids, and no link targets
	 * - the output meaning scene has one link target for each of the input link ids, and no other link sources or targets
	 * - link sources and targets should not be on the top scene nodes
	 * 
	 * @return true if the test passes
	 */
	public boolean checkWordLinks()
	{
		// list of all errors detected
		String errors = "";

		Hashtable<String,FSNode> sources = new Hashtable<String,FSNode>();
		
		// each input meaning scene should have exactly one link source with a distinct key, and no link targets
		for (int i = 0; i < children.size() - 1; i++)
		{
			FSNode input = getChild(i);
			if (input.getType().equals("scene"))
			{
				Hashtable<String,FSNode> st = input.getLinkRootTable(false);
				if (st.size() > 0) errors = errors + " input meaning scene " + i + " has " + st.size() + " link targets; ";

				Hashtable<String,FSNode> ss = input.getLinkRootTable(true);
				if (ss.size() != 1) errors = errors + " input meaning scene " + i + " has " + ss.size() + " link sources; ";
				else for (Enumeration<String> en = ss.keys(); en.hasMoreElements();)
				{
					String lKey = en.nextElement();
					if (sources.get(lKey) != null) errors = errors + " input meaning scene " + i + " has duplicate source key " + lKey + "; ";
					else sources.put(lKey, ss.get(lKey));
				}
				
				if (input.getSubsumingId() != null) errors = errors + " input meaning scene " + i + " has a source key on its top scene node; ";
			}
		}
		
		FSNode output = getLastChild();
		if (output.getType().equals("scene"))
		{
			Hashtable<String,FSNode> so = output.getLinkRootTable(true);
			if (so.size() > 0) errors = errors + " result meaning scene has " + so.size() + " link sources; ";

			Hashtable<String,FSNode> targets = output.getLinkRootTable(false);
			
			// count all discrepancies between sources and targets
			for (Enumeration<String> en = sources.keys(); en.hasMoreElements();)
			{
				String sKey = en.nextElement();
				if (targets.get(sKey) == null) errors = errors + " source key with no target key: " + sKey + "; ";
			}
			
			for (Enumeration<String> en = targets.keys(); en.hasMoreElements();)
			{
				String sKey = en.nextElement();
				if (sources.get(sKey) == null) errors = errors + " target key with no source key: " + sKey + "; ";
			}
			
			if (output.getSubsumedId() != null) errors = errors + " output meaning scene has a target key on its top scene node; ";
		}
		else errors = errors + " last child node is not a meaning scene; ";
		
		if (errors.length() > 0)
		{
			String struct = getTextTree(0);
			// Util.message(struct);
			return false;
		}
				
		return true;
	}
	

	
	/**
	 * clean up subsumption links resulting from generalisation of learning examples:
	 * 
	 * (a) If there are any link sources without targets, or targets without sources, remove them
	 * (b) If there are any link sources within the final meaning scene, remove them and the targets
	 * (c) If there are link sources and targets on top scene nodes, try to demote them
	 * 
	 * return false if (c) cannot be done
	 */
	public boolean cleanLinks() throws Exception
	{
		Hashtable<String,FSNode> sources = getLinkRootTable(true);
		Hashtable<String,FSNode> targets = getLinkRootTable(false);
		
		//clean up sources without targets
		for (Enumeration<String> en = sources.keys();en.hasMoreElements();)
		{
			String id = en.nextElement();
			if (targets.get(id) == null) 
			{
				sources.get(id).setSubsumingId(null);
				sources.remove(id);
			}
		}
		
		//clean up targets without sources
		for (Enumeration<String> en = targets.keys();en.hasMoreElements();)
		{
			String id = en.nextElement();
			if (sources.get(id) == null) 
			{
				targets.get(id).setSubsumedId(null);
				targets.remove(id);
			}
		}
		
		// clean up sources internal to the output meaning scene, and their targets
		for (Enumeration<String> en = sources.keys();en.hasMoreElements();)
		{
			String id = en.nextElement();
			FSNode s = sources.get(id);
			FSNode sm = s.getMeaningSceneAbove();
			// if sm is the output meaning scene
			if (sm.getChildIndex() == getChildren().size() -1)
			{
				s.setSubsumingId(null);
				FSNode t = targets.get(id);
				t.setSubsumedId(null);
				sources.remove(id);
				targets.remove(id);
			}
		}
		
		// demote sources and targets on scene nodes to the thing nodes below them (success if there is one possible target node)
		boolean success = true;
		for (Enumeration<String> en = sources.keys();en.hasMoreElements();)
		{
			String id = en.nextElement();
			FSNode s = sources.get(id);
			if (s.getType().equals("scene"))
			{
				s.setSubsumingId(null);
				FSNode t = targets.get(id);
				t.setSubsumedId(null);

				Vector<FSNode> scs = s.getTypeChildren("thing");
				if (scs.size() > 0)
				{
					FSNode sc0 = scs.get(0);
					sc0.setSubsumingId(id);
					Vector<FSNode> tcs = t.getTypeChildren("thing");
					int linkable = 0;
					for (FSNode tc: tcs) if (Ops.subsumes(sc0, tc))
					{
						linkable++;
						tc.setSubsumedId(id);
					}
					if (linkable != 1) success = false;
				}
				else success = false;
			}
		}
		
		return success;
		
	}
	
	/**
	 * 
	 * @return the meaning scene containing this node
	 */
	public FSNode getMeaningSceneAbove()
	{
		if (getParent().getType().equals("script")) return this;
		return (getParent().getMeaningSceneAbove());
	}

	
	/**
	 * 
	 * @return count of nodes in the subtree
	 */
	public int nodeCount()
	{
		int nodes = 1;		
		for (FSNode child: children) nodes = nodes + child.nodeCount();		
		return nodes;
	}
	
	/**
	 * 
	 * @return
	 */
	public int slotCount()
	{
		int slots = ownSlotValues.size();
		for (FSNode child: children) slots = slots + child.slotCount();	
		return slots;
	}
	
	
	/**
	 * 
	 * @return subtree summary of 'type[nodes,slots]'
	 */
	public String getSummary()
	{
		return (getType() + "[" + nodeCount() + "," + slotCount() + "] ");
	}
	
	
	/**
	 * simple nested write
	 * @param indent
	 */
	public void write(String indent)
	{
		if (indent.length() == 0) Util.message("\n" + getName());
		
		String line = indent + getType() + " " + getName() + " ";
		if (isUsed()) line = line + "used; ";
		if (!isViable()) line = line + "unviable; ";
		for (Enumeration<String> en = ownSlotValues.keys();en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String slotValue = ownSlotValues.get(slotName);
			if (slotName.equals("sound")) line = line + "'" + slotValue + "' ";
			else line = line + slotName + ":" + slotValue + " ";
		}
		Util.message(line);
		
		String nextIndent = indent + "   ";		
		for (FSNode child: children) child.write(nextIndent);
	}
	
	
	//----------------------------------------------------------------------------------
	//						Setters and getters
	//----------------------------------------------------------------------------------

	
	public LanguageModel getModel() {return model;}
	
	// Allowed node types of FS - the name in the box
	public String getType() {return type;}
	
	// name of the FS, e.g. the word with a postscript '_N' for word sense, or sentence text. Inherited by child FS when read in.
	public void setName(String name) 
	{
		this.name = name;
		for (FSNode child: children) child.setName(name);
	}
	public String getName() {return name;}
	
	// order of unifying words to understand a sentence
	public String getUnifyOrder() {return unifyOrder;}
	public void setUnifyOrder(String order) {unifyOrder = order;}
	
	// index of this child in the list of its parent's child FS
	public int getChildIndex() {return childIndex;}
	public void setChildIndex(int i) {childIndex = i;}
	// mark child indexes in their original order, before a permutation
	public void markChildIndexes() {for (int i = 0; i < children.size(); i++) children.get(i).setChildIndex(i);}
	
	// depth in the FS tree
	public int getDepth() {return depth;}
	
	// true of this FS is the result 'action' child of a word FS
	public boolean isResult() {return isResult;}
	public void setIsResult(boolean isResult) {this.isResult = isResult;}
	
	// true for parts of a sentence FS which have been unified with a word FS and should not be unified again
	public boolean isUsed() {return isUsed;}
	// set this node and all its descendants used or not used
	public void setUsed(boolean used) 
	{
		isUsed = used;
		for (FSNode child: children) child.setUsed(used);
	}
	
	// line colour depends on usage
	private Color lineColour()
	{
		if (isUsed) return Color.LIGHT_GRAY;
		return Color.BLACK;
	}
	
	// true if this FS is bring edited (attribute not saved)
	public void setEdited(boolean edited) {this.edited = edited;}
	public boolean isEdited() {return edited;}
	
	// if this is a learning example, the WordLearn object for which it is an example
	public void setWordLearnExampleParent(WordToLearn wordLearnParent) {this.wordLearnExampleParent = wordLearnParent;}
	public WordToLearn getWordLearnExampleParent() {return wordLearnExampleParent;}
	
	//  if this is the feature structure to be learnt for a word, the WordLearn object for which it is the structure to be learnt
	public void setWordLearnToLearnParent(WordToLearn wordLearnToLearnParent) {this.wordLearnToLearnParent = wordLearnToLearnParent;}
	public WordToLearn getWordLearnToLearnParent() {return wordLearnToLearnParent;}
	
	// if this FS is the result of unification or generalisation, the expression of the operation and inputs
	public String getExpression() {return expression;}
	public void setExpression(String expression) {this.expression = expression;}
	public void clearExpressions()
	{
		expression = null;
		for (FSNode child: children) child.clearExpressions();
	}
	
	// a description of the FS - only used at script level
	public String getDescription() {return description;}
	public void setDescription(String description) {this.description = description;}

	/*"
	 * true only if unification or generalisation gives a viable result
	 */
	public void setViable(boolean viable) {this.viable = viable;}
	public boolean isViable() {return viable;}
	
	// key = slot name; value = slot value
	public String getSlotValue(String slotName) {return ownSlotValues.get(slotName);}
	public void setSlotValue(String slotName,String value) {ownSlotValues.put(slotName, value);}
	public void removeSlotValue(String slotName) {ownSlotValues.remove(slotName);}

	public boolean hasSlots() {return (ownSlotValues.size() > 0);}
	public Hashtable<String,String> getOwnSlotValues() {return ownSlotValues;}
	
	// value of a slot, or null
	public String slotValue(String slotName) {return ownSlotValues.get(slotName);}
	
	// child feature structures
	public Vector<FSNode> getChildren() {return children;}
	public boolean hasChildren() {return (children.size() > 0);}
	public FSNode getChild(int c) {return children.get(c);}
	public FSNode getLastChild() 
	{
		if (children.size() == 0) return null;
		return children.get(children.size() - 1);
	}
	
	public void replaceLastChild(FSNode newChild)
	{
		if (children.size() > 0)
		{
			children.remove(getLastChild());
			addChild(newChild);
		}
	}
	
	/**
	 * 
	 * @param type
	 * @return
	 */
	public Vector<FSNode> getTypeChildren(String type)
	{
		Vector<FSNode> tc = new Vector<FSNode>();
		for (FSNode child: children) 
			if (child.getType().equals(type)) tc.add(child);
		return tc;
	}

	
	/**
	 * 
	 * @param typePath path like '/scene/thing/prop', 
	 * not including the type of this node,
	 * and having length > 0
	 * 
	 * @return all descendants got by traversing the path of node types
	 */
	public Vector<FSNode> getTypeNodes(String typePath)
	{
		Vector<FSNode> res = new Vector<FSNode>(); // to be the result
		
		StringTokenizer st = new StringTokenizer(typePath,"/");
		String type = st.nextToken(); // must exist
		Vector<FSNode> typeNodes = getTypeChildren(type); // child nodes of the first type
		String nextPath = typePath.substring(type.length() + 1); // path with first type removed

		if (!st.hasMoreTokens()) res = typeNodes; // last step of the path
		else if (st.hasMoreTokens())
			for (FSNode child: typeNodes)
			{
				// collect all descendants of the typed child nodes, with the shorter path
				Vector<FSNode> desc = child.getTypeNodes(nextPath);
				for (FSNode d: desc) res.add(d);
			}

		return res;		
	}


	// add a child FS, so it knows what its parent is
	public void addChild(FSNode child) 
	{
		children.add(child);
		child.setParent(this);
	}
	
	/* true position of a node in a tree walk of an entire script; propagate through cloning
	 * (cloning loses ancestors, so changes the position directly computed from a tree walk */
	public int getTruePosition() {return truePosition;}
	public void setTruePosition(int truePosition) {this.truePosition = truePosition;}
	public FSNode getNodeAtTruePosition(int pos)
	{
		if (pos == truePosition) return this;
		for (FSNode child: children)
			if (child.getNodeAtTruePosition(pos) != null) 
				return child.getNodeAtTruePosition(pos);
		return null;
	}
	
	/**
	 * position of an unused node in a full list of child nodes - used and unused
	 * @param pos
	 */
	public void setFullListPosition (int pos) {fullListPosition = pos;}
	public int getFullListPosition() {return fullListPosition;}
	
	/**
	 * set all true positions to tree-walk computed positions, from script level
	 */
	public void setAllTruePositions()
	{
		truePosition = getPosition();
		for (FSNode child: children) child.setAllTruePositions();
	}
	

	//------------------------------------------------------------------------------------------
	//                  Finding nodes paired by generalisation or unification
	//------------------------------------------------------------------------------------------
	
	/**
	 * generalise in1 and in2 together, to find the correspondences of their nodes,
	 * and return a node in in2 which corresponds to node p1 in in1.
	 * 
	 * @param in1 input to generalisation; true positions in in1 must have been set up in advance
	 * @param in2 input to generalisation
	 * @param p1 a node in in1 (which must have its true position set)
	 * 
	 * @return the node p2 in in2 that was paired with p1 in generalisation;
	 * or null if there was none
	 */
	public static FSNode findPairedNode(FSNode in1, FSNode in2, FSNode p1) throws Exception
	{
		FSNode paired = null;
		
		/* this generalisation leaves true positions on both inputs (whatever the top node type),
		 * and does not change the true positions preset on in1 */
		FSNode gen =  Ops.traceGeneralise(in1, in2);
		
		// true position in in1 of the node we are trying to pair
		int pos = p1.getTruePosition();
		
		// find all origins in the generalisation
		Vector<int[]> origins = allOrigins(gen);
		
		// find the true position of the paired node in in2
		int tp = -1;
		for (int[] origin: origins)
			if (origin[0] == pos) tp = origin[1];
		
		if (tp >= 0) paired = in2.getNodeAtTruePosition(tp);
		
		return paired;
	}
	
	/**
	 * unify in1 and in2 together, to find the correspondences of their nodes to nodes in the result,
	 * and return a node in the result which corresponds to node p1 in in1.
	 * 
	 * @param in1 input to unification; true positions in in1 must have been set up in advance
	 * @param in2 input to unification
	 * @param p1 a node in in1 (which must have its true position set)
	 * 
	 * @return the node p2 in the result of unification that was derived from with p1;
	 */
	public static FSNode findDerivedNode(FSNode in1, FSNode in2, FSNode p1) throws Exception
	{
		/* this generalisation leaves true positions on both inputs (whatever the top node type),
		 * and does not change the true positions preset on in1 */
		FSNode uni =  Ops.traceUnify(in1, in2, Ops.UNDERSTAND);
		
		// true position in in1 of the node we are trying to pair
		int pos = p1.getTruePosition();
				
		FSNode paired = getNodeWithOrigin(uni,pos,0);
		
		return paired;
	}
	
	/**
	 * 
	 * @param top
	 * @param pos
	 * @param side 0 or 1
	 * @return an FSNode somewhere under top whose origin has pos on side side
	 */
	public static FSNode getNodeWithOrigin(FSNode top,int pos, int side)
	{
		FSNode res = null;
		int[] origin = top.getOrigin();
		if ((origin != null) && (origin[side] == pos)) res = top;
		
		for (FSNode child: top.getChildren()) if (res == null)
			res = getNodeWithOrigin(child,pos, side);
		return res;
	}
	
	/**
	 * 
	 * @param gen
	 * @return
	 */
	public static Vector<int[]> allOrigins(FSNode gen)
	{
		Vector<int[]> origins = new Vector<int[]>();
		
		int[] origin = gen.getOrigin();
		origins.add(origin);
		
		for (FSNode child: gen.getChildren())
		{
			Vector<int[]> cOrigins = allOrigins(child);
			for (int[] orig: cOrigins) origins.add(orig);
		}
				
		return origins;
	}
	

	// replace the list of child nodes, so they all know what their parent is
	public void setChildren(Vector<FSNode> newChildren)
	{
		if (newChildren == null) 
		{
			Util.message("Cannot set child nodes all null");
		}
		children = new Vector<FSNode>();
		for (FSNode child:newChildren) addChild(child);
	}
	
	// parents and ancestors
	public FSNode getParent() {return parent;}
	public void setParent(FSNode parent) {this.parent = parent;}
	public FSNode getRootAncestor()
	{
		FSNode res = this;
		FSNode step = parent;
		while (step != null)
		{
			res = step;
			step = step.getParent();
		}
		return res;
	}
	
	// the path of node types from the root
	public String getPath()
	{
		String path = "";
		if (parent == null) path = getType();
		else path = parent.getPath() + "/" + getType();
		return path;
	}
	
	//free orders
	public Vector<int[]> getFreeOrder() {return freeOrder;}
	
	// for sending to the web client; the drawn row
	public void setRow(int row)
	{
		this.row = row;
		for (FSNode child: children) child.setRow(row);
	}

	
	// for sending to the web client; the drawn column
	public void setColumn(int column)
	{
		this.column = column;
		for (FSNode child: children) child.setColumn(column);
	}
	
	
	/**
	 * text form of the FSNode tree, showing all populated slots,
	 * in standard order of slots, so it can be used for comparison
	 * (except for variable order of child nodes - FIXME)
	 * @return
	 */
	public String getTextTree(int depth)
	{
		String used = "";
		if (isUsed) used = "_U";
		
		String text = "(" + getType() + used + ": ";
		for (int d = 0; d < depth; d++) text = "   " + text;
		NodeType nt = model.getNodeType(getType());
		
		// source and target of subsumption links
		String s = getSubsumingId();
		String t = getSubsumedId();
		if (s != null) text = text + "S[" + s + "] ";
		if (t != null) text = text + "T[" + t + "] ";
		
		for (Slot slot: nt.allowedSlots())
		{
			String slotName = slot.getName();
			String value = getSlotValue(slotName);
			if (value != null) text = text + slotName + "='" + value + "' ";
		}
		
		for (FSNode child: getChildren()) text = text + "\n" + child.getTextTree(depth + 1);
		
		text = text + ") ";
		return text;
	}
	
	/**
	 *  The name of a word - made up from its phonemes separated by '-'
	 * @return
	 */
	public String getSpacedName()
	{
		if (getSounds().size() == 0) 
		{
			return "no sounds in FS";
		}

		// the first and only sequence of sound scenes in the word
		Vector<String> sounds = getSounds().get(0);
		return getSpacedName(sounds);
	}
	
	
	/**
	 * convert a list of phoneme sounds [t,i,m] to a single spaced name t-i-m
	 * @param sounds
	 * @return
	 */
	public static String getSpacedName(Vector<String> sounds)
	{
		String name = "";
		for (String sound: sounds) name = name + "-" + sound;
		return name.substring(1); // remove first '-'
	}
	
	/**
	 * segments of unused sounds in a sentence 
	 * - each segment made up from its phonemes separated by '-'
	 */
	public Vector<String> getSoundSegments()
	{
		Vector<String> segments = new Vector<String>();
		
		for (Vector<String> sounds: getSounds())
		{
			String name = "";
			for (String sound: sounds) name = name + "-" + sound;
			segments.add(name.substring(1)); // remove first '-'
		}
		
		return segments;
	}
	
	/**
	 * if a sequence of sound nodes can be found, the number of input meaning scenes before it and after it;
	 * or if it cannot be found, return [-1,-1].
	 * Intended for a word or a learning example, with a final meaning scene which is not counted
	 * @param sounds
	 * @return
	 */
	public int[] meaningSceneCount(Vector<String> sounds)
	{
		int[] ms = new int[2];
		for(int i = 0; i < 2; i++) ms[i] = 0; // initial values for counts
		
		// segments of contiguous sound nodes
		Vector<String> seg = new Vector<String>();
		
		boolean found = false;
		for (int c = 0; c < children.size() - 1; c++) // ignore final meaning scene
		{
			FSNode child = children.get(c);
			if (child.getType().equals("scene"))
			{
				seg = new Vector<String>();// restart the segment
				if (found) ms[1]++;
				else ms[0]++;
			}
			if (child.getType().equals("sound"))
			{
				seg.add(child.getSlotValue("sound"));
				if (Util.includes(seg, sounds)) found = true;
			}
		}
		
		if (!found) for(int i = 0; i < 2; i++) ms[i] = -1; //failure
		return ms;
	}
	
	/**
	 * for a script FSNode only:
	 * @return a Vector of all input meaning scenes, 
	 * not including the final output meaning scene
	 */
	public Vector<FSNode> getInputMeaningScenes()
	{
		Vector<FSNode> res = new Vector<FSNode>();
		for (int i = 0; i < children.size() - 1; i++) // ignore the final scene
		{
			FSNode child = children.get(i);
			if (child.getType().equals("scene")) res.add(child);
		}
		return res;
	}
	
	
	
	/**
	 * applies to a scripts with input sounds and meanings, and an output meaning
	 * @param withLinks - if true, the scenes must be the start of subsumption links to the final meaning scene
	 * @return numbers of input meaning scenes, before the first sound and after it.
	 */
	public int[] scenesBeforeAfter(boolean withLinks)
	{
		int[] links = new int[2];
		for (int i = 0; i < 2; i++) links[i] = 0;
		
		int bef = 0;
		for (FSNode child: getChildren())
		{
			if (child.getType().equals("sound")) bef = 1;
			else if ((!withLinks)||(child.hasLink())) links[bef]++;
		}
		
		// exclude the final output meaning scene, if it was added (no outgoing link test)
		if (!withLinks) links[1]--;

		return links;
	}
	/**
	 * 
	 * @return unused sound segments separated by '|'
	 */
	public String getSegmentString()
	{
		String seg = "";
		
		for (String segment: getSoundSegments())
			seg = seg + "|" + segment;
		
		if (seg.length() > 0) seg = seg.substring(1);
		return seg;
	}

	
	//-------------------------------------------------------------------------
	//                 For use in  traceability, and subsumption links
	//-------------------------------------------------------------------------
	
	/**
	 *  summary of node types in subtree structure
	 * @return
	 */
	public String treeSummary()
	{
		String summary = getType();
		if (children.size() > 0)
		{
			summary  = summary + "(";
			for (FSNode child: children) 
				summary = summary + child.treeSummary() + ",";
			summary  = summary.substring(0,summary.length() -1) + ")";
		}
		return summary;
	}
	
	
	public int getHighlightedPosition() {return highlightedPosition;}
	public void setHighlightedPosition(int pos) 
	{
		highlightedPosition = pos;
		for (FSNode child:children) child.setHighlightedPosition(pos);
	}
	
	
	/**
	 * @return for a result of unification and generalisation, 
	 * the positions of nodes in the two inputs that gave rise to this node
	 */
	public int[] getOrigin() {return origin;}
	
	/**
	 * 
	 * @return the origins of all nodes, in tree walk order
	 */
	public  Vector<int[]> getOrigins()
	{
		Vector<int[]> origins = new Vector<int[]>();
		Vector<FSNode> fullWalk = treeWalk();
		for (FSNode node: fullWalk) origins.add(node.getOrigin());
		return origins;
	}
	
	/**
	 * 
	 * @param left
	 * @param right
	 */
	public void setOrigin(int left, int right)
	{
		origin = new int[2];
		origin[0] = left;
		origin[1] = right;
	}
	
	
	/**
	 * This FSNode is input to a unification. For the let-hand input , isLeft is true.
	 * 
	 * Set the origins of this node and all its descendants, 
	 * so that if any of them go into the result without unification, 
	 * the correct origin (from one input only) appears in the the result.
	 * 
	 * @param isLeft
	 */
	public void setAllOrigins(boolean isLeft)
	{
		if (isLeft) setOrigin(getPosition(),0);
		else setOrigin(0,getPosition());
		for (FSNode child: children) child.setAllOrigins(isLeft);
	}
	
	
	public void writePositionAndOrigin()
	{
		if (origin == null) Util.message("No origin; position " + getPosition());
		else Util.message ("Origin [" + origin[0] + "," + origin[1] + "]; position " + getPosition());

	}
	
	// id to be pointed at in subsumption links
	public String getSubsumedId() {return subsumedId;}
	public void setSubsumedId(String id) {this.subsumedId = id;}

	// id of FS pointed to in subsumption links
	public String getSubsumingId() {return subsumingId;}
	public void setSubsumingId(String subsumesId) {this.subsumingId = subsumesId;}
	
	// id of FS which subsumed this FS
	public String getFromId() {return fromId;}
	
	// subsumesId on FS which subsumed this FS
	public String getFromSubsumesId() {return fromSubsumesId;}
	
	/**
	 *  after successful subsumption, set from ids
	 * @param fs
	 */
	public void setFromIds(FSNode fs)
	{
		fromId = fs.getSubsumedId();
		fromSubsumesId = fs.getSubsumingId();
	}
	
	/**
	 * after failed subsumption, cancel all from ids
	 */
	public void cancelFromIds()
	{
		fromId = null;
		fromSubsumesId = null;		
		for (FSNode child:children) child.cancelFromIds();
	}
	
	// ids for generalisation or unification, to match FS to compute subsumption links
	public void setGenId(String id) {genId = id;}
	public String getGenId() {return genId;}
	
	//drawing subsumption links. Keys are key value + "_start" or "_end". Values are [x,y]
	public Hashtable<String,int[]> getLinkEnds() {return linkEnds;}
	

	/**
	 * 
	 * @return
	 */
	public String getScriptSummary()
	{
		if (!type.equals("script")) return treeSummary();

		String res = "";
		for (FSNode child: children)
		{
			String ct = child.getType();
			if (ct.equals("sound")) res = res + "'" + child.getSlotValue("sound") + "'";
			else res = res + child.treeSummary();
			
			boolean used = child.isUsed();
			if (used) res = res + "/";
			else res = res + ";";
		}
		return res;
	}
	
	/**
	 * 
	 * @param example
	 * @return the index of a learning example in the Vector of examples
	 */
	public int exampleIndex()
	{
		int index = -1;
		try 
		{
			StringTokenizer st = new StringTokenizer(getName(),":");
			String iString = st.nextToken().substring(2);
			index = Integer.parseInt(iString);
		} 
		catch (Exception ex) {}
		return index;
	}
	
	/**
	 * flatten a subtree to a list;
	 * (each member of the list may have a subtree)
	 * @return
	 */
	public Vector<FSNode> flatten()
	{
		Vector<FSNode> flat = new Vector<FSNode>();
		flat.add(this);
		
		for (FSNode child: children)
			for (FSNode c: child.flatten()) flat.add(c);

		return flat;
	}
	
	//---------------------------------------------------------------------------------
	// 	detect if a word FSNode creates a nested context 
	//   (so the word should not be used to learn other words)
	//---------------------------------------------------------------------------------
	
	/**
	 * 
	 * @return true if there is any subsumption link between nodes of different depth
	 */
	public boolean makesNestedContext()
	{
		boolean nests = false;
		Hashtable<String,Integer> startDepths = new Hashtable<String,Integer>();
		Hashtable<String,Integer> endDepths = new Hashtable<String,Integer>();
		
		recordLinkEnds(0,true,startDepths);
		recordLinkEnds(0,false,endDepths);
		
		for (Enumeration<String> en = startDepths.keys();en.hasMoreElements();)
		{
			String key = en.nextElement();
			Integer ss = startDepths.get(key);
			Integer ee = endDepths.get(key);
			if (ee != null)
			{
				int diff = ss.intValue() - ee.intValue();
				if (diff != 0) nests = true;
			}
		}
		
		return nests;
	}
	
	/**
	 * record the depths of subsumption link ends
	 * @param depth
	 * @param start
	 * @param lDepths
	 */
	public void recordLinkEnds(int depth,boolean start,Hashtable<String,Integer> lDepths)
	{
		String key = null;
		if (start) key = getSubsumingId();
		else key = getSubsumedId();
		if (key != null) lDepths.put(key, Integer.valueOf(depth));
		
		for (FSNode child: children) child.recordLinkEnds(depth + 1, start, lDepths);
	}



}
