package modelPack;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import utilPack.Util;

public class Ops {
	
	
	// types of unification
	public static int PLAIN = 10;
	public static int UNDERSTAND = 11;
	public static int GENERATE = 12;
	
	// operations
	public static int UNIFY = 0;
	public static int GENERALISE = 1;
	public static int SUBSUMES = 2;
	
	// symbols for the operations in expressions
	public static String[] SYMBOL = {" U "," ^ ", " < "};
	
	
	//--------------------------------------------------------------------------
	//                      Operations - All
	//--------------------------------------------------------------------------
	
	/**
	 * after any operation, note in the result how it was derived
	 * @param result
	 * @param input1
	 * @param input2
	 * @param opType
	 */
	public static String getExpression(FSNode input1, FSNode input2, int opType)
	{
		String arg1 = input1.getExpression();
		if (arg1 == null) arg1 = input1.getName();
		
		String arg2 = input2.getExpression();
		if (arg2 == null) arg2 = input2.getName();
		
		return("(" + arg1 + SYMBOL[opType] + arg2 + ")");
	}
	
	/**
	 * set the unify order of the result of an operation,
	 * to the set union of the unify orders of the two inputs.
	 * This is used for dynamic computation of sound segments in compact drawing of feature structures.
	 * @param result
	 * @param input1
	 * @param input2
	 */
	private static void setResultUnifyOrder(FSNode result, FSNode input1, FSNode input2)
	{
		if (result.getType().equals("script"))
		{
			Vector<String> order1 = input1.getUnifyWords();
			Vector<String> order2 = input2.getUnifyWords();
			String uOrder = "";
			
			for (String w1:order1) uOrder = uOrder + w1 + " ";
			for (String w2:order2) if (!Util.inVector(w2, order1))
				uOrder = uOrder + w2 + " ";
			
			if (uOrder.length() > 0) result.setUnifyOrder(uOrder);

		}
	}
	
	
	//--------------------------------------------------------------------------
	//                      Operations - Unification
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param other
	 * @return the unification of this FS with the other FS (which may not be viable)
	 */
	public static FSNode unify(FSNode first,FSNode other, int unifyType) throws Exception
	{
		String type = first.getType();
		boolean isScript = (type.equals("script"));

		// start making the result
		FSNode result = new FSNode(first.getModel(),type);
		result.setViable(false);
		
		// if the types do not match, return a non-viable result
		if (!type.equals(other.getType())) return result;
					
		// make clones of both input Feature Structures
		FSNode res1 = new FSNode(first);
		FSNode res2 = new FSNode(other);
		
		// result takes subsumption links from the first input (the word). Other non-slots needed?
		result.setSubsumingId(first.getSubsumingId());
		result.setSubsumedId(first.getSubsumedId());
		result.setTruePosition(first.getTruePosition()); // ?? seems to serve no purpose
		
		if (isScript)
		{
			//labels on the top script node only
			result.setName("Result");
			result.setDescription("Result of unification");
			// set the result expression to '(first U other)'
			result.setExpression(getExpression(first,other,UNIFY));

			// set true positions of all input nodes (which survive cloning of descendant nodes) from tree-walk positions
			res1.setAllTruePositions();
			res2.setAllTruePositions();
			
			// set origins of all the input nodes, so that when they are used but not unified, the correct origin appears in the result.
			res1.setAllOrigins(true); // true means the left input
			res2.setAllOrigins(false); // false means the right input
		}
		
		// check that all the inputs to unification have origins and true positions > 0
		checkInputOrigins(res1,1);
		checkInputOrigins(res2,2);
						
		// if the slots are not compatible, the result remains not viable
		String conflicts = unifySlots(result, res1,res2);
		boolean slotsCompatible = (conflicts.length() == 0);
		
		if (slotsCompatible)
		{
			// unify the child nodes
			Vector<FSNode> childFS = unifyChildFS(res1, res2, unifyType);
			
			// if unifying child nodes has succeeded, or has failed in specified non-viable child nodes
			if (childFS != null)
			{
				result.setViable(true);
				result.setChildren(childFS);
				
				// if some child nodes are not viable, make the parent result not viable, and record why
				String failures = "";
				for (FSNode child: childFS) if (!child.isViable()) 
					failures = failures + child.getDescription() + ";";
				if (failures.length () > 0)
				{
					result.setViable(false);
					result.setDescription(failures);
				}
				
				// record the origin of this node
				result.setOrigin(res1.getTruePosition(), res2.getTruePosition());
				
				setResultUnifyOrder(result,res1,res2);
			}
		}
		
		else result.setDescription("Failed to unify slots on " + type + " node: " + conflicts);
		
		// sometimes a script appears to be viable, but has no child nodes; so will not pass the subsume tests
		boolean scriptTest = true;
		if (isScript) scriptTest = ((result != null) && (result.getChildren().size() > 0));

		// Both inputs should subsume the result of unification, at any level. Test to throw an exception, or allow a break point
		if ((result != null) && (result.isViable()))
		{
			if (!subsumes(first,result))
			{
				writeSubsumeError(first,other, result);
				boolean sub = subsumes(first,result);
				throw new Exception("First unification input of type " + first.getType() + " does not subsume the unify result");
			}
			if (!subsumes(other,result))
			{
				writeSubsumeError(first,other, result);
				boolean  sub = subsumes(other,result);
				throw new Exception("Second unification input of type " + other.getType() + " does not subsume the unify result");
			}
			
			// check the origins of all nodes in this result, at all levels
			try
			{
				// clone inputs res1 and res2 have had true positions set
				checkOperationOrigins(result,res1,res2,false);
			}
			catch(Exception ex)
			{
				boolean writeExceptions = false;
				if (writeExceptions) Util.message(ex.getMessage());
			}
			
			// finally at script level, patch up origins if needed; and remove any subsumption links
			if (isScript) 
			{
				patchResultOrigins(result,res1,res2);
				result.removeAllLinks();
			}
		}

		return result;
	}
	
	
	/**
	 * Usually, true positions of FSNodes are set zero all the way down the tree, 
	 * and unification only sets up true positions once at script level, in clones of the two inputs,
	 * to put the true positions in origins in the result.
	 * 
	 * Use this form of unification if:
	 * (a) the top nodes you are unifying might not be script nodes
	 * (b) you want true positions to be left set on the inputs, 
	 * so true positions can be used with origins in the result, to match nodes between the inputs
	 * 
	 * @param first
	 * @param other
	 * @return
	 * @throws Exception
	 */
	public static FSNode traceUnify(FSNode first, FSNode other, int type) throws Exception
	{
		first.setAllTruePositions();
		other.setAllTruePositions();
		
		return unify(first, other,type);
	}


	
	/**
	 * 
	 * @param first
	 * @param other
	 * @param result
	 */
	private static void writeSubsumeError(FSNode first, FSNode other, FSNode result) throws Exception
	{
		Util.message("\nSubsume test failure");
		Util.message("First: " + first.getTextTree(0));
		Util.message("Other: " + other.getTextTree(0));
		Util.message("Result: " + result.getTextTree(0));
		
		// repeat to find out what happened
		unify(first,other,UNDERSTAND);
	}
	
	/**
	 * check that all nodes in an input have a true position and an origin
	 * @param res
	 * @param side  = 1, 2, or 3; used only for diagnostics
	 */
	private static void checkInputOrigins(FSNode res, int side)
	{
		String type = res.getType();

		// check that the input has a true position set
		if (res.getTruePosition() < 1)
		{
			Util.message("No true position on node of type " + type + " in side " + side);
		}
		
		// check that the input has an origin set (usually one-sided)
		if (res.getOrigin() == null)
		{
			boolean originMessage = false;
			if (originMessage) Util.message("No origin on node " + res.getTruePosition() + " of type " + type + " in side " + side);
		}
		
		for (FSNode child: res.getChildren()) checkInputOrigins(child,side);
	}
	
	/**
	 * 
	 * @param res1
	 * @param res2
	 * @return a string describing any failures of unification of slots;
	 * empty if unification succeeds
	 * side-effect - leave the unified slot values in FSNode
	 */
	private static String unifySlots(FSNode result, FSNode res1, FSNode res2)
	{
		// any conflict in slot values between res1 and the res2 means the unification is not viable.
		String failures = "";
		boolean compatible = true;

		for (Enumeration<String>  en = res1.getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String ownValue = res1.getSlotValue(slotName);
			String otherValue = res2.getSlotValue(slotName);
						
			// if both inputs have the slot
			if (otherValue != null)
			{
				Slot slot = res1.getModel().getSlot(slotName);
				compatible = (slot.compatible(ownValue, otherValue));
				
				if (compatible)
				{
					// put the more specific value in the result
					if (slot.isSubValue(otherValue, ownValue)) result.getOwnSlotValues().put(slotName, otherValue);
					if (slot.isSubValue(ownValue, otherValue)) result.getOwnSlotValues().put(slotName, ownValue);
				}
				
				else failures = failures + "slot '" + slotName + "' has incompatible values " + ownValue + " and " + otherValue + ";";
			}
			
			// if only the first input has the slot
			else if (otherValue == null) result.getOwnSlotValues().put(slotName, ownValue);
		}
		
		// for slots where res1 has no value, set the result value to the value from res2
		for (Enumeration<String>  en = res2.getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String ownValue = res1.getSlotValue(slotName);
			String otherValue = res2.getSlotValue(slotName);
			
			if (ownValue == null) result.getOwnSlotValues().put(slotName, otherValue);
		}
		
		// possible break point
		if (failures.length() > 0)
		{
			// Util.message("Slot unify failure: " + failures);
		}
		
		return failures;
	}
	
	/**
	 * 
	 * @param p1
	 * @param p2
	 * @param unifyType
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> unifyChildFS(FSNode p1, FSNode p2, int unifyType) throws Exception
	{
	   if (unifyType == UNDERSTAND) return unifyChildUnderstand(p1,p2);
	   if (unifyType == GENERATE) return unifyChildGenerate(p1,p2);
	   if (unifyType == PLAIN) return unifyChildPlain(p1,p2);
	   return null;
	}
	


	/**
	 * 
	 * @param fWord
	 * @param fSent
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> unifyChildGenerate(FSNode fWord, FSNode fSent) throws Exception
	{
		
		Vector<FSNode> bestUnified = new Vector<FSNode>(); // returned if no success
		
		// the result meaning part of the word must be unified with one of the unused meaning parts of the (partially understood) sentence
		
		// get the unused meaning parts of the sentence
		Vector<FSNode> unusedInSentence = new Vector<FSNode>();
		for (FSNode child: fSent.getChildren())
			if (child.getType().equals("scene") && (!child.isUsed())) unusedInSentence.add(child);
		
		// get the meaning result part of the word
		FSNode resChild  = null;
		for (FSNode wChild: fWord.getChildren())
			if (wChild.getType().equals("scene") && (wChild.isResult())) resChild = wChild;
		if (resChild == null) throw new Exception("No result meaning scene in word to be unified for generation.");
		
		/* try out all unused meaning parts of the sentence, picking the first part which 
		 * (a) the word result subsumes and (b) unifies with the word result.
		 * The second condition is redundant.  */
		FSNode unifiedChild = null;
		FSNode usedChild = null;
		for (FSNode sChild: unusedInSentence) if (unifiedChild == null)
		{
			/* The word result scene must subsume the sentence scene; that is , 
			 * the sentence meaning must include all information in the word result meaning. 
			 * The word meaning scene must not add anything to the sentence meaning scene. */
			if (subsumes(resChild,sChild))
			{
				unifiedChild = unify(resChild,sChild,PLAIN);
				if (unifiedChild.isViable())
				{
					sChild.setUsed(true);
					unifiedChild.setUsed(true);
					usedChild = sChild;
					usedChild.setUsed(true);
				}					
				else unifiedChild = null;
			}
		}
		
		// build up the result vector of child nodes
		if (unifiedChild != null)
			for (FSNode sChild: fSent.getChildren())
			{
				// if the sentence child has been unified, replace it by the whole word FS, with subsumption links applied
				if (sChild.equals(usedChild))
				{
					/* the further use of the unified child is to find the pairings between the nodes of its inputs,
					 * and feed information from the pairings back along the links */
					Vector<int[]> pairs = getPairedOrigins(unifiedChild);
					for (int[] pair: pairs)
					{
						FSNode wordResultSource = resChild.getNodeAtTruePosition(pair[0]);
						FSNode sentenceSource = usedChild.getNodeAtTruePosition(pair[1]);
						if (wordResultSource == null) throw new Exception("Cannot find unify source node in word");
						if (sentenceSource == null) throw new Exception("Cannot find unify source node in sentence");
						
						// if the node in the word result is the end of a subsumption link, find where in the word FS it came from
						String sid = wordResultSource.getSubsumedId();
						if (sid != null)
						{
							boolean sourceFound = false;

							// look for the source of the link in each non-result meaning part of the word
							Vector<FSNode> newChildren = new Vector<FSNode>();
							for (FSNode wChild: fWord.getChildren())
							{
								if (wChild.getType().equals("scene") && (!wChild.isResult()))
								{
									// every non-sound non-result child of a word FS should be the source of one link
									FSNode subsumer = getSubsumingChild(wChild);
									if (subsumer == null) throw new Exception("Word child node is not the source of a link");
									
									// if you find the matching link
									if (subsumer.getSubsumingId().equals(sid))
									{
										// apply the subsumption link
										Util.message("Generation subsumption link unifies " + subsumer.treeSummary() + " with " + sentenceSource.treeSummary());
										FSNode linkApplied = unify(subsumer,sentenceSource,PLAIN);
										// Util.message("Unify result: " + linkApplied.treeSummary());

										// reinsert the result in the child scene of the Word FS (which has been cloned)
										wChild = substitute(wChild,linkApplied);
										// Util.message("Substitution: " + wChild.treeSummary());
										
										// make sure all parts will show in the result, in spite of the link unification
										wChild.setUsed(false);
										
										sourceFound = true;
										
									} // end of subsuming id case
									
								}  // end of scene case
								
								newChildren.add(wChild);
								
							} // end of loop over fWord.getChildren()
							
							fWord.setChildren(newChildren);
							
							if (!sourceFound) throw new Exception("Cannot find the source of link " + sid);
							
						} // end of if (sid != null)

					} // end of for (int[] pair: getPairedOrigins(unifiedChild))

					
					// stitch in the (expanded by links, and cloned) child nodes of the word FS
					String addSums = "";
					for (FSNode wChild: fWord.getChildren())
					{
						// Util.message("word child: " + wChild.treeSummary() + "-" + wChild.isUsed());
						if (wChild.equals(resChild)) 
							{bestUnified.add(unifiedChild); addSums = addSums + "1";}
						else 
							{bestUnified.add(new FSNode(wChild)); addSums = addSums + "2";}
					}
					// Util.message("Choices: " + addSums);

				}  // end of if (sChild.equals(usedChild))

				// otherwise, carry over the sentence child node
				else bestUnified.add(sChild);

			} // end of for (FSNode sChild: fSent.getChildren())
				
		// diagnostic
		String bestSum = "Result summaries: ";
		for (FSNode best: bestUnified) {bestSum = bestSum + best.treeSummary() + "-" + best.isUsed() + ", ";}
		// Util.message(bestSum);
		
		return bestUnified;
		
	}
	
/**
 * g must be one end of a subsumption link - I think the subsumed end.
 * somewhere in f must be the other end of the link (maybe at the top).
 * Wherever in f you find the other end, substitute g for f or for the descendant of f
 * @param f
 * @param g
 * @throws Exception
 */
private static FSNode substitute(FSNode f, FSNode g) throws Exception
{
	// look for either kind of id on f
	String fid = f.getSubsumedId();
	if (fid  == null) fid = f.getSubsumingId();

	// look for either kind of id on g - there should be one of them
	String gid = g.getSubsumingId();
	if (gid == null) gid = g.getSubsumedId(); 
	if (gid == null) throw new Exception("Cannot find either link end on head node to be substituted");

	boolean matched = ((fid != null) && (fid.equals(gid)));	
	if (matched) return g;
	
	// if this node is not a match, look for a match further down the tree of f
	Vector<FSNode> newChildren = new Vector<FSNode>();
	for (FSNode child: f.getChildren()) newChildren.add(substitute(child,g));	
	f.setChildren(newChildren);	
	
	// pass any reasons for failure back up the tree
	for (FSNode child: f.getChildren()) if (!child.isViable())
	{
		f.setViable(false);
		f.setDescription(child.getDescription());
	}
	
	return f;
	
}
	
	/**
	 * 
	 * @param unified
	 * @return
	 */
	private static Vector<int[]>  getPairedOrigins(FSNode unified)
	{
		Vector<int[]> pairings = new Vector<int[]>();
		addPairings(pairings,unified);		
		return pairings;
	}
	
	/**
	 * 
	 * @param pairings
	 * @param unified
	 */
	private static void addPairings(Vector<int[]> pairings,FSNode unified)
	{
		int[] pair = unified.getOrigin();
		if ((pair[0] > 0) && (pair[1] > 0)) pairings.add(pair);		
		for (FSNode child: unified.getChildren()) addPairings(pairings, child);
	}


	
	/**
	 * 
	 * @param fWord
	 * @param fSent
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> unifyChildUnderstand(FSNode fWord, FSNode fSent) throws Exception
	{
		Vector<FSNode> result = null; // to indicate failure, if matches has no elements
		
		// make a list of sentence parts that are not marked as used
		Vector<FSNode> unused = removeUsed(fSent);
		
		// try to find a full match of the word with no meaning gaps
		Vector<int[]> matches = getMatchNoGaps(fWord,unused);
		
		// if there is no full match, use a match with one gap
		if (matches.size() == 0) matches = getMatchOneGap(fWord,unused);
		
		if (matches.size() > 0)
		{
			// unify input meanings of the word with meanings of the sentence
			result = unifyList(matches,fWord,unused);
			
			// find the positions of nodes from the word, for subsumption links
			Vector<int[]> wordPos = wordPositions(matches);
			int wordResultIndex = wordPos.get(wordPos.size() -1)[1];
			FSNode wordResult = result.get(wordResultIndex);
			Vector<FSNode> wordInputs = findWordInputs(wordPos,result);
			
			// apply subsumption links from unified word input scenes, to change wordResult
			FSNode fullWordResult = applySubsumptionLinks(wordInputs, wordResult);
			fullWordResult.setFullListPosition(-1);
			
			// put the changed word result back in the list of results
			result.removeElementAt(wordResultIndex);
			result.insertElementAt(fullWordResult, wordResultIndex);
			
			// restore previously used nodes to the sentence node vector
			result = restoreUsedNodes(result,fSent,wordPos);
		}
				
		return result;
	}
	
	/**
	 * 
	 * @param matches
	 * @return a vector of pairs: [index in word FS, index in result FS]
	 */
	private static Vector<int[]> wordPositions(Vector<int[]> matches)
	{
		Vector<int[]> res = new Vector<int[]>();
		
		for (int i = 0; i < matches.size(); i++)
		{
			int[] pair = matches.get(i);
			if (pair[0] > -1)
			{
				pair[1] = i;
				res.add(pair);
			}
		}		
		return res;
	}
	
	/**
	 * From a unified result vector, find the scenes that are the result of unifying with 
	 * input scenes of a word.
	 * @param wordPos
	 * @param result
	 * @return
	 */
	private static Vector<FSNode> findWordInputs(Vector<int[]> wordPos,Vector<FSNode> result)
	{
		Vector<FSNode> res = new Vector<FSNode>();
		
		for (int[] pair: wordPos)
		{
			FSNode fs = result.get(pair[1]);
			if (fs.getType().equals("scene")) res.add(fs);
		}
		
		return res;
	}
			

	
	/**
	 * 
	 * @param matches
	 * @param fWord
	 * @param unused
	 * @return
	 */
	private static Vector<FSNode> unifyList(Vector<int[]>  matches,FSNode fWord,Vector<FSNode>  unused) throws Exception
	{
		Vector<FSNode> result = new Vector<FSNode>();
		Vector<FSNode> wParts = fWord.getChildren();
		
		for (int[] match: matches)
		{
			// nodes from the sentence, not affected by the word
			if (match[0] < 0) 
			{
				// clone from the sentence list, and do not change the used status
				result.add(new FSNode(unused.get(match[1])));
			}
			
			// nodes from the word only (word result, and any gaps)
			else if (match[1] < 0)
			{
				// clone from the word
				FSNode fromWord = wParts.get(match[0]);
				FSNode res = new FSNode(fromWord);
				// set used, except for the word result
				if (match[0] < wParts.size() -1) res.setUsed(true);
				// gaps and the word result have no position from the sentence list
				res.setFullListPosition(-1);
				result.add(res);
			}
			
			// nodes from unification
			else
			{
				// unify from the word and from the sentence (sound or meaning node) and set used
				FSNode fromWord = wParts.get(match[0]);
				FSNode fromSentence = unused.get(match[1]);
				FSNode res = unify(fromWord,fromSentence,PLAIN);
				res.setFullListPosition(fromSentence.getFullListPosition());
				res.setUsed(true);
				result.add(res);
			}
		}
		
		return result;
		
	}

	
	/**
	 * 
	 * @param fWord a word FS
	 * @param unused all child nodes of the sentence script that are unused
	 * @param allowed number of gaps allowed
	 * 
	 * @return list of pairs [word position, sentence position for unification];
	 * or -1 if in one half if there is to be no unification
	 */
	private static Vector<int[]>  getMatchNoGaps(FSNode fWord, Vector<FSNode> unused) throws Exception
	{
		Vector<int[]> res = new Vector<int[]>();
		int wordSize = fWord.getChildren().size();
		
		// find all start points in the sentence for matches between the word sounds and the sentence sounds
		Vector<Integer> allMatchStarts = matchSounds(fWord, unused);
		int [] ends = soundEnds(fWord);
		
		// try each match start position, to see if the input meaning scenes of the word can be matched
		for (Integer matchStart: allMatchStarts) if (res.size() == 0)
		{
			int start  = matchStart.intValue();
			int wordStart = start - ends[0]; // the node in the sentence which should match the first node in the word
			
			// see how many of the meaning input scenes of the word can or cannot be matched with the sentence
			int gaps = 0;
			Vector<int[]> wRes = new Vector<int[]>(); // will be wordSize pairs for the word
			for (int j = 0; j < wordSize; j++) 
			{
				FSNode wordChild = fWord.getChild(j);
				String wType = wordChild.getType();
				// the last child of the word is its result scene
				boolean resultScene = (j == wordSize -1);

				// find the implied position of the scene in the sentence list
				int sentPos = wordStart + j;
				boolean inSentence = ((sentPos > -1) && (sentPos < unused.size()));
				FSNode sentChild = null;
				if (inSentence) sentChild = unused.get(sentPos);
				
				int[] pair = new int[2];
				pair[0] = j;
				pair[1] = -1;
				
				if ((wType.equals("scene")) && (!resultScene)) // do not try to match the word result scene
				{
					// if the sentence position is out of range, there must be a gap for a word meaning scene
					if (!inSentence) gaps++;
					
					else if (inSentence)
					{
						// cannot match a word meaning node with a sentence sound node
						if (sentChild.getType().equals("sound")) gaps++;
						else pair[1] = sentPos;
					}
				}
				
				// always match sounds; check that they match (as they should)
				else if (wType.equals("sound")) 
				{
					String ws = wordChild.getSlotValue("sound");
					if (sentChild == null) throw new Exception("Failed match of word sounds: sentence position " + sentPos + "is out of range ");
					if (!(sentChild.getType().equals("sound"))) 
						throw new Exception("Failed match of word sounds: sentence position " + sentPos + " is not a sound node ");
					String ss = sentChild.getSlotValue("sound");
					if (!(ws.equals(ss))) throw new Exception("Failed match of word sounds: '" + ws + "' and '" + ss + "'");
					
					pair[1] = sentPos;
				}
				
				// add one pair for each node in the  word
				wRes.add(pair); // the last pair (for the word result) is a non-match
			}
			
			// if there are no gaps, make the list of matches between the word and the sentence
			if (gaps == 0)
			{
				// loop over all positions in the sentence
				for (int p = 0; p < unused.size(); p++)
				{
					int[] pair = new int[2];
					pair[0] = -1; // outside the word limits, no match
					pair[1] = p;
					
					// before the start of the word - non-match
					if (p < wordStart) res.add(pair);
					
					// during the word - wordSize entries (the last is a non-match for the word result)
					else if (p == wordStart) 
						for (int i = 0; i < wordSize; i++)
							res.add(wRes.get(i));
					
					// after the end of the word - non-match
					else if (p > wordStart + wordSize - gaps - 2) res.add(pair);					
				}
			}
		}
		
		return res;
	}
	/**
	 * 
	 * @param fWord a word FS
	 * @param unused all child nodes of the sentence script that are unused
	 * 
	 * @return list of pairs [word position, sentence position for unification];
	 * or -1 if in one half if there is to be no unification
	 */
	private static Vector<int[]>  getMatchOneGap(FSNode fWord, Vector<FSNode> unused) throws Exception
	{
		Vector<int[]> res = new Vector<int[]>();
		int wordSize = fWord.getChildren().size();
		
		// find all start points in the sentence for matches between the word sounds and the sentence sounds
		Vector<Integer> allMatchStarts = matchSounds(fWord, unused);
		// find positions of first and last sound nodes in the word
		int [] ends = soundEnds(fWord);
		
		// try each match start position, to see if the input meaning scenes of the word can be matched
		int gaps = 0 ;
		// do not try another start if the previous start gave 1 gap, as required
		for (Integer matchStart: allMatchStarts) if (gaps != 1)
		{
			int start  = matchStart.intValue();
			int wordStart = start - ends[0]; // the node in the sentence which could match the first child node 0 of the word
			Vector<int[]> hungover = new Vector<int[]>(); // for sound nodes in the sentence which do not match word scenes after the word sounds
			boolean inSentence = true;
			
			// loop over unused child nodes of the sentence, 
			for (int sp = 0; sp < unused.size(); sp++) // sp = position in the sentence
			{
				FSNode sentChild = unused.get(sp);
				boolean isScene = (sentChild.getType().equals("scene"));
				int wp = sp - wordStart; // wp = position in the word
				
				// before the start of the word, there is no possible match
				if (wp < 0) addMatch(res,-1,sp);
				
				// first node of the word  - might be a scene or a sound
				else if (wp == 0)
				{
					// word FS starts with a sound; must match a sound in the sentence
					if ((ends[0] == 0) && (!isScene)) addMatch(res,wp,sp);
					// word FS starts with a scene, and the sentence has a scene to match it
					else if ((ends[0] == 1) && (isScene))  addMatch(res,wp,sp);
					// word FS starts with a scene, and the sentence does not match it 
					else
					{
						gaps++;
						addMatch(res,-1,sp); // add the sentence sound first
						addMatch(res,wp,-1); // add the word meaning scene as a gap
					}
					
					// now consume all the rest of the word, and add its meaning scene
					for (wp = 1; wp < wordSize; wp++)
					{
						sp = wp + wordStart; // keep sentence position in step
						if (sp < unused.size())
						{
							sentChild = unused.get(sp);
							isScene = (sentChild.getType().equals("scene"));
						}
						else inSentence = false;
						
						// other sounds in the word - must match
						if (wp < ends[1] + 1) addMatch(res,wp,sp);
						
						// input meaning scenes in the word, after the word sounds 
						else if (wp < wordSize -1)
						{
							if  (isScene) 
							{
								// if the word has overshot the sentence
								if (!inSentence) addMatch(res,wp,-1);
								// otherwise match the meaning scenes
								else addMatch(res,wp,sp);
							}
							// sound scenes in the sentence do not match a meaning scene in the word
							else if  (!isScene)
							{
								gaps++;
								addMatch(res,wp,-1);
								if (inSentence) addMatch(hungover,-1,sp); // will need to add unmatched sounds from the sentence, after the word result scene
							}
						}

						// result scene of the word - not matched
						else if (wp == wordSize -1) 
						{
							addMatch(res,wp,-1);
							for (int[] next: hungover) res.add(next); // the sounds which did not match the post-sound input scenes of the word
							
							sp--; // the result scene of the word does not correspond to a sentence scene
						}
						
					} // end of loop over wp; sp has been increased to (wordStart + wordSize - 2)
					
				} // end of if (wp == 0)
				
				// rest of the sentence, if there are any sp > (wordStart + wordSize - 2)
				else addMatch(res,-1,sp);
				
			} // end of loop over sp
			
		} // end of loop over allMatchStarts
		
		// too many gaps - return an empty result
		if (gaps > 1) res = new Vector<int[]>();
		
		return res;
	}
	
	/**
	 * add an instruction to add or unify a child
	 * @param res
	 * @param wp
	 * @param sp
	 */
	private static void addMatch(Vector<int[]> res, int wp, int sp) 
	{
		int[] next = new int[2];
		next[0] = wp;
		next[1] = sp;
		res.add(next);
	}
	
	/**
	 * find the indexes of the first and last sound nodes in a word FS
	 * @param fWord
	 * @return
	 */
	private static int [] soundEnds(FSNode fWord)
	{
		int[] ends = new int[2];
		ends[0] = -1;
		ends[1] = -1;
		for (int i = 0; i < fWord.getChildren().size(); i++)
		{
			FSNode child = fWord.getChild(i);

			// in a run of sound nodes
			if (child.getType().equals("sound"))
			{
				//set ends[1] for every sound node, so it ascends
				ends[1] = i;
				// set ends[0] only for the first sound node
				if (ends[0] == -1) ends[0] = i;
			}

		}		
		return ends;
	}


	
	/**
	 * 
	 * @param fWord
	 * @return a list of the unused child nodes of a sentence script FSNode
	 * (in both the full list of sentence child nodes, and the shortened list)
	 */
	private static Vector<FSNode> removeUsed(FSNode fWord)
	{
		Vector<FSNode> result = new Vector<FSNode>();
		
		for (int i = 0; i <  fWord.getChildren().size(); i++)
		{
			FSNode child = fWord.getChild(i);
			// mark the list position of all child node
			child.setFullListPosition(i);
			// add unused child nodes to the result
			if (!child.isUsed()) result.add(child);
		}				
		return result;		
	}

	/**
	 * 
	 * @param unified
	 * @param Sent
	 * @param wordPos
	 * @return
	 */
	private static Vector<FSNode> restoreUsedNodes(Vector<FSNode> unified,FSNode sent,Vector<int[]>wordPos)
	{
		Vector<FSNode> result = new Vector<FSNode>();
		Vector<FSNode> original = sent.getChildren();
		int lastPos = 0;
		
		for (FSNode res: unified)
		{
			int oPos = res.getFullListPosition(); // position of this node in the original list

			// the word result or gaps in the word are added immediately
			if (oPos < 0) result.add(res);
			
			// if the unified list gets ahead of the original full list, catch up
			else if (oPos > -1)
			{
				if (oPos > lastPos)
				{
					for (int pos = lastPos; pos < oPos; pos++)
					{
						FSNode toAdd = original.get(pos);
						// safety check to avoid duplication - add back only used nodes
						if (toAdd.isUsed()) result.add(toAdd);
					}
					// record that you have caught up
					lastPos = oPos;
				}
				
				// always add from the unified result
				result.add(res);
			}
		}
				
		return result;		
	}


	
	
	/**
	 * apply subsumption links at script level in unification for understanding
	 * 
	 * @param unified Vector of child nodes of the top script node, following unification wit input scenes of the word
	 * @param wordResult the result scene of the word
	 * 
	 * it does not alter elements of Unified, but returns an altered wordResult
	 */
	private static FSNode applySubsumptionLinks(Vector<FSNode> unified, FSNode wordResult) throws Exception
	{
		for (FSNode uChild: unified)
		{
			checkInputOrigins(uChild,3);

			// find any child which has a subsumption link to the word result
			FSNode subsumer = getSubsumingChild(uChild);
			if (subsumer != null)
			{
				checkInputOrigins(subsumer,4);
				
				String id = subsumer.getSubsumingId();
				
				// find the target of the subsumption link in the word result
				FSNode subsumed = getSubsumedChild(wordResult,id);
				
				if (subsumed != null)
				{
					// unify for the subsumption link
					FSNode newSubsumed = unify(subsumer,new FSNode(subsumed), PLAIN);
					newSubsumed.setUsed(false);

					// to get left and right sides of the origin right after a subsumption link unification
					// swapOrigins(newSubsumed);

					// replace the subsumed subtree of wordResult, using getId() to find what to replace
					// (this assumes that a later subsumption link does not dominate an earlier one)
					if (newSubsumed.isViable()) wordResult = new FSNode(wordResult,newSubsumed,UNDERSTAND);
					
					// if there is a problem, propagate the reason upwards
					else if (!newSubsumed.isViable())
					{
						wordResult.setViable(false);
						wordResult.setDescription(newSubsumed.getDescription());
					}
				}
				else if (subsumed == null) throw new Exception("Cannot find end of subsumption link "  + id);
			}
		}
		
		return wordResult;
	}
	
	/**
	 * 
	 * @param parent
	 * @return
	 */
	public static FSNode getSubsumingChild(FSNode first)
	{
		if (first.getSubsumingId() != null) return first;
		
		for (FSNode child: first.getChildren())
			if (getSubsumingChild(child) != null)
				return getSubsumingChild(child);
		
		return null;
	}
	
	/**
	 * 
	 * @param id
	 * @return
	 */
	public static FSNode getSubsumedChild(FSNode first, String id)
	{
		if (id.equals(first.getSubsumedId())) return first;
		
		for (FSNode child: first.getChildren())
		{
			FSNode res = getSubsumedChild(child,id);
			if (res != null) return res;
		}
		
		return null;
	}

	
	/**
	 * 
	 * @param childFS
	 * @return index of the first child FS which is a sound
	 */
	public static  int firstSound(Vector<FSNode> childFS)
	{
		int firstSound = 0;
		boolean found = false;
		for (int c = 0; c < childFS.size(); c++)
		{
			if (!found) firstSound = c;
			if(childFS.get(c).getType().equals("sound")) found = true;
		}
		return firstSound;
	}
	
	
/**
 * find all possible ways to match all the consecutive sounds of the word with consecutive sounds (unused) of the sentence
 * @param fWord
 * @param fSent
 * @return a Vector of Integers like {0, 5,..} etc. 
 * each N is the start position in the sentence FS of a sequence of unused sounds which match the sounds of the word
 */
public static Vector<Integer> matchSounds(FSNode fWord, Vector<FSNode> fromSent) throws Exception
{
	Vector<Integer> matches = new Vector<Integer>();
	
	// find the sequence of sounds of the word
	Vector<String> sounds = wordSounds(fWord);
	
	// try out all possible start positions for the word in the sentence
	if (sounds.size() > 0)
	{
		int size = fromSent.size();
		int sLast = size - sounds.size() + 1;
		for (int sStart = 0; sStart < sLast ; sStart++)
		{
			boolean matched = true;
			for (int ws = 0; ws < sounds.size(); ws++)
			{
				FSNode cPos = fromSent.get(sStart + ws);
				
				if (!cPos.getType().equals("sound")) matched = false;
				else
				{
					if (cPos.slotValue("sound") == null) throw new Exception("Missing sound slot in FS ");
					if (!cPos.slotValue("sound").equals(sounds.get(ws))) matched = false;
					// if (cPos.isUsed()) matched = false;
				}
			}
			
			// if all sounds have matched, record another valid start point in the sentence
			if (matched) matches.add(Integer.valueOf(sStart));
		}
	}
	
	return matches;
}

		
	/**
	 * find all possible ways to match all the consecutive sounds of the word with consecutive sounds (unused) of the sentence
	 * @param fWord
	 * @param fSent
	 * @return a Vector of Integers like {0, 5,..} etc. 
	 * each N is the start position in the sentence FS of a sequence of unused sounds which match the sounds of the word
	 */
	public static Vector<Integer> matchSounds(FSNode fWord, FSNode fSent) throws Exception
	{
		Vector<Integer> matches = new Vector<Integer>();
		
		// find the sequence of sounds of the word
		Vector<String> sounds = wordSounds(fWord);
		
		// try out all possible start positions for the word in the sentence
		if (sounds.size() > 0)
		{
			int size = fSent.getChildren().size();
			int sLast = size - sounds.size() + 1;
			for (int sStart = 0; sStart < sLast ; sStart++)
			{
				boolean matched = true;
				for (int ws = 0; ws < sounds.size(); ws++)
				{
					FSNode cPos = fSent.getChild(sStart + ws);
					
					if (!cPos.getType().equals("sound")) matched = false;
					else
					{
						if (cPos.slotValue("sound") == null) throw new Exception("Missing sound slot in FS " + fSent.getName());
						if (!cPos.slotValue("sound").equals(sounds.get(ws))) matched = false;
						if (cPos.isUsed()) matched = false;
					}
				}
				
				// if all sounds have matched, record another valid start point in the sentence
				if (matched) matches.add(Integer.valueOf(sStart));
			}
		}
		
		return matches;
	}
	
	/**
	 * 
	 * @param fWord
	 * @return
	 * @throws Exception
	 */
	public static Vector<String> wordSounds(FSNode fWord) throws Exception
	{
		Vector<String> sounds = new Vector<String>();
		boolean started  = false;
		boolean ended = false;
		for (int ws = 0; ws < fWord.getChildren().size() ;ws++)
		{
			FSNode child = fWord.getChild(ws);
			if (child.getType().equals("sound"))
			{
				// record start of sounds in the word FS
				if (!started) started = true;
				if (started) sounds.add(child.getSlotValue("sound"));
				
				// a sound after a non-sound after a sound
				if (ended) throw new Exception("Word sounds are not contiguous for word " + fWord.getName());
			}
			
			// a non-sound after a sound; word sounds have ended
			else if (started) ended  = true;
		}
		return sounds;
	}

	
	/**
	 * 
	 * @param fWord
	 * @return
	 * @throws Exception
	 */
	public static Vector<FSNode> wordSoundFS(FSNode fWord) throws Exception
	{
		Vector<FSNode> sounds = new Vector<FSNode>();
		boolean started  = false;
		boolean ended = false;
		for (int ws = 0; ws < fWord.getChildren().size() ;ws++)
		{
			FSNode child = fWord.getChild(ws);
			if (child.getType().equals("sound"))
			{
				// record start of sounds in the word FS
				if (!started) started = true;
				if (started) sounds.add(child);
				
				// a sound after a non-sound after a sound
				if (ended) throw new Exception("Word sounds are not contiguous for word " + fWord.getName());
			}
			
			// a non-sound after a sound; word sounds have ended
			else if (started) ended  = true;
		}
		return sounds;
	}

	/**
	 * 
	 * @param p1
	 * @param p2
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> unifyChildPlain(FSNode p1, FSNode p2) throws Exception
	{
		Vector<FSNode> result = new Vector<FSNode>();
		
		// vectors of child FS  of inputs
		Vector<FSNode> cc1 = p1.getChildren();
		Vector<FSNode> cc2 = p2.getChildren();
		
		// find all types of child node in both lists of children
		Hashtable<String,String> allTypes = new Hashtable<String,String>();
		for (FSNode c1:cc1) allTypes.put(c1.getType(), "1");
		for (FSNode c2:cc2) allTypes.put(c2.getType(), "1");
		
		// loop over child node types
		for (Enumeration<String> en = allTypes.keys();en.hasMoreElements();)
		{
			String type = en.nextElement();
			Vector<FSNode> tc1 = p1.getTypeChildren(type);
			Vector<FSNode> tc2 = p2.getTypeChildren(type);
			
			// if there are no FS of this type in one of the lists, pass across the other list (no unification)
			if (tc1.size() == 0) for (FSNode t2: tc2) result.add(t2);
			else if (tc2.size() == 0) for (FSNode t1:tc1) result.add(t1);
			
			// if both lists have some FS, permute the longer list to find the best match (fewest failed unifications, least information score), and add them to the result
			else
			{
				Vector<FSNode> toAdd = bestUnify(tc1,tc2);
				for (FSNode added: toAdd) result.add(added);
			}
		}
		
		return result;
		
	}
	
	/**
	 * tc1 and tc2 are non-empty lists of FS of the same type.
	 * Permute the longer list, and unify its first FS with the un-permuted shorter list.
	 * Pass across the rest of the longer list.
	 * Out of all permutations, keep the result that has the fewest failed unifications; 
	 * and within that set, the least total information score.
	 * 
	 * This forces you to try to unify all FS in the shorter list - some may fail.
	 * @param tc1
	 * @param tc2
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> bestUnify(Vector<FSNode> tc1, Vector<FSNode> tc2) throws Exception
	{
		Vector<FSNode> best = new Vector<FSNode>();
		int s1 = tc1.size();
		int s2 = tc2.size();
		int leastFails = 10;
		double leastScore = 10000;
		
		// tc2 is the longer or equal list; permute it and try to unify the first FS in it with those in tc1
		if (s2 > s1-1)
		{
			Vector<Vector<FSNode>> perms = Util.permutations(tc2);
			for (Vector<FSNode> perm:perms)
			{
				// do all unifications or carry-overs for this permutation
				Vector<FSNode> current = new Vector<FSNode>();
				for (int i = 0; i < s2; i++)
				{
					if (i < s1) current.add(unify(tc1.get(i),perm.get(i),PLAIN));
					else current.add(perm.get(i));
				}
				
				// find the number of failed unifications, and information scores
				int fails = failures(current);
				double score = infoScore(current);
				
				// if there are more failures than a previous try, do nothing
				
				// if there are fewer failures than before, ignore the score
				if (fails < leastFails)
				{
					best = current;
					leastFails = fails;
					leastScore = score;
				}
				
				// if failures are equal, use least information score as a tie-breaker
				if ((fails == leastFails) && (score < leastScore))
				{
					best = current;
					leastFails = fails;
					leastScore = score;
				}
			}
		}
		
		// tc1 is the longer list; permute it and try to unify the first FS in it with those in tc2
		if (s1 > s2)
		{
			Vector<Vector<FSNode>> perms = Util.permutations(tc1);
			for (Vector<FSNode> perm:perms)
			{
				// do all unifications or carry-overs for this permutation
				Vector<FSNode> current = new Vector<FSNode>();
				for (int i = 0; i < s1; i++)
				{
					if (i < s2) current.add(unify(perm.get(i),tc2.get(i),PLAIN));
					else current.add(perm.get(i));
				}
				
				// find the number of failed unifications, and information scores
				int fails = failures(current);
				double score = infoScore(current);
				
				// if there are more failures than a previous try, do nothing
				
				// if there are fewer failures than before, ignore the score
				if (fails < leastFails)
				{
					best = current;
					leastFails = fails;
					leastScore = score;
				}
				
				// if failures are equal, use least information score as a tie-breaker
				if ((fails == leastFails) && (score < leastScore))
				{
					best = current;
					leastFails = fails;
					leastScore = score;
				}
			}
		}
		
		return best;		
	}
	

	// the number of failed unifications in a list of FS
	private static int failures(Vector<FSNode> current)	
	{
		int fails = 0;
		for (FSNode c:current) if (!c.isViable()) fails++;
		return fails;
	}
	
	// the total information score of a list of FS
	private static double infoScore(Vector<FSNode> current)	
	{
		double info = 0;
		for (FSNode c:current) info = info + c.getInformation(null);
		return info;
	}
	

	
	//--------------------------------------------------------------------------
	//                      Operations - generalisation
	//--------------------------------------------------------------------------
	
	static void trace(int i) {Util.message("trace " + i);}
	
	/**
	 * @param first
	 * @param other
	 * @return the generalisation of the two feature structures
	 */
	public static FSNode generalise(FSNode first, FSNode other) throws Exception
	{
		// int before = Util.freeBefore();
		
		// make clones for safety (e.g. to mark them with ids)
		FSNode f1 = new FSNode(first);
		FSNode f2 = new FSNode(other);
		
		// set true positions 1..N to compute origins
		String type = first.getType();
		if (type.equals("script"))
		{
			f1.setAllTruePositions();
			f2.setAllTruePositions();
		}
		
		// make a basic empty feature structure
		FSNode result = new FSNode(first.getModel(), first.getType());
		result.setName("Result"); // name to be set later to something more descriptive
		result.setViable(false);
		result.setOrigin(f1.getTruePosition(), f2.getTruePosition());
		
		// fail if types do not match
		if (!type.equals(f2.getType())) return result;
		
		
		// remove any child nodes that have been used (e.g. in unification to partly understand learning examples)
		removeUsedChildNodes(f1);
		removeUsedChildNodes(f2);
		
		// mark with generalisation ids, to compute order constraints, and to find and check subsumption links
		if (type.equals("script"))
		{
			Util.markIds(f1,"k1");
			Util.markIds(f2,"l1");
		}
		
		// generalise slot values - empty table if the result has zero information
		Hashtable<String,String> genSlotValues = generaliseSlots(f1, f2);
		
		// make the result viable if it has any generalised slot values
		for (Enumeration<String> en = genSlotValues.keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String value = genSlotValues.get(slotName);
			result.setSlotValue(slotName, value);
			result.setViable(true);
		}
		
		// generalise child FS, permuting to maximise information content
		Vector<FSNode> genChildren = new Vector<FSNode>();
		if (first.getType().equals("script")) {genChildren = generaliseChildScript(f1,f2);}
		else genChildren = generaliseChildPlain(f1,f2);
		// remove empty child nodes, except 'thing' and 'scene' nodes which are always kept
		genChildren = removeEmptyChildren(genChildren);
		result.setChildren(genChildren);
		
		// the result is viable if it has any slots or any viable children
		if (!result.isViable()) 
		{
			boolean viable = false;
			for (FSNode child:genChildren) if (child.isViable()) viable = true;
			result.setViable(viable);
		}
		
		// set distinct generalisation id,so each FS can be traced to the FS it came from
		String joinedId = f1.getGenId() + "*" + f2.getGenId();
		result.setGenId(joinedId);
		
		// use the joined ids of child FS to compute the order constraints for the result
		// Vector<int[]> genOrder = Util.combineOrders(result,f1,f2);
		
		// for words with some sound scenes in the result (not frames) make or check the subsumption links - always from the top level
		String content = "";
		if ((type.equals("script")) && (result.soundOnly().getChildren().size() > 0))
		{
			findSubsumptionLinks(result,f1,f2);
			
			// find the information content of the inputs and the result
			double fullInfo = result.getInformation(null);
			FSNode rScene = result.getLastChild();
			double resultInfo = rScene.getInformation(null);
			double inputInfo = fullInfo - resultInfo; // all child nodes except the meaning result
			int i1 = (int)inputInfo;
			int i2 = (int)resultInfo;
			result.setInfoContent(i1, i2);
			content = ": input information " + i1 + " bits; final meaning information " + i2 + " bits.";
		}
		
		// record the information contents and the expression on the top node
		result.setDescription("Result of generalisation" + content);
		result.setExpression(getExpression(first,other,GENERALISE));
		
		// set the unify order of the result (for compact drawing of FS)
		setResultUnifyOrder(result,first,other);
		
		// the result of generalisation should subsume both inputs at any level. Test to throw an exception, or allow a break point
		if ((result != null) && (result.isViable()))
		{
			if (!subsumes(result,first))
			{
				throw new Exception("First generalisation input of type " + first.getType() + " is not subsumed by the result");
			}
			if (!subsumes(result, other))
			{
				throw new Exception("Second generalisation input of type " + other.getType() + " is not subsumed by the result");
			}
			
			// check that origins have been correctly computed, at every level
			try {checkOperationOrigins(result, f1, f2, true);}
			catch (Exception ex)
			{
				boolean writeOriginMessages = false;
				if (writeOriginMessages) Util.message(ex.getMessage());
			}
		}
		
		// Util.freeUsed(before, "Generalise " + type);
				
		return result;
	}
	
	/**
	 * Usually, true positions of FSNodes are set zero all the way down the tree, 
	 * and generalisation only sets up true positions once at script level, in clones of the two inputs,
	 * to put the true positions in origins in the result.
	 * 
	 * Use this form of generalisation if:
	 * (a) the top nodes you are generalising might not be script nodes
	 * (b) you want true positions to be left set on the inputs, 
	 * so true positions can be used with origins in the result, to match nodes between the inputs
	 * 
	 * @param first
	 * @param other
	 * @return
	 * @throws Exception
	 */
	public static FSNode traceGeneralise(FSNode first, FSNode other) throws Exception
	{
		first.setAllTruePositions();
		other.setAllTruePositions();
		
		return generalise(first, other);
	}

	
	/**
	 *  remove used child nodes from a script
	 * @param f
	 */
	private static void removeUsedChildNodes(FSNode f)
	{
		Vector<FSNode> newChildren = new Vector<FSNode>();
		
		for (FSNode child: f.getChildren())
			if (!child.isUsed()) newChildren.add(child);
		
		f.setChildren(newChildren);
	}
	
	/**
	 * remove any empty children from a generalise result 
	 * (of type not  = thing or scene, which are always kept) 
	 * @param genChildren
	 */
	private static Vector<FSNode> removeEmptyChildren(Vector<FSNode> genChildren)
	{
		Vector<FSNode> result  = new Vector<FSNode>();		
		for (FSNode f:genChildren)
		{
			String type = f.getType();
			if (type.equals("scene")) result.add(f);
			else if (type.equals("thing")) result.add(f);
			else if (!isEmpty(f)) result.add(f);
		}
		
		return result;
	}
	
	/**
	 * 
	 * @param f
	 * @return true if f and its subtree are empty
	 */
	private static boolean isEmpty(FSNode f)
	{
		boolean empty = true;
		if (f.getOwnSlotValues().size() > 0) empty = false;
		for (FSNode child: f.getChildren())
			if (!isEmpty(child)) empty = false;
		return empty;
	}


	
	/**
	 * 
	 * @param f1
	 * @param f1
	 * @return generalised slot values - or null if there are any incompatible slots
	 */
	private static Hashtable<String,String>  generaliseSlots(FSNode f1, FSNode f2) throws Exception
	{
		String type = f1.getType(); // for debugging
		
		Hashtable<String,String> result = new Hashtable<String,String>();
		
		// the result can only have a slot value if f1 has a slot value
		for (Enumeration<String> en = f1.getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			Slot slot = f1.getModel().getSlot(slotName);
			if (slot == null) throw new Exception("Cannot find slot " + slotName + " from FS " + f1.getName() + " in the model");
			String val1 = f1.getSlotValue(slotName); // known to be be non-null
			String val2 = f2.getSlotValue(slotName);
			
			// the result can only have a slot value if both f1 and f2 have a slot value
			if (val2 != null) 
			{
				String genValue =  slot.generaliseValue(val1, val2);
				if (genValue != null) result.put(slotName, genValue);
			}
		}
		
		return result;
	}
	
	/**
	 * generalise the child nodes at script level, where there may be many 'sound' child nodes.
	 * @param p1
	 * @param p2
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> generaliseChildScript(FSNode p1, FSNode p2) throws Exception
	{
		Vector<FSNode> res = new Vector<FSNode>(); // to indicate failure, if no sounds match
		
		// find the start positions of the longest contiguous sequence of matching sound nodes
		int[] bestMatch = matchSoundSequence(p1,p2);
		// longest sequence of matching sounds
		int length = bestMatch[2];
		
		// word learning, where there is a matching sound sequence 
		if (length > 0)
		{
			// form compressed versions of both input FS, removing sound nodes which do not match
			FSNode q1 = compress(p1,bestMatch[0],length);
			FSNode q2 = compress(p2,bestMatch[1],length);
			
			// generalise of the child  nodes (sound or scene) of the compressed FS, in order
			res = generaliseScriptChildNodes(q1,q2);
		}
		
		// frame learning, where neither input has any sound sequence - only meaning scenes
		else if ((p1.soundOnly().getChildren().size() == 0) && (p2.soundOnly().getChildren().size() == 0))
		{
			// generalise of the child scene nodes in order
			res = generaliseScriptChildNodes(p1,p2);
		}


		return res;
		
	}
	
	/**
	 * compress a feature structure by:
	 *  - remove all sound child FS except the matching sequence
	 *  
	 * @param input
	 * @param soundStart index of the first matching sound child node
	 * @param soundLen number of consecutive matching sound nodes
	 * @return
	 */
	private static FSNode compress(FSNode input, int soundStart, int soundLen)
	{
		// clone for safety
		FSNode res = new FSNode(input);
		Vector<int[]> orders = res.getFreeOrder();
		
		// build up the compressed child list
		Vector<FSNode> newChildren = new Vector<FSNode>();
		
		for (int oldPos = 0; oldPos < input.getChildren().size();oldPos++)
		{
			boolean added = false;
			FSNode child = input.getChildren().get(oldPos);
			if (child.getType().equals("sound"))
			{
				// sounds in the matching sequence
				if ((oldPos > soundStart - 1) && (oldPos < soundStart + soundLen))
				{
					newChildren.add(child);
					added = true;
				}
				// otherwise added stays false
			}
			
			// add any non-sound children
			else
			{
				newChildren.add(child);
				added = true;
			}
			
			// the child at position i has been either removed (added = false) or added at position newPos (added= true)			
			int newPos = newChildren.size( )- 1;
			// update the order constraints to go on the parent for this change
			orders = Util.updateOrders(orders, oldPos, newPos,added);
		}
		
		// reset the child nodes
		res.setChildren(newChildren);
		
		return res;
	}
	
	
	
	/**
	 * find the longest matching contiguous sound sequence in 2 FS
	 * 
	 * @param p1
	 * @param p2
	 * @return an int[] with 3 elements:
	 * 
	 * 0 : index of first matching sound in p1
	 * 1 : index of first matching sound in p2
	 * 2 : length of sequence of matching sounds (the maximum)
	 */
	public static int[] matchSoundSequence(FSNode p1, FSNode p2)
	{
		// set failure values for result
		int[] result = new int[3];
		result[0] = -1;
		result[1] = -1;
		result[2] = 0;

		int maxLen = 0;
		
		int s1 = p1.getChildren().size();
		int s2 = p2.getChildren().size();
		
		for (int i = 0; i < s1; i++)
		{
			FSNode c1 = p1.getChild(i);
			if (c1.getType().equals("sound"))
			{
				for (int j = 0; j < s2; j++)
				{
					FSNode c2 = p2.getChild(j);
					// only try for a sequence if the initial sound matches
					if ((c2.getType().equals("sound")) && (c1.getSlotValue("sound").equals(c2.getSlotValue("sound"))))						
					{
						boolean ended = false;
						int len = 0; // increment length until ended
						while (!ended)
						{
							FSNode d1 = p1.getChild(i + len);
							FSNode d2 = p2.getChild(j + len);
							// we already know that the initial sounds (for len = 0) match
							if ((d1.getType().equals("sound")) && 
								(d2.getType().equals("sound")) && 
								(d1.getSlotValue("sound").equals(d2.getSlotValue("sound"))))
							{
								len++;
								// some improvement over previous best
								if (len > maxLen)
								{
									maxLen = len; // length starts at 1 if there is any match
									result[0] = i; // record start position in p1
									result[1] = j; // record start position in p2
								}
								// avoid overshoots of child nodes of either FS
								ended = ((i + len > s1 -1)||(j + len > s2 -1));
							}
							
							// non-sounds or non-matching sounds end this attempt
							else ended = true;							
						} // end of loop to extend a sequence
					} 	// end of 'if this is a possible sequence'		
				} // end of loop over children of p2
				
			}	// end of 'if c1 is a sound'		
		} // end of loop over children of p1
		
		result[2] = maxLen; //zero for failure
		return result;
	}

	
	/**
	 * child nodes of a script node can only be <sound> or <scene>
	 * the sound nodes have been pre-filtered so they are only one sequence of sounds which match
	 * Generalise the scene nodes in three stages:
	 * 
	 * (1) those before the sound nodes
	 * (2) those after the sound nodes, except the final meaning scene
	 * (3) the final meaning scene.
	 * 
	 * Usually the number of scenes in (1) and(2) matches, so there is no choice of pairings; 
	 * but this method copes with non-matching numbers, 
	 * finding the pairing with highest information content	  
	 *
	 * @param p1
	 * @param p2
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> generaliseScriptChildNodes(FSNode p1, FSNode p2) throws Exception
	{		
		// mark position indexes of child nodes. 
		p1.markChildIndexes();
		p2.markChildIndexes();
		
		Vector<FSNode> res = new Vector<FSNode>();
		
		// vectors of child FS
		Vector<FSNode> c1 = p1.getChildren();
		Vector<FSNode> c2 = p2.getChildren();
		
		// best match of pre-sound scenes
		Vector<FSNode> a1 = nonSound(c1,true);
		Vector<FSNode> a2 = nonSound(c2,true);		
		Vector<FSNode> pre = bestSceneMatch(a1,a2);
		for (FSNode n:pre) res.add(n);
		
		// unique match of sounds (pre-computed, but need to set origins; and check the match at the same time)
		int s1 = a1.size(); // start of sounds in c1
		int s2 = a2.size(); // start of sounds in c2
		Vector<FSNode> sounds = getSounds(c1);
		for (int s = 0; s < sounds.size(); s++)
		{
			FSNode ss1 = c1.get(s + s1);
			FSNode ss2 = c2.get(s + s2);
			String so1 = ss1.getSlotValue("sound");
			String so2 = ss2.getSlotValue("sound");
			if (!so1.equals(so2)) throw new Exception("Non-matching sounds in generalisation: " + so1 + " and " + so2);
			
			FSNode ss = new FSNode(sounds.get(s));
			ss.setOrigin(ss1.getTruePosition(), ss2.getTruePosition());
			res.add(ss);
		}
				
		// best match of post-sound scenes
		Vector<FSNode> b1 = nonSound(c1,false);
		Vector<FSNode> b2 = nonSound(c2,false);
		Vector<FSNode> post = bestSceneMatch(b1,b2);
		for (FSNode n:post) res.add(n);
		
		// match the final meaning scenes
		FSNode last = generalise(c1.get(c1.size() -1),c2.get(c2.size() -1));
		res.add(last);
		
		return res;
	}
	
	/**
	 * extract sound nodes only, and clone them
	 * @param c
	 * @return
	 */
	private static Vector<FSNode> getSounds(Vector<FSNode> c)
	{
		Vector<FSNode> res = new Vector<FSNode>();
		
		for (FSNode child: c) 
			if (child.getType().equals("sound"))
				res.add(new FSNode(child));
		
		return res;
	}

	
	/**
	 * 
	 * @param c
	 * @param preSound
	 * @return scene child nodes (except the last), either before any sounds, or after
	 */
	private static Vector<FSNode> nonSound(Vector<FSNode> c, boolean preSound)
	{
		Vector<FSNode> res = new Vector<FSNode>();
		boolean soundStarted = false;
		
		// miss out the last meaning scene
		for (int i = 0; i < c.size() - 1; i++)
		{
			FSNode child = c.get(i);
			if (child.getType().equals("sound")) soundStarted = true;
			else if ((child.getType().equals("scene")) && (preSound != soundStarted)) 
				res.add(child);
		}
		
		return res;
	}
	
	/**
	 * the lists of scene nodes c1 and c2 are normally of equal length.
	 * In this case they must be generalised in order, with no choice.
	 * 
	 * However, if one list is longer than the other, all scenes in the shorter list must be generalised,
	 * but there is some choice of scenes from the longer list.
	 * Choose them to maximise the information in the result.
	 * @param c1 list of scene nodes
	 * @param c2 list of scene nodes
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> bestSceneMatch(Vector<FSNode> c1, Vector<FSNode> c2) throws Exception
	{
		// try out all possible ordered pairings between the child lists, looking for the largest information content
		Vector<FSNode> bestResult = new Vector<FSNode>();
		double bestInformation  = 0.0;

		int s1 = c1.size();
		int s2 = c2.size();
		
		// if there can be any generalisation scene nodes
		if ((s1 > 0) && (s2 > 0))
		{
			
			boolean ordered = true;
			Vector<Vector<int[]>> allPairings  = Util.pairings(s1,s2,ordered);
			
			// each Vector<int[]> pairing has length = min(s1,s2). Each int[] has two elements, in the ranges 0..s1-1 and 0..s2-1
			for (Vector<int[]> pairing: allPairings)
			{
				double info = 0.0;
				Vector<FSNode> aResult = new Vector<FSNode>();
				for (int[] pair: pairing)
				{
					FSNode gen = generalise(c1.get(pair[0]),c2.get(pair[1]));
					info = info + gen.getInformation(null);
					aResult.add(gen);
				}
				
				if  (info > bestInformation - 0.01)
				{
					bestInformation = info;
					bestResult = aResult;
				}
				
			} // end of loop over possible pairings of scene nodes
			
		} // end of if ((s1 > 0) && (s2 > 0))
		
		return bestResult;
	}


	
	/**
	 * 
	 * @param p1
	 * @param p2
	 * @return
	 * @throws Exception
	 */
	private static Vector<FSNode> generaliseChildPlain(FSNode p1, FSNode p2) throws Exception
	{
		Vector<FSNode> res = new Vector<FSNode>();
		
		// mark position indexes of child nodes. 
		p1.markChildIndexes();
		p2.markChildIndexes();
		
		// vectors of child FS
		Vector<FSNode> c1 = p1.getChildren();
		Vector<FSNode> c2 = p2.getChildren();
		
		Hashtable<String,String> types = allTypes(c1,c2);
		
		// loop over child node types in either list
		for (Enumeration<String> en = types.keys(); en.hasMoreElements();)
		{
			String type = en.nextElement();
			c1 = p1.getTypeChildren(type);
			c2 = p2.getTypeChildren(type);
			
			int s1 = c1.size();
			int s2 = c2.size();
			
			// if there can be any generalisation nodes of this type
			if ((s1 > 0) && (s2 > 0))
			{
				 /* construct the Cartesian product of all possible pairwise generalisations;
				 * after this, no more child generalisations are needed */
				FSNode [][] cartesian = new FSNode [s1][s2];
				for (int i = 0; i < s1; i++)
					for (int j = 0; j < s2; j++)
					{
						// clone to avoid contamination
						FSNode d1 = new FSNode(c1.get(i));
						FSNode d2 = new FSNode(c2.get(j));
						cartesian[i][j] = generalise(d1,d2);
					}
				
				// try out all possible pairings between the child lists, looking for the largest information content
				Vector<FSNode> bestResult = new Vector<FSNode>();
				double bestInformation  = 0.0;
				
				boolean ordered = false;
				Vector<Vector<int[]>> allPairings  = Util.pairings(s1,s2,ordered);
				
				// each Vector<int[]> pairing has length min(s1,s2). Each int[] has two elements, in the ranges 0..s1-1 and 0..s2-1
				for (Vector<int[]> pairing: allPairings)
				{
					double info = 0.0;
					Vector<FSNode> aResult = new Vector<FSNode>();
					for (int[] pair: pairing)
					{
						FSNode gen = cartesian[pair[0]][pair[1]];
						info = info + gen.getInformation(null);
						aResult.add(gen);
					}
					
					if  (info > bestInformation - 0.01)
					{
						bestInformation = info;
						bestResult = aResult;
					}
					
				} // end of loop over possible pairings of child nodes of one type
				
				// add the result of the best pairing of this type to the full result
				for (FSNode best: bestResult) res.add(best);
				
			} // end of if ((s1 > 0) && (s2 > 0))
			
		} // end of loop over child node types
		
		return res;
			
	}
	
			

	
	/**
	 * all types in two lists of FSNode
	 * @param v1
	 * @param v2
	 * @return
	 */
	private static Hashtable<String,String> allTypes(Vector<FSNode> v1, Vector<FSNode> v2)
	{
		Hashtable<String,String> types = new Hashtable<String,String>();		
		for (FSNode f1: v1) types.put(f1.getType(),"1");
		for (FSNode f2: v2) types.put(f2.getType(),"1");		
		return types;
	}

	/**
	 * 
	 * @param result of generalisation  = f1^f2
	 * @param f1 one input
	 * @param f2 the other input
	 */
	private static void findSubsumptionLinks(FSNode result,FSNode f1,FSNode f2) throws Exception
	{
		// find all nodes in the result which are not 'sound' or 'script'
		Vector<FSNode> nodes = nonSoundNodes(result);
		
		// increment to make all ids unique
		int id = 1;
		
		// find all pairs of nodes in the result which are of the same type (not sound or script)
		for (int m = 0; m < nodes.size(); m++)
		{
			// first node in the result
			FSNode g = nodes.get(m);

			// find the nodes in the inputs which were generalised to get this node
			FSNode g1 = findSameGenId(g,f1);
			FSNode g2 = findSameGenId(g,f2);
			if ((g1 == null)||(g2==null)) throw new Exception("Null first source " +  m + " of node type " + g.getType());
			
			// find a different second node in the result
			for (int n = m + 1; n < nodes.size(); n++)
			{
				FSNode h = nodes.get(n);
				
				// find the nodes in the inputs which were generalised to get this node
				FSNode h1 = findSameGenId(h,f1);
				FSNode h2 = findSameGenId(h,f2);
				if ((h1 == null)||(h2==null)) throw new Exception("Null second source " +  n + " of node type " + h.getType());
								
				if (g.getType().equals(h.getType()))
				{
					// ids keep ascending even if the subsumption test is not passed
					String linkId = "k_" + id;
					id++;

					// g subsumes h in both the inputs
					if (subsumes(g1,h1) && subsumes(g2,h2) && (canBeLinkSource(g)))
					{
						g.setSubsumingId(linkId);
						h.setSubsumedId(linkId);
					}
					
					// h subsumes g in both the inputs (do not allow the link to go both ways
					else if (subsumes(h1,g1) && subsumes(h2,g2) && (canBeLinkSource(h)))
					{
						h.setSubsumingId(linkId);
						g.setSubsumedId(linkId);
					}
				}
			}
		}
		// remove any subsumption links if there is also a link between the parent nodes
		removeChildLinks(result);
	}
	
	/**
	 * If a node has no slots of its own, and has only one child node,
	 * and appears eligible to be the source of a subsumption link,
	 * then its one child is bound to be eligible to be a source, as well;
	 * so there is no need to make this node a source.
	 * 
	 * @param g
	 * @return
	 */
	private static boolean canBeLinkSource(FSNode g)
	{
		boolean canBeSource = true;
		
		if ((g.getOwnSlotValues().size() == 0) && (g.getChildren().size() == 1)) canBeSource = false;
		
		return canBeSource;
	}
	
	/**
	 * 
	 * @param f
	 * @return all nodes in f which are not of type 'sound' or 'script'
	 */
	private static Vector<FSNode> nonSoundNodes(FSNode f)
	{
		Vector<FSNode> nodes = new Vector<FSNode>();		
		addNonSoundNodes(nodes,f);
		return nodes;
	}
	
	/**
	 * recursive addition of non-sound nodes
	 * @param nodes
	 * @param f
	 */
	private static void addNonSoundNodes(Vector<FSNode> nodes,FSNode f)
	{
		String type = f.getType();
		if ((!type.equals("script")) && (!type.equals("sound"))) nodes.add(f);
		
		for (FSNode fc: f.getChildren()) addNonSoundNodes(nodes,fc);
	}
	
	/**
	 * g has a genId which is a concatenation of two ids, like  k5*m6.
	 * One of these two genIds (k5 or m6) appears somewhere in f. Find it.
	 * (assume all nodes have genIds)
	 * 
	 * @param g
	 * @return the node in f (should never return null)
	 */
	private static FSNode findSameGenId(FSNode g, FSNode f)
	{
		FSNode res = null;
		String gid = g.getGenId();
		String fid = f.getGenId();
		
		// see if the genId of f matches any part of the genId of g
		StringTokenizer st = new StringTokenizer(gid,"*");		
		while (st.hasMoreTokens())
			if (fid.equals(st.nextToken())) res = f;
		
		// try descendants of f
		for (FSNode fc: f.getChildren())
			if (res == null) res = findSameGenId(g,fc);
		
		return res;
	}
	
	/**
	 *  remove any subsumption links if there is also a link between the parent nodes
	 * @param result
	 */
	private static void removeChildLinks(FSNode result)
	{
		// find all nodes at each end of a link
		Hashtable<String,FSNode> starts = findSubsumers(result, true);
		Hashtable<String,FSNode> ends = findSubsumers(result, false);
		
		// build up a list of link ids to remove, because the parent nodes have a link between them
		Vector<String> toRemove = new Vector<String>();
		for (Enumeration<String> en = starts.keys(); en.hasMoreElements();)
		{
			String id = en.nextElement();
			FSNode start = starts.get(id);
			FSNode end = ends.get(id);
			if (end != null)
			{
				FSNode sp = start.getParent();
				FSNode ep = end.getParent();
				String sid = sp.getSubsumingId();
				String eid = ep.getSubsumedId();
				// if the parents have a link, mark the child id for removal
				if ((sid != null) && (eid != null) && (sid.equals(eid))) toRemove.add(id);
			}
		}
		
		// remove the link ids from the nodes that have them
		for (String id:toRemove)
		{
			FSNode start = starts.get(id);
			start.setSubsumingId(null);
			FSNode end = ends.get(id);
			end.setSubsumedId(null);
		}
	}
	
	/**
	 * 
	 * @param result
	 * @param isSource
	 * @return a table with key = subsuming id or subsumed id; value = the node
	 */
	private static Hashtable<String,FSNode> findSubsumers(FSNode result, boolean isSource)
	{
		Hashtable<String,FSNode> nodes = new Hashtable<String,FSNode>();
		addSubsumers(nodes, result, isSource);
		return nodes;
	}

	/**
	 * recursively build up a table of (id,node) for subsuming id or subsumed id
	 * @param nodes
	 * @param res
	 * @param isSource
	 */
	private static void addSubsumers(Hashtable<String,FSNode> nodes, FSNode res, boolean isSource)
	{
		String id = null;
		if (isSource) id = res.getSubsumingId();
		else id = res.getSubsumedId();
		if (id != null) nodes.put(id, res);
		
		for (FSNode child: res.getChildren()) addSubsumers(nodes,child,isSource);
	}

	//------------------------------------------------------------------------------------------
	//        Try to patch up origins for unify - and make sure there are no howlers
	//------------------------------------------------------------------------------------------
	
	/**
	 * 
	 * @param result
	 * @param left
	 * @param right
	 * @throws Exception
	 */
	private static void patchResultOrigins(FSNode result, FSNode left, FSNode right) throws Exception
	{
		// do not check at script level, because the check does plain unify, and the origin of a script node is trivially [1,1].
		String type = result.getType();
		if (!type.equals("script")) patchNodeOrigin(result,left, right);
		
		for (FSNode child: result.getChildren()) 
			patchResultOrigins(child,left, right);
	}
	
	/**
	 * 
	 * @param result
	 * @param left
	 * @param right
	 * @throws Exception
	 */
	private static void patchNodeOrigin(FSNode result, FSNode left, FSNode right) throws Exception
	{
		boolean checked = false;
		int[] origin = result.getOrigin();
		
		// if there is an origin, check it
		if (origin != null)
		{
			FSNode leftIn = null;
			if (origin[0] > 0) leftIn = left.getNodeAtTruePosition(origin[0]);
			FSNode rightIn = null;
			if (origin[1] > 0) rightIn = right.getNodeAtTruePosition(origin[1]);
			
			checked = originCheck(result, leftIn, rightIn);
		}
		
		if (!checked)
		{
			origin = repairOrigin(result, left, right);
			result.setOrigin(origin[0],  origin[1]);
		}
	}
	
	/**
	 * 
	 * @param result
	 * @param leftIn
	 * @param rightIn
	 * @return
	 */
	private static boolean originCheck(FSNode result, FSNode leftIn, FSNode rightIn) throws Exception
	{
		boolean check = false;
		
		// if there are two inputs, the result node must be the unification of the two
		if ((leftIn != null) && (rightIn != null)) check = equalFS(result, unify(leftIn, rightIn,PLAIN));
		
		// if there is only a left input (word), the it may subsume the result node, which can have more information because of a subsumption link
		else if (leftIn != null) check = subsumes(leftIn,result);
		
		// if there is only a right input (from the sentence) it must subsume the result node, which can have more information because of a subsumption link
		else if (rightIn != null) check = subsumes(rightIn,result);	
		
		// tracing
		int[] origin = result.getOrigin();
		// Util.message("Check origin of " + result.getType() + " [" + origin[0] + "," + origin[1] + "] ; result " +check);
		
		// if there are no inputs, the origin must be wrong
		return check;
	}
	
	/**
	 * check for equality of two FSNodes - subsumption works both ways
	 * @param n1
	 * @param n2
	 * @return
	 * @throws Exception
	 */
	private static boolean equalFS(FSNode n1, FSNode n2) throws Exception
	{
		boolean s1 = subsumes(n1,n2);
		boolean s2 = subsumes(n2,n1);
		return (s1 & s2);
	}
	
	/**
	 * attempt to repair an origin
	 * @param result
	 * @param leftIn
	 * @param rightIn
	 * @return
	 */
	private static int[] repairOrigin(FSNode result, FSNode left, FSNode right) throws Exception
	{
		boolean tracing = false;
		int[] oldOrigin = result.getOrigin();
		if (tracing)
		{
			if (oldOrigin != null) Util.message("Repairing origin [" + oldOrigin[0] + "," + oldOrigin[1] +  "] of node of type " + result.getType());
			else Util.message("Repairing null origin of node of type " + result.getType());
		}
		
		int[] origin = new int[2];
		// stays this way if no unique repair can be fond
		origin[0] = 0;
		origin[1] = 0;
		
		// build lists of nodes in the inputs which subsume the result node
		Vector<FSNode> leftSubsumeList = new Vector<FSNode>();
		Vector<FSNode> rightSubsumeList = new Vector<FSNode>();
		buildSubsumingList(result, left, leftSubsumeList);
		buildSubsumingList(result, right, rightSubsumeList);
		int nl = leftSubsumeList.size();
		int nr = rightSubsumeList.size();
		if (tracing) Util.message("Candidate input nodes: " + nl + " and " + nr);
		
		// cases in which the origin can be repaired - an unique origin can be found
		
		// unique solution from one side or the other
		if ((nl == 1) && (nr== 0)) origin[0] = leftSubsumeList.get(0).getTruePosition();
		else if ((nl == 0) && (nr== 1)) origin[1] = rightSubsumeList.get(0).getTruePosition();
		
		// unique unify solution
		else if ((nl > 0) && (nr > 0))
		{
			int found = 0;
			int lp = 0;
			int rp = 0;
			
			// look for any pair [right,left] which unify to give the result node
			for (int il = 0; il < nl; il++)
			{
				FSNode ll = leftSubsumeList.get(il);
				for (int ir = 0; ir < nr; ir++)
				{
					FSNode rr = rightSubsumeList.get(ir);
					FSNode unit = unify(ll,rr,PLAIN);
					if ((unit.isViable()) && (equalFS(result,unit)))
					{
						found++;
						lp = ll.getTruePosition();
						rp = rr.getTruePosition();
					}
				}
			}
			
			// only if there is a unique solution, return it
			if (found == 1)
			{
				origin[0] = lp;
				origin[1] = rp;
			}
			else if ((found > 1) & tracing) Util.message("Found " + found + " possible unified origins.");

					
		}
		
		if (tracing) Util.message("Repaired to [" + origin[0] + "," + origin[1] + "]");
		return origin;
	}

	/**
	 * 
	 * @param result
	 * @param in
	 * @param subsumeList
	 */
	private static void buildSubsumingList(FSNode result, FSNode in, Vector<FSNode> subsumeList) throws Exception
	{
		// the input node should not be the start of a subsumption link - the end of the link will do instead
		if ((subsumes(in,result)) && (in.getSubsumingId() == null)) subsumeList.add(in);
		
		for (FSNode child: in.getChildren()) buildSubsumingList(result, child, subsumeList);
	}


	//------------------------------------------------------------------------------------------
	//    Checking that Origins have been validly assigned - throw an exception if not
	//------------------------------------------------------------------------------------------
	
	/**
	 * 
	 * @param result
	 * @param left left input
	 * @param right right input (one of these two may be null, for unify
	 * @param generalise false if unify, true if generalise
	 * @throws Exception
	 */
	private static void checkOperationOrigins(FSNode result, FSNode left, FSNode right, boolean generalise) throws Exception
	{
		boolean tracing = false;
		
		// for informative messages, if possible
		String path = result.getPath() + "(computed position " + result.getPosition() + ")";
		String type = result.getType();
		String fullType = type;
		if (type.equals("sound")) fullType = type + "(" + result.getSlotValue("sound") + ")";
		if (tracing) Util.message("Checking origin at path " + path + ", type " + fullType);
		
		// check that there is some input
		if ((left == null) && (right == null)) throw new Exception("No input nodes for result node at path " + path);
		
		// check that the result has an origin
		int[] origin = result.getOrigin();
		if (origin == null) throw new Exception("Result node at path " + path + " has no origin.");
		String oString = "[" + origin[0] + "," + origin[1] + "]";
		if (tracing) Util.message("Origin: " + oString);
		
		// check there are enough nonzero parts
		int parts = 0;
		for (int i = 0; i < 2; i++) if (origin[i] > 0) parts++;
		if ((generalise) && (parts < 2)) throw new Exception("Generalise result at path " + path + " must come from both inputs.");
		if ((!generalise) && (parts < 1)) throw new Exception("Unify result at path " + path + " must come from at least one input.");
		
		// check the types match, and the input true positions match the origin, and the slot subsumption holds
		if (origin[0] > 0) 
		{
			if (left == null) throw new Exception("Null left input at result path " + path + " with origin " + oString);
			if (!type.equals(left.getType())) 
				throw new Exception("Left origin node type " + left.getType() + " does not match result node at path " + path);
			if (origin[0] != left.getTruePosition()) 
				throw new Exception("Left input position " + left.getTruePosition() + " does not match origin " + origin[0] + " at path " + path);
			
			// check slot subsumption on the nodes
			boolean subTest = true;
			if (generalise) subTest = subsumesSlots(result,left);
			else subTest = subsumesSlots(left,result);
			if (!subTest) throw new Exception("Failed slot subsumption test with left input at result path " + path + " with origin " + oString);
		}
		
		if (origin[1] > 0) 
		{
			if (right == null) throw new Exception("Null right input at result path " + path+ " with origin " + oString);
			if (!type.equals(right.getType())) 
				throw new Exception("Right origin node type " + right.getType() + " does not match result node at path " + path);
			if (origin[1] != right.getTruePosition()) 
				throw new Exception("Right input position " + right.getTruePosition() + " does not match origin " + origin[1] + " at path " + path);
			
			// check slot subsumption on the nodes
			boolean subTest = true;
			if (generalise) subTest = subsumesSlots(result,right);
			else subTest = subsumesSlots(right,result);
			if (!subTest) throw new Exception("Failed slot subsumption test with right input at result path " + path + " with origin " + oString);
		}
		
		// re-check the child nodes (errors should have been detected already in previous calls, but just in case...)
		for (FSNode child: result.getChildren())
		{
			int[] cOrigin = child.getOrigin();
			if (cOrigin == null) throw new Exception("Child result node has no origin at path " + child.getPath());

			FSNode cLeft = null;
			if ((left != null) && (cOrigin[0] > 0)) cLeft = childAtPosition(left,cOrigin[0]);
			FSNode cRight = null;
			if ((right != null) && (cOrigin[1] > 0)) cRight = childAtPosition(right,cOrigin[1]);

			checkOperationOrigins(child,cLeft,cRight,generalise);				
		}
		
		if (tracing) Util.message("Passed at path " + path);
		
	}
	
	/**
	 * 
	 * @param n
	 * @param pos
	 * @return
	 */
	private static FSNode childAtPosition(FSNode n, int pos) 
	{
		FSNode res = null;
		for (FSNode child: n.getChildren()) 
			if (child.getTruePosition() == pos) res = child;
		return res;
	}


	//--------------------------------------------------------------------------
	//                      Operations - subsumption test
	//--------------------------------------------------------------------------
	
	/**
	 * A subsumes B, if B contains all the information of A, and possibly more
	 * @param other
	 * @return
	 */
	public static boolean subsumes(FSNode first, FSNode other) throws Exception
	{
		
		// types must match
		if (!first.getType().equals(other.getType())) return false;
		String type = first.getType();
		
		//check subsumption of slots
		boolean subsumes = subsumesSlots(first,other);
		
		// if slots pass, check subsumption of child feature structures 
		if (subsumes) subsumes = subsumesChildren(first,other);
		
		// mark the other feature structure for subsumption link tests
		if (subsumes) other.setFromIds(first);
		else other.cancelFromIds();

		// from the top level only, make subsumption link tests
		if (subsumes & (type.equals("script"))) subsumes = checkSubsumptionLinks(first);
		
		// clean-up
		if (type.equals("script")) other.cancelFromIds();
		
		return subsumes;
	}
	
	/**
	 * @param other
	 * @return true if slots subsume
	 */
	public static boolean subsumesSlots(FSNode first, FSNode other) throws Exception
	{
		boolean subsumes = true;
		
		// check subsumption of all own slots in the slots of other
		for (Enumeration<String> en = first.getOwnSlotValues().keys(); en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			String ownValue = first.getOwnSlotValues().get(slotName);
			
			if (subsumes)
			{
				// the other feature structure must have the same slot
				String otherValue = other.slotValue(slotName);
				if (otherValue == null)	 return false;	
				
				// slots with open sets of values
				if (first.getModel().openSlot(slotName))
				{
					if (!ownValue.equals(otherValue)) subsumes = false;
				}
				
				// slots with a hierarchy of values
				else
				{
					Slot theSlot = first.getModel().getSlot(slotName);
					if (theSlot != null) 
					{
						SlotValue theValue = theSlot.getSlotValueWith(ownValue);
						if (!theValue.containsValue(otherValue)) subsumes = false;
					}
					//  sometimes 'unviable' is present - only write a message
					else Util.message("Model does not have slot " + slotName);
				}
			}

		}
		
		return subsumes;
	}
	
	/**
	 * 
	 * @param other
	 * @return
	 */
	public static boolean subsumesChildren(FSNode first, FSNode other) throws Exception
	{
		// if this FS has no child FS, child subsumption passes
		if (first.getChildren().size() == 0) return true;
		
		// the other FS must have the same number of child FS as this, or more
		if (first.getChildren().size() > other.getChildren().size()) return false;

		/* there must be some valid reordered list of the child FS of this FS, 
		 * such that members of that list all subsume an ordered sublist of the other's child FS*/
		Vector<Vector<FSNode>> reorderings = Util.constrainedPermutations(first.getChildren(), first.getFreeOrder());
		
		// for all node types except 'scene', there should be no constraint on reorderings of child nodes
		if (!(first.getType().equals("script"))) reorderings = Util.permutations(first.getChildren());
		
		int perms = reorderings.size();
		if (perms > 6) 
		{
			Util.message(perms + " constrained permutations of " + first.getChildren().size() + " nodes");
		}
		
		String otherSummary = makeSummary(other.getChildren());
		
		// try out each possible reordering of the child FS, until you find one that works
		boolean subsumes = false;
		for(Vector<FSNode> reordering: reorderings) if (!subsumes)
		{
			String thisSummary = makeSummary(reordering);
			subsumes = childSubsume(reordering, other.getChildren());
		}
				
		return subsumes;
	}
	
	/**
	 * One-line summary of a Vector of FeatureStructures
	 * @param feats
	 * @return
	 */
	private static String makeSummary(Vector<FSNode> feats)
	{
		String summary = "";
		
		for (FSNode feat: feats)
		{
			String type = feat.getType();
			if (type.equals("sound")) summary = summary + "'" + feat.getSlotValue("sound") + "' ";
			else summary = summary + type + " ";
		}
		
		return summary;
	}
	
	/**
	 * To pass the test, every child on the list thisChildren must subsume some child on the longer list otherChildren;
	 * and those on the other list must be in the same order, so that the ordering constraints of the two lists are consistent,
	 * e.g. AB does not subsume BA, if both lists are ordered. But if either list is free, it does.
	 *  (the list thisChildren is a permutation respecting its ordering constraints; all such permutations are tried out.
	 *  Only one of the lists need be permuted)
	 *  
	 * @param thisChildren
	 * @param otherChildren
	 * @return
	 */
	private static boolean childSubsume(Vector<FSNode> thisChildren, Vector<FSNode> otherChildren) throws Exception
	{
		// index of the last child of the other FS list, which some child of this list subsumed
		int otherUsed = -1;
		boolean stillPossible = true;

		for (FSNode child: thisChildren) if (stillPossible)
		{
			boolean thisChildSubsumes = false;
			
			/* try all remaining children on the other list, to find the first 'other' child that this child subsumes -
			 * to leave the largest number of other children to test against later 'this' children. 
			 * It does not matter if this child also subsumes later other children */

			if (otherUsed + 1 < otherChildren.size()) // if there are still some 'other' children remaining to test
			{
				// test all remaining other children until you succeed
				for (int j = otherUsed + 1; j < otherChildren.size(); j++) if (!thisChildSubsumes)
				{
					FSNode otherChild = otherChildren.get(j);
					thisChildSubsumes = (subsumes(child,otherChild));
					if (thisChildSubsumes) otherUsed = j;
				}
			}
			
			stillPossible = stillPossible & thisChildSubsumes;
		}
		
		return stillPossible;
	}
		
	/**
	 * at script level only, find all subsumption links passed over from the subsuming FS, and check them
	 * @return
	 */
	private static boolean checkSubsumptionLinks(FSNode first) throws Exception
	{
		// sub-feature structures marked as subsumed by some feature structure with an id
		Hashtable<String,FSNode> fromIds = new Hashtable<String,FSNode>();
		// sub-feature structures marked as subsumed by some feature structure with a subsumesId
		Hashtable<String,FSNode> fromSubsumeIds = new Hashtable<String,FSNode>();
		// build up these two tables
		addSubsumeLinks(first,fromIds,fromSubsumeIds);
		
		boolean passed = true;
		for (Enumeration<String> en = fromIds.keys();en.hasMoreElements();)
		{
			String fromId = en.nextElement();
			if (passed)
			{
				// pass if FS subsumer subsumes FS subsumed
				FSNode subsumed = fromIds.get(fromId);
				FSNode subsumer = fromSubsumeIds.get(fromId);

				// this condition ( a missing fromSubsumeId) is really obscure; it no longer throws an Exception, just bypasses the test
				if (subsumer == null) 
				{
					Util.message(first.getTextTree(0));
					Util.message("Cannot find subsumer FS for key " + fromId  + " when checking subsumption links in a subsume test");
				}				
				else if (!subsumes(subsumer,subsumed)) passed = false;

				subsumed.cancelFromIds(); // leave no trace of the last subsumption test
			}
		}
		
		return passed;		
	}
	
	/**
	 * recursively find all subsumption links passed over to this FS from a subsuming FS in a successful subsumption
	 * @param fromIds
	 * @param fromSubsumeIds
	 */
	public static void addSubsumeLinks(FSNode first, Hashtable<String,FSNode> fromIds, Hashtable<String,FSNode> fromSubsumeIds)
	{
		if (first.getFromId() != null) fromIds.put(first.getFromId(), first);
		if (first.getFromSubsumesId() != null) fromSubsumeIds.put(first.getFromSubsumesId(), first);
		
		for (FSNode child: first.getChildren()) addSubsumeLinks(child,fromIds,fromSubsumeIds);
		
	}



}
