package modelPack;

import java.awt.Graphics;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import utilPack.Util;

public class Slot {
	
	LanguageModel model;
	
	private String name;
	public String getName() {return name;}
	public void setName(String name) {this.name = name;}
	
	private String description = "";
	public String getDescription() {return description;}
	public void setDescription(String desc) {description = desc;}
	
	// the top-level slot values (not all values; some are children of these values)
	private Vector<SlotValue> values = new Vector<SlotValue>();
	public Vector<SlotValue> getValues() {return values;}
	
	// node types this slot can appear on
	private Vector<NodeType> allowedTypes = new Vector<NodeType>();
	public Vector<NodeType> getAllowedTypes() {return allowedTypes;}
	
	public boolean isAllowedNodeType(String type)
	{
		boolean allowed = false;
		for (NodeType nt: allowedTypes) if (nt.getName().equals(type)) allowed = true;
		
		return allowed;
	}
	
	// slot value made temporarily to draw a node for the slot itself
	private SlotValue drawingValue = null;
	
	//--------------------------------------------------------------------------
	//                        Constructors
	//--------------------------------------------------------------------------

	
	/**
	 * constructor from XML
	 * @param model
	 * @param slotEl
	 * @throws Exception
	 */
	public Slot(LanguageModel model, Element slotEl) throws Exception
	{
		this.model = model;
		name = slotEl.getAttribute("name");
		description = slotEl.getAttribute("dsc");
		
		String types = slotEl.getAttribute("nodes");
		setAllowedTypes(types);
		
		for (Element valueEl: Util.namedChildElements(slotEl, "value"))
		{
			SlotValue val = new SlotValue(this,null, valueEl,model);
			values.add(val);
		}
	}
	
	/**
	 * create a new slot in the model editor
	 * @param model
	 * @param name
	 * @param nodeTypes
	 * @param description
	 */
	public Slot(LanguageModel model, String name, String nodeTypes, String description) throws Exception
	{
		this.model = model;
		this.name = name;
		this.description = description;
		
		setAllowedTypes(nodeTypes);
		
		values = new Vector<SlotValue>();
	}

	
	/**
	 * 
	 * @param types
	 * @throws Exception
	 */
	public void setAllowedTypes (String types) throws Exception
	{
		allowedTypes = new Vector<NodeType>();
		StringTokenizer st = new StringTokenizer(types,", ");
		while (st.hasMoreTokens())
		{
			String type = st.nextToken();
			NodeType nt = model.getNodeType(type);
			if (nt == null) throw new Exception("slot " + name + " has forbidden node type " + type);
			allowedTypes.add(nt);
		}
	}
	
	/**
	 * 
	 * @return names of node types this slot can appear on, space-separated
	 */
	public String getAllowedNodeTypeNames()
	{
		String names = "";
		
		for (NodeType type:allowedTypes)
			names = names + type.getName() + " ";
		
		return names;
		
	}
	
	/**
	 * 
	 * @return
	 */
	private String  nodeTypeString()
	{
		String type = "";
		for (NodeType nt: allowedTypes) type = type + nt.getName() + " ";
		return type;
	}
	
	//--------------------------------------------------------------------------
	//                         XML output
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param doc
	 * @return
	 * @throws Exception
	 */
	public Element xmlForm(Document doc) throws Exception
	{
		Element root = Util.newElement(doc, "slot");
		root.setAttribute("name", name);
		if (description.length() > 0) root.setAttribute("dsc", description);
		root.setAttribute("nodes",nodeTypeString());
		
		for (SlotValue sv: values) 
			root.appendChild(sv.xmlForm(doc));
		
		return root;
	}
	
	//--------------------------------------------------------------------------
	//                       Retrieval methods
	//--------------------------------------------------------------------------


	/**
	 * 
	 * @return all values of the slot, at any part of the hierarchy of values
	 */
	public Vector<String> getAllValues()
	{
		Vector<String> allValues = new Vector<String>();
		// for each top level slot value, add its value and the values of all its descendants
		for (SlotValue val: values) val.addValues(allValues);		
		return allValues;
	}
	
	/**
	 * get a random value, from anywhere in the tree of values,
	 * with equal probability for all values
	 * @return
	 */
	public String getRandomValue()
	{
		Vector<String> all = getAllValues();
		int chosen = (int)(all.size()*Math.random()); // rounds down to 0...all.size() -1
		return all.get(chosen);
	}
	
	/**
	 * 
	 * @param val
	 * @return true if the value is allowed
	 */
	public boolean isAllowedValue(String val)
	{
		return Util.inVector(val, getAllValues());
	}
	
	/**
	 * 
	 * @param val
	 * @return true  if this slot has the value, at any part of its hierarchy of values
	 */
	public boolean hasValue(String val)
	{
		return Util.inVector(val, getAllValues());
	}
	
	/**
	 * 
	 * @param val1
	 * @param val2
	 * @return true if the value val1 is one of the sub-values of val2
	 * (if val1 is below val2 in the hierarchy of values)
	 */
	public boolean isSubValue(String val1, String val2)
	{
		boolean implies = false;
		
		SlotValue v2 = getSlotValueWith(val2);
		if (v2 != null) implies  = v2.containsValue(val1);
		
		return implies;
	}
	
	/**
	 * 
	 * @param val1
	 * @param val2
	 * @return true if val1 and val2 are compatible allowed values of the slot
	 * (if either one is a sub-value of the other)
	 */
	public  boolean compatible(String val1, String val2)
	{
		return ((isSubValue(val1,val2))||(isSubValue(val2,val1)));
	}
	
	/**
	 * 
	 * @param val1
	 * @param val2
	 * @return the slot value which generalises val1 and val 2; 
	 * or null if there is no such value
	 */
	public String generaliseValue(String val1, String val2) throws Exception
	{
		String res = null;
		
		if (!hasValue(val1)) throw new Exception("Slot " + this.getName() + " has no value " + val1);
		if (!hasValue(val2)) throw new Exception("Slot " + this.getName() + " has not the value " + val2);
		
		// if one value is a sub-value of  the other, return the higher, more general value
		if (isSubValue(val1,val2)) res = val2;
		else if (isSubValue(val2,val1)) res = val1;
		
		else
		{
			// this slot value should exist
			SlotValue v1 = getSlotValueWith(val1);
			if (v1 == null) throw new Exception("Cannot find SlotValue of slot " + getName() + " for value " + val1);
			
			// ascend the hierarchy until some ancestor of val1 has val2 as a sub-value.
			SlotValue parent = v1.getParent();			
			while ((parent != null) && (res == null))
			{
				if (parent.containsValue(val2)) res = parent.getValue();
				else parent = parent.getParent();
			}
		}
		
		return res;
	}
	
	/**
	 * 
	 * @param val
	 * @return the SlotValue containing a specific value
	 * or null if there is none
	 */
	public SlotValue getSlotValueWith(String val)
	{
		SlotValue sVal = null;		
		for (SlotValue cand:values)
			if (cand.withValue(val) != null) 
				sVal = cand.withValue(val);
		
		return sVal;
	}
	
	/**
	 * 
	 * @return sum of weights of top-level values
	 */
	public double sumWeights()
	{
		double w = 0;
		for (SlotValue sub: values) w = w + sub.getWeight();		
		return w;
	}
	
	/**
	 * 
	 * @param val
	 * @return
	 */
	public double getInformation(String val)
	{
		double info = 0.0;
		
		SlotValue sv = getSlotValueWith(val);
		if (sv != null) info = sv.getinformation();
						
		return info;
	}
	
	
	public String getExplanation()
	{
		String explain = "Slot: " + getName();
		if (description.length() > 0) explain = explain + " - " + description;
		
		explain = explain + "\n\nValues:\n";
		String indent = "       ";
		
		for (SlotValue val:getValues())
			explain = explain + val.getExplanation(indent) + "\n";
		
		return explain;
	}

	
	//--------------------------------------------------------------------------
	//                       Recursive drawing
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param left
	 * @param top
	 * @param g
	 */
	public void draw(int left, int top, Graphics g)
	{
		// make a SlotValue to draw the top node, like any other SlotValue
		drawingValue = new SlotValue(this,model);
		
		// draw this SlotValue and all descendants
		drawingValue.draw(left, top, g);
	}

	

	

}
