package modelPack;

import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import utilPack.Util;

public class SlotValue {
	
	private Slot slot;
	public Slot getSlot() {return slot;}
	
	/* weight used to compute the information gain (in bits) from choosing this slot from its siblings.
	 * default is 1.0 unless specified in the model file. */
	private double weight = 1.0;
	public double getWeight() {return weight;}
	
	private String ownValue;
	public String getValue() {return ownValue;}
	public void setValue(String val) {ownValue = val;}
	
	private String description;
	public String getDescription() {return description;}
	public void setDescription(String desc) {description = desc;}
	
	private Vector<SlotValue> subValues = new Vector<SlotValue>();
	public Vector<SlotValue> getSubValues() {return subValues;}
	
	private SlotValue parent;
	public SlotValue getParent() {return parent;}
	
	private LanguageModel model;
	
	// true if this is a special SlotValue representing the slot as a whole, for drawing and selecting purposes
	private boolean representsSlot;
	public boolean representsSlot() {return representsSlot;}
	
	//--------------------------------------------------------------------------
	//                        Constructors
	//--------------------------------------------------------------------------

	
	/**
	 * constructor from XML input
	 * @param slot
	 * @param valueEl
	 */
	public SlotValue(Slot slot, SlotValue parent, Element valueEl,LanguageModel model)
	{
		this.slot = slot;
		this.model = model;
		ownValue = valueEl.getAttribute("val");
		description = valueEl.getAttribute("dsc");
		this.parent = parent;
				
		for (Element subValueEl: Util.namedChildElements(valueEl, "value"))
		{
			SlotValue subValue = new SlotValue(slot,this,subValueEl,model);
			subValues.add(subValue);
		}
		representsSlot = false;
	}
	
	/**
	 * special SlotValue used to draw the top node for a slot, then draw all the slot value tree 
	 * @param slot
	 */
	public SlotValue(Slot slot,LanguageModel model)
	{
		this.slot = slot;
		this.model = model;
		ownValue = slot.getName(); // the slot name will be written (like a value) in the top box
		description = slot.getExplanation();
		parent = null;
		
		for (SlotValue val: slot.getValues()) subValues.add(val);
		
		representsSlot = true;
		
	}
	
	/**
	 * new SlotValue made in the editor
	 * @param value
	 * @param descriptkon
	 * @param slot
	 * @param parent
	 */
	public SlotValue(String value, String description, Slot slot, SlotValue parent)
	{
		ownValue = value;
		this.description = description;
		this.slot = slot;
		this.parent  = parent;
		model = slot.model;
		// no subvalues
	}
	
	/**
	 * add the values from this slot, and all those it contains, to a set of values
	 * @param values
	 */
	public void addValues(Vector<String> values)
	{
		values.add(ownValue);
		for(SlotValue subValue: subValues) subValue.addValues(values);
	}
	
	
	//--------------------------------------------------------------------------
	//                         XML output
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param doc
	 * @return
	 * @throws Exception
	 */
	public Element xmlForm(Document doc) throws Exception
	{
		Element root = Util.newElement(doc, "value");
		root.setAttribute("val", ownValue);
		if (description.length() > 0) root.setAttribute("dsc", description);
		
		for (SlotValue subValue:subValues) 
			root.appendChild(subValue.xmlForm(doc));
		
		return root;
	}
	
	//--------------------------------------------------------------------------
	//                     Retrieval methods
	//--------------------------------------------------------------------------

	
	/**
	 * 
	 * @param val
	 * @return true if val is this slotValue or one it contains
	 */
	public boolean containsValue(String val)
	{
		boolean contains = (ownValue.equals(val));		
		for (SlotValue subVal: subValues)
			if ((!contains) && (subVal.containsValue(val))) contains = true;

		return contains;
	}
	
	/**
	 * 
	 * @param val
	 * @return this SlotValue or a descendant which has a specific  value
	 */
	public SlotValue withValue(String val)
	{
		if (ownValue.equals(val)) return this;
		for (SlotValue subVal: subValues)
			if (subVal.withValue(val) != null) 
				return subVal.withValue(val);
		return null;
	}
	
	/**
	 * 
	 * @return information content of this value, in bits
	 */
	public double getinformation()
	{
		// sum of weights of all sibling values
		double sumW;
		if (parent != null) sumW = parent.sumWeights();
		else sumW = slot.sumWeights();
		
		// both weight and sumW must be non-zero
		double info = Math.log(sumW/weight)/Math.log(2.0);
		
		// recursion back up the tree
		if (parent != null) info = info + parent.getinformation();
		
		return info;		
	}
	
	/**
	 * 
	 * @return
	 */
	public double sumWeights()
	{
		double w = 0;
		for (SlotValue sub: subValues) w = w + sub.getWeight();		
		return w;
	}
	
	/**
	 * explanation of this slot value and its descendants
	 * @param indent
	 * @return
	 */
	public String getExplanation(String indent)
	{
		String explain  = indent + getValue();
		if (description.length() > 0) explain = explain + ": " + description;
		
		String nextIndent = indent + "      ";		
		for (SlotValue sub: subValues)
			explain = explain + "\n" + sub.getExplanation(nextIndent);
		
		return explain;
	}

	
	//--------------------------------------------------------------------------
	//                       Recursive drawing
	//--------------------------------------------------------------------------
	
	/*
	 * true if child slot values are shown
	 */
	private boolean expanded = true;
	public void setExpanded(boolean expanded) {this.expanded = expanded;}
	public boolean isExpanded() {return expanded;}

	
	// height of boxes
	private static int boxHeight = 20;
	
	// height of slot rows
	private static int rowHeight = 16;
	
	// gap between FS heights 
	private static int gapHeight = 8;
		
	// box width per character in the type name and link names (estimate for font size 12)
	private static int boxWidthPerChar = 7;
	
	/**
	 * 
	 * @param left
	 * @param top
	 * @param g
	 */
	public void draw(int left, int top, Graphics g)
	{
		// height of the middle and top of the root node
		int rootMiddle = top + getDrop();
		int rootBoxTop = rootMiddle - boxHeight/2;
		int textBase = rootBoxTop + boxHeight - 5;
		
		// draw the root node box, including possibly a [+-] box to the right of it
		int boxWidth = drawBox(left, rootBoxTop, g);
		
		// write the text to the right of boxes with no children
		writeText(left + boxWidth, textBase, description, g);
		
		// draw child nodes and lines to them
		if (expanded)
		{
			int right = left + boxWidth;
			int childLeft = right + 50;
			int childTop = top;
			for (SlotValue child: subValues)
			{
				// draw the child slot and the line to it
				child.draw(childLeft, childTop, g);
				g.drawLine(right, rootMiddle, childLeft, childTop + child.getDrop());
				
				// ready for the next child
				int childHeight = child.getDrawingHeight();
				childTop = childTop + childHeight;
			}
		}		
	}


	/**
	 * 
	 * @param g
	 * @param left
	 * @param boxTop
	 * 
	 * @return the width of the box
	 */
	private int drawBox(int left, int boxTop, Graphics g)
	{
		
		// draw the box for the root node with the value
		// safety margin on box width
		int boxWidth = boxWidthPerChar*(ownValue.length() + 1) + 10;
		int totalWidth = boxWidth;

		// draw the box
		g.drawRect(left, boxTop,boxWidth, boxHeight);
				
		// record this node for mouse clicking
		Rectangle clickable = new Rectangle(left, boxTop,boxWidth, boxHeight);
		model.addValueLocation(clickable, this);
		
		// bold if the box has been selected
		SlotValue clicked = model.getFrame().getDrawArea().getJustClickedValue();		
		if ((clicked != null) && (clicked.equals(this)))
			g.drawRect(left - 1, boxTop -1, boxWidth + 2, boxHeight + 2);

		// draw text inside the box
		int boxTextBase = boxTop + boxHeight - 5;
		g.drawChars(ownValue.toCharArray(),0,ownValue.length(),left + 5,boxTextBase);
		
		// if there are subvalues, draw a clickable [+ -] box to the right of the main box
		if (subValues.size() > 0)
		{
			// small  box for + or -
			String plusMinus = "+";
			if (expanded) plusMinus = "-";
			int pmWidth = 2*boxWidthPerChar;
			int pmLeft = left + boxWidth;
			g.drawRect(pmLeft, boxTop, pmWidth, boxHeight);
			g.drawChars(plusMinus.toCharArray(),0,plusMinus.length(),pmLeft + 5,boxTextBase);

			// record the position of the [+-] box for mouse clicks
			Rectangle pmClickable = new Rectangle(pmLeft, boxTop, pmWidth, boxHeight);
			model.addpmBoxLocation(pmClickable, this);

			totalWidth = boxWidth + pmWidth;
		}
		
		
		return totalWidth;		
	}
	
	/**
	 * 
	 * @param left
	 * @param textBase
	 * @param text
	 * @param g
	 */
	private void writeText(int left, int textBase, String text, Graphics g)
	{
		boolean noText = ((subValues.size() > 0) && (expanded));
		if (!noText) g.drawChars(text.toCharArray(),0,text.length(),left + 5,textBase);			
	}


	
	/**
	 * height in pixels of the slot value and its children
	 * @return
	 */
	public int getDrawingHeight()
	{
		if (!expanded) return boxHeight + gapHeight;
		else if (subValues.size() == 0) return boxHeight + gapHeight;
		else
		{
			int h = 0;
			for (SlotValue child: subValues) h = h + child.getDrawingHeight();
			return h;
		}
	}
	
	/**
	 * 
	 * @return drop in pixels from the top of the area occupied by this slot value and its children,
	 * to the centre of the box for this slot value
	 */
	public int getDrop()
	{
		// if there are no children, or they are not shown
		int centreHeight =  (boxHeight/2);
		
		/*
		 *  otherwise, the centre of this box's node is the mean of the centres of its children's boxes 
		 *  */
		if ((expanded) && (subValues.size() > 0))
		{
			int childTop = 0;
			int sumCentre = 0;
			for (SlotValue child: subValues)
			{
				int childCentre = child.getDrop() + childTop;
				sumCentre = sumCentre + childCentre;
				childTop = childTop + child.getDrawingHeight();
			}
			centreHeight = sumCentre/subValues.size();
		}
		
		return centreHeight;
	}
	

}
