package modelPack;

import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import learnerPack.ExampleWord;
import learnerPack.LearningExample;
import learnerPack.LearntWord;
import utilPack.Util;

public class WordToLearn {
	
	// edited language model
	private LanguageModel model;
	
	// the stage of learning
	private LearningStage stage;
	public LearningStage getStage() {return stage;}
	public String getStageName() {return stage.getName();}
	
	// name of the word being learned
	private String wordName;
	public String getWordName() {return wordName;}
	
	// feature structure of the word being learnt - to display before learning
	private FSNode wordFS = null;
	public FSNode getWordFS() {return wordFS;}
	public void setWordFS(FSNode f) {wordFS = f;}
	
	// sentence-meaning pairs which are learning examples
	private  Vector<FSNode> examples = new  Vector<FSNode>();
	public Vector<FSNode> getExamples() {return examples;}
	
	// cyceNo is 0 for the static language model, positive after any automatic learning cycle
	public int getCycleNo() {return stage.getCycleNo();}
	
	public LearntWord getLearntWord(String spacedName) {return stage.getLearner().getBestWord(spacedName);}
		
	// when showing a learnt word to re-learn it, use up to this number of learning examples
	int maxExamples = 4;
	
	public FSNode getExample(String exampleName)
	{
		FSNode lex = null;
		
		for (int i = 0; i < examples.size(); i++)
			if (examples.get(i).getName().equals(exampleName))
				lex = examples.get(i);
		
		return lex;
	}
	
	/**
	 * 
	 * @return
	 */
	public Vector<String> getExampleNames()
	{
		Vector<String> names = new Vector<String>();
		for (FSNode example: examples) names.add(example.getName());
		return names;
	}
	
	//--------------------------------------------------------------------------
	//                Empty Constructor for use in the editor
	//--------------------------------------------------------------------------
	
	
	/**
	 * set up a new word to learn, with no learning examples and an empty structure to be learnt
	 * @param stage
	 * @param wordName
	 * @param stageName
	 * @throws Exception
	 */
	public WordToLearn(LearningStage stage, String wordName) throws Exception
	{
		this.stage = stage;
		model = stage.getModel();
		this.wordName = wordName;
		
		// set up an empty feature structure to be learnt (to be edited later)
		wordFS = new FSNode(model,"script");
		wordFS.setName(wordName);
		wordFS.setWordLearnToLearnParent(this);
	}
	
	
	//--------------------------------------------------------------------------
	//                  Constructor from a LearntWord
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param stage
	 * @param learner
	 * @param lw
	 * @throws Exception
	 */
	public WordToLearn(LearningStage stage, LearntWord lw) throws Exception
	{
		this.stage = stage;
		model = stage.getModel();
		
		// set up the learnt word so it can be displayed
		wordFS = lw.getWord();
		wordName = lw.getSpacedWordName();
		wordFS.setName(wordName);
		wordFS.setWordLearnToLearnParent(this);
		
		// set up the last few learning examples (up to maxExamples)
		int maxEx = lw.getExampleCount();
		int minEx = maxEx - maxExamples;
		
		// include no learning examples before the example which promoted the word to its current productivity
		if (minEx < lw.getpromotedExNo()[0]) minEx = lw.getpromotedExNo()[0];

		// iterate over recent learning examples used by the word
		for (int  exW = minEx; exW < maxEx; exW++)
		{
			ExampleWord exPos = lw.getExamplePositions().get(exW);
			int exampleNo = exPos.getExNo(); // ascending with gaps
			int versionNo = exPos.getFSNo(); // 0 if there are no unifications of the example

			LearningExample lex = stage.getLearner().generator().getExample(exampleNo);

			//  use any example in its initial un-unified version, with an informative name
			FSNode example = lex.getUnifyResult(0);
			examples.add(example);
			
			// set up the unify order for the example, in phonetic spelling of words
			String unifyOrder = "";
			Vector<LearntWord> lwv = lex.getUnifiedWords();
			/* unify all words up to the version# of this example used in this word;
			 * if versionNo = 1, there can be 1 word in the unify order */
			for (int u = 0; u < versionNo; u++)
			{
				if (u < lwv.size())
				{
					LearntWord lx = lwv.get(u);
					String spacedName = lx.getSpacedWordName();
					// do not put the word itself in the unify order
					if (!spacedName.equals(wordName)) unifyOrder = unifyOrder + spacedName + " ";
				}
				else
				{
					Util.message("BBBB Learning example overflow");
				}
			}
			example.setUnifyOrder(unifyOrder);
		}
	}
	
	
	
	//--------------------------------------------------------------------------
	//                    Constructor from XML
	//--------------------------------------------------------------------------
	
	
		

	
	/**
	 * constructor from an XML data file
	 * @param stage
	 * @param learnEl
	 * @param stageName
	 * @throws Exception
	 */
	public WordToLearn(LearningStage stage, Element learnEl, String stageName) throws Exception
	{
		this.stage = stage;
		model = stage.getModel();
		wordName = learnEl.getAttribute("word");
		Vector<Element> structureEls = Util.namedChildElements(learnEl, "structure");
		// the structure to be learnt started out optional, but there is always exactly one of them
		if (structureEls.size() > 0)
		{
			Element scriptEl = Util.firstNamedChild(structureEls.get(0), "script");
			wordFS = new FSNode(model,wordName,scriptEl,0,null);
			wordFS.setWordLearnToLearnParent(this);
		}
		
		Vector<Element> exampleEls = Util.namedChildElements(learnEl,"example");
		
		for (Element exampleEl:exampleEls)
		{
			// the sentence used for learning
			String sentenceName = exampleEl.getAttribute("name");
						
			/* each example feature structure has a top-level attribute unifyOrder,
			 * which defines the words already known at the time of learning, and the order in which they are unified
			 * for partial understanding before generalisation */
			Element scriptEl = Util.firstNamedChild(exampleEl, "script");
			FSNode example = new FSNode(model,sentenceName,scriptEl,0,null);
			addExample(example)	;	
		}
	}
	
	
	// add a learning example
	public void addExample(FSNode example)
	{
		examples.add(example);
		example.setWordLearnExampleParent(this);
	}
	
	//--------------------------------------------------------------------------
	//     Getting a word to be used for partial parsing of a learning example
	//--------------------------------------------------------------------------
	
	public FSNode getWord(String wordName)
	{
		// before any learning cycle, return a word from the static language model
		if (getCycleNo() == 0) return model.getWords().get(wordName);
		// for any learning cycle, return a word learnt in or before that cycle
		return stage.getLearner().getKnownWord(getCycleNo(), wordName);
	}
	
	
	//--------------------------------------------------------------------------
	//                        XML saving
	//--------------------------------------------------------------------------
	

	
	/**
	 * 
	 * @param doc
	 * @return
	 * @throws Exception
	 */
	public Element xmlForm(Document doc, boolean forClient) throws Exception
	{
		Element learnEl = Util.newElement(doc, "learn");
		learnEl.setAttribute("word", getWordName());
		
		if (wordFS != null)
		{
			Element structureEl = Util.newElement(doc, "structure");
			learnEl.appendChild(structureEl);
			// in case the user on the client selects a node
			if (forClient)
			{
				wordFS.setRow(-1);
				wordFS.setColumn(0);
				wordFS.setAllTruePositions();
			}
			structureEl.appendChild(wordFS.xmlForm(doc,forClient));
		}
		
		for (int i = 0; i < examples.size();i++)
		{
			FSNode example = examples.get(i);
			String eName = example.getName();
			
			Element exampleEl = Util.newElement(doc, "example");
			exampleEl.setAttribute("name", eName);
			learnEl.appendChild(exampleEl);
			// examples are not used on the client, so there is no need to set their row, column, or tree positions (no need to send them at all, really)
			exampleEl.appendChild(example.xmlForm(doc,forClient));
		}
		return learnEl;		
	}

}
